/*
 *                      ***** TA Pattern Search Client *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#define WITH_SRCH_INTERFACE
#include "srch.h"

TaTable g_tab_query, g_tab_results;
SearchQuery g_query;
SearchResults g_results;
FILE *g_fd;
TaLongLong g_repeat;
int my_sid_number;

void DoNextQuery(TaTable *table, void *par);

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 17;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	char name[TA_NAME_LENGTH];
	int i, stations_amount = 2;
	my_sid_number = 1;

	taInitialize();
	taDefaultAreaQos(&aqos);
	taDefaultStationQos(&sqos);
	taDefaultChannelQos(&cqos);
	name[0] = 0;
	aqos.tp_cycle = 100000;
	strcpy(sqos.hostname, "127.0.0.1");
	sqos.first_port = 9000;
	sqos.enable_mask = TA_ENABLE_TCP_CLIENT;
	sqos.send_strat = TA_SEND_STRAT_CLIENT;
	cqos.assign_sids = _True;
	taInitArea(aid, &aqos, "srch", &area);
	taReallocStations(MAX_STATIONS, &area);
	g_fd = NULL;

	for (i = 1; i < argc; i++) {
		if (!strcmp("-n", argv[i])) {
			if (argv[i+1])
				my_sid_number = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-t", argv[i]) || !strcmp("-trace_level", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-e", argv[i]) || !strcmp("-protocol", argv[i])) {
			if (argv[i+1]) {
				aqos.enable_mask = sqos.enable_mask 
					= atoi(argv[i+1]);
				taInitArea(aid, &aqos, "srch", &area);
				taReallocStations(MAX_STATIONS, &area);
			}
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-a", argv[i])) {
			if (argv[i+1])
				area.aid = aid = (TaLong)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-port", argv[i])) {
			if (argv[i+1])
				sqos.first_port = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-hostname", argv[i])) {
			if (argv[i+1])
				strcpy(sqos.hostname, argv[i+1]);
			++i;
		} else if (!strcmp("-name", argv[i])) {
			if (argv[i+1])
				strcpy(name, argv[i+1]);
			++i;
		} else if (!strcmp("-n_stations", argv[i])) {
			if (argv[i+1])
				stations_amount = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-repeat", argv[i])) {
			if (argv[i+1])
				g_repeat = 1000000*atoi(argv[i+1]);
			++i;
		} else {
			if (g_query.pattern[0] == 0)
				strcpy(g_query.pattern, argv[i]);
			else {
				g_fd = fopen(argv[i], "r");
				if (!g_fd) {
					printf("file not found '%s'\n", argv[i]);
					return 4;
				}
			}
		}
	}

	for (i = 0; i < stations_amount; i++)
		taInitStation(i, &sqos, &cqos, name, &area, &stations[i]);

	if (my_sid_number < 0 || my_sid_number >= stations_amount)
		return 2;
	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);

	i = taParseToRTTI(g_SearchQuery, "Tests.SRCH", &g_rtti_SearchQuery, 0);
	i |= taParseToRTTI(g_FoundString, "Tests.SRCH", &g_rtti_FoundString, 0);	
	i |= taParseToRTTI(g_SearchResults, "Tests.SRCH",
		&g_rtti_SearchResults, 0);
	if (i < 0)
		return 3;
	if (!g_fd) {
		printf("usage: %s -op val [-op val] pattern filename\n", argv[0]);
		return 4;
	}
	taNewTable(&area, &g_tab_query);
	taNewTable(&area, &g_tab_results);
	taAttachTable(&g_tab_query, &g_rtti_SearchQuery, 0, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 1, NULL, TA_REALLOC_EXACT);
	taAttachTable(&g_tab_results, &g_rtti_SearchResults, 0, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 1, NULL, TA_REALLOC_EXACT);

	g_query.text = (TaChar *)
		taAllocWithRTTI(sizeof(TaChar), 0);
	taWaitTable(&g_tab_results, 3000000);
	taCallTable(&g_tab_results, DoNextQuery, (void *)1);
	taRun(&area, &my_sid_number);

	taCloseStationsData(&area);
	return 0;
}

/* ------------------------------------------------------- */
void DoNextQuery(TaTable *table, void *par)
{
	char str[MAX_STRING_LENGTH], 
		text[N_STRINGS_AT_ONCE*MAX_STRING_LENGTH];
	int i, l;
	SearchResults *sr;
	if (!par) {
		sr = (SearchResults *) table->recent_data;
		l = taLengthWithRTTI(sr->found);
		for (i = 0; i < l; i++)
			printf("#%d\t%s\n", sr->found[i].pos,
				sr->found[i].result_str);
	}
	if (!g_fd)
		return;
	text[0] = 0;
	for (i = 0; i < N_STRINGS_AT_ONCE; i++) {
		if (!fgets(str, MAX_STRING_LENGTH, g_fd)) {
			if (g_repeat) {
				fseek(g_fd, 0, SEEK_SET);
				taWaitTable(&g_tab_results, g_repeat);
				taCallTable(&g_tab_results, DoNextQuery, 
					(void *)1);
			} else {
				fclose(g_fd);
				g_fd = NULL;
				table->area->finish_flag = 3;
			}
			break;
		}
		strcat(text, str);
	}
	i = strlen(text);
	g_query.text = (TaChar *) taReallocWithRTTI(g_query.text, i+1);
	strcpy(g_query.text, text);
	taWriteTable(&g_tab_query, &g_query, TA_TO_ADDR(1-my_sid_number), 
		_True);
	taWaitTable(&g_tab_results, TA_WAIT_ANY_MESSAGE);
	taCallTable(&g_tab_results, DoNextQuery, NULL);
}

