/*
 *                      ***** TA Pattern Search Client *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#define WITH_SRCH_INTERFACE
#include "fsrch.h"

TaTable g_tab_search;
SearchResults g_search;
TaLongLong g_repeat;
int my_sid_number;

void DoNextQuery(TaTable *table, void *par);
void OnErrorMessage(const void *data, TaULong msg_type,
        const struct TaTable *t);


/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 16;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	char name[TA_NAME_LENGTH];
	int i, stations_amount = 2;
	my_sid_number = 1;

	taInitialize();
	taDefaultAreaQos(&aqos);
	taDefaultStationQos(&sqos);
	taDefaultChannelQos(&cqos);
	name[0] = 0;
	aqos.tp_cycle = 100000;
	strcpy(sqos.hostname, "127.0.0.1");
	sqos.first_port = 9000;
	sqos.enable_mask = TA_ENABLE_TCP_CLIENT;
	sqos.send_strat = TA_SEND_STRAT_CLIENT;
	cqos.assign_sids = _True;
	taInitArea(aid, &aqos, "fsrch", &area);
	taReallocStations(MAX_STATIONS, &area);
	g_search.start_pos = 0;
	g_search.end_pos = 0;

	for (i = 1; i < argc; i++) {
		if (!strcmp("-n", argv[i])) {
			if (argv[i+1])
				my_sid_number = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-t", argv[i]) || !strcmp("-trace_level", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-e", argv[i]) || !strcmp("-protocol", argv[i])) {
			if (argv[i+1]) {
				aqos.enable_mask = sqos.enable_mask 
					= atoi(argv[i+1]);
				taInitArea(aid, &aqos, "fsrch", &area);
				taReallocStations(MAX_STATIONS, &area);
			}
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-a", argv[i])) {
			if (argv[i+1])
				area.aid = aid = (TaLong)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-port", argv[i])) {
			if (argv[i+1])
				sqos.first_port = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-hostname", argv[i])) {
			if (argv[i+1])
				strcpy(sqos.hostname, argv[i+1]);
			++i;
		} else if (!strcmp("-name", argv[i])) {
			if (argv[i+1])
				strcpy(name, argv[i+1]);
			++i;
		} else if (!strcmp("-repeat", argv[i])) {
			if (argv[i+1])
				g_repeat = 1000000*atoi(argv[i+1]);
			++i;
		} else {
			if (g_search.pattern[0] == 0)
				strcpy(g_search.pattern, argv[i]);
			else
				strcpy(g_search.file_name, argv[i]);
		}
	}

	for (i = 0; i < stations_amount; i++)
		taInitStation(i, &sqos, &cqos, name, &area, &stations[i]);

	if (my_sid_number < 0 || my_sid_number >= stations_amount)
		return 2;
	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);

	i = taParseToRTTI(g_FoundString, "Tests.FSRCH", 
		&g_rtti_FoundString, 0);	
	i |= taParseToRTTI(g_SearchResults, "Tests.FSRCH",
		&g_rtti_SearchResults, 0);
	if (i < 0)
		return 3;
	if (!g_search.file_name[0]) {
		printf("usage: %s -op val [-op val] pattern filename\n", argv[0]);
		return 4;
	}
	taNewTable(&area, &g_tab_search);
	taAttachTable(&g_tab_search, &g_rtti_SearchResults, 12, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 1, OnErrorMessage, TA_REALLOC_EXACT);

	taWaitTable(&g_tab_search, 3000000);
	taCallTable(&g_tab_search, DoNextQuery, (void *)1);
	taRun(&area, &my_sid_number);

	taCloseStationsData(&area);
	return 0;
}

/* ------------------------------------------------------- */
void DoNextQuery(TaTable *table, void *par)
{
	char str[MAX_STRING_LENGTH];
	int i, l;
	SearchResults *sr;
	if (!par) {
		sr = (SearchResults *) table->recent_data;
		l = taLengthWithRTTI(sr->found);
		printf("~ %d nfound:%d pos:%d %d\n", sr->request_number, l, 
			sr->start_pos, sr->end_pos);
		for (i = 0; i < l; i++)
			printf("#%d\t%s\n", sr->found[i].pos,
				sr->found[i].result_str);
	} else
		sr = &g_search;
	if (sr->end_pos < 0) {
		taClearTable(table);
		if (g_repeat > 0) {
			sr->end_pos = 0;
			taWaitTable(&g_tab_search, g_repeat);
			taCallTable(&g_tab_search, DoNextQuery, 
				(void *)1);
		} else {
			table->area->finish_flag = 3;
		}
	} else {
		sr->start_pos = sr->end_pos;
		taRequestTable(&g_tab_search, sr, 
			TA_TO_ADDR(1-my_sid_number), _True, DoNextQuery, NULL);
	}
}

/* ------------------------------------------------------- */
void OnErrorMessage(const void *data, TaULong msg_type,
        const struct TaTable *t)
{
	if (msg_type == taErrorDataMsgType()) {
		TaErrorData *edp = (TaErrorData *) data;
		printf("*** error:%d table:%d\n", 
			edp->errcode, t->dst_oid);
	}
}
