/*
 *                      ***** TA Global Weather SOAP Server *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */
#define MAX_STATIONS	2
#include "weather.h"

TaTable g_tab_GetCitiesByCountry, g_tab_GetWeather;
GetCitiesByCountry g_var_GetCitiesByCountry;
GetWeather g_var_GetWeather;
int my_sid_number;

int ClientInit(TaArea *area, int match_mode);
void DoGetCitiesByCountry(TaTable *table, void *par);
void DoGetWeather(TaTable *table, void *par);

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 22;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	int i, stations_amount = MAX_STATIONS, match_mode;
	TaChar *country = NULL, *city = NULL;

	my_sid_number = 1;
	match_mode = TA_SOAP_GET_CLIENT;

	taInitialize();
	taDefaultAreaQos(&aqos);
	taDefaultStationQos(&sqos);
	taDefaultChannelQos(&cqos);
	aqos.tp_cycle = 1000000;
	strcpy(sqos.hostname, "127.0.0.1");
	sqos.first_port = -80;
	sqos.prot_format = TA_PROT_FORMAT_SOAP;
	sqos.enable_mask = TA_ENABLE_TCP_CLIENT;
	sqos.ping_mask = 0;
	sqos.send_strat = TA_SEND_STRAT_CLIENT;
	cqos.assign_sids = _True;
	cqos.max_send_polls = 0;
	taInitArea(aid, &aqos, "gweather", &area);
	taReallocStations(MAX_STATIONS, &area);

	for (i = 1; argv[i]; i++) {
		if (!strcmp("-t", argv[i]) || 
		!strcmp("-trace_level", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-hostname", argv[i])) {
			if (argv[i+1])
				strcpy(sqos.hostname, argv[i+1]);
			++i;
		} else if (!strcmp("-port", argv[i])) {
			if (argv[i+1])
				sqos.first_port = -atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-soap", argv[i])) {
			if (argv[i+1]) {
				if (!strcmp("11", argv[i+1]))
					match_mode = TA_SOAP_11_CLIENT;
				else if (!strcmp("12", argv[i+1]))
					match_mode = TA_SOAP_12_CLIENT;
				else if (!strcmp("GET", argv[i+1]))
					match_mode = TA_SOAP_GET_CLIENT;
				else if (!strcmp("POST", argv[i+1]))
					match_mode = TA_SOAP_POST_CLIENT;
			}
			++i;
		} else {
			if (!country)
				country = argv[i];
			else
				city = argv[i];
		}
	}
	if (!country) {
		printf("usage: %s -op val [-op val] country [city]\n",
			argv[0]);
		return 4;
	}
	for (i = 0; i < stations_amount; i++)
		taInitStation(i, &sqos, &cqos, NULL, &area, 
			&stations[i]);

	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);

	i = ClientInit(&area, match_mode);
	if (i < 0) {
		printf("*init error rc:%d\n", i);
		return 1;
	}

	if (!city) {
		taWaitTable(&g_tab_GetCitiesByCountry, 5000000);
		taChar2WChar(g_var_GetCitiesByCountry.CountryName, country, 128);
		taCallTable(&g_tab_GetCitiesByCountry, DoGetCitiesByCountry, 
			(void *)1);
	} else {
		taWaitTable(&g_tab_GetWeather, 5000000);
		taChar2WChar(g_var_GetWeather.CountryName, country, 128);
		taChar2WChar(g_var_GetWeather.CityName, city, 128);
		taCallTable(&g_tab_GetWeather, DoGetWeather, (void *)1);
	}
	taRun(&area, &my_sid_number);

	taCloseStationsData(&area);
	return 0;
}

/* ------------------------------------------------------- 
 *
 * Client Logic
 *
 * ------------------------------------------------------- */
int ClientInit(TaArea *area, int match_mode)
{
	int i;
	i = taParseToRTTI(g_CountryList, "globalweather.asmx", 
		&g_rtti_CountryList, 0);
	i |= taParseToRTTI(g_NDSet, "globalweather.asmx", 
		&g_rtti_NDSet, 0);
	i |= taParseToRTTI(g_GetCitiesByCountry, "globalweather.asmx", 
		&g_rtti_GetCitiesByCountry, 0);
	i |= taParseToRTTI(g_CWeather, "globalweather.asmx", 
		&g_rtti_CWeather, 0);
	i |= taParseToRTTI(g_GetWeather, "globalweather.asmx", 
		&g_rtti_GetWeather, 0);
	if (i < 0)
		return i;

	taNewTable(area, &g_tab_GetCitiesByCountry);
	taNewTable(area, &g_tab_GetWeather);
	g_tab_GetCitiesByCountry.mqos.tp_timeout = 20000000;
	g_tab_GetCitiesByCountry.mqos.tp_send_timeout = 10000000;
	g_tab_GetWeather.mqos.tp_timeout = 20000000;
	g_tab_GetWeather.mqos.tp_send_timeout = 10000000;
	i |=
	taAttachTable(&g_tab_GetCitiesByCountry, &g_rtti_GetCitiesByCountry, 
		1-my_sid_number, 0, match_mode, 1, 
		NULL, TA_REALLOC_EXACT);
	i |=
	taAttachTable(&g_tab_GetWeather, &g_rtti_GetWeather, 
		1-my_sid_number, 1, match_mode, 1, 
		NULL, TA_DONT_REALLOC);

	return i;
}

/* ------------------------------------------------------- */
void DoGetCitiesByCountry(TaTable *table, void *par)
{
	GetCitiesByCountry *gcc;
	int i, l;
	char buf1[128], buf2[128];
	if (!par) {
		gcc = (GetCitiesByCountry *) table->recent_data;
		l = taLengthWithRTTI(gcc->NewDataSet.Table);
		printf("'%s' ncities:%d \n", 
			taWChar2Char(buf1,
			g_var_GetCitiesByCountry.CountryName, 128), l);
		for (i = 0; i < l; i++)
			printf("  '%s' '%s'\n",
				taWChar2Char(buf1,
					gcc->NewDataSet.Table[i].City, 128),
				taWChar2Char(buf2,
					gcc->NewDataSet.Table[i].Country, 128));
	} else {
		taRequestTable(table, &g_var_GetCitiesByCountry,
			TA_TO_ADDR(1-my_sid_number), _False, 
			DoGetCitiesByCountry, NULL);
	}
}

/* ------------------------------------------------------- */
void DoGetWeather(TaTable *table, void *par)
{
	GetWeather *gw;
	char buf[128];
	if (!par) {
		gw = (GetWeather *) table->recent_data;
		printf("'%s' ", 
			taWChar2Char(buf,
			g_var_GetWeather.CountryName, 128));
		printf("'%s'\n", 
			taWChar2Char(buf,
			g_var_GetWeather.CityName, 128));
		printf("  Location: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.Location, 128));
		printf("  Time: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.Time, 128));
		printf("  Wind: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.Wind, 128));
		printf("  Visibility: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.Visibility, 128));
		printf("  SkyConditions: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.SkyConditions, 128));
		printf("  Temperature: '%g'\n", gw->CurrentWeather.Temperature);
		printf("  DewPoint: '%g'\n", gw->CurrentWeather.DewPoint);
		printf("  RelativeHumidity: '%d'\n", 
			gw->CurrentWeather.RelativeHumidity);
		printf("  Pressure: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.Pressure, 128));
		printf("  Status: '%s'\n", 
			taWChar2Char(buf,
			gw->CurrentWeather.Status, 128));
	} else {
		taRequestTable(table, &g_var_GetWeather,
			TA_TO_ADDR(1-my_sid_number), _False, 
			DoGetWeather, NULL);
	}
}

