/*
 *                      ***** TA Pattern Search Client *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#define WITH_SRCH_INTERFACE
#include "fsrch.h"

TaTable g_tab_search;
SearchResults g_search;
int my_sid_number;
int dst_oid;
TaBool controlled;

void DoNextQuery(TaTable *table, void *par);
void OnErrorMessage(const void *data, TaULong msg_type,
        const struct TaTable *t);


/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 19;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	char name[TA_NAME_LENGTH];
	int i, stations_amount = 2;
	my_sid_number = 1;
	dst_oid = 100;
	controlled = _True;

	taInitialize();
	taDefaultAreaQos(&aqos);
	taDefaultStationQos(&sqos);
	taDefaultChannelQos(&cqos);
	name[0] = 0;
	aqos.tp_cycle = 100000;
	strcpy(sqos.hostname, "127.0.0.1");
	sqos.first_port = 8000;
	sqos.enable_mask = TA_ENABLE_UDP_MCAST;
	taInitArea(aid, &aqos, "psrch", &area);
	taReallocStations(MAX_STATIONS, &area);
	g_search.start_pos = 0;
	g_search.end_pos = 0;
	controlled = _True;

	for (i = 1; i < argc; i++) {
		if (!strcmp("-n", argv[i])) {
			if (argv[i+1])
				my_sid_number = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-t", argv[i]) || !strcmp("-trace_level", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-e", argv[i]) || !strcmp("-protocol", argv[i])) {
			if (argv[i+1]) {
				aqos.enable_mask = sqos.enable_mask 
					= atoi(argv[i+1]);
				taInitArea(aid, &aqos, "psrch", &area);
				taReallocStations(MAX_STATIONS, &area);
			}
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-a", argv[i])) {
			if (argv[i+1])
				area.aid = aid = (TaLong)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-port", argv[i])) {
			if (argv[i+1])
				sqos.first_port = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-hostname", argv[i])) {
			if (argv[i+1])
				strcpy(sqos.hostname, argv[i+1]);
			++i;
		} else if (!strcmp("-name", argv[i])) {
			if (argv[i+1])
				strcpy(name, argv[i+1]);
			++i;
		} else if (!strcmp("-dst_oid", argv[i])) {
			if (argv[i+1])
				dst_oid = atoi(argv[i+1]);
			++i;
		} else {
			if (g_search.pattern[0] == 0)
				strcpy(g_search.pattern, argv[i]);
			else
				strcpy(g_search.file_name, argv[i]);
		}
	}

	for (i = 0; i < stations_amount; i++)
		taInitStation(i, &sqos, &cqos, name, &area, &stations[i]);

	if (my_sid_number < 0 || my_sid_number >= stations_amount)
		return 2;
	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);

	i = taParseToRTTI(g_FoundString, "Tests.PSRCH", 
		&g_rtti_FoundString, 0);	
	i |= taParseToRTTI(g_SearchResults, "Tests.PSRCH",
		&g_rtti_SearchResults, 0);
	if (i < 0)
		return 3;
	if (dst_oid < 100 || dst_oid > 104) {
		printf("dst_oid must be within [100,104]\n");
		return 4;
	}
	if (dst_oid == 100 || g_search.file_name[0] == 0)
		controlled = _False;
	taNewTable(&area, &g_tab_search);
	g_tab_search.mqos.tp_timeout = 10000000;
	taAttachTable(&g_tab_search, &g_rtti_SearchResults, dst_oid, dst_oid, 
		TA_MSG_SRPS_MATCH_DST, 1, OnErrorMessage, TA_REALLOC_EXACT);

	taWaitTable(&g_tab_search, 3000000);
	if (!controlled) {
		taSubscribeTable(&g_tab_search, _False, DoNextQuery, NULL);
	} else {
		taWriteTable(&g_tab_search, &g_search, 
			TA_TO_ADDR(1-my_sid_number), _True);
		taSubscribeTable(&g_tab_search, _True, DoNextQuery, NULL);
	}
	taRun(&area, &my_sid_number);

	taCloseStationsData(&area);
	return 0;
}

/* ------------------------------------------------------- */
void DoNextQuery(TaTable *table, void *par)
{
	char str[MAX_STRING_LENGTH];
	int i, l;
	SearchResults *sr;
	sr = (SearchResults *) table->recent_data;
	l = taLengthWithRTTI(sr->found);
	printf("~ %d nfound:%d pos:%d %d\n", sr->request_number, l, 
		sr->start_pos, sr->end_pos);
	for (i = 0; i < l; i++)
		printf("#%d\t%s\n", sr->found[i].pos,
			sr->found[i].result_str);
	if (dst_oid == 100) {
		// again and again
	} else if (sr->end_pos < 0 || sr->end_pos >= 10000) {
		if (controlled)
			taUnsubscribeTable(&g_tab_search, 
				TA_TO_ADDR(1-my_sid_number));
		else
			taUnsubscribeTable(&g_tab_search, 0);
		table->area->t_expired = taGetTime() + 3000000;
	}
}

/* ------------------------------------------------------- */
void OnErrorMessage(const void *data, TaULong msg_type,
        const struct TaTable *t)
{
	if (msg_type == taErrorDataMsgType()) {
		TaErrorData *edp = (TaErrorData *) data;
		printf("*** error:%d table:%d\n", 
			edp->errcode, t->dst_oid);
	}
}
