/*
 *                      ***** TA Noahs Ark *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ------------------------------------------------------- */
#define MAX_STATIONS 2
int areaUpdate(TaArea *area, TaLongLong t_updating, void *client_data);

int usage()
{
	printf(
"usage: noahs_ark -n number ?-option value? station-0 .. station-N\n"
"where:\n"
"'station-1 .. station-N' is valid address: hostname[:port], max amount=%d\n"
"'-n number' is current station number [0-%d]\n"
"'-e enable' enables protocol by bits:\n"
"\t1 - UDP multicast, 2 - UDP, 4 - TCP Server, 8 - TCP Client\n"
"'-p ping' sets protocol for ping by bits:\n"
"\t1 - UDP multicast, 2 - UDP, 4 - TCP Server, 8 - TCP Client\n"
"'-c cycle' sets new runtime cycle in mcs [1000-10000000]\n"
"'-f frame' sets initial frame to [1-10] or current station if negative\n"
"'-r 1' sets redundand sending on\n"
"'-a area' sets area number [4,1000000]\n"
"'-p phase_correct' sets phase_correct interval [0,..], def:60000000\n"
"'-m max_poll_fails' sets max_poll_fails def:10\n"
"'-t trace' enables tracing by bits:\n"
"\t1 - RUNTIME, 2 - TRANSPORT, 4 - POLLING, 8 - MESSAGE, 16 - MESSAGE DATA\n",
	MAX_STATIONS, MAX_STATIONS-1);
	return 1;
}

/* ------------------------------------------------------- */
TA_STRUCT(Troyka,
	TaChar horses[3];
	TaShort chariot;
)
TA_STRUCT(MyString,
	TaChar *str;
)
TA_STRUCT(NoahsArk,
	TaChar dove_female;
	TaWChar dove_male;
	TaOctet raven_female;
	TaShort raven_male;
	TaBool olive;
	TaLongLong domestic;
	TaLong wild;
	TaFloat human_female;
	TaDouble human_male;
	TaOctet bees[6];
	TaLongLong cows[2];
	TaWChar name[13];
	TaChar sname[11];
	Troyka troyka;
	TaBool *olives;
	Troyka *troykas;
	MyString *strings;
	Troyka tro[3];
)

NoahsArk noahsArk, sendArk;
TaTable table;
void HandleMessage(const void *data, TaULong msg_type,
	const struct TaTable *t);

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 69;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	int i, my_sid_number, stations_amount;
	int sqos_first_port;

	if (argc < 3)
		return usage();

	taInitialize();
	taDefaultAreaQos(&aqos);
	aqos.tp_cycle = 1000000;
	taDefaultStationQos(&sqos);
	sqos_first_port = sqos.first_port;

	my_sid_number = -1;
	stations_amount = 0;
	for (i = 1; i < argc && stations_amount < MAX_STATIONS; i++) {
		if (!strcmp("-n", argv[i])) {
			if (argv[i+1])
				my_sid_number = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-t", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-e", argv[i])) {
			if (argv[i+1]) {
				aqos.enable_mask = sqos.enable_mask = 
					sqos.ping_mask = atoi(argv[i+1]);
				taInitArea(aid, &aqos, "NoahsArk", &area);
				taReallocStations(MAX_STATIONS, &area);
			}
			++i;
		} else if (!strcmp("-p", argv[i])) {
			if (argv[i+1])
				sqos.ping_mask = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-f", argv[i])) {
			if (argv[i+1])
				area.qos.justify_frame = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-a", argv[i])) {
			if (argv[i+1])
				area.aid = aid = (TaLong)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-p", argv[i])) {
			if (argv[i+1])
				area.qos.tp_phase_correct = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-m", argv[i])) {
			if (argv[i+1])
				area.qos.max_poll_fails = atoi(argv[i+1]);
			++i;
		} else {
			char s_name[32];
			sprintf(s_name, "%ds%d", (int)aid, stations_amount);
			if (*argv[i] == '*') {
				strcpy(sqos.hostname, argv[i]+1);
				if (my_sid_number < 0)
					my_sid_number = stations_amount;
			} else
				strcpy(sqos.hostname, argv[i]);
			taInitStation(stations_amount, &sqos, NULL, s_name, 
				&area, &stations[stations_amount]);
			++stations_amount;
		}
	}

	if (my_sid_number < 0 || my_sid_number >= stations_amount)
		return usage();
	if (!taSetStationHostBySid(my_sid_number, &area))
		return TA_ERR_ARGUMENT;
	taInitStationsData(&area);

	i = taParseToRTTI(g_Troyka, "Tests.INoahsArk", &g_rtti_Troyka, 0);
	i |= taParseToRTTI(g_MyString, "Tests.INoahsArk", &g_rtti_MyString, 0);
	i |= taParseToRTTI(g_NoahsArk, "Tests.INoahsArk", &g_rtti_NoahsArk, 0);
	if (i < 0) {
		printf("*** parse error:%d\n", i);
		return i;
	}
	taNewTable(&area, &table);
	taAttachTable(&table, &g_rtti_NoahsArk, 0, 0, TA_MSG_SRPS_ACCEPT_ANY, 
		1, HandleMessage, TA_REALLOC_LARGEST);

	sendArk.dove_female = 'd';
	sendArk.dove_male = 'D';
	sendArk.raven_female = 6;
	sendArk.raven_male = 9;
	sendArk.olive = _True;
	sendArk.wild = my_sid_number;
	sendArk.domestic = area.host_station->channels[0].port;
	sendArk.human_female = 1.234;
	sendArk.human_male = 4.321;
	taChar2WChar(sendArk.name, "Noahs Ark", 13);
	printf("~domestic:%d horses:'%c'/%d '%c' '%c' ch:%d ~olives:%d %x\n", 
((char*)&sendArk.domestic-(char*)&sendArk),
sendArk.troyka.horses[0], sendArk.troyka.horses[0], 
		sendArk.troyka.horses[1], sendArk.troyka.horses[2], sendArk.troyka.chariot,
		((char*)&sendArk.olives-(char*)&sendArk), sendArk.olives);

	strcpy(sendArk.sname, "noahs ark");
	sendArk.olives = (TaBool *) taAllocWithRTTI(sizeof(TaBool), 3);
	sendArk.olives[0] = _True; sendArk.olives[1] = _False; 
	sendArk.olives[2] = _True;
	sendArk.troykas = (Troyka *) taAllocWithRTTI(sizeof(Troyka), 2);
//printf("%x troyka~%x troyka.chariot~%x olives~%x:%x troykas~%x:%x\n", &sendArk, 
//&sendArk.troyka, &sendArk.troyka.chariot, &sendArk.olives, 
//sendArk.olives, &sendArk.troykas, sendArk.troykas);
	sendArk.troykas[0].horses[0] = 'x'; sendArk.troykas[0].horses[1] = 'y';
	sendArk.troykas[0].horses[2] = 'z'; sendArk.troykas[0].chariot = 11;
	sendArk.troykas[1].horses[0] = 'X'; sendArk.troykas[1].horses[1] = 'Y';
	sendArk.troykas[1].horses[2] = 'Z'; sendArk.troykas[1].chariot = 12;
	sendArk.strings = (MyString *) taAllocWithRTTI(sizeof(MyString), 2);
	sendArk.strings[0].str = taDupStringWithRTTI("first");
	sendArk.strings[1].str = taDupStringWithRTTI("the second string"); 
	sendArk.tro[0].horses[0] = '1'; sendArk.tro[0].horses[1] = '2';
	sendArk.tro[0].horses[2] = '3'; sendArk.tro[0].chariot = 31;
	sendArk.tro[1].horses[0] = '4'; sendArk.tro[1].horses[1] = '5';
	sendArk.tro[1].horses[2] = '6'; sendArk.tro[1].chariot = 32;
	sendArk.tro[2].horses[0] = '7'; sendArk.tro[2].horses[1] = '8';
	sendArk.tro[2].horses[2] = '9'; sendArk.tro[2].chariot = 33;

	area.on_update = areaUpdate;
	taRun(&area, &my_sid_number);

	return 0;
}


/* ------------------------------------------------------- */
void HandleMessage(const void *data, TaULong msg_type,
	const struct TaTable *t)
{
	int prev_counter;
	if (msg_type == t->msg_type) {
		NoahsArk *ark = (NoahsArk *) data;
		int i, l;
		char buf[13];
		noahsArk = *ark;
		if (ta_trace_level & 64)
			printf("dove~%d~%d human_male~%d bees~%d cows~%d"
			" name~%d sname~%d troyka~%d olives~%d troykas~%d\n",
				(char *)(&ark->dove_female)-(char *)ark,
				(char *)(&ark->dove_male)-(char *)ark,
				(char *)(&ark->human_male)-(char *)ark,
				(char *)(&ark->bees)-(char *)ark,
				(char *)(&ark->cows)-(char *)ark,
				(char *)(&ark->name)-(char *)ark,
				(char *)(&ark->sname)-(char *)ark,
				(char *)(&ark->troyka)-(char *)ark,
				(char *)(&ark->olives)-(char *)ark,
				(char *)(&ark->troykas)-(char *)ark);
		printf("dove:%c:%c raven:%d:%d olive:%d wild:%d domestic:%d"
		" human:%g:%g bees:%d:%d cows:%d:%d %s:%s :%c%c%c%d", 
			ark->dove_female, (TaChar)ark->dove_male, 
			ark->raven_female, ark->raven_male, 
			ark->olive, ark->wild, (int)ark->domestic,
			ark->human_female, (float)ark->human_male,
			(int)ark->bees[0], (int)ark->bees[1], 
			(int)ark->cows[0], (int)ark->cows[1], 
			taWChar2Char(buf, ark->name, 13), ark->sname, (char)ark->troyka.horses[0],
			(char)ark->troyka.horses[1],
			(char)ark->troyka.horses[2],
			(char)ark->troyka.chariot);
		for (i = 0; i < 3; i++)
		printf(" |%c%c%c %d", (char)ark->tro[i].horses[0],
			(char)ark->tro[i].horses[1], 
			(char)ark->tro[i].horses[2],
			ark->tro[i].chariot);
		printf(" olives:%x troykas:%x strings:%x ", 
			ark->olives, ark->troykas, ark->strings);
		for (i = 0, l = taLengthWithRTTI(ark->olives);
		i < l; i++)
			printf("$%s", (ark->olives[i] == 0 ? "FALSE" : "TRUE"));
		for (i = 0, l = taLengthWithRTTI(ark->troykas);
		i < l; i++)
			printf(".%c%c%c%d", ark->troykas[i].horses[0],
			ark->troykas[i].horses[1], ark->troykas[i].horses[2],
			ark->troykas[i].chariot);
		for (i = 0, l = taLengthWithRTTI(ark->strings);
		i < l; i++)
			printf(":%s", ark->strings[i].str);
		printf("\n");
	} else if (msg_type == taErrorDataMsgType()) {
		TaErrorData *edp = (TaErrorData *) data;
		printf("*** error:%d \n", edp->errcode);
	}
}

/* ------------------------------------------------------- */
int areaUpdate(TaArea *area, TaLongLong t_crtu, void *client_data)
{
	TaWChar dove; TaShort raven;
	if (area->n_msgs_not_found)
		printf("*** not_found: %d\n", area->n_msgs_not_found);
	sendArk.olive = !sendArk.olive;
	dove = sendArk.dove_male; sendArk.dove_male = sendArk.dove_female;
		sendArk.dove_female = (TaChar) dove;
	raven = sendArk.raven_male; sendArk.raven_male = sendArk.raven_female;
		sendArk.raven_female = (TaOctet) raven;
	if (sendArk.olive) {
		sendArk.human_female = 1.234;
		sendArk.human_male = 4.321;
		sendArk.bees[0] = 1;
		sendArk.bees[1] = 2;
		sendArk.cows[0] = 123;
		sendArk.cows[1] = 456;
		sendArk.troyka.horses[0] = 'A';
		sendArk.troyka.horses[1] = 'B';
		sendArk.troyka.horses[2] = 'C';
		sendArk.troyka.chariot = 99;
		taChar2WChar(sendArk.name, "Noahs Ark", 13);
	} else {
		sendArk.human_female = 4.321;
		sendArk.human_male = 1.234;
		sendArk.bees[0] = 2;
		sendArk.bees[1] = 1;
		sendArk.cows[0] = 456;
		sendArk.cows[1] = 123;
		sendArk.troyka.horses[0] = 'a';
		sendArk.troyka.horses[1] = 'b';
		sendArk.troyka.horses[2] = 'c';
		sendArk.troyka.chariot = 88;
		taChar2WChar(sendArk.name, "Noahs Ark1", 13);
	}
	taWriteTable(&table, &sendArk, TA_TO_ADDR(1-area->host_station->sid), 
		_False);
	return 0;
}
