/*
 *                      ***** TA nodex *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ------------------------------------------------------- */
#define MAX_STATIONS	32
int areaUpdate(TaArea *area, TaLongLong t_updating, void *client_data);
int printStatistics(TaStation *station, TaLongLong t_crtu, void *client_data);
TaMessageQos g_mqos;

int usage()
{
	printf(
"usage: nodex -n number ?-option value? station-0 .. station-N\n"
"where:\n"
"'station-1 .. station-N' is valid address: hostname[:port], max amount=%d\n"
"'-n number' is current station number [0-%d]\n"
"'-e enable' enables protocol by bits:\n"
"\t1 - UDP multicast, 2 - UDP, 4 - TCP Server, 8 - TCP Client\n"
"'-p ping' sets protocol for ping by bits:\n"
"\t1 - UDP multicast, 2 - UDP, 4 - TCP Server, 8 - TCP Client\n"
"'-c cycle' sets new runtime cycle in mcs [1000-10000000]\n"
"'-f frame' sets initial frame to [1-10] or current station if negative\n"
"'-r 1' sets redundand sending on\n"
"'-a area' sets area number [4,1000000]\n"
"'-p phase_correct' sets phase_correct interval [0,..], def:60000000\n"
"'-m max_poll_fails' sets max_poll_fails def:10\n"
"'-t trace' enables tracing by bits:\n"
"\t1 - RUNTIME, 2 - TRANSPORT, 4 - POLLING, 8 - MESSAGE, 16 - MESSAGE DATA\n",
	MAX_STATIONS, MAX_STATIONS-1);
	return 1;
}

/* ------------------------------------------------------- */
TA_STRUCT(NodeStatus,
	TaLongLong t_changed;
	TaLongLong t_display;
	TaULongLong counters_mask;
	TaLong counter;
	TaLong host_counter;
)

TaTable nodexTable;
void HandleMessage(const void *data, TaULong msg_type,
	const struct TaTable *table);

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 3;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	int i, my_sid_number, stations_amount;
	int sqos_first_port;

	if (argc < 3)
		return usage();

	taInitialize();
	taDefaultAreaQos(&aqos);
	aqos.tp_cycle = 100000;
	taInitArea(aid, &aqos, "nodex", &area);
	taReallocStations(MAX_STATIONS, &area);
	taDefaultStationQos(&sqos);
	sqos_first_port = sqos.first_port;
	taDefaultMessageQos(&g_mqos);

	sqos.enable_mask = TA_ENABLE_UDP_MCAST;
	sqos.ping_mask = TA_ENABLE_ALL;
	my_sid_number = -1;
	stations_amount = 0;
	for (i = 1; i < argc && stations_amount < MAX_STATIONS; i++) {
		if (!strcmp("-n", argv[i])) {
			if (argv[i+1])
				my_sid_number = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-t", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-e", argv[i])) {
			if (argv[i+1])
				sqos.enable_mask = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-p", argv[i])) {
			if (argv[i+1])
				sqos.ping_mask = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-f", argv[i])) {
			if (argv[i+1])
				area.qos.justify_frame = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-r", argv[i])) {
			if (argv[i+1])
				g_mqos.is_redundant = (TaBool)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-a", argv[i])) {
			if (argv[i+1])
				area.aid = aid = (TaLong)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-p", argv[i])) {
			if (argv[i+1])
				area.qos.tp_phase_correct = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-m", argv[i])) {
			if (argv[i+1])
				area.qos.max_poll_fails = atoi(argv[i+1]);
			++i;
		} else {
			char s_name[32];
			sprintf(s_name, "%ds%d", (int)aid, stations_amount);
			if (*argv[i] == '*') {
				strcpy(sqos.hostname, argv[i]+1);
				if (my_sid_number < 0)
					my_sid_number = stations_amount;
			} else
				strcpy(sqos.hostname, argv[i]);
			taInitStation(stations_amount, &sqos, NULL, s_name, 
				&area, &stations[stations_amount]);
			++stations_amount;
		}
	}

	if (my_sid_number < 0 || my_sid_number >= stations_amount)
		return usage();
	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);

	i = taParseToRTTI(g_NodeStatus, "Tests.INodex", &g_rtti_NodeStatus, 0);
	if (i < 0)
		return i;
	taNewTable(&area, &nodexTable);
	taAttachTable(&nodexTable, &g_rtti_NodeStatus, 0, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 2, HandleMessage, TA_DONT_REALLOC);

	area.on_update = areaUpdate;
	area.host_station->on_update = printStatistics;
	area.host_station->qos.tp_cycle = 10000000;
	taRun(&area, &my_sid_number);

	return 0;
}

/* ------------------------------------------------------- */
NodeStatus g_nodes[MAX_STATIONS];
typedef struct {
	int number_of_lost;
	float number_of_lost_pct;
	int not_just_in_time;
	float not_just_in_time_pct;
	int n_t_overtime, n_2t_overtime, n_gaps, n_msgs_at_once;
	int n_lost, n_done, must_recv;
} NodeStatistics;
NodeStatistics g_statistics[MAX_STATIONS];

int g_update_counter;
float g_deviation_sum, g_deviation_avg, cycle_time_sum, cycle_time;
TaLongLong g_t_start, last_crtu;	// wait 3 seconds
#define START_DELAY 10000000

/* ------------------------------------------------------- */
void HandleMessage(const void *data, TaULong msg_type,
	const struct TaTable *t)
{
	int prev_counter;
	if (msg_type == t->msg_type) {
		NodeStatus *node, *nodenew = (NodeStatus *) data;
		node = &g_nodes[t->message.src_sid];
		prev_counter = node->counter;
		*node = *nodenew;
		node->host_counter = g_update_counter;
		if (!(node->counter == prev_counter+1
		|| (node->counter == 1 && prev_counter == 1000))
		&& t->area->t_crtu >= g_t_start+START_DELAY) {
			++g_statistics[t->message.src_sid].n_gaps;
		}
		if (t->area->t_crtu >= g_t_start+START_DELAY)
			++g_statistics[t->message.src_sid].n_done;
		++g_statistics[t->message.src_sid].n_msgs_at_once;
	} else if (msg_type == taErrorDataMsgType()) {
		TaErrorData *edp = (TaErrorData *) data;
		printf("*** error:%d \n", edp->errcode);
	}
}

/* ------------------------------------------------------- */
int areaUpdate(TaArea *area, TaLongLong t_crtu, void *client_data)
{
	TaMessage message;
	NodeStatus *node;
	int msg_ind = 0, rc, prev_counter, i;
	int *my_sid_number = (int *) client_data;
	TaULongLong counters_mask;
	float dev;

	if (g_t_start == 0)
		g_t_start = t_crtu;
	// check skips and overtime
	if (t_crtu >= g_t_start+START_DELAY)
	for (i = 0; i < area->n_stations; i++) {
		if (i == *my_sid_number || !area->stations[i] /*||
		area->stations[i]->is_disconnected*/)
			continue;
		++g_statistics[i].must_recv;
		node = &g_nodes[i];
		if (g_statistics[i].must_recv > g_statistics[i].n_done)
			g_statistics[i].number_of_lost =
				g_statistics[i].must_recv
					- g_statistics[i].n_done;
		else
			g_statistics[i].number_of_lost = 0;
		if (g_statistics[i].n_msgs_at_once != 1)
			++g_statistics[i].not_just_in_time;
		g_statistics[i].n_msgs_at_once = 0;
		if (t_crtu-node->t_changed >= area->qos.tp_cycle)
			++g_statistics[i].n_t_overtime;
		if (t_crtu-node->t_changed >= 2*area->qos.tp_cycle)
			++g_statistics[i].n_2t_overtime;
		if (!(node->counters_mask & 1<<*my_sid_number))
			++g_statistics[i].n_lost;
	}
	// send
	counters_mask = 0;
	for (i = 0; i < area->n_stations; i++) {
		if (i == *my_sid_number)
			continue;
		node = &g_nodes[i];
		if ((node->t_changed >= g_nodes[*my_sid_number].t_changed
		-4000000000LL) && (node->t_changed <=
		g_nodes[*my_sid_number].t_changed+4000000000LL))
			counters_mask |= 1<<i;
	}
	// send
	node = &g_nodes[*my_sid_number];
        node->t_changed = node->t_display = t_crtu;
	node->counters_mask = counters_mask;
	++node->counter;
	if (node->counter > 1000)
		node->counter = 1;

	taWriteTable(&nodexTable, node, TA_ADDR_ALL, _False);

	++g_update_counter;
	dev = 100.*(area->t_crt0 - area->t_crtu)/area->qos.tp_cycle;
	if (dev < 0)
		dev = -dev;
	g_deviation_sum += dev;
	g_deviation_avg = g_deviation_sum/g_update_counter;
	if (g_update_counter > 1) {
		cycle_time_sum += (t_crtu - last_crtu);
		cycle_time = cycle_time_sum/(g_update_counter-1);
	}
	last_crtu = t_crtu;
	return 0;
}

/* ------------------------------------------------------- */
int printStatistics(TaStation *station, TaLongLong t_updating, 
	void *client_data)
{
	int i;
	int *my_sid_number = (int *) client_data;
	printf("deviation: %f%% cycle_time: %fmcs\n", 
		g_deviation_avg, cycle_time);
	for (i = 0; i < station->area->n_stations; i++) {
		NodeStatus *node = &g_nodes[i];
		NodeStatistics *ns = &g_statistics[i];
		if (i == *my_sid_number)
			continue;
		if (!ns->must_recv)
			return 0;
		ns->number_of_lost_pct = 100.*ns->number_of_lost
			/ns->must_recv;
		ns->not_just_in_time_pct = 100.*ns->not_just_in_time
			/ns->must_recv;
		printf("#%d number_of_lost:%d %f%% blind:%d\n", i+1,
			ns->number_of_lost, ns->number_of_lost_pct, ns->n_lost);
		printf(
		"#%d not_just_in_time:%d %f%% no_order:%d must_recv:%d\n", i+1,
			ns->not_just_in_time, ns->not_just_in_time_pct, 
			ns->n_gaps, ns->must_recv);
	}
	return 1;
}
