/*
 *                      ***** PING Example *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* ------------------------------------------------------- */
#define MAX_STATIONS	40
int g_transmitted[MAX_STATIONS], g_received[MAX_STATIONS], g_errs[MAX_STATIONS];
int stationUpdate(TaStation *station, TaLongLong t_updating, void *client_data);
void printStatistics(TaArea *area);

int usage()
{
	printf(
"usage: pings ?-option value? station-0 .. station-N\n"
"where:\n"
"'station-1 .. station-N' is list of hosts\n"
"'-c count' Stop  after  sending  count  ECHO_REQUEST packets\n"
"'-i interval' Wait interval seconds between sending each packet\n"
"'-w deadline' Specify a timeout, in seconds, before ping exits\n"
"'-I interface or address' Set  source address to specified interface address\n"
"'-t trace' enables tracing by bits:\n"
"\t1 - RUNTIME, 2 - TRANSPORT, 4 - POLLING, 8 - MESSAGE, 16 - MESSAGE DATA\n");
	return 1;
}

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 1983;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	int i, stations_amount;
	int count = 0;
        char my_addr[64];
        strcpy(my_addr, "127.0.0.1");
	if (argc < 2)
		return usage();

	taInitialize();
	taDefaultAreaQos(&aqos);
	aqos.tp_cycle = 10000;
        taInitArea(aid, &aqos, "pings", &area);
	taReallocStations(MAX_STATIONS, &area);
	taDefaultStationQos(&sqos);
	sqos.tp_cycle = sqos.tp_ping_timeout = 1000000;
	taDefaultChannelQos(&cqos);
	//cqos.tp_discovery_timeout=10000000;
#ifdef _WINDOWS
	cqos.ping_set_err = _False;
#endif

	sqos.enable_mask = TA_ENABLE_ICMP;
	sqos.ping_mask = 0;
	area.ctypes[0] = TA_CHANNEL_ICMP;
	area.n_ctypes = 1;
	stations_amount = 0;

	for (i = 1; i < argc && stations_amount < MAX_STATIONS; i++) {
		if (!strcmp("-t", argv[i])) {
			if (argv[i+1])
                                taTrace(atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				count = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-i", argv[i])) {
			float interval = 1;
			if (argv[i+1])
				interval = atoi(argv[i+1]);
			sqos.tp_cycle = sqos.tp_ping_timeout = (TaLongLong)
				(1000000 * interval);
			++i;
		} else if (!strcmp("-I", argv[i])) {
			if (argv[i+1])
				strcpy(my_addr, argv[i+1]);
			++i;
		} else if (!strcmp("-w", argv[i])) {
			if (argv[i+1])
				area.t_expired = taGetTime()+
					1000000*atoi(argv[i+1]);
			++i;
		} else {
			if (stations_amount == 0) {
				strcpy(sqos.hostname, my_addr);
				taInitStation(0, &sqos, &cqos, my_addr,
					&area, &stations[0]);
				stations_amount = 1;
			}
			strcpy(sqos.hostname, argv[i]);
			taInitStation(stations_amount, &sqos, &cqos, argv[i], 
				&area, &stations[stations_amount]);
			++stations_amount;
		}
	}

        taSetStationHostBySid(0, &area);
	taInitStationsData(&area);

	stations[0].on_update = stationUpdate;
	taRun(&area, &count);

	taCloseStationsData(&area);
	printStatistics(&area);
	return 0;
}

/* ------------------------------------------------------- */
int stationUpdate(TaStation *station, TaLongLong t_crtu, void *client_data)
{
	int *count = (int *) client_data;
	if (*count && station->n_updates >= *count)
		station->area->finish_flag = 1;
	else
		taSetMessage(station->area, NULL, 0, 
                        TA_MSG_TYPE_USER, 0, 0, NULL, TA_ADDR_ALL, 0, NULL);
	if (station->n_updates > 0) {
		int i;
		for (i = 1; i < station->area->n_stations; i++) {
			TaStation *dstation = station->area->stations[i];
			TaChannel *dstchannel = &dstation->channels[0];
			if (dstchannel->complete_send_rc > 0)
				++g_transmitted[i];
			if (dstchannel->complete_recv_rc > 0)
				++g_received[i];
			if (dstchannel->recv_errno || dstchannel->send_errno)
				++g_errs[i];
			printf("%d/%d from %s icmp_seq=%d errno=%d/%d %s\n",
				dstchannel->complete_send_rc, 
				dstchannel->complete_recv_rc,
				dstation->name, (int)station->n_updates,
				dstchannel->send_errno, dstchannel->recv_errno,
				taReadableChannelErrno(dstchannel->recv_errno,
					dstchannel));
			dstchannel->complete_send_rc = 
			dstchannel->complete_recv_rc = 0;
		}
	}
	return 0;
}

/* ------------------------------------------------------- */
void printStatistics(TaArea *area)
{
	int i;
	char buffer[128];
	TaStation *station;
	printf("-----         pings statistics         -----\nHOSTS        ");
	for (i = 1; i < area->n_stations; i++) {
		station = area->stations[i];
		printf("%-16s", station->name);
	}
	printf("\nsnd/rcv/err  ");
	for (i = 1; i < area->n_stations; i++) {
		station = area->stations[i];
		sprintf(buffer, "%d/%d/%d", g_transmitted[i], g_received[i],
			g_errs[i]);
		printf("%-16s", buffer);
	}
	printf("\ndone %%       ");
	for (i = 1; i < area->n_stations; i++) {
		station = area->stations[i];
		if (!g_transmitted[i])
			sprintf(buffer, "0 %%");
		else
			sprintf(buffer, "%d %%", 
				100*g_received[i]/g_transmitted[i]);
		printf("%-16s", buffer);
	}
	printf("\n");
}
