/*
 *                      ***** TA RPS Example *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#ifdef WITH_PYTHON
#include <Python.h>
#endif
#include <ta.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* ------------------------------------------------------- */
TA_STRUCT(Output,
	TaChar output;
	TaChar input;
	TaShort result;
	TaLong totals;
	TaLong dc_counter;
	TaLong myc_counter;
)
TA_STRUCT(Control,
	TaChar command;
)

/* ------------------------------------------------------- */
#define MAX_STATIONS	40
TaMessageQos g_mqos;
int initHostScript(TaStation *station);

void MakeYourChoice(TaTable *table, void *par);
void DoCalculate(TaTable *table, void *par);
void DoControl(TaTable *table, void *par);
void OnErrorMessage(const void *data, TaULong msg_type,
        const struct TaTable *t);

/* ------------------------------------------------------- */
Output g_outputs[MAX_STATIONS];
Control g_control;
TaTable g_tab_outputs[MAX_STATIONS];
TaTable g_tab_control;

int g_enable_emgr;
const int MAX_N_MOVES = 1000;
#define START_DELAY 10000000

/* ------------------------------------------------------- */
int g_played[MAX_STATIONS];
#ifdef WITH_PYTHON
PyObject *g_moves[MAX_STATIONS];
#endif

int totals;
float done_pct;

int usage()
{
	printf(
"usage: rps -n number ?-option value? station-0 .. station-N\n"
"where:\n"
"'station-1 .. station-N' is address: hostname:script[:port], max amount=%d\n"
"'-n number' is current station number [0-%d]\n"
"'-e enable' enables protocol by bits:\n"
"\t2 - UDP, 4 - TCP Server, 8 - TCP Client\n"
"'-c cycle' sets new runtime cycle in mcs [1000-10000000]\n"
"'-f frame' sets initial frame to [1-10] or current station if negative\n"
"'-r 1' sets redundand sending on\n"
"'-a area' sets area number [4,1000000]\n"
"'-p phase_correct' sets phase_correct interval [0,..], def:60000000\n"
"'-m max_poll_fails' sets max_poll_fails def:10\n"
"'-t trace' enables tracing by bits:\n"
"\t1 - RUNTIME, 2 - TRANSPORT, 4 - POLLING, 8 - MESSAGE, 16 - MESSAGE DATA\n",
	MAX_STATIONS, MAX_STATIONS-1);
	return 1;
}

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 5;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	int i, my_sid_number, stations_amount;
	int sqos_first_port;

	if (argc < 3)
		return usage();

#ifdef WITH_PYTHON
	Py_SetProgramName(argv[0]);
	Py_Initialize();
#endif

	taInitialize();
	taDefaultAreaQos(&aqos);
	aqos.tp_cycle = 60000;
	aqos.messages_array_size = 1000;
	taInitArea(aid, &aqos, "rps", &area);
	taReallocStations(MAX_STATIONS, &area);
	taDefaultStationQos(&sqos);
	sqos_first_port = sqos.first_port;
	taDefaultChannelQos(&cqos);
	cqos.max_send_polls = 10;
//cqos.tp_receiving_timeout=10000000;
	taDefaultMessageQos(&g_mqos);
	g_mqos.tp_timeout = 60000000;
	g_mqos.tp_send_timeout = 30000000;
	//g_mqos.n_send_attempts = 5;

	//sqos.tp_discovery_timeout = 30000000;
	//sqos.tp_ping_timeout = 20000000;
	my_sid_number = -1;
	stations_amount = 0;
	g_enable_emgr = 3;
	for (i = 1; i < argc && stations_amount < MAX_STATIONS; i++) {
		if (!strcmp("-n", argv[i])) {
			if (argv[i+1])
				my_sid_number = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-t", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-e", argv[i])) {
			if (argv[i+1]) {
				aqos.enable_mask = sqos.enable_mask =
					sqos.ping_mask = atoi(argv[i+1]);
				taInitArea(aid, &aqos, "rps", &area);
				taReallocStations(MAX_STATIONS, &area);
			}
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		} else if (!strcmp("-f", argv[i])) {
			if (argv[i+1])
				area.qos.justify_frame = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-r", argv[i])) {
			if (argv[i+1])
				g_mqos.is_redundant = (TaBool)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-a", argv[i])) {
			if (argv[i+1])
				area.aid = aid = (TaLong)atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-p", argv[i])) {
			if (argv[i+1])
				area.qos.tp_phase_correct = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-m", argv[i])) {
			if (argv[i+1])
				area.qos.max_poll_fails = atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-x", argv[i])) {
			if (argv[i+1])
				g_enable_emgr = atoi(argv[i+1]);
			++i;
		} else {
			char s_name[128], *p;
			if (*argv[i] == '*') {
				strcpy(sqos.hostname, argv[i]+1);
				if (my_sid_number < 0)
					my_sid_number = stations_amount;
			} else
				strcpy(sqos.hostname, argv[i]);
			p = strchr(sqos.hostname, ':');
			if (!p)
				return usage();
			*p = 0;
			strcpy(s_name, p+1);
			p = strchr(s_name, ':');
			if (p) {
				*p = 0;
				sqos.first_port = atoi(p+1);
			} else
				sqos.first_port = sqos_first_port;
			taInitStation(stations_amount, &sqos, &cqos, s_name, 
				&area, &stations[stations_amount]);
			++stations_amount;
		}
	}

	if (my_sid_number < 0 || my_sid_number >= stations_amount)
		return usage();
	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);
#ifdef WITH_PYTHON
	initHostScript(area.host_station);
#endif

	i = taParseToRTTI(g_Output, "Tests.IRPS", &g_rtti_Output, 0);
	i |= taParseToRTTI(g_Control, "Tests.IRPS", &g_rtti_Control, 0);

	taNewTable(&area, &g_tab_control);
	g_tab_control.mqos = g_mqos;
	taAttachTable(&g_tab_control, &g_rtti_Control, 0, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 1, OnErrorMessage, TA_DONT_REALLOC);
	for (i = 0; i < stations_amount; i++) {
		if (i == my_sid_number)
			continue;
		taNewTable(&area, &g_tab_outputs[i]);
		g_tab_outputs[i].mqos = g_mqos;
		taAttachTable(&g_tab_outputs[i], &g_rtti_Output, (TaUShort) i, 
			(TaUShort) my_sid_number, TA_MSG_SRPS_MATCH_SID, 
			1, OnErrorMessage, TA_DONT_REALLOC);
	}

	if (my_sid_number == 0) {
		g_control.command = 'I';
		taWaitTable(&g_tab_control, START_DELAY);
		taWriteTable(&g_tab_control, &g_control, 
			TA_ADDR_ALL, _False);
		taCallTable(&g_tab_control, DoControl, &my_sid_number);
	} else {
		taWaitTable(&g_tab_control, TA_WAIT_ANY_MESSAGE);
		taCallTable(&g_tab_control, DoControl, &my_sid_number);
	}
	totals = 0;
	done_pct = 0;

	taRun(&area, &my_sid_number);

#ifdef WITH_PYTHON
	Py_Finalize();
#endif
	taCloseStationsData(&area);
	return 0;
}

/* ------------------------------------------------------- */
void DoPrintResults()
{
	TaArea *area;
	int i, ndone, nerr, my_sid;
	float f;
	if (g_tab_outputs[0].area)
		area = g_tab_outputs[0].area;
	else
		area = g_tab_outputs[1].area;
	my_sid = area->host_station->sid;

	// Is Complete
	for (i = ndone = nerr = 0; i < area->n_stations; i++) {
		if (g_played[i] == MAX_N_MOVES)
			++ndone;
		else if (g_played[i] != 0)
			++nerr;
	}
	if (ndone+nerr < area->n_stations-1) {
		printf("...%s:%d ndone:%d nerr:%d\n", area->host_station->name,
			(int) my_sid, ndone, nerr);
		return;
	}
	for (i = ndone = 0; i < area->n_stations; i++) {
		if (i == my_sid || g_outputs[i].dc_counter == 0)
			continue;
		f = 50+50.*g_outputs[i].totals/g_outputs[i].dc_counter;
		printf("%s%s:%d=vs=%s:%d played:%d scope:%g%% totals:%d "
		"dc/myc:%d/%d smid/dmid:%d/%d\n",
			(g_outputs[i].dc_counter == MAX_N_MOVES ? "" : "~"),
			area->host_station->name, (int) my_sid,
			area->stations[i]->name, i,
			g_played[i], f, (int)g_outputs[i].totals, 
			(int)g_outputs[i].dc_counter, 
			(int)g_outputs[i].myc_counter,
			g_tab_outputs[i].src_mid, g_tab_outputs[i].message.mid);
		if(g_outputs[i].dc_counter == MAX_N_MOVES)
			++ndone;
	}
	printf("===%s:%d ndone:%d\n", area->host_station->name, 
		(int) my_sid, ndone);
	area->t_expired = taGetTime() + 1000000;
}

/* ------------------------------------------------------- */
void Evaluate(TaTable *table, TaShort sid, TaChar input, TaChar *output)
{
	TaStation *station;
	//station = table->area->stations[sid];
	station = table->area->host_station;
	if (!strcmp(station->name, "r"))
		*output = 'R';
	else if (!strcmp(station->name, "p"))
		*output = 'P';
	else if (!strcmp(station->name, "s"))
		*output = 'S';
	else if (!strcmp(station->name, "rand")) {
		int ir = rand();
		if (ir <= RAND_MAX/3)
			*output = 'R';
		else if (ir <= RAND_MAX/3*2)
			*output = 'P';
		else
			*output = 'S';
	} else if (!strcmp(station->name, "brand")) {
		int bias = 0;
		int ir;
		if (input == 'R')
			bias = 0;
		else if (input == 'P')
			bias = RAND_MAX/2;
		else
			bias = -RAND_MAX/2;
		ir = rand()/2+bias;
		if (ir <= RAND_MAX/3)
			*output = 'R';
		else if (ir <= RAND_MAX/3*2)
			*output = 'P';
		else
			*output = 'S';
	} else {
#ifdef WITH_PYTHON
		PyObject *pArgs, *pValue;
		char buf[2], *s;
		buf[0] = input;
		buf[1] = 0;
		pArgs = PyTuple_New(1);
		pValue = PyString_FromString(buf);
		PyTuple_SetItem(pArgs, 0, pValue);
		pValue = PyObject_CallObject(g_moves[sid], pArgs);
		if (pValue) {
			s = PyString_AsString(pValue);
			*output = *s;
			Py_DECREF(pValue);
		} else {
			printf("err input:%c\n", 
				(input=='\0' ? ' ' : input));
			*output = 0;
			PyErr_Print();
		}
		Py_DECREF(pArgs);
#else
		*output = '\0';
#endif
	}
}

/* ------------------------------------------------------- */
void MakeYourChoice(TaTable *table, void *par)
{
	Output *outp, *outhost;
        outp = (Output *) table->recent_data;
        outhost = &g_outputs[table->dst_oid];
//if(0&&table->src_oid==0)
//printf("=MakeYourChoice %s src:%d dst:%d myc:%d/%d dc:%d/%d\n",
//table->area->host_station->name, table->src_oid, table->dst_oid,
//outp->myc_counter, outhost->myc_counter, outp->dc_counter, outhost->dc_counter);
	if (outp->dc_counter != outhost->dc_counter) {
		printf("__unchoicable: %s src:%d dst:%d myc:%d/%d dc:%d/%d "
		"src_sid:%d dst_sid:%d m.dst:%d\n", 
			table->area->host_station->name, table->src_oid, 
			table->dst_oid, outp->myc_counter, 
			outhost->myc_counter, outp->dc_counter, 
			outhost->dc_counter, table->message.src_sid, 
			table->message.dst_sid, table->message.dst_oid);
		return;
	}
	Evaluate(table, table->dst_oid, g_outputs[table->dst_oid].input, 
		&g_outputs[table->dst_oid].output);
	++g_outputs[table->dst_oid].myc_counter;
	taWriteTable(&g_tab_outputs[table->dst_oid], 
		&g_outputs[table->dst_oid], TA_TO_ADDR(table->dst_oid), 
		_True);
	taWaitTable(&g_tab_outputs[table->dst_oid], TA_WAIT_ANY_MESSAGE);
	taCallTable(&g_tab_outputs[table->dst_oid], DoCalculate, NULL);
}

/* ------------------------------------------------------- */
void DoCalculate(TaTable *table, void *par)
{
	Output *outp, *outhost;
	int i, all;
	outp = (Output *) table->recent_data;
	outhost = &g_outputs[table->dst_oid];
//if(0&&table->src_oid==0)
//printf("=DoCalculate %s src:%d dst:%d myc:%d/%d dc:%d/%d\n",
//table->area->host_station->name, table->src_oid, table->dst_oid,
//outp->myc_counter, outhost->myc_counter, outp->dc_counter, outhost->dc_counter);
	if (outp->myc_counter == outhost->myc_counter) {
		if (!outhost->output || !outp->output)
			outhost->result = 0;
		else if (outhost->output == outp->output)
			outhost->result = 0;
		else if (!strchr("RPS", outhost->output))
			outhost->result = -1;
		else if (!strchr("RPS", outp->output))
			outhost->result = 1;
		else if (outhost->output == 'R') {
			if (outp->output == 'S')
				outhost->result = 1;
			else
				outhost->result = -1;
		} else if (outhost->output == 'S') {
			if (outp->output == 'R')
				outhost->result = -1;
			else 
				outhost->result = 1;
		} else {
			if (outp->output == 'R')
				outhost->result = 1;
			else
				outhost->result = -1;
		}
		outhost->totals += outhost->result;
		outhost->input = outp->output;
		++outhost->dc_counter;
		if (outhost->dc_counter >= MAX_N_MOVES) {
			taWriteTable(&g_tab_outputs[table->dst_oid], 
				&g_outputs[table->dst_oid], 
				TA_TO_ADDR(table->dst_oid), 
				_True);
			g_played[table->dst_oid] = outhost->dc_counter;
			DoPrintResults();
		} else {
			taWriteTable(&g_tab_outputs[table->dst_oid], 
				&g_outputs[table->dst_oid], 
				TA_TO_ADDR(table->dst_oid), 
				_True);
			taWaitTable(&g_tab_outputs[table->dst_oid],
				TA_WAIT_ANY_MESSAGE);
			taCallTable(&g_tab_outputs[table->dst_oid], 
				MakeYourChoice, NULL);
		}
	} else {
		printf("__unexpected: %s src:%d dst:%d myc:%d/%d "
		"dc:%d/%d src_sid:%d dst_sid:%d m.dst:%d\n", 
			table->area->host_station->name, table->src_oid, 
			table->dst_oid, outp->myc_counter, 
			outhost->myc_counter, outp->dc_counter, 
			outhost->dc_counter, table->message.src_sid, 
			table->message.dst_sid, table->message.dst_oid);
		return;
	}
	all = 0; totals = 0;
	for (i = 0; i < table->area->n_stations; i++) {
		if (i == table->area->host_station->sid)
			continue;
		all += g_outputs[i].dc_counter;
		totals += g_outputs[i].totals;
	}
	done_pct = 100.0*all/(table->area->n_stations-1)/MAX_N_MOVES;
	if (ta_trace_level & 128 || ((ta_trace_level & 64) && done_pct >= 100))
		printf("%s totals:%d done_pct:%g\n", 
			table->area->host_station->name, totals, done_pct);
}

/* ------------------------------------------------------- */
void DoControl(TaTable *table, void *par)
{
	Control *node;
	int i, *my_sid_number = (int *) par;
//	printf("DoControl\n");
	if (*my_sid_number == 0)
		node = &g_control;
	else
		node = (Control *) table->recent_data;
	if (ta_trace_level & 64) {
		printf("%d command %c\n", *my_sid_number, node->command);
		taDumpStationLinks(table->area);
	}
	if (node->command == 'I') {
		for (i = 0; i < table->area->n_stations; i++) {
			if (i == *my_sid_number)
				continue;
			g_outputs[i].input = '\0';
			taCallTable(&g_tab_outputs[i], MakeYourChoice, NULL);
		}
	}

}

/* ------------------------------------------------------- */
void OnErrorMessage(const void *data, TaULong msg_type,
        const struct TaTable *t)
{
	if (msg_type == taErrorDataMsgType()) {
                TaErrorData *edp = (TaErrorData *) data;
                printf("*** error:%d %s t:%u src:%d dst:%d\n", 
			edp->errcode, t->area->host_station->name,
			(int)(t->area->t_crtu/1000000), 
			t->src_oid, t->dst_oid);
		// finish on error
		// t == &g_tab_control
		if (t->dst_oid == t->src_oid) {
			printf("===%s src:%d finished by timeout\n", 
				t->area->host_station->name, 
				t->area->host_station->sid);
			t->area->t_expired = taGetTime() + 1000000;
		} else {
			g_played[t->dst_oid] = edp->errcode;
			DoPrintResults();
		}
        }
}

/* ------------------------------------------------------- */
#ifdef WITH_PYTHON
int initHostScript(TaStation *station)
{
	PyObject *pName, *pModule, *pFunc;
	if (	!strcmp("r", station->name)
	||	!strcmp("p", station->name)
	||	!strcmp("s", station->name)
	||	!strcmp("rand", station->name)
	||	!strcmp("brand", station->name))
			return 0;
	pName = PyString_FromString("rpsupport");
	pModule = PyImport_Import(pName);
	Py_DECREF(pName);
	pFunc = PyObject_GetAttrString(pModule, "load_bots");
//	printf("pModule=%x,pFunc=%x ver:%s\n", pModule, pFunc, Py_GetVersion());
	if (pFunc && PyCallable_Check(pFunc)) {
		PyObject *pArgs, *pValue;
		char buffer[64];
		int i;
		sprintf(buffer, "py/%s.py", station->name);
		pArgs = PyTuple_New(1);
		pValue = PyString_FromString(buffer);
		PyTuple_SetItem(pArgs, 0, pValue);
		// one instance for each partnew
		for (i = 0; i < station->area->n_stations; i++) {
			TaStation *st = (TaStation *)station->area->stations[i];
			if (!st)
				continue;
			if (st == station)
				g_moves[i] = NULL;
			else {
				pValue = PyObject_CallObject(pFunc, pArgs);
				g_moves[i] = PyObject_GetAttrString(pValue, 
					"get_move");
			}
		}

		Py_DECREF(pValue);
		Py_DECREF(pArgs);
		Py_DECREF(pFunc);
	}

	return 0;
}
#endif

