/*
 *                      ***** TA Pattern Search Server Implementation *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */
#define WITH_SRCH_INTERFACE
#include "srch.h"

TaTable g_tab_query_, g_tab_results_;
SearchQuery g_query_;
SearchResults g_results_;

EXTERN int taGetLibStationData (TaLibStationData *lsd);

void ProcessSearchQuery(TaTable *table, void *par);

/* ------------------------------------------------------- 
 *
 * Server Logic
 *
 * ------------------------------------------------------- */
int SearchServerInit(TaArea *area)
{
	int i;
	i = taParseToRTTI(g_SearchQuery, "Tests.SRCH", &g_rtti_SearchQuery, 0);
	i |= taParseToRTTI(g_FoundString, "Tests.SRCH", &g_rtti_FoundString, 0);
	i |= taParseToRTTI(g_SearchResults, "Tests.SRCH", 
		&g_rtti_SearchResults, 0);
	if (i < 0)
		return i;

	taNewTable(area, &g_tab_query_);
	g_tab_query_.mqos.tp_timeout = -1;
	taNewTable(area, &g_tab_results_);
	taAttachTable(&g_tab_query_, &g_rtti_SearchQuery, 0, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 1, NULL, TA_REALLOC_EXACT);
	taAttachTable(&g_tab_results_, &g_rtti_SearchResults, 0, 0, 
		TA_MSG_SRPS_ACCEPT_ANY, 1, NULL, TA_REALLOC_EXACT);

	g_results_.found = (FoundString *)
		taAllocWithRTTI(sizeof(FoundString), 0);
	taRepeatTable(&g_tab_query_, ProcessSearchQuery, NULL);
	return 0;
}

/* ------------------------------------------------------- */
void ProcessSearchQuery(TaTable *table, void *par)
{
	int i, j, r, poss[N_STRINGS_AT_ONCE*MAX_STRING_LENGTH];
	char *p;
	SearchQuery *sq;
	sq = (SearchQuery *) table->recent_data;
	i = j = 0;
	do {
		p = strstr(sq->text+i, sq->pattern);
		if (!p)
			r = -1;
		else
			r = (int)(p-sq->text);
		if (r >= 0) {
			poss[j] = r;
			j++;
			i = r+1;
		}
	} while (r > 0);
	g_results_.found = (FoundString *)
		taReallocWithRTTI(g_results_.found, j);
	for (i = 0; i < j; i++) {
		int dr;
		g_results_.found[i].pos = poss[i];
		for (dr = 0; dr > -MAX_STRING_LENGTH && poss[i]+dr > 0; dr--)
			if (sq->text[poss[i]+dr] == '\n') {
				dr++;
				break;
			}
		for (r = 0; r < MAX_STRING_LENGTH; r++) {
			char c = sq->text[poss[i]+r+dr];
			if (c == '\n')
				break;
			g_results_.found[i].result_str[r] = c;
		}
		g_results_.found[i].result_str[r] = 0;
	}
	taWriteTable(&g_tab_results_, &g_results_, 
		TA_TO_ADDR(table->message.src_sid), _True);
}

/* ------------------------------------------------------- 
 *
 * Dynamic Library Implementation
 *
 * ------------------------------------------------------- */

int taGetLibStationData (TaLibStationData *lsd)
{
        /*lsd->libarea.on_update = onUpdateSrch;
	lsd->client_data = lsd;*/
	return SearchServerInit(&lsd->libarea);
}

