/*
 *                      ***** TA Global Weather SOAP Server *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */
#define MAX_STATIONS	3
#include "weather.h"

TaTable g_tab_GetCitiesByCountry, g_tab_GetWeather;
GetCitiesByCountry g_var_GetCitiesByCountry;
GetWeather g_var_GetWeather;

int ServerInit(TaArea *area);
void ProcessGetCitiesByCountry(TaTable *table, void *par);
void ProcessGetWeather(TaTable *table, void *par);

/* ------------------------------------------------------- */
int main(int argc, char **argv)
{
	TaArea area;
	TaAreaQos aqos;
	TaLong aid = 22;
	TaStation stations[MAX_STATIONS];
	TaStationQos sqos;
	TaChannelQos cqos;
	int i, my_sid_number = 0, stations_amount = MAX_STATIONS;

	taInitialize();
	taDefaultAreaQos(&aqos);
	taDefaultStationQos(&sqos);
	taDefaultChannelQos(&cqos);
	aqos.tp_cycle = 100000;
	strcpy(sqos.hostname, "127.0.0.1");
	sqos.first_port = -80;
	sqos.prot_format = TA_PROT_FORMAT_SOAP;
	sqos.enable_mask = TA_ENABLE_TCP_SERVER;
	sqos.ping_mask = 0;
	sqos.send_strat = TA_SEND_STRAT_SERVER;
	cqos.assign_sids = _True;
	cqos.check_host_ip = _False;
	cqos.max_send_polls = 0;
	taInitArea(aid, &aqos, "gweather", &area);
	taReallocStations(MAX_STATIONS, &area);

	for (i = 0; argv[i]; i++) {
		if (!strcmp("-t", argv[i]) || 
		!strcmp("-trace_level", argv[i])) {
			if (argv[i+1])
				taTrace(ta_trace_level = atoi(argv[i+1]), NULL);
			++i;
		} else if (!strcmp("-port", argv[i])) {
			if (argv[i+1])
				sqos.first_port = -atoi(argv[i+1]);
			++i;
		} else if (!strcmp("-c", argv[i])) {
			if (argv[i+1])
				area.qos.tp_cycle = atoi(argv[i+1]);
			if (area.qos.tp_cycle < 1000)
				area.qos.tp_cycle = 1000;
			if (area.qos.tp_cycle > 10000000)
				area.qos.tp_cycle = 10000000;
			++i;
		}
	}
	for (i = 0; i < stations_amount; i++)
		taInitStation(i, &sqos, &cqos, NULL, &area, 
			&stations[i]);

	taSetStationHostBySid(my_sid_number, &area);
	taInitStationsData(&area);

	ServerInit(&area);

	taRun(&area, &my_sid_number);

	taCloseStationsData(&area);
	return 0;
}

/* ------------------------------------------------------- 
 *
 * Server Logic
 *
 * ------------------------------------------------------- */
int ServerInit(TaArea *area)
{
	int i;
	i = taParseToRTTI(g_CountryList, "globalweather.asmx", 
		&g_rtti_CountryList, 0);
	i |= taParseToRTTI(g_NDSet, "globalweather.asmx", 
		&g_rtti_NDSet, 0);
	i |= taParseToRTTI(g_GetCitiesByCountry, "globalweather.asmx", 
		&g_rtti_GetCitiesByCountry, 0);
	i |= taParseToRTTI(g_CWeather, "globalweather.asmx", 
		&g_rtti_CWeather, 0);
	i |= taParseToRTTI(g_GetWeather, "globalweather.asmx", 
		&g_rtti_GetWeather, 0);
	if (i < 0)
		return i;

	taNewTable(area, &g_tab_GetCitiesByCountry);
	taNewTable(area, &g_tab_GetWeather);
	taAttachTable(&g_tab_GetCitiesByCountry, &g_rtti_GetCitiesByCountry, 
		0, 0, TA_MSG_SOAP_MATCH_INTERFACE, 1, NULL, TA_REALLOC_EXACT);
	taAttachTable(&g_tab_GetWeather, &g_rtti_GetWeather, 
		0, 1, TA_MSG_SOAP_MATCH_INTERFACE, 1, NULL, TA_DONT_REALLOC);
	taRepeatTable(&g_tab_GetCitiesByCountry, ProcessGetCitiesByCountry, 
		NULL);
	taRepeatTable(&g_tab_GetWeather, ProcessGetWeather, NULL);
	return 0;
}

/* ------------------------------------------------------- */
void ProcessGetCitiesByCountry(TaTable *table, void *par)
{
	char buffer[256];
	GetCitiesByCountry *gcc;
	gcc = (GetCitiesByCountry *) table->recent_data;
	printf("ProcessGetCitiesByCountry CountryName:'%s'\n",
		taWChar2Char(buffer, gcc->CountryName, 256));
	g_var_GetCitiesByCountry.NewDataSet.Table = (CountryList *)
		taAllocWithRTTI(sizeof(CountryList), 3);
	taWCharCpy(g_var_GetCitiesByCountry.NewDataSet.Table[0].Country,
		gcc->CountryName, 128);
	taChar2WChar(g_var_GetCitiesByCountry.NewDataSet.Table[0].City,
		"Kitezh", 128);
	taWCharCpy(g_var_GetCitiesByCountry.NewDataSet.Table[1].Country,
		gcc->CountryName, 128);
	taChar2WChar(g_var_GetCitiesByCountry.NewDataSet.Table[1].City,
		"Tmutarakanh", 128);
	taWCharCpy(g_var_GetCitiesByCountry.NewDataSet.Table[2].Country,
		gcc->CountryName, 128);
	taChar2WChar(g_var_GetCitiesByCountry.NewDataSet.Table[2].City,
		"Vassuki", 128);

	taWriteTable(table, &g_var_GetCitiesByCountry, 
		TA_TO_ADDR(table->message.src_sid), _False);
}

/* ------------------------------------------------------- */
void ProcessGetWeather(TaTable *table, void *par)
{
	GetWeather *gw;
	char buf1[256], buf2[256];
	gw = (GetWeather *) table->recent_data;
	printf("ProcessGetWeather CityName:'%s' CountryName:'%s'\n",
		taWChar2Char(buf1, gw->CityName, 256), 
		taWChar2Char(buf1, gw->CountryName, 256));

	taChar2WChar(g_var_GetWeather.CurrentWeather.Location, 
		"Moscow, Russia", 128);
	taChar2WChar(g_var_GetWeather.CurrentWeather.Time, 
		"23/12/2012 00:00:00", 128);
	taChar2WChar(g_var_GetWeather.CurrentWeather.Wind, 
		"Very Strong", 128);
	taChar2WChar(g_var_GetWeather.CurrentWeather.Visibility, 
		"Low", 128);
	taChar2WChar(g_var_GetWeather.CurrentWeather.SkyConditions, 
		"Burning Sky", 128);
	g_var_GetWeather.CurrentWeather.Temperature = 38.5;
	g_var_GetWeather.CurrentWeather.DewPoint = 23.4;
	g_var_GetWeather.CurrentWeather.RelativeHumidity = 78;
	taChar2WChar(g_var_GetWeather.CurrentWeather.Pressure, 
		"Very High", 64);
	taChar2WChar(g_var_GetWeather.CurrentWeather.Status, 
		"Testing", 32);

	taWriteTable(table, &g_var_GetWeather, 
		TA_TO_ADDR(table->message.src_sid), _False);
}

