/*
 *                      ***** TA Pattern Search Server Implementation *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */
#define WITH_SRCH_INTERFACE
#include "fsrch.h"

TaTable g_tab_search[MAX_STATIONS];
SearchResults g_search[MAX_STATIONS];
TaLong g_request_numbers[MAX_STATIONS];

extern const char *pattern;
extern const char *file_name;
void ProcessSearchQuery(TaTable *table, void *par);

/* ------------------------------------------------------- 
 *
 * Server Logic
 *
 * ------------------------------------------------------- */
int SearchServerInit(TaArea *area)
{
	int i;
	i = taParseToRTTI(g_FoundString, "Tests.PSRCH", 
		&g_rtti_FoundString, 0);
	i |= taParseToRTTI(g_SearchResults, "Tests.PSRCH", 
		&g_rtti_SearchResults, 0);
	if (i < 0)
		return i;

	for (i = 0; i < area->n_stations; i++) {
		taNewTable(area, &g_tab_search[i]);
		taAttachTable(&g_tab_search[i], &g_rtti_SearchResults, 100+i, 
			100+i, TA_MSG_SRPS_MATCH_DST, 1,
			NULL, TA_REALLOC_EXACT);
		if (area->stations[i] != area->host_station)
			taPublishTable(&g_tab_search[i],_False, 
				ProcessSearchQuery, NULL);
		else {
			strncpy(g_search[i].pattern, pattern,
				MAX_STRING_LENGTH);
			g_search[i].pattern[MAX_STRING_LENGTH-1] = 0;
			strncpy(g_search[i].file_name, file_name, 
				MAX_STRING_LENGTH);
			g_search[i].file_name[MAX_STRING_LENGTH-1] = 0;
			taPublishTable(&g_tab_search[i],_True, 
				ProcessSearchQuery, NULL);
		}
	}
	return 0;
}

/* ------------------------------------------------------- */
#define POS_EOF	-1
#define NOT_FOUND -2
void ProcessSearchQuery(TaTable *table, void *par)
{
	int i, j = 0, r, nstr, pos, ind;
	TaBool permanent;
	char *p, text[MAX_STRING_LENGTH];
	SearchResults *sr;
	FILE *fd;
	FoundString found[N_STRINGS_ONCE];

	ind = table->dst_oid-100;
	permanent = (ind == (int)table->area->host_station->sid);
	if (permanent) {
		if (g_search[ind].end_pos < 0) {
			g_search[ind].start_pos = 0;
			g_search[ind].end_pos = 0;
		}
	} else if (table->got_message) {
		sr = (SearchResults *) table->recent_data;
		strncpy(g_search[ind].pattern, sr->pattern,
			MAX_STRING_LENGTH);
		g_search[ind].pattern[MAX_STRING_LENGTH-1] = 0;
		strncpy(g_search[ind].file_name, sr->file_name, 
			MAX_STRING_LENGTH);
		g_search[ind].file_name[MAX_STRING_LENGTH-1] = 0;
		g_search[ind].start_pos = 0;
	} else if (g_search[ind].end_pos < 0) {
		taClearTable(table);
		return;
	} else {
		g_search[ind].start_pos = g_search[ind].end_pos;
	}
	sr = &g_search[ind];
printf("process %d pattern:%s file_name:%s pos:%d %d\n", 
table->dst_oid, sr->pattern, sr->file_name, sr->start_pos, sr->end_pos);
	fd = fopen(sr->file_name, "r");
	if (!fd)
		sr->end_pos = NOT_FOUND;
	else {
		pos = sr->start_pos;
		sr->end_pos = POS_EOF;
		fseek(fd, pos, SEEK_SET);
		nstr = 0;
		while (fgets(text, MAX_STRING_LENGTH, fd)) {
			p = strstr(text, "\n");
			if (p) *p = 0;
			i = 0;
			do {
				p = strstr(text+i, sr->pattern);
				if (!p)
					r = -1;
				else
					r = (int)(p-text);
				if (r >= 0) {
					strncpy(found[j].result_str, text,
						MAX_STRING_LENGTH);
					found[j].pos = r;
					j++;
					if (j < N_STRINGS_ONCE)
						i = r+1;
					else
						r = -4;
				}
			} while (r > 0);
			if (r == -4 || nstr >= N_STRINGS_FROM_FILE) {
				sr->end_pos = i+pos;
				break;
			}
			r = 0;
			pos = (int)ftell(fd);
		}
		fclose(fd);
	}
	if (sr->found)
		sr->found = (FoundString *)
			taReallocWithRTTI(sr->found, j);
	else
		sr->found = (FoundString *)
                	taAllocWithRTTI(sizeof(FoundString), j);
	for (i = 0; i < j; i++)
		memcpy(&sr->found[i], &found[i],
			sizeof(FoundString));
	++g_request_numbers[ind];
	sr->request_number = g_request_numbers[ind];
printf("resulted in %d matched, pos:%d %d\n", j, sr->start_pos, sr->end_pos);
	taWriteTable(table, sr, TA_ADDR_ALL, _False);
}
