/*
 *                      ***** TA Pattern Search Server Implementation *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */
#define WITH_SRCH_INTERFACE
#include "fsrch.h"

TaTable g_tab_search[MAX_STATIONS];
SearchResults g_search[MAX_STATIONS];
TaLong g_request_numbers[MAX_STATIONS];

void ProcessSearchQuery(TaTable *table, void *par);

/* ------------------------------------------------------- 
 *
 * Server Logic
 *
 * ------------------------------------------------------- */
int SearchServerInit(TaArea *area)
{
	int i;
	i = taParseToRTTI(g_FoundString, "Tests.FSRCH", 
		&g_rtti_FoundString, 0);
	i |= taParseToRTTI(g_SearchResults, "Tests.FSRCH", 
		&g_rtti_SearchResults, 0);
	if (i < 0)
		return i;

	for (i = 0; i < area->n_stations; i++) {
		TaArea *tarea;
		if (area->stations[i] == area->host_station
		|| !area->stations[i]->ttd)
			continue;
		tarea = area->stations[i]->ttd->area;
		taNewTable(tarea, &g_tab_search[i]);
		g_tab_search[i].mqos.tp_timeout = -1;
		taAttachTable(&g_tab_search[i], &g_rtti_SearchResults, 
			(TaUShort) i, 0, TA_MSG_SRPS_MATCH_SID, 1,
			NULL, TA_REALLOC_EXACT);
		taRepeatTable(&g_tab_search[i], ProcessSearchQuery, NULL);
	}
	return 0;
}

/* ------------------------------------------------------- */
#define POS_EOF	-1
#define NOT_FOUND -2
void ProcessSearchQuery(TaTable *table, void *par)
{
	int i, j = 0, r, nstr, pos, ind;
	char *p, text[MAX_STRING_LENGTH];
	SearchResults *sr;
	FILE *fd;
	FoundString found[N_STRINGS_ONCE];

	sr = (SearchResults *) table->recent_data;
printf("process #%d pattern:%s file_name:%s pos:%d %d\n", 
table->message.src_sid, sr->pattern, sr->file_name, 
sr->start_pos, sr->end_pos);
	fd = fopen(sr->file_name, "r");
	if (!fd)
		sr->end_pos = NOT_FOUND;
	else {
		pos = sr->start_pos;
		sr->end_pos = POS_EOF;
		fseek(fd, pos, SEEK_SET);
		nstr = 0;
		while (fgets(text, MAX_STRING_LENGTH, fd)) {
			i = 0;
			p = strstr(text, "\n");
			if (p) *p = 0;
			do {
				p = strstr(text+i, sr->pattern);
				if (!p)
					r = -1;
				else
					r = (int)(p-text);
				if (r >= 0) {
					strncpy(found[j].result_str, text,
						MAX_STRING_LENGTH);
					found[j].pos = r;
					j++;
					if (j < N_STRINGS_ONCE)
						i = r+1;
					else
						r = -4;
				}
			} while (r > 0);
			if (r == -4 || nstr >= N_STRINGS_FROM_FILE) {
				sr->end_pos = i+pos;
				break;
			}
			r = 0;
			pos = (int)ftell(fd);
		}
		fclose(fd);
	}
	if (sr->found)
		sr->found = (FoundString *)
			taReallocWithRTTI(sr->found, j);
	else
		sr->found = (FoundString *)
                	taAllocWithRTTI(sizeof(FoundString), j);
	for (i = 0; i < j; i++)
		memcpy(&sr->found[i], &found[i],
			sizeof(FoundString));
	++g_request_numbers[table->message.src_sid];
	sr->request_number = g_request_numbers[table->message.src_sid];
printf("resulted in %d matched, pos:%d %d\n", j, sr->start_pos, sr->end_pos);
	taWriteTable(table, sr, TA_TO_ADDR(table->message.src_sid), _True);
}
