/*
 *                          ***** TA *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1 Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following disclaimer.
 *
 * 2 Redistributions in binary form must reproduce the
 *   above copyright notice, this list of conditions
 *   and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * 3 Neither the name of the General Energy Technologies
 *   nor the names of its contributors may be used to
 *   endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 *
 */

#ifndef __TA_H__
#define __TA_H__

#define TA_MAJOR_VERSION 1
#define TA_MINOR_VERSION 0

#include <stdio.h>
#include <stdlib.h>
#include <ta_types.h>
#ifndef _WINDOWS
#include <sys/socket.h>
#include <netinet/in.h>
#include <pthread.h>
#else
#pragma comment(lib, "wininet.lib")
#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")

#include <winsock2.h>
#include <ws2tcpip.h>
#endif

/* ------------------------------------------------------- */
#define TA_N_CHANNELS		8
#define TA_CHANNEL_UDP_MCAST	0
#define TA_CHANNEL_UDP		1
#define TA_CHANNEL_TCP_SERVER	2
#define TA_CHANNEL_TCP_CLIENT	3
#define TA_CHANNEL_UNIX_DOMAIN	4
#define TA_CHANNEL_ICMP		5
#define TA_CHANNEL_LIB		16
#define TA_CHANNEL_SHM_MCAST	17
#define TA_CHANNEL_SHM		18
#define TA_CHANNEL_MSGQ		19
#define TA_ENABLE_UDP_MCAST	1
#define TA_ENABLE_UDP		2
#define TA_ENABLE_TCP_SERVER	4
#define TA_ENABLE_TCP_CLIENT	8
#define TA_ENABLE_UNIX_DOMAIN	16
#define TA_ENABLE_ICMP		32
#define TA_ENABLE_LIB		0x10000
#define TA_ENABLE_SHM_MCAST	0x20000
#define TA_ENABLE_SHM		0x40000
#define TA_ENABLE_MSGQ		0x80000
#define TA_ENABLE_ALL		0xFFFFFFFF
#define TA_TYPE_MASK(ct)	(1<<ct)

#define TA_FLAGS_MCAST		1
#define TA_FLAGS_DGRAM		2
#define TA_FLAGS_STREAM		4
#define TA_FLAGS_LISTEN		8
#define TA_FLAGS_MEMORY		16
#define TA_FLAGS_MESSAGE	32
#define TA_FLAGS_UNIX_DOMAIN	64
#define TA_FLAGS_LIB		128
#define TA_FLAGS_ICMP		256

#define TA_MAX_MASK_ADDRS	32
EXTERN const TaShort ta_addr_all;
#define TA_ADDR_ALL		-1
#define TA_TO_ADDR(addr)	1<<(addr)
#define TA_SIDS_ALL		(TaShort)-1
#define TA_N_SIDS_DEFAULT	-1

#define TA_MAX_SD_AMOUNT	300
#define TA_MAX_POLL_CHANNELS	300

#define TA_TRANS_SEND	1
#define TA_TRANS_RECV	2
#define TA_TRANS_INDIRECT	4

/* ------------------------------------------------------- */
typedef struct {
	TaShort n_repeats;
	TaULong mem_bufsize;	// mem allocated to mem_bufsize*2
	TaBool is_less_be;	// less than best effort
	TaUShort max_send_polls;
	TaBool tcp_nodelay;	// setsockopt(sd, SOL_TCP, TCP_NODELAY, 1)
	TaBool so_keepalive;	// setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, 1)
	TaShort so_reuseaddr;	// -1 never, 0 if bind, 1 always
	TaLongLong tp_connect_timeout;
	TaLongLong tp_disconnect_timeout;
	TaBool shm_rm_onexit;
	TaBool ping_set_err;
	TaBool assign_sids;
	TaBool check_host_ip;
} TaChannelQos;

/* ------------------------------------------------------- */
struct TaLibStationData;
typedef enum {TA_LINK_NONE=0, TA_LINK_CLOSED=1, TA_LINK_ERR=2,
	TA_LINK_INIT=3, TA_LINK_WAIT=4, TA_LINK_LISTEN=5, TA_LINK_OK=6} TaLink;
typedef struct {
	TaChannelQos qos;
	int cid;
	int ctype;		// TA_CHANNEL_*
	int flags;
	int sd;			// POSIX socked descriptor
	char hostname[TA_NAME_LENGTH];
	int port;
	int partner_cid;
	TaBool sid_acked;
	unsigned int s_addr0;
	struct sockaddr_in sa_from;
	struct sockaddr_in sa_msg;
	socklen_t sa_len;
	socklen_t sa_msg_len;

	int send_id;
	int recv_id;
	TaULong recv_offset;
	TaUShort recv_mid;
	char *send_buffer;
	char *recv_buffer;
	struct TaLibStationData *lib;
	// current status data
	TaBool is_enabled;
	TaLink link_stat;
	TaOctet trans_flags;
	// dynamic aux data
	int poll_events;
	int recv_rc;
	int send_rc;
	int complete_recv_rc;
	int complete_send_rc;
	int recv_errno;
	int send_errno;
	TaShort n_send_polls;
	TaLongLong t_csend;
	TaLongLong t_crecv;
	TaBool is_receiving;
	TaBool is_receiving2;
} TaChannel;

/* ------------------------------------------------------- */
typedef enum {TA_SEND_STRAT_SERVER = 0, TA_SEND_STRAT_CLIENT = 1, 
	TA_SEND_STRAT_MIN_ZVAL = 2, TA_SEND_STRAT_CR_SW = 3} TaSendStrat;
typedef enum {TA_PROT_FORMAT_SRPS = 0, TA_PROT_FORMAT_SOAP = 1} TaProtFormat;
#define CONNECT_READY_NOCHECK _False
#define CONNECT_WHEN_REQUEST _True
#define READY_WHEN_REQUEST (TaOctet)2
typedef struct {
	char hostname[TA_NAME_LENGTH];
	int first_port;
	int delta_port;
	TaULong enable_mask;	// TA_ENABLE* 4 bits
	TaULong ping_mask;	// TA_ENABLE* 4 bits
	TaLongLong tp_cycle;
	TaLongLong tp_discovery_timeout;
	TaLongLong tp_ping_timeout;	// normally 2 times less discovery
	TaSendStrat send_strat;
	char group_addr[TA_NAME_LENGTH];
	TaProtFormat prot_format;
	TaOctet connect_on_request;
} TaStationQos;

struct TaArea;
struct TaStation;
struct TaTable;
struct TaThreadData;

typedef int (TaUpdateStationData)(struct TaStation *station, 
	TaLongLong t_updating, void *client_data);
typedef int (TaInitStationData)(const struct TaArea *area, 
	struct TaStation *station);
typedef int (TaUpdateAreaData)(struct TaArea *area, TaLongLong t_updating,
	void *client_data);

/* ------------------------------------------------------- */
typedef struct TaStation {
	TaShort sid;		// > 0 unique station identifier
	char name[TA_NAME_LENGTH];
	TaStationQos qos;
	TaChannel channels[TA_N_CHANNELS];	// max fixed, no more
	int n_channels;
	struct TaArea *area;
	TaShort zval;
	// runtime control
	TaLongLong t_crt0;
	TaLongLong n_updates;
	TaBool is_initialized;
	TaUpdateStationData *on_update;
	TaInitStationData *on_init;
	int lib_cid;
	// online monitoring
	TaLongLong t_send;
	TaLongLong t_recv;
	TaLongLong t_mid_send;
	TaLongLong t_mid_recv;
	TaUShort last_mid;	// is it correct?
	TaBool is_online;
	TaBool is_user_locked;
	TaBool is_disconnected;
	TaUShort recv_bits;
	int n_ready_for_sending;
	int n_available_for_sending;
	struct TaThreadData *ttd;
} TaStation;

/* ------------------------------------------------------- */
typedef enum {TA_PROT_ANY=0, TA_PROT_RELIABLE=1, TA_PROT_UNRELIABLE=2,
	TA_PROT_MCAST=3} TaProtReliability;
typedef enum {TA_ACK_NONE=0, TA_ACK_BY_RESPONSE=1, TA_ACK_BY_SUBSCRIPTION=2} 
	TaAcknowledgement;
typedef struct {
	TaProtReliability reliability;	// select reliable protocol
	TaBool is_redundant;		// message is sent n_refs=2 times
	TaULong send_mask;		// sending to addrs in mask
	TaLongLong tp_timeout;
	TaLongLong tp_send_timeout;
	TaUShort n_send_attempts;
	TaBool send_once;
	TaAcknowledgement ack;
	TaBool all_channels_ready;
	TaBool all_sids_ready;
	TaLong dim_chunk;
} TaMessageQos;

/* ------------------------------------------------------- */
typedef struct {
	TaUShort mid;
	TaUShort mid_resp;
	TaShort src_sid;		// sender station Id
	TaShort dst_sid;		// sender station Id
	TaULong msg_type;
	TaUShort dst_oid;
	TaUShort iface_code;
	TaMessageQos qos;
	char *buffer;
	TaUShort basic_length;
	int buffer_length;
	TaLongLong t_set;
	TaLongLong t_got;
	TaShort sendto_sids[TA_MAX_MASK_ADDRS];	// more 8 is not effective
	int n_sendto_sids;
	int n_refs;	// discarded when n_refs==0
	TaULong mask;
} TaMessage;
#define taMessageBody(message) (message.buffer+sizeof(TaSRPSHeader))
#define taPMessageBody(message) (message->buffer+sizeof(TaSRPSHeader))
#define taMessageLength(message) (message.buffer_length-sizeof(TaSRPSHeader))
#define taPMessageLength(message) (message->buffer_length-sizeof(TaSRPSHeader))

/* ------------------------------------------------------- */
typedef struct {
	TaLongLong tp_cycle;
	TaShort justify_frame;	// justify_frame: [1-10] or -1 if host_sid+1
	int messages_array_size;
	int send_buffer_size;
	int recv_buffer_size;
	TaLongLong tp_phase_correct;
	TaULong max_poll_fails;
	TaULong enable_mask;
	int soap_buffer_size;
} TaAreaQos;

/* ------------------------------------------------------- */
typedef struct TaArea {
	TaLong aid;		// > 0 unique area identifier
	char name[TA_NAME_LENGTH];
	TaAreaQos qos;
	int n_stations;
	int max_stations;
	TaStation **stations;
	TaStation *host_station;
	TaStation *send_station;
	int n_ctypes;
	int ctypes[TA_N_CHANNELS];
	// runtime control
	int finish_flag;	// nonzero means return code
	TaLongLong t_crt;
	TaLongLong t_crt0;
	TaLongLong t_crtu;
	TaLongLong tp_cdt;
	TaLongLong tp_deadzone;
	TaLongLong t_expired;
	TaUpdateAreaData *on_update;
	TaULong n_skips;
	TaULong n_poll_fails;
	// message data
	char *send_buffer;
	char *recv_buffer;
	int n_send_messages;
	int n_send_used;
	TaMessage *send_messages;
	TaUShort current_mid;
	TaLongLong t_last_msg;
	int n_recv_messages;
	int n_recv_used;
	TaMessage *recv_messages;
	// phase correction and second
	TaLongLong t_phase_last;
	TaUShort second_phase;
	TaUShort phase_recv;
	TaUShort n_phase_update;
	TaShort phase_correction;
	TaBool with_second;
	//
	struct TaTable *tablist;
	int n_msgs_not_found;
	TaLongLong t_waiting_expired;
} TaArea;

/* i----- channel protocol description ----- */
#define TA_N_PROTOCOLS  32
typedef int (TaInitProc)(TaBool is_datahost, TaBool is_send_station,
	const TaStation *station, TaChannel *channel);
typedef void (TaCloseProc)(TaChannel *channel);
typedef int (TaConnectProc)(TaArea *area, TaStation *dstation, int cid);
typedef int (TaSendProc)(const TaArea *area, const char *buffer,
	int buffer_length, TaChannel *channel);
typedef int (TaRecvProc)(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);
typedef const char * (TaReadableErrno)(int ierrno);
typedef struct {
	char name[32];
	int ctype;
	int flags;
	TaInitProc *init_proc;	// taInitChannelData
	TaCloseProc *close_proc;	// taCloseChannelData
	TaCloseProc *shutdown_proc;	// taShutdownChannelData
	TaConnectProc *connect_proc;	// taTryConnectChannel
	TaSendProc *send_proc;	// taSendByChannel
	TaRecvProc *recv_proc;	// taRecvByChannel
	TaReadableErrno *readable_errno;	// taReadableChannelErrno
} TaChannelProtocol;
typedef struct {
	int wyear, wmonth, wday, hh, mm, ss;
} TaTmStruct;

/* ----- Library Station Interface ----- */
typedef struct TaLibStationData {
	void *plib;
	TaStation *station;
	int cid;
	void *client_data;
	TaArea *area;
	TaArea libarea;
	TaStation *pstations[2];
	TaStation libstations[2];
} TaLibStationData;

typedef int (TaGetLibStationDataProc) (TaLibStationData *lsd);

#ifdef _WINDOWS
#define pthread_t DWORD
#define pthread_mutex_t CRITICAL_SECTION
#define pthread_mutex_lock(m) EnterCriticalSection(m)
#define pthread_mutex_unlock(m) LeaveCriticalSection(m)
#endif
typedef struct TaThreadData {
	void *client_data;
	TaStation *station;
	TaArea *area;
	void *handle;
	TaArea threadarea;
	pthread_t f_thread;
	pthread_mutex_t send_mutex;
	pthread_mutex_t recv_mutex;
} TaThreadData;

/* ----- Init Routines ----- */
EXTERN int taInitialize();
EXTERN int taInitMemProts();
EXTERN int taInitLibProts();
EXTERN int taInitPingProts();
EXTERN int taTrace(TaULong trace_level, const char *filename);
EXTERN void taPuts(TaChar *str, TaBool newline);
EXTERN void taLocaltime(TaLongLong t_crt, TaTmStruct *tloc); 
EXTERN void taMaketime(const TaTmStruct *tloc, TaLongLong *t_crt);
EXTERN double taRand();
EXTERN TaULong taCrc32(const char *buf, TaULong crc0, TaULong nbits);
EXTERN char **taArgsFromConfig(const char *cfg_name);
EXTERN void taClearArgs(char **args);
EXTERN TaChar *taWChar2Char(TaChar *buffer, const TaWChar *wstr, int length);
EXTERN TaWChar *taChar2WChar(TaWChar *wbuffer, const TaChar *str, int length);
EXTERN TaWChar *taWCharCpy(TaWChar *wbuffer, const TaWChar *src, int length);

/* ----- Channel Routines ----- */
EXTERN int taAddChannelProtocol(const TaChannelProtocol *protocol);
EXTERN const TaChannelProtocol *taChannelProtocol(int ctype);
EXTERN void taDefaultChannelQos(TaChannelQos *qos);
EXTERN int taInitChannel(int ctype, const TaChannelQos *qos,
	TaChannel *channel);
EXTERN int taInitChannelData(TaBool is_datahost, TaBool is_send_station,
	const TaStation *station, TaChannel *channel);
EXTERN void taCloseChannelData(TaChannel *channel);
EXTERN void taShutdownChannelData(TaChannel *channel);
EXTERN int taPollChannels(TaArea *area);
EXTERN int taSendByChannel(const TaArea *area, const char *buffer, 
	int buffer_length, TaChannel *channel);
EXTERN int taTryAcceptChannels(TaArea *area, int cid);
EXTERN int taTryConnectChannel(TaArea *area, TaStation *dstation, int cid);
EXTERN int taRecvByChannel(TaArea *area, char *buffer, int buffer_length, 
	TaChannel *channel);
EXTERN TaBool taCheckChannels(TaChannel *msg_channel, 
	TaChannel *sid_channel);
EXTERN const char * taReadableChannelErrno(int ierrno, TaChannel *channel);
EXTERN void taClearChannelRc(TaArea *area, int sid, int cid);
//
/* ----- Station Routines ----- */
EXTERN void taDefaultStationQos(TaStationQos *qos);
EXTERN int taInitStation(TaShort sid, const TaStationQos *qos, 
	const TaChannelQos *cqos, const char *name,
	TaArea *area, TaStation *station);
EXTERN int taInitStationData(const TaArea *area, TaStation *station);
EXTERN int taReinitStation(const TaStationQos *qos, TaStation *station);
EXTERN void taCloseStationData(TaStation *station);
EXTERN int taUpdateStationData(const TaArea *area, TaStation *station,
	void *client_data);
EXTERN void taDumpStationLinks(const TaArea *area);

/* ----- Area Routines ----- */
EXTERN void taDefaultAreaQos(TaAreaQos *qos);
EXTERN int taInitArea(TaLong aid, const TaAreaQos *qos, const char *name, 
	TaArea *area);
EXTERN int taInitLibArea(const TaArea *area, TaLibStationData *lsd,
	TaArea *libarea);
EXTERN int taInitThreadData(TaAreaQos *qos, TaStation *station);
EXTERN int taInitThreadsData(TaAreaQos *qos, TaArea *area);
EXTERN void taInitAreaBuffers(TaArea *area);
EXTERN int taInitAreaCtypes(TaULong enable_mask, TaArea *area);
EXTERN void taReallocStations(int max_stations, TaArea *area);
EXTERN TaStation *taSetStationHostByName(const char *name, TaArea *area);
EXTERN TaStation *taSetStationHostBySid(TaShort sid, TaArea *area);
EXTERN void taRun(TaArea *area, void *client_data);
EXTERN int taUpdateData(TaArea *area, void *client_data);
EXTERN int taSecondUpdateData(TaArea *area);
EXTERN int taUpdateLibData(TaLibStationData *lsd);
EXTERN int taThreadStart(TaThreadData *ttd, void *client_data);
EXTERN int taThreadsStart(TaArea *area, void *client_data);
EXTERN int taUpdateThreadData(TaThreadData *ttd, void *client_data);
EXTERN int taInitStationsData(TaArea *area);
EXTERN void taCloseStationsData(TaArea *area);
EXTERN int taProcessPhaseMessages(TaArea *area);
EXTERN void taPhaseRecvCorrect(TaArea *area, TaUShort phase);
EXTERN void taPhaseCorrect(TaArea *area, TaUShort crt_phase);

/* ----- Message Routines ----- */
EXTERN void taDefaultMessageQos(TaMessageQos *qos);
EXTERN void taReadyForSendMessage(TaArea *area, const TaShort *sendto_sids, 
	int n_sendto_sids, int buf_size, const TaMessageQos *qos, 
	TaBool *is_ready);
EXTERN int taSetMessage(TaArea *area, const void *buffer, int buffer_length,
	TaULong msg_type, TaUShort dst_oid, TaUShort iface_code,
	const TaShort *sendto_sids, int n_sendto_sids, 
	TaUShort mid_resp, const TaMessageQos *qos);
EXTERN int taAddToMessage(TaArea *area, const void *buffer, int buffer_length,
	TaULong msg_type, TaUShort dst_oid, TaUShort iface_code);
EXTERN int taSendMessages(TaArea *area, TaBool is_second);
EXTERN int taSendMessage(TaArea *area, TaMessage *message);
EXTERN int taSendPingMessage(TaArea *area, TaStation *station);
EXTERN int taSendPhaseMessage(TaArea *area, TaStation *station, TaUShort phase);
EXTERN int taSendPingMessages(TaArea *area);
EXTERN int taSendPhaseMessages(TaArea *area, TaUShort phase_msg);
EXTERN int taGetMessage(TaArea *area, int index, TaMessage *message);
EXTERN int taRecvMessages(TaArea *area, TaBool is_second);
EXTERN int taRecvChannelMessages(TaArea *area, TaChannel *channel, TaShort sid);

#include <ta_marsh.h>

#endif
