/*
 *                          ***** TA *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#ifndef __TA_MARSH_H__
#define __TA_MARSH_H__

/* basic types identifiers */
#define TaNo_TYPE	0
#define TaBool_TYPE	1
#define TaChar_TYPE	2
#define TaOctet_TYPE	3
#define TaWChar_TYPE	4
#define TaShort_TYPE	5
#define TaLong_TYPE	6
#define TaLongLong_TYPE	7
#define TaFloat_TYPE	8
#define TaDouble_TYPE	9
#define TaLongDouble_TYPE	10
#define TaArr_TYPE	11
#define TaRec_TYPE	12
#define TaPtr_TYPE	13
#define TaProc_TYPE	14

#define TaUser_TYPE	20

/* stuctures marchalling */
#ifndef TA_EXTERN
#define TA_STRUCT(struct_name, struct_descr) typedef struct { struct_descr } struct_name; const char *g_##struct_name = "typedef struct { " #struct_descr " } " #struct_name; TaRTTI g_rtti_##struct_name;
#define TA_RTTI(struct_name, iface_name) taParseToRTTI(g_##struct_name, iface_name, &g_rtti_##struct_name)
#else
#define TA_STRUCT(struct_name, struct_descr) typedef struct { struct_descr } struct_name;
#endif

/*
TA_STRUCT(NodeStatus,
typedef struct {
        TaLongLong t_changed;
        TaLongLong t_display;
        TaULongLong counters_mask;
        int counter;
        int host_counter;
} NodeStatus)

TA_RTTI(NodeStatus, "Tests.INodex");
=> taParseToRTTI(g_NodeStatus, "Tests.INodex", g_rtti_NodeStatus);
*/

/* ----- RTTI Structures ----- */
typedef struct TaRTTItem {
	struct TaRTTItem *contains;
	struct TaRTTItem *next;
	char item_name[TA_NAME_LENGTH];		// t_changed, for example
	char type_name[TA_NAME_LENGTH];		// TaLongLong
	int dim, item_type_id, item_type_size;
	int offset, size, size_packed, number;
	TaLong names_crc;	/* for pointers */
	TaShort types_crc;	/* for pointers */
} TaRTTItem;

typedef struct {
	TaRTTItem mainitem;
	char iface_name[TA_NAME_LENGTH];	// Tests.INodex
	int max_alignment, pack_alignment, 
		sizeof_struct, sizeof_packed, columns;
	TaLong names_crc;
	TaShort types_crc;
} TaRTTI;

typedef struct {
	TaLong size, length, elemsize, dim;
} TaRTTIadata;

/* ----- RTTI Routines ----- */
#ifdef TA_ZP_VALUE
// set this def in case of /Zp1../Zp16 compiler options
#define TA_PACK_ALIGN	TA_ZP_VALUE
#else
#ifdef __osf
#define TA_PACK_ALIGN	8
#else
#ifdef _WINDOWS
#define TA_PACK_ALIGN	8
#else
#define TA_PACK_ALIGN	4
#endif
#endif
#endif
EXTERN int taParseToRTTI(const char *struct_descr, const char *iface_name,
	TaRTTI *rtti, int pack_alignment);
// nonzero pack_alignment if you set pragma pack for specific struct
EXTERN void taClearRTTI(TaRTTI *rtti);

EXTERN void *taAllocWithRTTI(TaLong itemsize, TaLong dim);
EXTERN void taFreeWithRTTI(void *ptr);
EXTERN void *taReallocWithRTTI(void *ptr, TaLong newdim);
EXTERN TaLong taLengthWithRTTI(void *ptr);
EXTERN TaLong taLengthAllocatedRTTI(void *ptr);
EXTERN TaChar *taDupStringWithRTTI(const TaChar *src);
EXTERN TaWChar *taDupWStringWithRTTI(const TaWChar *src);

/* ----- Table Structures ----- */
typedef void (TaOnDataHandler)(const void *data, TaULong msg_type, 
	const struct TaTable *table);

#define TA_MAX_KEEP_HISTORY	10
#define TA_MAX_PREALLOCED	5
struct TaTable;
typedef void (TaTableCommand)(struct TaTable *table, void *par);
#define TA_ELEM_NOT_USED	0
#define TA_ELEM_PREALLOCED	1
#define TA_ELEM_MALLOCED	2

#define TA_MAX_MID		32000
#define TA_PERMANENT_MID	32001
#define TA_SOAP_MID		32010
#define TA_SOAP_11_MID		32011
#define TA_SOAP_12_MID		32012
#define TA_SOAP_GET_MID		32013
#define TA_SOAP_POST_MID	32014
#define TA_SOAP_11_CLIENT       32021
#define TA_SOAP_12_CLIENT       32022
#define TA_SOAP_GET_CLIENT      32023
#define TA_SOAP_POST_CLIENT     32024
#define SOAP_MSG(mid)	(mid>=TA_SOAP_MID && mid<=TA_SOAP_POST_CLIENT)

#define TA_MSG_SRPS_ACCEPT_ANY	0
#define TA_MSG_SRPS_MATCH_DST	1
#define TA_MSG_SRPS_MATCH_SID	2
#define TA_MSG_SOAP_MATCH_INTERFACE	32020
//define TA_SOAP_11_CLIENT		32021
//define TA_SOAP_12_CLIENT		32022
//define TA_SOAP_GET_CLIENT		32023
//define TA_SOAP_POST_CLIENT		32024

typedef struct TaElem {
	struct TaElem *next;
	int op;
	struct TaTable *table;
	int row;
	TaRTTItem *item;
	TaTableCommand *tcmd;
	void *par;
	TaShort write_sids[TA_MAX_MASK_ADDRS];
	int n_write_sids;
	TaUShort current_mid;
	TaShort waiting;
	TaLongLong t_waiting_expired;
	TaOctet used;
} TaElem;

typedef struct TaTable {
	int rows, columns, res;
	TaMessageQos mqos;
	const TaRTTI *rtti;
	char *recent_data;
	TaBool changed, got_message;
	// stdTable
	struct TaTable *next;
	TaArea *area;
	TaULong msg_type, keep_history;
	TaUShort dst_oid, src_oid, iface_code, src_mid, mid_resp;
	TaUShort match_mode, realloc_mode;
	TaMessage message;
	char *history_data[TA_MAX_KEEP_HISTORY];
	TaOnDataHandler *on_data;
	char *buffer, *dyn_buffer;
	TaShort sendto_sids[TA_MAX_MASK_ADDRS];
	int n_sendto_sids;
	TaUShort send_attempts_counter;
	int buffer_size, dyn_buffer_size, dyn_buffer_len;
	TaElem *queue, *tail, prealloced[TA_MAX_PREALLOCED];
} TaTable;

typedef struct {
	int errcode;
	int auxdata;
} TaErrorData;

EXTERN void taNewTable(TaArea *area, TaTable *table);
EXTERN int taAttachTable(TaTable *table, const TaRTTI *rtti, TaUShort dst_oid, 
	TaUShort src_oid, TaUShort match_mode, TaULong keep_history, TaOnDataHandler *on_data,
	TaUShort realloc_mode);
EXTERN int taSendtoSids(TaTable *table, TaShort *sendto_sids, int n_sendto_sids);
EXTERN int taUpdateAreaTables(TaArea *area);
EXTERN TaULong taErrorDataMsgType();

#define TA_NO_WAIT		0
#define TA_WAIT_WRITE_PERMISSION	-1
#define TA_WAIT_UNTIL_TIMEOUT	-2
#define TA_WAIT_ANY_MESSAGE	-3
#define TA_WAIT_FOR_RESPONSE	-4
#define TA_REPEAT_FOREACH_RESPONSE	-5
#define TA_REQUEST	-6
#define TA_PUBLISH	-7
#define TA_SUBSCRIBE	-8

#define TA_DONT_REALLOC		0
#define TA_REALLOC_LARGEST	1
#define TA_REALLOC_EXACT	2

EXTERN TaBool taAvailable(TaTable *t);
EXTERN void taCallTable(TaTable *t, TaTableCommand *command,
	void *par);
EXTERN void taWriteTable(TaTable *t, void *data, 
	TaULong sids, TaBool wait_write_permission);
EXTERN void taWaitTable(TaTable *t, TaLongLong t_waiting_expired);
EXTERN void taRepeatTable(TaTable *t, TaTableCommand *command, void *par);
EXTERN void taClearTable(TaTable *t);
EXTERN void taRequestTable(TaTable *t, void *data, TaULong sids, 
	TaBool wait_write_permission, TaTableCommand *command, void *par);
EXTERN void taPublishTable(TaTable *t, TaBool *permanent, TaTableCommand *command,
	void *par);
EXTERN void taSubscribeTable(TaTable *t, TaBool *controlled, TaTableCommand *command,
	void *par);
EXTERN void taUnsubscribeTable(TaTable *t, TaULong sids);

// common.c byte ofdering routines
EXTERN TaShort taH2N16(TaShort iv);
EXTERN TaShort taN2H16(TaShort iv);
EXTERN TaLong taH2N32(TaLong iv);
EXTERN TaFloat taH2Nr32(TaFloat fv);
EXTERN TaLong taN2H32(TaLong iv);
EXTERN TaFloat taN2Hr32(TaFloat fv);
EXTERN TaLongLong taH2N64(TaLongLong iv);
EXTERN TaDouble taH2Nr64(TaDouble fv);
EXTERN TaLongLong taN2H64(TaLongLong iv);
EXTERN TaDouble taN2Hr64(TaDouble fv);

// parser routines
#define TA_PARSE_COUPLE_BEGIN	1
#define TA_PARSE_COUPLE_END	2
#define TA_PARSE_COUPLE_BOTH	3
#define TA_PARSE_COUPLE_DIM	4
#define TA_MAX_STACK 32
typedef struct {
	char name[TA_NAME_LENGTH];
	void *user_data;
	int index, dim;
	char *recent_data;
} TaNamesStack;
typedef int (TaOnParseCouple)(int parse_flag, TaNamesStack *nstack, 
	int nlevel, const char *value, void *client_data);
typedef struct {
	char *message_buffer;
	TaUShort message_mid; 
	const char *type_name; 
	char *table_buffer;
	int soap_buffer_size;
	TaOnParseCouple *on_read_soap;
	void *client_data;
} TaOnParseSoap;
EXTERN int taSimpleXmlParser(const char *str, TaOnParseCouple *parser,
	void *client_data);
EXTERN int taSimpleLineParser(const char *str, TaOnParseCouple *parser,
	void *client_data);
EXTERN int taOnReadSoap(int parse_flag, TaNamesStack *nstack, int nlevel, 
	const char *value, void *client_data);
EXTERN int taReadSoap(TaOnParseSoap *ops);
EXTERN int taWriteSoapItem(const TaRTTItem *ritem, TaTable *t, const char *data,
	char *item_addr, int buffer_size);
EXTERN int taUnquoteBuffer(char *buffer, int buffer_length, int content_length);
#endif
