/*
 *                          ***** TA *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 */

#ifndef __TA_TYPES_H__
#define __TA_TYPES_H__

/* basic definitions */
#ifdef _WINDOWS
#define DLLEXPORT __declspec(dllexport)
#define STDCALL __stdcall
#define CALLBACK
#else
#define DLLEXPORT
#define STDCALL
#define CALLBACK
#endif
#ifdef __cplusplus
#define EXTERN extern "C" DLLEXPORT
#else
#define EXTERN DLLEXPORT
#endif

#define TA_NAME_LENGTH 64

/* basic types */
typedef unsigned char TaBool;
#define _False (TaOctet)0
#define _True (TaOctet)1
typedef char TaChar;
typedef unsigned char TaOctet;
//#ifdef _WINDOWS
// short && for unicode ??
typedef short TaWChar;
//typedef wchar_t TaWChar;
//#else
//typedef wchar_t TaWChar;
//#endif
typedef short TaShort;
typedef unsigned short TaUShort;
typedef long TaLong;
typedef unsigned long TaULong;
typedef long long TaLongLong;
typedef unsigned long long TaULongLong;
typedef float TaFloat;
typedef double TaDouble;
typedef double TaLongDouble;

#ifdef TA_OLD_TIME_FORMAT
#define TA_SEC(lsec) (TaULong)(lsec/1000000)
#define TA_MCS(lsec) (TaLong)(lsec%1000000)
#define TA_TP(lval) lval
#else
EXTERN TaULong TA_SEC(TaLongLong t);
#define TA_MCS(lsec) (TaLong)(lsec%10000000/10)
#define TA_TP(lval) (lval*10)
#endif

/* error codes */
#define TA_ERR_ARGUMENT		-1
#define TA_ERR_TIMEOUT		-2
#define TA_ERR_OUT_OF_MEMORY	-3
#define TA_NOT_FOUND		-4
#define TA_ERR_WRONG_DATA	-5
#define TA_ERR_TRANSPORT	-6
#define TA_ERR_LINK		-7
#define TA_PARSING		-8
#define TA_MSG_CHECK_ERR	-9
#define TA_MSG_INCOMPATIBLE	-10
#define TA_SEND_TIMEOUT		-11
#define TA_RESPONSE_TIMEOUT	-12
#define TA_RESPONSE_RESEND	-13
#define TA_RESPONSE_MISMATCH	-14
#define TA_TO_BE_CONTINUED	-15

/* trace level */
EXTERN TaULong ta_trace_level;
EXTERN FILE *ta_file;
#define TA_TRAL_RUNTIME		1
#define TA_TRAL_TRANSPORT	2
#define TA_TRAL_POLLING		4
#define TA_TRAL_MESSAGE		8
#define TA_TRAL_MESSAGE_DATA	16
#define TA_TRAL_ERRORS		32

/* common routines */
EXTERN int taSleep(TaLongLong tp);
EXTERN TaLongLong taGetTime();

#define TA_MAX_N_FRAMES		10
/*#define TA_MSG_TYPE_PING	70
#define TA_MSG_TYPE_PHASE1	71
#define TA_MSG_TYPE_PHASE_MAX	(TA_MSG_TYPE_PHASE1-1+TA_MAX_N_FRAMES)
*/
#define TA_MSG_TYPE_USER	100
#define TA_N_SKIPPED_MID	100

#define TA_QB_RECEIVING		1
#define TA_QB_RECEIVING2	2
#define TA_MSG_TYPE(bits)	(bits&0x70)
#define TA_PING_MSG_TYPE	0x10
#define TA_PHASE_MSG_TYPE	0x20
#define TA_SYSTEM_MSG_TYPE	0x30
#define TA_USER_MSG_TYPE	0x40

#define TA_SHMMIN		128
#define TA_MIN_RECV_BUFFER_LENGTH       100
/* SRPS protocol message header */
typedef struct {
	char protname[4];	// SRPS
	TaOctet major_version;	// minor changes do not change interface
	TaOctet spec_bits;	// various qos or diag enabled
	TaUShort iface_code;	// last 16 bits of Interface Id
	TaULong msg_type;	// valid message type
	TaUShort dst_oid;	// message destination oid
	TaUShort basic_length;	// of header and static part
	TaULong length;		// of whole message with dymamic part
	TaUShort mid;		// message Id
	TaUShort mid_resp;	// response to message Id
	TaLong aid;		// area Id
	TaShort src_sid;	// sender Id
	TaShort dst_sid;	// destination Id
} TaSRPSHeader;

#ifdef _WINDOWS
        #ifdef errno
                #undef errno
        #endif
        #define errno WSAGetLastError()
        #define TA_EC(errcode) WSA##errcode
#define strerror(errnum) gai_strerror(errnum)
#else
        #define TA_EC(errcode) errcode
#endif

#endif
