/*
 *                       ***** TA Common *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *
 */

#include <ta.h>
#include <time.h>
#ifndef _WINDOWS
#include <sys/time.h>

#if __BYTE_ORDER == __BIG_ENDIAN
#define TA_BIG_ENDIAN_BYTE_ORDER
#elif __BYTE_ORDER == __LITTLE_ENDIAN
#define TA_LITTLE_ENDIAN_BYTE_ORDER
#endif

#else
#include <ctype.h>
#include <Windows.h>
#define MAX_SLEEP_SECONDS (DWORD)4294966        // (2**32-2)/1000

#define TA_LITTLE_ENDIAN_BYTE_ORDER

#endif

TaULong ta_trace_level;
FILE *ta_file, *ta_tfile;
static TaULong g_crc_table[256];
void taRTTI_Init();

#ifndef TA_OLD_TIME_FORMAT
const TaLongLong EPOCH_DIFF = 11644473600L;
#endif

/* ------------------------------------------------------- */
int taInitialize()
{
	TaULong i, j, cr;
#ifdef _WINDOWS
	WORD wVersionRequested;
	WSADATA wsaData;
	wVersionRequested = MAKEWORD( 2, 2 );
	if (WSAStartup( wVersionRequested, &wsaData ) != 0)
		return TA_ERR_ARGUMENT;
#endif
	if (taInitMemProts() < 0)
		return TA_ERR_ARGUMENT;
	if (taInitPingProts() < 0)
		return TA_ERR_ARGUMENT;
	taRTTI_Init();

	ta_file = stdout;
	ta_tfile = NULL;
	srand((unsigned int) (taGetTime() % 1000000000));
	for (i = 0; i < 256; i++) {
		cr = i;
		for (j = 0; j < 8; j++)
			cr = cr & 1 ? (cr >> 1) ^ 0xEDB88320UL : cr >> 1;
		
		g_crc_table[i] = cr;
	}        
	return 1;
}

/* ------------------------------------------------------- */
int taTrace(TaULong trace_level, const char *filename)
{
	if (trace_level == (TaULong)-1)
		return (int) ta_trace_level;
	if (ta_tfile) {
		fclose(ta_tfile);
		ta_tfile = NULL;
	}
	if (filename && *filename)
		ta_tfile = fopen(filename, "w");

	if (ta_tfile)
		ta_file = ta_tfile;
	else
		ta_file = stdout;
	ta_trace_level = trace_level;
	return (int) ta_trace_level;
}

/* ------------------------------------------------------- */
void taPuts(TaChar *str, TaBool newline)
{
	if (newline)
		fprintf(ta_file, "%s\n", str);
	else
		fprintf(ta_file, "%s", str);
	fflush(ta_file);
}

/* ------------------------------------------------------- */
int taSleep(TaLongLong tp)
{
#ifdef _WINDOWS
        unsigned long secs = TA_SEC(tp);
        unsigned long mcs = TA_MCS(tp);
        DWORD no_of_max_sleeps, i;
        if (secs <= MAX_SLEEP_SECONDS) {
                Sleep(secs * 1000 + mcs/1000);
                return 0;
        }
        no_of_max_sleeps = secs / MAX_SLEEP_SECONDS;
        for (i = 0; i < no_of_max_sleeps; i++)
                Sleep(MAX_SLEEP_SECONDS * 1000);
        Sleep((secs % MAX_SLEEP_SECONDS) * 1000 + mcs/1000);
        return 0;
#else
        struct timespec req, rem;
        req.tv_sec = (time_t) TA_SEC(tp);
        req.tv_nsec = (long) TA_MCS(tp);
#ifdef TA_NANOSLEEP
	req.tv_nsec = req.tv_nsec * 1000;
        return nanosleep(&req, &rem);
#else
	while (req.tv_sec >= 1000) {
		usleep(1000000000);
		req.tv_sec -= 1000;
	}
	usleep(req.tv_sec * 1000000 + req.tv_nsec);
	return 0;
#endif
#endif
}

/* ------------------------------------------------------- */
#ifdef TA_OLD_TIME_FORMAT
TaLongLong taGetTime()
{
#ifdef _WINDOWS
        static int days_in_preceding_months[12]
                = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };
        static int days_in_preceding_months_leap[12]
                = { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335 };
        SYSTEMTIME st;
        DWORD days;
        unsigned long abs_sec;
        GetSystemTime(&st);
        // this formula should work until 1st March 2100
        days = ((st.wYear - 1970) * 365 + (st.wYear - 1969) / 4
                + ((st.wYear % 4)
                        ? days_in_preceding_months[st.wMonth - 1]
                        : days_in_preceding_months_leap[st.wMonth - 1])
                + st.wDay - 1);
        abs_sec = st.wSecond + 60 * (st.wMinute + 60 * (st.wHour + 24 * days));
        return ((TaLongLong) abs_sec)*1000000 + st.wMilliseconds * 1000;
#else
        struct timeval tv;
        gettimeofday(&tv, NULL);
        unsigned int isec = (unsigned int)tv.tv_sec;
        return ((TaLongLong) isec)*1000000 + tv.tv_usec;
#endif
}
#else
TaULong TA_SEC(TaLongLong t)
{
	TaLongLong li;
	li = t / 10000000;
	if (li < EPOCH_DIFF)
		return (TaULong) li;
	return (TaULong) (li - EPOCH_DIFF);
	
}

#ifdef _WINDOWS
typedef union {
	FILETIME ft;
	TaLongLong lv;
} FTime;
TaLongLong taGetTime()
{
        FTime t;
        GetSystemTimeAsFileTime(&t.ft);
	return t.lv;
#else
TaLongLong taGetTime()
{
        struct timeval tv;
        gettimeofday(&tv, NULL);
	TaLongLong t = (unsigned int)tv.tv_sec;
	t = (t + EPOCH_DIFF) * 10000000 + tv.tv_usec * 10;
        return t;
#endif
}
#endif

/* ------------------------------------------------------- */
void taLocaltime(TaLongLong t_crt, TaTmStruct *tloc)
{
	TaULong t_t;
	struct tm l_tm, *p_tm;
#ifdef _WINDOWS
	__time64_t t64 = (__time64_t) TA_SEC(t_crt);
	if (_localtime64_s(&l_tm, &t64))
		return;
	p_tm = &l_tm;
#else
	t_t = TA_SEC(t_crt);
	p_tm = localtime((const time_t *)&t_t);
#endif
	tloc->wyear = p_tm->tm_year;
	tloc->wmonth = p_tm->tm_mon;
	tloc->wday = p_tm->tm_mday;
	tloc->hh = p_tm->tm_hour;
	tloc->mm = p_tm->tm_min;
	tloc->ss = p_tm->tm_sec;
}

/* ------------------------------------------------------- */
void taMaketime(const TaTmStruct *tloc, TaLongLong *t_crt)
{
	TaULong t_t;
	struct tm l_tm, *p_tm;

	if (tloc->wday == 0) {
#ifdef _WINDOWS
		__time64_t t64 = time(NULL);
		if (_localtime64_s(&l_tm, &t64))
			return;
#else
		t_t = time(NULL);
		p_tm = localtime((const time_t *)&t_t);
		l_tm = *p_tm;
#endif
	} else {
		l_tm.tm_year = tloc->wyear;
		l_tm.tm_mon = tloc->wmonth;
		l_tm.tm_mday = tloc->wday;
	}
	l_tm.tm_hour = tloc->hh;
	l_tm.tm_min = tloc->mm;
	l_tm.tm_sec = tloc->ss;
	t_t = mktime(&l_tm);
#ifdef TA_OLD_TIME_FORMAT
	*t_crt = t_t;
	*t_crt = (*t_crt)*1000000;
#else
	*t_crt = t_t;
	*t_crt = (*t_crt + EPOCH_DIFF) * 10000000;
#endif
}

/* ------------------------------------------------------- */
double taRand()
{
	return 1.0 * rand() / RAND_MAX;
}

/* ------------------------------------------------------- */
TaULong taCrc32(const char *buf, TaULong crc, TaULong nbits)
{
	while (*buf)
		crc = g_crc_table[(crc ^ *buf++) & 0xFF] ^ (crc >> 8);
	crc = crc ^ 0xFFFFFFFFUL;
	if (nbits == 8)
		return crc & 0xFFUL;
	else if (nbits == 16)
		return crc & 0xFFFFUL;
	return crc;
}

/*
 * ----------------------------------------------------------
 * Byte order transformations
 *    __BYTE_ORDER and __FLOAT_WORD_ORDER
 *
 *  __LITTLE_ENDIAN 1234	- x86 byte order
 *  __BIG_ENDIAN    4321	- network byte order
 *  __PDP_ENDIAN    3412
 *
 * ----------------------------------------------------------
 */
typedef union {
        TaChar buffer[2];
        TaShort iv;
} Un2;
typedef union {
        TaChar buffer[4];
        TaLong iv;
        TaFloat fv;
} Un4;
typedef union {
        TaChar buffer[8];
        TaLongLong iv;
        TaDouble fv;
} Un8;

/* ------------------------------------------------------- */
TaShort taH2N16(TaShort iv)
{
	Un2 src, dst;
	src.iv = iv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.iv = iv;
#else
        dst.buffer[1] = src.buffer[0];
        dst.buffer[0] = src.buffer[1];
#endif
	return dst.iv;
}
TaShort taN2H16(TaShort iv)
{ return taH2N16(iv); }

/* ------------------------------------------------------- */
TaLong taH2N32(TaLong iv)
{
	Un4 src, dst;
	src.iv = iv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.iv = iv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[3] = src.buffer[0];
        dst.buffer[2] = src.buffer[1];
        dst.buffer[1] = src.buffer[2];
        dst.buffer[0] = src.buffer[3];
#else
        dst.buffer[2] = src.buffer[0];
        dst.buffer[3] = src.buffer[1];
        dst.buffer[0] = src.buffer[2];
        dst.buffer[1] = src.buffer[3];
#endif
#endif
	return dst.iv;
}

TaFloat taH2Nr32(TaFloat fv)
{
	Un4 src, dst;
	src.fv = fv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.fv = fv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[3] = src.buffer[0];
        dst.buffer[2] = src.buffer[1];
        dst.buffer[1] = src.buffer[2];
        dst.buffer[0] = src.buffer[3];
#else
        dst.buffer[2] = src.buffer[0];
        dst.buffer[3] = src.buffer[1];
        dst.buffer[0] = src.buffer[2];
        dst.buffer[1] = src.buffer[3];
#endif
#endif
	return dst.fv;
}

TaLong taN2H32(TaLong iv)
{
	Un4 src, dst;
	src.iv = iv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.iv = iv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[3] = src.buffer[0];
        dst.buffer[2] = src.buffer[1];
        dst.buffer[1] = src.buffer[2];
        dst.buffer[0] = src.buffer[3];
#else
        dst.buffer[1] = src.buffer[0];
        dst.buffer[0] = src.buffer[1];
        dst.buffer[3] = src.buffer[2];
        dst.buffer[2] = src.buffer[3];
#endif
#endif
	return dst.iv;
}

TaFloat taN2Hr32(TaFloat fv)
{
	Un4 src, dst;
	src.fv = fv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.fv = fv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[3] = src.buffer[0];
        dst.buffer[2] = src.buffer[1];
        dst.buffer[1] = src.buffer[2];
        dst.buffer[0] = src.buffer[3];
#else
        dst.buffer[1] = src.buffer[0];
        dst.buffer[0] = src.buffer[1];
        dst.buffer[3] = src.buffer[2];
        dst.buffer[2] = src.buffer[3];
#endif
#endif
	return dst.fv;
}

/* ------------------------------------------------------- */
TaLongLong taH2N64(TaLongLong iv)
{
	Un8 src, dst;
	src.iv = iv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.iv = iv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[7] = src.buffer[0];
        dst.buffer[6] = src.buffer[1];
        dst.buffer[5] = src.buffer[2];
        dst.buffer[4] = src.buffer[3];
        dst.buffer[3] = src.buffer[4];
        dst.buffer[2] = src.buffer[5];
        dst.buffer[1] = src.buffer[6];
        dst.buffer[0] = src.buffer[7];
#else
        dst.buffer[6] = src.buffer[0];
        dst.buffer[7] = src.buffer[1];
        dst.buffer[4] = src.buffer[2];
        dst.buffer[5] = src.buffer[3];
        dst.buffer[2] = src.buffer[4];
        dst.buffer[3] = src.buffer[5];
        dst.buffer[0] = src.buffer[6];
        dst.buffer[1] = src.buffer[7];
#endif
#endif
	return dst.iv;
}

TaDouble taH2Nr64(TaDouble fv)
{
	Un8 src, dst;
	src.fv = fv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.fv = fv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[7] = src.buffer[0];
        dst.buffer[6] = src.buffer[1];
        dst.buffer[5] = src.buffer[2];
        dst.buffer[4] = src.buffer[3];
        dst.buffer[3] = src.buffer[4];
        dst.buffer[2] = src.buffer[5];
        dst.buffer[1] = src.buffer[6];
        dst.buffer[0] = src.buffer[7];
#else
        dst.buffer[6] = src.buffer[0];
        dst.buffer[7] = src.buffer[1];
        dst.buffer[4] = src.buffer[2];
        dst.buffer[5] = src.buffer[3];
        dst.buffer[2] = src.buffer[4];
        dst.buffer[3] = src.buffer[5];
        dst.buffer[0] = src.buffer[6];
        dst.buffer[1] = src.buffer[7];
#endif
#endif
	return dst.fv;
}

TaLongLong taN2H64(TaLongLong iv)
{
	Un8 src, dst;
	src.iv = iv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.iv = iv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[7] = src.buffer[0];
        dst.buffer[6] = src.buffer[1];
        dst.buffer[5] = src.buffer[2];
        dst.buffer[4] = src.buffer[3];
        dst.buffer[3] = src.buffer[4];
        dst.buffer[2] = src.buffer[5];
        dst.buffer[1] = src.buffer[6];
        dst.buffer[0] = src.buffer[7];
#else
        dst.buffer[1] = src.buffer[0];
        dst.buffer[0] = src.buffer[1];
        dst.buffer[3] = src.buffer[2];
        dst.buffer[2] = src.buffer[3];
        dst.buffer[5] = src.buffer[4];
        dst.buffer[4] = src.buffer[5];
        dst.buffer[7] = src.buffer[6];
        dst.buffer[6] = src.buffer[7];
#endif
#endif
	return dst.iv;
}

TaDouble taN2Hr64(TaDouble fv)
{
	Un8 src, dst;
	src.fv = fv;
#ifdef TA_BIG_ENDIAN_BYTE_ORDER
	dst.fv = fv;
#else
#ifdef TA_LITTLE_ENDIAN_BYTE_ORDER
        dst.buffer[7] = src.buffer[0];
        dst.buffer[6] = src.buffer[1];
        dst.buffer[5] = src.buffer[2];
        dst.buffer[4] = src.buffer[3];
        dst.buffer[3] = src.buffer[4];
        dst.buffer[2] = src.buffer[5];
        dst.buffer[1] = src.buffer[6];
        dst.buffer[0] = src.buffer[7];
#else
        dst.buffer[1] = src.buffer[0];
        dst.buffer[0] = src.buffer[1];
        dst.buffer[3] = src.buffer[2];
        dst.buffer[2] = src.buffer[3];
        dst.buffer[5] = src.buffer[4];
        dst.buffer[4] = src.buffer[5];
        dst.buffer[7] = src.buffer[6];
        dst.buffer[6] = src.buffer[7];
#endif
#endif
	return dst.fv;
}


/* ------------------------------------------------------- */
#define MAX_LINES_IN_CONFIG	500
#define MAX_STR TA_NAME_LENGTH*2
char **taArgsFromConfig(const char *cfg_name)
{
	char **args, str[MAX_STR], *p;
	int i;
	FILE *fd = fopen(cfg_name, "r");
	if (!fd)
		return NULL;
	args = malloc(MAX_LINES_IN_CONFIG * sizeof(char *));
	memset(args, 0, MAX_LINES_IN_CONFIG * sizeof(char *));
	i = 0;
	while(fgets(str, MAX_STR, fd)) {
		if (str[0] == '#')
			continue;
		p = strchr(str, '\n');
		if (p)	*p = 0;
		p = strchr(str, '=');
		if (!p)
			continue;
		*p = 0;
		args[i++] = strdup(str);	// option
		args[i++] = strdup(p+1);	// value
	}
	fclose(fd);
	return args;
}

/* ------------------------------------------------------- */
void taClearArgs(char **args)
{
	int i;
	for (i = 0; args[i]; i++) {
		free(args[i]);
		args[i] = NULL;
	}
	free(args);
}

/* ------------------------------------------------------- */
TaChar *taWChar2Char(TaChar *buffer, const TaWChar *wstr, int length)
{
	int i;
	for (i = 0; i < length; i++) {
		buffer[i] = (TaChar)wstr[i];
		if (!wstr[i])
			break;
	}
	return buffer;
}

/* ------------------------------------------------------- */
TaWChar *taChar2WChar(TaWChar *wbuffer, const TaChar *str, int length)
{
	int i;
	for (i = 0; i < length; i++) {
		wbuffer[i] = (TaWChar)str[i];
		if (!str[i])
			break;
	}
	return wbuffer;
}

/* ------------------------------------------------------- */
TaWChar *taWCharCpy(TaWChar *wbuffer, const TaWChar *src, int length)
{
	int i;
	for (i = 0; i < length; i++) {
		wbuffer[i] = src[i];
		if (!src[i])
			break;
	}
	return wbuffer;
}
/* ------------------------------------------------------- */
#ifdef MY_INTEL_X86_ARCHITECTURE
extern "C" { 
    __inline__ uint64_t rdtsc() 
    { 
        uint32_t lo, hi; 
        __asm__ __volatile__ ( 
            "xorl %%eax,%%eax \n        cpuid" 
            ::: "%rax", "%rbx", "%rcx", "%rdx"); 
        __asm__ __volatile__ ("rdtsc" : "=a" (lo), "=d" (hi)); 
        return (uint64_t)hi << 32 | lo; 
    } 
} 
#endif
