
/*
 *                       ***** TA Message *****
 *
 *                      Transparent Architecture
 *
 *   Copyright (c) 2012 Dmitry Dagaev. All rights reserved.
 *
 *
 */

#include <ta.h>
#include <stdio.h>
#include <string.h>

TaBool BoolVal(const char *addr);
void PutBoolVal(char *addr, TaBool b);
TaOctet I1Val(const char *addr);
void PutI1Val(char *addr, TaOctet i1);
TaChar CharVal(const char *addr);
void PutCharVal(char *addr, TaChar c);
TaShort I2Val(const char *addr);
void PutI2Val(char *addr, TaShort i2);
TaWChar WCharVal(const char *addr);
void PutWCharVal(char *addr, TaWChar wc);
TaLong I4Val(const char *addr);
void PutI4Val(char *addr, TaLong i4);
TaLongLong I8Val(const char *addr);
void PutI8Val(char *addr, TaLongLong i8);
TaFloat R4Val(const char *addr);
void PutR4Val(char *addr, TaFloat r4);
TaDouble R8Val(const char *addr);
void PutR8Val(char *addr, TaDouble r8);

static void ReadItemString(TaRTTItem *ritem, const char *item_addr,
	char *recent_data, TaBool *changed);
static int UnquoteAndSimpleXmlParser(char *str0, const char *str, TaOnParseCouple *parser,
	void *client_data);

/* ------------------------------------------------------- */
void ReadItemString(TaRTTItem *ritem, const char *item_addr,
	char *recent_data, TaBool *changed)
{
	TaOctet i1;
	TaBool b;
	TaChar c;
	TaShort i2;
	TaWChar wc;
	TaLong i4;
	TaLongLong i8;
	TaFloat r4;
	TaDouble r8;
	TaRTTItem *item1;
	TaLong nb;

	switch (ritem->item_type_id) {
	case TaBool_TYPE:
		i1 = (TaOctet) strcmp("0", item_addr);
		b = ((i1 == 0) ? _False : _True);
		if (b != BoolVal(recent_data+ritem->offset)) {
			*changed = _True;
			PutBoolVal(recent_data+ritem->offset, b);
		}
		break;
	case TaOctet_TYPE:
		i1 = (TaOctet) atoi(item_addr);
		if (i1 != I1Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI1Val(recent_data+ritem->offset, i1);
		}
		break;
	case TaChar_TYPE:
		c = *item_addr;
		if (c != CharVal(recent_data+ritem->offset)) {
			*changed = _True;
			PutCharVal(recent_data+ritem->offset, c);
		}
		break;
	case TaShort_TYPE:
		i2 = (TaShort) atoi(item_addr);
		if (i2 != I2Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI2Val(recent_data+ritem->offset, i2);
		}
		break;
	case TaWChar_TYPE:
		taChar2WChar(&wc, item_addr, 1);
		if (wc != WCharVal(recent_data+ritem->offset)) {
			*changed = _True;
			PutWCharVal(recent_data+ritem->offset, wc);
		}
		break;
	case TaLong_TYPE:
		i4 = (TaLong) atoi(item_addr);
		if (i4 != I4Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI4Val(recent_data+ritem->offset, i4);
		}
		break;
	case TaLongLong_TYPE:
		i8 = (TaLongLong) atoi(item_addr);
		if (i8 != I8Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutI8Val(recent_data+ritem->offset, i8);
		}
		break;
	case TaFloat_TYPE:
		r4 = (TaFloat) atoi(item_addr);
		if (r4 != R4Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutR4Val(recent_data+ritem->offset, r4);
		}
		break;
	case TaDouble_TYPE:
		r8 = (TaDouble) atoi(item_addr);
		if (r8 != R8Val(recent_data+ritem->offset)) {
			*changed = _True;
			PutR8Val(recent_data+ritem->offset, r8);
		}
		break;
	case TaArr_TYPE:
		for (i4 = 0; i4 < ritem->dim; i4++) {
			ReadItemString(ritem->contains, item_addr+i4,
			recent_data + i4 * ritem->contains->size,
				changed);
			if (item_addr+i4 == '\0')
				break;
		}
		break;
	case TaRec_TYPE:
		if (ritem->contains)
			item1 = ritem->contains;
		else
			item1 = ritem->next;
		while (item1) {
			ReadItemString(item1, item_addr, recent_data,
				changed);
			item1 = item1->next;
		}
		break;
	case TaPtr_TYPE:
		// to be continued
		break;
	}
}

/* ------------------------------------------------------- */
int taOnReadSoap(int parse_flag, TaNamesStack *nstack,
        int nlevel, const char *value, void *client_data)
{
	TaTable *t = (TaTable *) client_data;
	TaRTTItem *ritem, *parent_item, *ri, *item1;
	TaBool changed = _False;
	char **addr, *recent_data;
	TaRTTIadata *adt;
	int dim;

//printf("taOnReadSoap %d nl:%d '%s' '%s' item:'%s'\n", parse_flag, nlevel, 
	//nstack[nlevel].name, value, t->rtti->mainitem.type_name);

	// ----------------------------------
	if (parse_flag & TA_PARSE_COUPLE_BEGIN) {
		// get parent item
		if (nlevel == 0) {
			ritem = parent_item = &t->rtti->mainitem;
			nstack[0].recent_data = NULL;
		} else {
			parent_item = (TaRTTItem *)nstack[nlevel-1].user_data;
			ritem = NULL;
			switch (parent_item->item_type_id) {
			case TaArr_TYPE:
				/*if (!parent_item->contains)
					return TA_PARSING;
				if (ritem->contains->item_type_id == TaRec_TYPE)
					ritem = ritem->contains;*/
				break;
			case TaRec_TYPE:
				if (parent_item->contains)
					ritem = parent_item->contains;
				else
					ritem = parent_item->next;
				break;
			case TaPtr_TYPE:
				ritem = parent_item->contains;
				if (!ritem 
				|| ritem->item_type_id != TaArr_TYPE
				|| !ritem->contains)
					return TA_PARSING;
				// pointer to array of records
				if (ritem->contains->item_type_id == TaRec_TYPE)
				{
					ritem = ritem->contains;
					if (ritem->contains)
						ritem = ritem->contains;
					else
						ritem = ritem->next;
				}
				// or pointer to array of simple
				break;
			default:
				return TA_PARSING;
			}
		}
		// find this rtti item
		nstack[nlevel].user_data = NULL;
		ri = ritem;
		while (ri) {
			if (!strcmp(nstack[nlevel].name, ri->item_name)) {
				nstack[nlevel].user_data = ri;
				break;
			}
			ri = ri->next;
		}
		if (!nstack[nlevel].user_data)
			return 0;
			//return TA_NOT_FOUND;
		if (ri->item_type_id == TaPtr_TYPE) {
			// realloc ptr
			addr = (const char **) (t->recent_data + ri->offset);
			if (!*addr) {
				adt = NULL;
				dim = 0;
			} else {
				adt = (TaRTTIadata *) 
					(*addr-sizeof(TaRTTIadata));
				dim = adt->dim;
			}
			item1 = ri->contains;
			if (nstack[nlevel].index >= dim) {
				if (t->realloc_mode == TA_DONT_REALLOC)
					return TA_ERR_OUT_OF_MEMORY;
				if (t->mqos.dim_chunk <= 0)
					nstack[nlevel].dim = nstack[nlevel].index+1;
				else
					nstack[nlevel].dim 
						= nstack[nlevel].index 
						/ t->mqos.dim_chunk
						* t->mqos.dim_chunk 
						+ t->mqos.dim_chunk;
				if (!adt) {
					*addr = (char *) taAllocWithRTTI(
						item1->size, 
						nstack[nlevel].dim);
					adt = (TaRTTIadata *)
						(*addr-sizeof(TaRTTIadata));
				} else
					*addr = (char *) taReallocWithRTTI(
						*addr, nstack[nlevel].dim);
			} else
				nstack[nlevel].dim = dim;
			nstack[nlevel].recent_data = 
				*addr + nstack[nlevel].index
					* item1->contains->size;
		} else {
			if (nlevel > 0)
				nstack[nlevel].recent_data = 
					nstack[nlevel-1].recent_data;
			nstack[nlevel].dim = 0;
		}
	}

	// ----------------------------------
	if (parse_flag & TA_PARSE_COUPLE_END) {
		if (!*value)
			return 0;
		if (!nstack[nlevel].user_data)
			return 0;
			//return TA_NOT_FOUND;
		ritem = (TaRTTItem *)nstack[nlevel].user_data;
		recent_data = t->recent_data;
		if (nstack[nlevel].recent_data)
			recent_data = nstack[nlevel].recent_data;
		ReadItemString(ritem, value, recent_data, &changed);
	}

	// ----------------------------------
	if (parse_flag & TA_PARSE_COUPLE_DIM) {
		ritem = (TaRTTItem *)nstack[nlevel].user_data;
		if (!ritem)
			return 0;
		addr = (const char **) (t->recent_data + ritem->offset);
		nstack[nlevel].dim = nstack[nlevel].index+1;
		if (t->realloc_mode == TA_REALLOC_EXACT)
			*addr = (char *) taReallocWithRTTI(*addr, nstack[nlevel].dim);
	} 
	return (int) changed;
}


/* ------------------------------------------------------- */
int taSimpleXmlParser(const char *str, TaOnParseCouple *parser,
	void *client_data)
{
	TaNamesStack nm[TA_MAX_STACK];
	char value[4096];
	const char *p0, *p1;
	int rc = 0, i, l, nst = 0, izz=0;
	int pcf;
	TaBool changed;

	//printf("taSimpleXmlParseR %x\n", client_data);
	p1 = str;
	value[0] = 0;
	memset(nm, 0, sizeof(TaNamesStack)*TA_MAX_STACK);
	while (1) {
		p0 = strchr(p1, '<');
		if (!p0)
			break;
		++p0;
		if (*p0 == '/') {
			++p0;
			pcf = TA_PARSE_COUPLE_END;
		} else if (*p0 == '?') {
			p1 = p0;
			continue;
		} else {
			value[0] = 0;
			pcf = TA_PARSE_COUPLE_BEGIN;
		}
		changed = _False;
		for (i = l = 0; p0[i]; i++) {
			if (p0[i] == '>') {
				l = i;
				break;
			}
			if (isspace(p0[i]))
				l = i;
			else {
				if (nm[nst].name[i] != p0[i])
					changed = _True;
				nm[nst].name[i] = p0[i];
			}
		}
		if (l >= TA_NAME_LENGTH)
			l = TA_NAME_LENGTH-1;
		nm[nst].name[l] = 0;
		p1 = p0+l+1;

		if (pcf == TA_PARSE_COUPLE_BEGIN) {
			if (!changed)
				++nm[nst].index;
			else {
				nm[nst].index = 0;
				if (nm[nst].dim > 0 
				&& nm[nst].dim != nm[nst].index+1)
					(*parser)(TA_PARSE_COUPLE_DIM,
						nm, nst, value, client_data);
			}
			++nst;
		}
		if (nst <= 0)
			break;
		i =
		(*parser)(pcf, nm, nst-1, value, client_data);
		if (pcf == TA_PARSE_COUPLE_END) {
			if (nm[nst].dim > 0
			&& nm[nst].dim != nm[nst].index+1)
				(*parser)(TA_PARSE_COUPLE_DIM,
					nm, nst, value, client_data);
			--nst;
			if (nst < 0)
				break;
		}
		if (i < 0)
			return i;
		rc |= i;

		i = l = 0;
		while (*p1 && *p1 != '<') {
			if (!isspace(*p1)) {
				l = i+1;
				value[i++] = *p1;
			} else if (l > 0) {
				value[i++] = *p1;
			}
			++p1;
		}
		value[l] = 0;
	}
	return rc;
}

/* ------------------------------------------------------- */
int taSimpleLineParser(const char *str, TaOnParseCouple *parser,
	void *client_data)
{
	TaNamesStack nm[TA_MAX_STACK];
	char value[4096], sv[3];
	const char *p;
	int rc = 0, l, i, j, nv;
	sv[2] = 0;
	memset(nm, 0, sizeof(TaNamesStack)*TA_MAX_STACK);
	while (1) {
		p = strchr(str, '=');
		if (!p)
			break;
		l = (int)(p-str);
		if (l >= TA_NAME_LENGTH)
			l = TA_NAME_LENGTH-1;
		strncpy(nm[0].name, str, l);
		nm[0].name[l] = 0;
		str = p+1;
		p = strchr(str, '&');
		if (!p)
			p = strchr(str, ' ');
		if (!p)
			p = strchr(str, '\n');
		if (!p)
			l = strlen(str);
		else
			l = (int)(p-str);
		for (i = j = 0; i < l; i++, j++) {
			switch(str[i]) {
			case '+':
				value[j] = ' ';
				break;
			case '%':
				strncpy(sv, &str[i+1], 2);
				sscanf(sv, "%x", &nv);
				value[j] = (char) nv;
				i += 2;
				break;
			default:
				value[j] = str[i];
			}
		}
		value[j] = 0;

		i =
		(*parser)(TA_PARSE_COUPLE_BOTH, nm, 0, value,
			client_data);
		if (i < 0)
			return i;
		rc |= i;

		if (!p)
			break;
		str = p+1;
	}
	return rc;
}

/* ------------------------------------------------------- */
int UnquoteAndSimpleXmlParser(char *str0, const char *str, TaOnParseCouple *parser,
	void *client_data)
{
	int rc;
	rc = taUnquoteBuffer(str0, 0, 0);
	if (rc < 0)
		return rc;
	return taSimpleXmlParser(str, parser, client_data);
}

/* ------------------------------------------------------- */
int taReadSoap(TaOnParseSoap *ops)
{
	int i;
	char *p, buf[256], *buffer;
//printf("taReadSoap %d %x\n", ops->message_mid, ops->client_data);
	switch (ops->message_mid) {
	case TA_SOAP_11_MID:
		p = strstr(ops->message_buffer, "<soap:Body");
		if (!p) return TA_ERR_WRONG_DATA;
		p = strstr(p, ops->type_name);
		if (!p) return TA_ERR_WRONG_DATA;
		p = strchr(p, '<');
		if (!p) return TA_ERR_WRONG_DATA;
		return UnquoteAndSimpleXmlParser(ops->message_buffer,
			p, ops->on_read_soap, ops->client_data);
	case TA_SOAP_12_MID:
		p = strstr(ops->message_buffer, "<soap12:Body");
		if (!p) return TA_ERR_WRONG_DATA;
		p = strstr(p, ops->type_name);
		if (!p) return TA_ERR_WRONG_DATA;
		p = strchr(p, '<');
		if (!p) return TA_ERR_WRONG_DATA;
		return UnquoteAndSimpleXmlParser(ops->message_buffer,
			p, ops->on_read_soap, ops->client_data);
	case TA_SOAP_GET_MID:
		p = strchr(ops->message_buffer, '?');
		if (!p) return TA_ERR_WRONG_DATA;
		return taSimpleLineParser(p+1, ops->on_read_soap, ops->client_data);
	case TA_SOAP_POST_MID:
		p = ops->message_buffer; i = 0;
		while (*p) {
			if (*p == '\n') {
				if (i == 0)
					break;
				else
					i = 0;
			} else if (!isspace(*p))
				++i;
			++p;
		}
		if (!p || !*p) return TA_ERR_WRONG_DATA;
		while (isspace(*p))
			p++;
		return taSimpleLineParser(p, ops->on_read_soap, ops->client_data);
	case TA_SOAP_11_CLIENT:
		p = strstr(ops->message_buffer, "<soap:Body");
		if (!p) return TA_ERR_WRONG_DATA;
		sprintf(buf, "<%sResult>", ops->type_name);
		p = strstr(p, buf);
		if (!p) return TA_ERR_WRONG_DATA;
		i = strlen(buf);
		// check if complete xml
		sprintf(buf, "</%sResult>", ops->type_name);
		if (strstr(p, buf))
			return UnquoteAndSimpleXmlParser(ops->message_buffer,
				p+i, ops->on_read_soap, ops->client_data);
//fprintf(ta_file,"Append 11\n");
		strncpy(ops->table_buffer, ops->message_buffer, ops->soap_buffer_size);
		ops->table_buffer[ops->soap_buffer_size-1] = 0;
		return TA_TO_BE_CONTINUED;
	case TA_SOAP_12_CLIENT:
		p = strstr(ops->message_buffer, "<soap12:Body");
		if (!p) return TA_ERR_WRONG_DATA;
		sprintf(buf, "<%sResult>", ops->type_name);
		p = strstr(p, buf);
		if (!p) return TA_ERR_WRONG_DATA;
		i = strlen(buf);
		// check if complete xml
		sprintf(buf, "</%sResult>", ops->type_name);
		if (strstr(p, buf))
			return UnquoteAndSimpleXmlParser(ops->message_buffer,
				p+i, ops->on_read_soap, ops->client_data);
//fprintf(ta_file,"Append 12\n");
		strncpy(ops->table_buffer, ops->message_buffer, ops->soap_buffer_size);
		ops->table_buffer[ops->soap_buffer_size-1] = 0;
		return TA_TO_BE_CONTINUED;
	case TA_SOAP_POST_CLIENT:
		p = strstr(ops->message_buffer, "<string");
		if (!p) return TA_ERR_WRONG_DATA;
		p = strchr(p, '>');
		if (!p) return TA_ERR_WRONG_DATA;
		// check if complete xml
		if (strstr(p, "</string>"))
			return UnquoteAndSimpleXmlParser(ops->message_buffer,
				p+1, ops->on_read_soap, ops->client_data);
//fprintf(ta_file,"Append POST\n");
		strncpy(ops->table_buffer, ops->message_buffer, ops->soap_buffer_size);
		ops->table_buffer[ops->soap_buffer_size-1] = 0;
		return TA_TO_BE_CONTINUED;
	case TA_SOAP_MID:
		// append to the previous
		i = strlen(ops->table_buffer);
		strncpy(ops->table_buffer+i, ops->message_buffer, ops->soap_buffer_size-i);
		ops->table_buffer[ops->soap_buffer_size-1] = 0;
		sprintf(buf, "<%sResult>", ops->type_name);
		p = strstr(ops->table_buffer, buf);
		if (p)
			sprintf(buf, "<%sResult>", ops->type_name);
		else if (!(p = strstr(ops->table_buffer, "<string>")))
			return TA_ERR_WRONG_DATA;
		else
			sprintf(buf, "</string>");
		if (!strstr(ops->table_buffer, buf))
			break;
		p = strchr(p, '>');
		if (!p) return TA_ERR_WRONG_DATA;
//fprintf(ta_file,"Process Appended\n");
		return UnquoteAndSimpleXmlParser(ops->table_buffer,
			p+1, ops->on_read_soap, ops->client_data);
	default:
		return TA_ERR_ARGUMENT;
	}
	return 0;
}

/* ------------------------------------------------------- */
int WriteXmlSoapItem(const TaRTTItem *ritem, TaTable *t, 
	const char *recent_data, char *item_addr, int buffer_size, int dim)
{
	TaOctet i1;
	TaChar c;
	TaShort i2;
	TaWChar wc;
	TaLong i4, i, l;
	TaLongLong i8;
	TaFloat r4;
	TaDouble r8;
	TaRTTItem *item1;
	char buffer[4096];
	const char *addr;
	const TaRTTIadata *adt;

	switch (ritem->item_type_id) {
	case TaBool_TYPE:
		if (BoolVal(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>1</%s>\n",
				ritem->item_name, ritem->item_name);
		return 0;
	case TaOctet_TYPE:
		if (i1 = I1Val(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%d</%s>\n",
				ritem->item_name, i1, ritem->item_name);
		return 0;
	case TaChar_TYPE:
		if (c = CharVal(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%c</%s>\n",
				ritem->item_name, c, ritem->item_name);
		return 0;
	case TaShort_TYPE:
		if (i2 = I2Val(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%d</%s>\n",
				ritem->item_name, i2, ritem->item_name);
		return 0;
	case TaWChar_TYPE:
		if (wc = WCharVal(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%c</%s>\n",
				ritem->item_name, 
				*taWChar2Char(buffer, &wc, 1), 
				ritem->item_name);
		return 0;
	case TaLong_TYPE:
		if (i4 = I4Val(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%d</%s>\n",
				ritem->item_name, i4, ritem->item_name);
		return 0;
	case TaLongLong_TYPE:
		if (i8 = I8Val(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%ld</%s>\n",
				ritem->item_name, i8, ritem->item_name);
		return 0;
	case TaFloat_TYPE:
		if (r4 = R4Val(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%g</%s>\n",
				ritem->item_name, r4, ritem->item_name);
		return 0;
	case TaDouble_TYPE:
		if (r8 = R8Val(recent_data+ritem->offset))
			return sprintf(item_addr, "<%s>%lg</%s>\n",
				ritem->item_name, r8, ritem->item_name);
		return 0;
	case TaArr_TYPE:
		if (dim <= 0)
			dim = ritem->dim;
		switch (ritem->contains->item_type_id) {
		case TaChar_TYPE:
			if (strlen(recent_data+ritem->offset) == 0)
				return 0;
			return sprintf(item_addr, "<%s>%s</%s>\n",
				ritem->item_name, recent_data+ritem->offset,
				ritem->item_name);
		case TaWChar_TYPE:
			if (strlen(recent_data+ritem->offset) == 0)
				return 0;
			return sprintf(item_addr, "<%s>%s</%s>\n",
				ritem->item_name, 
				taWChar2Char(buffer, 
					recent_data+ritem->offset, dim+1),
				ritem->item_name);
		default:
			for (i = l = 0; i < dim; i++)
				l +=
				WriteXmlSoapItem(ritem->contains, t,
					recent_data + i * ritem->contains->size,
					item_addr + l, buffer_size-l, 0);
		}
		return l;
	case TaRec_TYPE:
		if (ritem->contains)
			item1 = ritem->contains;
		else
			item1 = ritem->next;
		if (ritem->item_name[0])
			l = sprintf(item_addr, "<%s>\n", ritem->item_name);
		else
			l = 0;
		i = l;
		while (item1) {
			l +=
			WriteXmlSoapItem(item1, t, recent_data,
				item_addr + l, buffer_size-l, 0);
			item1 = item1->next;
		}
		if (i == l) {
			l = 0;
			item_addr[0] = 0;
		} else if (ritem->item_name[0])
			l += sprintf(item_addr+l, "</%s>\n", 
				ritem->item_name);
		return l;
	case TaPtr_TYPE:
		addr = *(const char **)(recent_data+ritem->offset);
		if (addr == NULL)
			return 0;
		adt = (const TaRTTIadata *) (addr-sizeof(TaRTTIadata));
		item1 = ritem->contains;
		if (!item1)
			return 0;
		return WriteXmlSoapItem(item1, t, addr, item_addr, buffer_size,
			adt->dim);
	}

	return 0;
}

/* ------------------------------------------------------- */
int WriteLineSoapItem(const TaRTTItem *ritem, TaTable *t, 
	const char *recent_data, char *item_addr, int buffer_size, int dim)
{
	TaOctet i1;
	TaChar c;
	TaShort i2;
	TaWChar wc, *wca;
	TaLong i4, i, l;
	TaLongLong i8;
	TaFloat r4;
	TaDouble r8;
	TaRTTItem *item1;
	char buffer[4096];
	const char *addr;
	const TaRTTIadata *adt;

	switch (ritem->item_type_id) {
	case TaBool_TYPE:
		if (BoolVal(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=1",
				ritem->item_name);
		return 0;
	case TaOctet_TYPE:
		if (i1 = I1Val(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%d",
				ritem->item_name, i1);
		return 0;
	case TaChar_TYPE:
		if (c = CharVal(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%c",
				ritem->item_name, c);
		return 0;
	case TaShort_TYPE:
		if (i2 = I2Val(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%d",
				ritem->item_name, i2);
		return 0;
	case TaWChar_TYPE:
		if (wc = WCharVal(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%c",
				ritem->item_name, 
				*taWChar2Char(buffer, &wc, 1));
		return 0;
	case TaLong_TYPE:
		if (i4 = I4Val(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%d",
				ritem->item_name, i4);
		return 0;
	case TaLongLong_TYPE:
		if (i8 = I8Val(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%ld",
				ritem->item_name, i8);
		return 0;
	case TaFloat_TYPE:
		if (r4 = R4Val(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%g",
				ritem->item_name, r4);
		return 0;
	case TaDouble_TYPE:
		if (r8 = R8Val(recent_data+ritem->offset))
			return sprintf(item_addr, "%s=%lg",
				ritem->item_name, r8);
		return 0;
	case TaArr_TYPE:
		switch (ritem->contains->item_type_id) {
		case TaChar_TYPE:
			if (recent_data[ritem->offset] == 0)
				return 0;
			l = sprintf(item_addr, "%s=", ritem->item_name);
			if (dim <= 0)
				dim = ritem->dim;
			for (i = 0; i < dim; i++) {
				c = recent_data[ritem->offset+i];
				switch (c) {
				case ' ':
					item_addr[l] = '+';
					++l;
					break;
				case '+':
				case '?':
					l += sprintf(item_addr+l,
						"%%%x", (int) c);
					break;
				case '\0':
					break;
				default:
					item_addr[l] = c;
					++l;
					break;
				}
			}
			item_addr[l] = 0;
			return l;
		case TaWChar_TYPE:
			if (dim <= 0)
				dim = ritem->dim;
			taWChar2Char(buffer,
				recent_data+ritem->offset, dim+1);
			if (buffer[0] == 0)
				return 0;
			l = sprintf(item_addr, "%s=", ritem->item_name);
			for (i = 0; i < dim; i++) {
				c = buffer[i];
				switch (c) {
				case ' ':
					item_addr[l] = '+';
					++l;
					break;
				case '+':
				case '?':
				case '&':
					l += sprintf(item_addr+l,
						"%%%x", (int) c);
					break;
				case '\0':
					break;
				default:
					item_addr[l] = c;
					++l;
					break;
				}
			}
			item_addr[l] = 0;
			return l;
		default:
			if (dim <= 0)
				dim = ritem->dim;
			for (i = l = 0; i < dim; i++)
				l +=
				WriteLineSoapItem(ritem->contains, t,
					recent_data + i * ritem->contains->size,
					item_addr + l, buffer_size-l, 0);
		}
		return l;
	case TaRec_TYPE:
		if (ritem->contains)
			item1 = ritem->contains;
		else
			item1 = ritem->next;
		l = 0;
		while (item1) {
			i =
			WriteLineSoapItem(item1, t, recent_data,
				item_addr + l, buffer_size-l, 0);
			if (i > 0) {
				item_addr[l+i] = '&';
				l += (i+1);
			}
			item1 = item1->next;
		}
		return l;
	case TaPtr_TYPE:
		addr = *(const char **)(recent_data+ritem->offset);
		if (addr == NULL)
			return 0;
		adt = (const TaRTTIadata *) (addr-sizeof(TaRTTIadata));
		item1 = ritem->contains;
		if (!item1)
			return 0;
		return WriteLineSoapItem(item1, t, addr, item_addr, buffer_size,
			adt->dim);
	}

	return 0;
}

/* ------------------------------------------------------- */
int taWriteSoapItem(const TaRTTItem *ritem, TaTable *t, const char *data,
	char *item_addr, int buffer_size)
{
	int n = 0, nlen = -1, n0, rc;
	char buf[16];
	switch (t->mid_resp) {
	case TA_SOAP_11_MID:
		n = sprintf(item_addr, "HTTP/1.1 200 OK\nContent-Type: ");
		n += sprintf(item_addr+n,
"text/xml; charset=utf-8\nContent-Length:       ");
		nlen = n-6;
		n += sprintf(item_addr+n, "\n\n");
		n0 = n;
		n += sprintf(item_addr+n, 
"<?xml version='1.0' encoding='utf-8'?>\n"
"<soap:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
"xmlns:xsd='http://www.w3.org/2001/XMLSchema' "
"xmlns:soap='http://schemas.xmlsoap.org/soap/envelope/'>\n"
"  <soap:Body>\n    <%sResponse xmlns='http://%s'>\n      <%sResult>\n", 
			ritem->type_name, t->area->send_station->qos.hostname, ritem->type_name);
		rc = WriteXmlSoapItem(ritem, t, data, item_addr+n, 
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		n += sprintf(item_addr+n,
"      </%sResult>\n    </%sResponse>\n  </soap:Body>\n</soap:Envelope>\n",
			ritem->type_name, ritem->type_name);
		break;
	case TA_SOAP_12_MID:
		n = sprintf(item_addr, "HTTP/1.1 200 OK\nContent-Type: ");
		n += sprintf(item_addr+n, 
"application/soap+xml; charset=utf-8\nContent-Length:       ");
		nlen = n-6;
		n += sprintf(item_addr+n, "\n\n");
		n0 = n;
		n += sprintf(item_addr+n, 
"<?xml version='1.0' encoding='utf-8'?>\n"
"<soap12:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
"xmlns:xsd='http://www.w3.org/2001/XMLSchema' "
"xmlns:soap12='http://www.w3.org/2003/05/soap-envelope'>\n"
"  <soap12:Body>\n    <%sResponse xmlns='http://%s'>\n      <%sResult>\n", 
			ritem->type_name, t->area->send_station->qos.hostname, ritem->type_name);
		rc = WriteXmlSoapItem(ritem, t, data, item_addr+n, 
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		n += sprintf(item_addr+n,
"      </%sResult>\n    </%sResponse>\n  </soap12:Body>\n</soap12:Envelope>\n",
			ritem->type_name, ritem->type_name);
		break;
	case TA_SOAP_GET_MID:
	case TA_SOAP_POST_MID:
		n = sprintf(item_addr, "HTTP/1.1 200 OK\nContent-Type: ");
		n += sprintf(item_addr+n,
"text/xml; charset=utf-8\nContent-Length:       ");
		nlen = n-6;
		n += sprintf(item_addr+n, "\n\n");
		n0 = n;
		n += sprintf(item_addr+n, 
"<?xml version='1.0' encoding='utf-8'?>\n<string xmlns='http://%s'>\n",
			t->area->send_station->qos.hostname);
		rc = WriteXmlSoapItem(ritem, t, data, item_addr+n, 
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		n += sprintf(item_addr+n, "</string>\n");
		break;
	case TA_SOAP_11_CLIENT:
		n = sprintf(item_addr, "POST /%s HTTP/1.1\nHost: %s\n", 
			t->rtti->iface_name, t->area->send_station->qos.hostname);
		n += sprintf(item_addr+n, 
"Content-Type: text/xml; charset=utf-8\nContent-Length:       ");
		nlen = n-6;
		n += sprintf(item_addr+n,
"\nSOAPAction: \"http://%s/%s\"\n\n",
			t->area->send_station->qos.hostname, ritem->type_name);
		n0 = n;
		n += sprintf(item_addr+n, 
"<?xml version='1.0' encoding='utf-8'?>\n"
"<soap:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
"xmlns:xsd='http://www.w3.org/2001/XMLSchema' "
"xmlns:soap='http://schemas.xmlsoap.org/soap/envelope/'>\n"
"  <soap:Body>\n    <%s xmlns='http://%s'>\n", 
			ritem->type_name, t->area->send_station->qos.hostname);
		rc = WriteXmlSoapItem(ritem, t, data, item_addr+n, 
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		n += sprintf(item_addr+n,
"    </%s>\n  </soap:Body>\n</soap:Envelope>\n",
			ritem->type_name);
		break;
	case TA_SOAP_12_CLIENT:
		n = sprintf(item_addr, "POST /%s HTTP/1.1\nHost: %s\n", 
			t->rtti->iface_name, t->area->send_station->qos.hostname);
		n += sprintf(item_addr+n, 
"Content-Type: application/soap+xml; charset=utf-8\nContent-Length:       ");
		nlen = n-6;
		n += sprintf(item_addr+n, "\n\n");
		n0 = n;
		n += sprintf(item_addr+n, 
"<?xml version='1.0' encoding='utf-8'?>\n"
"<soap12:Envelope xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
"xmlns:xsd='http://www.w3.org/2001/XMLSchema' "
"xmlns:soap12='http://www.w3.org/2003/05/soap-envelope'>\n"
"  <soap12:Body>\n    <%s xmlns='http://%s'>\n", 
			ritem->type_name, t->area->send_station->qos.hostname);
		rc = WriteXmlSoapItem(ritem, t, data, item_addr+n, 
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		n += sprintf(item_addr+n,
"    </%s>\n  </soap12:Body>\n</soap12:Envelope>\n",
			ritem->type_name);
		break;
	case TA_SOAP_GET_CLIENT:
		n = sprintf(item_addr, "GET /%s/%s?", 
			t->rtti->iface_name, ritem->type_name);
		rc = WriteLineSoapItem(ritem, t, data, item_addr+n,
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		if (item_addr[n-1] == '&') {
			item_addr[n-1] = 0;
			--n;
		}
//		n += sprintf(item_addr+n, " HTTP/1.1\nHost: %s\n",
	//		t->area->name);
		n += sprintf(item_addr+n, " HTTP/1.1\n");
		n += sprintf(item_addr+n, 
"Accept: text/html, application/xhtml+xml, */*\n"
"Host: %s\n\n", t->area->send_station->qos.hostname);
		break;
//"Accept-Encoding: gzip, deflate\n"
//"Accept-Language: ru-RU\n"
//"User-Agent: Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)\n"
//"Connection: Keep-Alive\n"
	case TA_SOAP_POST_CLIENT:
		n = sprintf(item_addr, "POST /%s/%s HTTP/1.1\nHost: %s\n", 
			t->rtti->iface_name, ritem->type_name,
			t->area->send_station->qos.hostname);
		n += sprintf(item_addr+n, 
"Accept: text/html, application/xhtml+xml, */*\n"
"Cache-Control: no-cache\n");
		n += sprintf(item_addr+n, 
"Content-Type: application/x-www-form-urlencoded\nContent-Length:       ");
		nlen = n-6;
		n += sprintf(item_addr+n, "\n\n");
		n0 = n;
		rc = WriteLineSoapItem(ritem, t, data, item_addr+n,
			buffer_size, 0);
		if (rc < 0)
			break;
		n += rc;
		if (item_addr[n-1] == '&') {
			item_addr[n-1] = 0;
			--n;
		}
		break;
	default:
		rc = TA_NOT_FOUND;
	}
	if (rc >= 0 && nlen >= 0) {
		sprintf(buf, "%d", n-n0);
		strncpy(item_addr+nlen, buf, strlen(buf));
	}
	if (ta_trace_level & TA_TRAL_MESSAGE_DATA) {
		printf("--- OUT SOAP len:%d ---\n", n);
		printf("%s\n", item_addr);
		printf("--- END OF SOAP '%s' ---\n", ritem->type_name);
	}
//	printf(
//"taWriteSoapItem rc:%d mid_resp:%d name:'%s' item_type:'%s' size:%d/%d\n",
//		rc, t->mid_resp, ritem->item_name, ritem->type_name, n, 
//		buffer_size);
	if (rc < 0)
		return rc;
	//return n+1;
	return n;
}

/* ------------------------------------------------------- */
int taUnquoteBuffer(char *buffer, int buffer_length, int content_length)
{
	char *ps, *pd, *pc;
	int nns;
	//if (content_length <= 0)
		//return 0;
	
	ps = strstr(buffer, "Content-Length:");
	if (!ps)
		return 0;
	if (content_length <= 0) {
		ps +=15;
		while (isspace(*ps)) ++ps;
		content_length = atoi(ps);
		if (content_length <= 0)
			return 0;
	}
	for (pd = ps, pc = NULL, nns = 0; *ps; ps++) {
		if (*ps == '\n') {
			if (nns == 0) {
				if (!pc) {
					pc = ps+1;
					pc[content_length] = 0;
				}
			}
			nns = 0;
			*pd++ = *ps;
		} else if (isspace(*ps)) {
			*pd++ = *ps;
		} else if (*ps == '&') {
			if (!strncmp(ps+1, "lt;", 3)) {
				*pd++ = '<'; ps += 3;
			} else if (!strncmp(ps+1, "gt;", 3)) {
				*pd++ = '>'; ps += 3;
			} else if (!strncmp(ps+1, "amp;", 4)) {
				*pd++ = '&'; ps += 4;
			} else
				*pd++ = *ps;
			++nns;
		} else {
			*pd++ = *ps;
			++nns;
		}
	}
	*pd = 0;
	/*if (pc) {
		if ((long)pc+content_length >= (long)buffer+buffer_length)
			return TA_ERR_OUT_OF_MEMORY;
printf("pc:%c%c%c\n", pc[content_length]-1, pc[content_length], pc[content_length]+1);
		pc[content_length] = 0;
	}*/
	if (ta_trace_level & TA_TRAL_MESSAGE_DATA) {
		fprintf(ta_file,"&&& IN unquoted SOAP &&&\n");
		fprintf(ta_file,"%s\n", buffer);
		fprintf(ta_file,"&&& END OF SOAP &&&\n");
	}
	return 1;
}
