MODULE Ta_Areas;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Ti := Ta_TInfo, Ts := Ta_Stations, Log;
	
	CONST 
		AS_INPROC_LIB* = 1; AS_TARUN_LIB* = 2; AS_CONSOLE_APP* = 3; AS_GUI_APP* = 4;
		NORMAL_FINISH* = 1; FINISH_EXPIRED* = 2; EMERGENCY_FINISH* = 3; 
		APPEND_AS_LIST* = -6;
		noWait* = 0; waitWritePermission* = -1; waitUntilTimeout* = -2; waitAnyMessage* = -3; 
			waitForResponse* = -4; repeatForeachResponse* = -5; request* = -6; 
			publish* = -7; subscribe* = -8;
		
	
	TYPE
		AreaQos* = RECORD (Tb.Msg)
			messages_array_size* : Tb.Int32;
			send_buffer_size* : Tb.Int32;
			recv_buffer_size* : Tb.Int32;
			max_poll_fails* : Tb.Int32;
			soap_buffer_size* : Tb.Int32;
			thread_mem_size* : Tb.Int32;
			polling*: Tb.Name16;
		END;
		DirectoryQos* = RECORD (Tb.Msg)
			cycle*, expired*, collect_cycle*: LONGINT;
			with_compensation*: BOOLEAN;
			finish_within*: LONGINT;
		END;
		GetDirectoryQos* = RECORD (Tb.Msg)
			dqos*: DirectoryQos
		END;
		
		Table* = POINTER TO TableDesc;
		Area* = POINTER TO AreaDesc;
		Directory* = POINTER TO DirectoryDesc;

		TableDesc* =  RECORD
			ind*: INTEGER;
			ptr*: Tb.AnyPtr;
			ti*: Ti.TInfo;
		END;
		ProData* = RECORD
			sid*, res*: INTEGER;
			got_message*, changed*: BOOLEAN
		END;
		AreaDesc* =  RECORD
		END;
		GetAreaRunTimes* = RECORD (Tb.Msg)
			t_setpoint*, t_crt*, t_crtu*, tp_cdt*: Tb.Time
		END;
		
		ModuleData* = RECORD
			name-, appname-: Tb.Name;
			initproc-: Tb.CallProc;
			ti-: Ti.TInfo;
		END;
		
		DirectoryDesc* =  RECORD
			(* ---------- Application Data ---------- *)
			exe_file_name-, current_dir-, command_line-: Tb.Name;
			ind_command_line- : INTEGER;
			(* ---------- Module Data ---------- *)
			default_mod-: ModuleData;
		END;
		
		OnDataHandler* = PROCEDURE (t: Table; prod: ProData);
		TableCommand* = PROCEDURE (t: Table; prod: ProData; par: Tb.AnyPtr);
		
		AreaUpdater* = PROCEDURE (area: Area; par: Tb.AnyPtr);

		
	VAR
		dir-, stdDir-: Directory;
		defaultAQos-: AreaQos;
		defaultDQos-: DirectoryQos;
		ALL-: SET;
		
		
	(* ========== Table interface ========== *)
	PROCEDURE (t: Table) TInsert* (ind: INTEGER; name: ARRAY OF CHAR; 
	typ: Tb.Type; vis, length: INTEGER; tinfo: Ti.TInfo);
	BEGIN END TInsert;
	PROCEDURE (t: Table) Available* (): BOOLEAN;
	BEGIN RETURN FALSE END Available;
	PROCEDURE (t: Table) Clear* (VAR data: Tb.AnyRec);
	BEGIN END Clear;
	PROCEDURE (t: Table) Call* (command: TableCommand; par: Tb.AnyPtr);
	BEGIN END Call;
	PROCEDURE (t: Table) Write* (data-: Tb.AnyRec; wait_write_permission: BOOLEAN);
	BEGIN END Write;
	PROCEDURE (t: Table) WriteSid* (data-: Tb.AnyRec; wait_write_permission: BOOLEAN;
	sid: INTEGER); BEGIN END WriteSid;
	PROCEDURE (t: Table) WriteTo* (data-: Tb.AnyRec; sids: SET);
	BEGIN END WriteTo;
	PROCEDURE (t: Table) Wait* (waiting_expired: LONGINT);
	BEGIN END Wait;
	PROCEDURE (t: Table) Repeat* (command: TableCommand; par: Tb.AnyPtr);
	BEGIN END Repeat;
	PROCEDURE (t: Table) Request* (data-: Tb.AnyRec; wait_write_permission: BOOLEAN;
	command: TableCommand; par: Tb.AnyPtr);
	BEGIN END Request;
	PROCEDURE (t: Table) Publish* (permanent: BOOLEAN; command: TableCommand; par: Tb.AnyPtr);
	BEGIN END Publish;
	PROCEDURE (t: Table) Subscribe* (controlled: BOOLEAN; command: TableCommand; par: Tb.AnyPtr);
	BEGIN END Subscribe;
	PROCEDURE (t: Table) Unsubscribe* (inform_publisher: BOOLEAN);
	BEGIN END Unsubscribe;
	PROCEDURE (t: Table) HandleMsg* (VAR msg: Tb.Msg); 
	BEGIN END HandleMsg;

	(* ========== Area interface ========== *)
	PROCEDURE (area: Area) Start* ();
	BEGIN END Start;
	PROCEDURE (area: Area) StartNoUpdate* ();
	BEGIN END StartNoUpdate;
	PROCEDURE (area: Area) Finish* (finish_flag: INTEGER);
	BEGIN END Finish;
	PROCEDURE (area: Area) Update* ();
	BEGIN END Update;
	PROCEDURE (area: Area) AddStations* ( name-: ARRAY OF CHAR; 
	n_appended, host_sid: INTEGER; VAR cqos: Tc.ChannelQosBase;
	(*OUT*)VAR res: INTEGER) ;
	BEGIN END AddStations;
	PROCEDURE (area: Area) NStations* (): INTEGER;
	BEGIN RETURN 0 END NStations;
	PROCEDURE (area: Area) HostSid* (): INTEGER;
	BEGIN RETURN 0 END HostSid;
	PROCEDURE (area: Area) Station* (sid: INTEGER): Ts.Station;
	BEGIN RETURN NIL END Station;
	PROCEDURE (area: Area) AddUpdater* (on_update: AreaUpdater; par: Tb.AnyPtr;
	period: LONGINT);
	BEGIN END AddUpdater;
	PROCEDURE (area: Area) NewTable* (VAR rec: Tb.AnyRec; 
	 iface_name-: ARRAY OF CHAR;  tname-: ARRAY OF CHAR;
	sid, oid: INTEGER; VAR tqos: Ts.TableQosBase; 
	on_data: OnDataHandler): Table; 
	BEGIN RETURN NIL END NewTable;
	PROCEDURE (area: Area) Message* (sid, src_oid: INTEGER; VAR msg: Tb.Msg);
	BEGIN END Message; 

	(* ========== Directory interface ========== *)
	PROCEDURE (d: Directory) NewArea* (aid: INTEGER; name: ARRAY OF CHAR;
	max_stations: INTEGER;  aqos-: AreaQos) : Area;
	BEGIN RETURN NIL END NewArea;
	PROCEDURE (d: Directory) InitHostModule* ( dqos-: DirectoryQos);
	BEGIN END InitHostModule;
	PROCEDURE (d: Directory) LoadedAs* () : INTEGER;
	BEGIN RETURN 0 END LoadedAs;

	PROCEDURE (d: Directory) InitModule* (modl, app_fn: ARRAY OF CHAR; 
	initproc: Tb.CallProc; ti: Ti.TInfo;  dqos-: DirectoryQos) ;
	BEGIN
		Tb.Copy(modl, d.default_mod.name);
		Tb.Copy(app_fn, d.default_mod.appname);
		d.default_mod.initproc := initproc;
		d.default_mod.ti := ti;
		d.InitHostModule(dqos);
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Log.String("InitModule "); Log.String(modl); Log.Char(' '); 
			Log.String(app_fn); Log.Ln;
			Log.String(d.exe_file_name); Log.Char('|');
			Log.String(d.current_dir); Log.Char('|');
			Log.String(d.command_line); Log.Char('|');
			Log.Int(d.ind_command_line); Log.Ln
		END
	END InitModule;

	(* ========== Qos Management ========== *)		
	PROCEDURE DefaultAreaQos (VAR qos: AreaQos);
	BEGIN
		qos.messages_array_size := 100;
		qos.send_buffer_size := 65536;
		qos.recv_buffer_size := 65536;
		qos.max_poll_fails := 1000;
		qos.soap_buffer_size := 32768;
		qos.thread_mem_size := 32768;
		qos.polling := "Poll";
	END DefaultAreaQos;

	PROCEDURE DefaultDirectoryQos (VAR qos: DirectoryQos);
	BEGIN
		qos.cycle := 1000000;
		qos.expired := 0;
		qos.collect_cycle := -1;
		qos.with_compensation := TRUE;
		qos.finish_within := 3000000
	END DefaultDirectoryQos;

	(* ========== Installation ========== *)		
	PROCEDURE SetDir* (d: Directory; exe_file_name, current_dir, command_line: ARRAY OF CHAR; 
	ind_command_line: INTEGER);
	BEGIN
		ASSERT(d # NIL, Tb.NIL_POINTER);
		dir := d;
		IF stdDir = NIL THEN stdDir := d END;

		Tb.Copy(exe_file_name, dir.exe_file_name); Tb.Copy(current_dir, dir.current_dir); 
		Tb.Copy(command_line, dir.command_line); dir.ind_command_line := ind_command_line;
	END SetDir;

	PROCEDURE Initialize;
		VAR j: INTEGER;
	BEGIN
		DefaultAreaQos(defaultAQos);
		DefaultDirectoryQos(defaultDQos);
		ALL := Tb.Int32_Set(-1)
	END Initialize;
	
BEGIN
	Initialize
END Ta_Areas.
