MODULE Ta_Basics;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, O2Strings, ConvTypes, WholeStr, RealStr, LongStr, Timelib := time, 
	Tl := Ta_Lib, Log;

	CONST
		(** type classes **)
		undef* = 0; boolTyp* = 1; char8Typ* = 2; char16Typ* = 3;
		int8Typ* = 4; int16Typ* = 5; int32Typ* = 6;
		real32Typ* = 7; real64Typ* = 8; setTyp* = 9;
		int64Typ* = 10; anyRecTyp* = 11; anyPtrTyp* = 12;
		timeTyp* = int64Typ; sysPtrTyp = 13; 
		procTyp* = 16; recTyp* = 17; arrTyp* = 18; ptrTyp* = 19;
		charTyp* = char8Typ; intTyp* = int16Typ; longTyp* = int32Typ; realTyp* = real32Typ;
		(* visibility *)
		hidden* = 1; readOnly* = 2; private = 3; exported* = 4;
		MAX_INT32 = 2147483647; MAX_UINT32 = 4294967295;
		NUM_MCS_IN_1 = 429496729; NUM_MS_IN_1 = 429496;
		WINDOWS_TICKS=10000000; UINT32_TICKS = 429.4967295;
		EPOCH_DIFF=11644473600;
		tYY* = 1; tMM* = 2; tDD* = 3; tWW* = 4; thh* = 5; tmm* = 6; tss* = 7; tms* = 8; 
		tloc* = 9; tdwh* = 10; tdwl* = 11; tyy* = 12; t0* = 13;
		tsDateTime* = {tYY,tMM,tDD,thh,tmm,tss,tloc}; tsDate* = {tYY,tMM,tDD,tloc}; 
		tsThm* = {thh,tmm}; tsThms* = {thh,tmm,tss}; tsThmst* = {thh,tmm,tss,tms};

		NAME_LENGTH* = 256; NAME_64_LENGTH* = 64; NAME_16_LENGTH* = 16;

		(* ---------- Trace level bits ---------- *)
		TRAL_RUNTIME* = 0; TRAL_TRANSPORT* = 1; TRAL_POLLING* = 2;
		TRAL_MESSAGE* = 3; TRAL_MESSAGE_DATA* = 4; TRAL_ERRORS* = 5;

		(*	---------- ASSERTIONS in TA: ---------- *)
		WRONG_TYPE* = 20; STRING_CONVERSION_ERROR* = 21; 
		NOT_FOUND* = 22; CHECK_RANGE_ERROR* = 23; OBJECT_INTEGRITY* = 24;
		PARSE_ERROR* = 25; ZERO_VAL* = 26; NIL_POINTER* = 27; NOT_EXPORTED* = 28; 
		ALREADY_INITIALIZED* = 29; INVALID_VAL* = 30; UNSUPPORTED* = 31;
		ZERO_RESULT* = 60; NIL_RESULT* = 61; INIT_FAILED* = 62; 
		INVALID_EXTENSION* = 63; TRUNCATED* = 64;
		INTERNAL_ERROR* = 100; 
		
		littleEndian* = TRUE;
		
		
	TYPE
		(* ---------- Basic types ---------- *)
		Bool* = BOOLEAN;
		Time_t = SYSTEM.CARD32;
		FILETIME = RECORD
			dwLowDateTime, dwHighDateTime: Time_t;
		END;
		SystemTime* = RECORD
			wYear, wMonth, wDay, wDayOfWeek, wHour, wMinute, 
			wSecond, wMilliseconds:
				LONGINT
		END;
		Tm = Timelib.tm_ptr;

		Int8* = SHORTINT;
		Char8* = CHAR;
		Char16* = INTEGER;
		Int16* = INTEGER;
		Int32* = LONGINT;
		UInt32* = SYSTEM.CARD32;
		Int64* = RECORD
			l*, h*: LONGINT
		END;
		Real32* = REAL;
		Real64* = LONGREAL;
		Time* = FILETIME;
		Address* = LONGINT;
		AnyRec* = RECORD END;
		Empty* = RECORD (AnyRec) END;
		AnyPtr* = POINTER TO AnyRec;
		Msg* = RECORD (AnyRec) END;
		Buffer* = ARRAY OF SHORTINT;
		PBuffer* = POINTER TO Buffer;
		
		ClientData* = Address;
		Type* = INTEGER;
		Name* = ARRAY NAME_LENGTH OF CHAR;
		Name64* = ARRAY NAME_64_LENGTH OF CHAR;
		Name16* = ARRAY NAME_16_LENGTH OF CHAR;
		SString* = ARRAY NAME_LENGTH OF Char8;
		String* = ARRAY NAME_LENGTH OF Char16;
		CallProc* = PROCEDURE;
		CallProc1* = PROCEDURE (s: ARRAY OF CHAR);
		PStr8* = POINTER TO ARRAY OF Char8;
		PStr16* = POINTER TO ARRAY OF Char16;
		PStr* = POINTER TO ARRAY OF CHAR;
		EmptyPtr* = POINTER TO Empty;
		Bits8* = ARRAY 1 OF Int8;
		Bits16* = ARRAY 2 OF Int8;
		Bits32* = ARRAY 4 OF Int8;
		Bits64* = ARRAY 8 OF Int8;
		
		CompatibleTypes* = ARRAY ptrTyp+1 OF Name16;
		
	VAR
		trace-: SET;
		trace_level-: Int32;
		INFINITE_TIME-, ZERO_TIME-: Time;
		crc_table: ARRAY 256 OF Int32;
		no_rec*: AnyRec;
		compatible_types-: CompatibleTypes;
		z: Int32;


		(* ---------- Compatible types conversions ---------- *)
	PROCEDURE Clean* (addr, len: Int32);
		VAR j: Int32; z0: Int8;
	BEGIN
		z0 := 0;
		FOR j := 0 TO len-1 DO SYSTEM.PUT(addr+j, z0) END
	END Clean;
	
	PROCEDURE ["C"] / __errno_location (): LONGINT;
	
	PROCEDURE Errno* () : LONGINT;
		VAR addr, errno: LONGINT;
	BEGIN
		(* RETURN errno.errno *)
		addr := __errno_location();
		SYSTEM.GET(addr, errno);
		RETURN errno
	END Errno;

	PROCEDURE Long_Int64* (lv: LONGINT) : Int64;
		VAR lv64: Int64;
	BEGIN lv64.h := 0; lv64.l := lv; RETURN lv64 END Long_Int64;

	PROCEDURE Long_Int32* (lv: LONGINT) : Int32;
	BEGIN RETURN lv END Long_Int32;
	
	PROCEDURE Int64_Int32* (lv: Int64) : Int32;
	BEGIN RETURN lv.l END Int64_Int32;

	PROCEDURE Int_Int16* (lv: INTEGER) : Int16;
	BEGIN RETURN lv END Int_Int16;
		
	PROCEDURE C16_C* (ch: Char16): CHAR;
	BEGIN RETURN SYSTEM.VAL(CHAR, ch) END C16_C;
	
	PROCEDURE C_C16* (ch: CHAR): Char16;
	BEGIN RETURN ORD(ch) END C_C16;
	
	PROCEDURE C_C8* (ch: CHAR): CHAR;
	BEGIN RETURN ch END C_C8;

	PROCEDURE Char16_Char* (src-: ARRAY OF Char16; 
	VAR dst: ARRAY OF CHAR);
		VAR j: INTEGER;
	BEGIN
		j := 0;
		WHILE src[j] # 0 DO
			dst[j] := SYSTEM.VAL(CHAR, src[j]); INC(j)
		END;
		dst[j] := 0X
	END Char16_Char;

	PROCEDURE Char_Char16* (src-: ARRAY OF CHAR; 
	VAR dst: ARRAY OF Char16);
		VAR j: INTEGER;
	BEGIN
		j := 0;
		WHILE src[j] # 0X DO
			dst[j] := ORD(src[j]); INC(j)
		END;
		dst[j] := 0
	END Char_Char16;

	PROCEDURE Char8_Char* (src-: ARRAY OF Char8; 
	VAR dst: ARRAY OF CHAR);
	BEGIN
		COPY(src, dst)
	END Char8_Char;

	PROCEDURE Char_Char8* (src-: ARRAY OF Char8; 
	VAR dst: ARRAY OF CHAR);
	BEGIN
		COPY(src, dst)
	END Char_Char8;

	PROCEDURE Int32_Set* (iv: Int32) : SET;
		VAR i: INTEGER; sv: SET;
	BEGIN
		RETURN SYSTEM.VAL(SET, iv)
	END Int32_Set;

	PROCEDURE Set_Int32* (sv: SET) : Int32;
	BEGIN
		RETURN SYSTEM.VAL(Int32, sv)
	END Set_Int32;

	PROCEDURE Trace* (trace_lev: Int32);
	BEGIN
		trace_level := trace_lev;
		trace := Int32_Set(trace_lev);
		Log.String("New trace_level is "); Log.Int(trace_level); Log.Ln
	END Trace;
	
			
		(* ---------- Ta routines for string manipulations ---------- *)
	PROCEDURE Concat* (src1, src2: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR);
		VAR i, j: INTEGER;
	BEGIN
		i := 0; j := 0;
		WHILE src1[i] # 0X DO dst[i] := src1[i]; INC(i) END;
		WHILE src2[j] # 0X DO dst[j+i] := src2[j]; INC(j) END;
		dst[i+j] := 0X
	END Concat;
	
	PROCEDURE Search* (p: ARRAY OF CHAR;  s-: ARRAY OF CHAR; start, M, N: INTEGER;  
	VAR r: INTEGER);
		VAR  i, j, k: INTEGER; d: ARRAY 128 OF INTEGER;
	BEGIN 
		FOR  i := 0  TO  127  DO  d[i] := M  END; 
		FOR  j := 0  TO  M-2  DO  d[ ORD( p[j] ) ] := M-j-1  END;

		i := M; j := M;  k := i;
		LOOP IF  (j > 0) & (i <= N) & (s[start+k-1] = p[j-1])  THEN
			DEC(k); DEC(j);
		ELSIF  (j > 0) & (i <= N)  THEN
			i := i + d[ ORD( s[start+i-1] ) ];  j := M;  k := i;
		ELSE EXIT END END;
		
		IF  j = 0  THEN  r := start+k  ELSE  r := -1  END;
	END Search;

	PROCEDURE CapEqual* (s1, s2: ARRAY OF CHAR): BOOLEAN;
		VAR j: INTEGER;
	BEGIN 
		j := -1;
		REPEAT
			INC(j)
		UNTIL (CAP(s1[j]) # CAP(s2[j])) OR (s1[j] = 0X);
		RETURN CAP(s1[j]) = CAP(s2[j])
	END CapEqual;

	PROCEDURE Equal* (s1, s2: ARRAY OF CHAR): BOOLEAN;
	BEGIN RETURN s1 = s2 END Equal;
	
	PROCEDURE Copy* (src: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR);
	BEGIN COPY(src, dst) END Copy;

	PROCEDURE Copy8* (src: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR);
	BEGIN COPY(src, dst) END Copy8;

	PROCEDURE Copy16* (src: ARRAY OF Char16; VAR dst: ARRAY OF Char16);
		VAR j: INTEGER;
	BEGIN
		j := 0;
		WHILE src[j] # 0 DO dst[j] := src[j]; INC(j) END;
		dst[j] := 0
	END Copy16;

	PROCEDURE Len* ( str-: ARRAY OF CHAR) : INTEGER;
	BEGIN RETURN O2Strings.Length(str) END Len;
	(** Returns the number of characters in s up to and excluding the first 0X *)
	PROCEDURE Len8* ( str-: ARRAY OF CHAR) : INTEGER;
	BEGIN RETURN O2Strings.Length(str) END Len8;
	(** Returns the number of characters in s up to and excluding the first 0X *)
	
	PROCEDURE Insert* (src: ARRAY OF CHAR; pos: INTEGER; VAR dst: ARRAY OF CHAR);
	BEGIN O2Strings.Insert(src, pos, dst) END Insert;
	(** Inserts the string s into the string d at position pos
  	(0 <= pos <= Length(d)). If pos = Length(d), s is appended to d.
  	If the size of d is not large enough to hold the result of operation,
  	the result is truncated so that d is always terminated with 0X.
	*)

	PROCEDURE Append* (src: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR);
	BEGIN O2Strings.Append(src, dst) END Append;
	(** Has the same effect as Insert(s,Length(d),d) *)
	
	PROCEDURE Delete* (VAR str: ARRAY OF CHAR; pos, n: INTEGER);
	BEGIN O2Strings.Delete(str, pos, n) END Delete;
	(** Deletes n characters from str starting at position pos
  	(0 <= pos < Length(d)). If n > Length(s)-pos, the new
  	length of s is pos.
	*)
	
	PROCEDURE Replace* (s: ARRAY OF CHAR; pos: INTEGER; VAR d: ARRAY OF CHAR);
	BEGIN O2Strings.Replace(s, pos, d) END Replace;
	(** Has the same effect as Delete(d,pos,Length(s)) followed by
  	an Insert(s,pos,d).
	*)
	
	PROCEDURE Extract* (s: ARRAY OF CHAR; pos, n: INTEGER; VAR d: ARRAY OF CHAR);
	BEGIN O2Strings.Extract(s, pos, n, d) END Extract;
	(** Extracts a substring d with n characters from position pos
  	(0 <= pos < Length(s)) in s. If n > Length(s) - pos, d is only
  	the part of s from pos to Length(s)-1.
  	If the size of d is not large enough to hold the result of operation,
  	the result is truncated so that d is always terminated with 0X.
	*)
	
	PROCEDURE Pos* (p: ARRAY OF CHAR;  s-: ARRAY OF CHAR; start: INTEGER): INTEGER;
	BEGIN RETURN O2Strings.Pos(p, s, start) END Pos;
	(** Returns the position of the first occurences of the pattern p
  	in s after position pos (inclusive). If p is not found, -1 is
  	returned.
	*)
	
	PROCEDURE Cap* (VAR str: ARRAY OF CHAR);
	BEGIN O2Strings.Cap(str) END Cap;
	(** Replaces each lower case letter within s by its upper case
  	equivalent.
	*)
	
	PROCEDURE StringToInt* ( s-: ARRAY OF CHAR; (*OUT*)VAR x: INTEGER; (*OUT*)VAR res: INTEGER);
		VAR cr: ConvTypes.ConvResults; lx: LONGINT;
	BEGIN 
		WholeStr.StrToInt (s, lx, cr);
		x := SYSTEM.VAL(INTEGER, lx);
		res := SYSTEM.VAL(INTEGER, cr);
	END StringToInt;
	
	PROCEDURE StringToLInt* ( s-: ARRAY OF CHAR; (*OUT*)VAR x: LONGINT; (*OUT*)VAR res: INTEGER);
		VAR cr: ConvTypes.ConvResults;
	BEGIN 
		WholeStr.StrToInt (s, x, cr);
		res := SYSTEM.VAL(INTEGER, cr);
	END StringToLInt;
	
	PROCEDURE StringToInt32* ( s-: ARRAY OF CHAR; (*OUT*)VAR x: LONGINT; (*OUT*)VAR res: INTEGER);
		VAR cr: ConvTypes.ConvResults;
	BEGIN 
		WholeStr.StrToInt (s, x, cr);
		res := SYSTEM.VAL(INTEGER, cr);
	END StringToInt32;
	
	PROCEDURE StringToReal* ( s-: ARRAY OF CHAR; (*OUT*)VAR x: REAL; (*OUT*)VAR res: INTEGER);
		VAR cr: ConvTypes.ConvResults;
	BEGIN 
		RealStr.StrToReal (s, x, cr);
		res := SYSTEM.VAL(INTEGER, cr);
	END StringToReal;
	
	PROCEDURE StringToLReal* ( s-: ARRAY OF CHAR; (*OUT*)VAR x: LONGREAL; 
	(*OUT*)VAR res: INTEGER);
		VAR cr: ConvTypes.ConvResults;
	BEGIN 
		LongStr.StrToReal (s, x, cr);
		res := SYSTEM.VAL(INTEGER, cr);
	END StringToLReal;

	PROCEDURE IntToString* (x: LONGINT; (*OUT*)VAR s: ARRAY OF CHAR);
	BEGIN
		WholeStr.IntToStr(x, s)
	END IntToString;
	
	PROCEDURE RealToString* (x: REAL; (*OUT*)VAR s: ARRAY OF CHAR);
	BEGIN 
		RealStr.RealToStr(x, s); s[15] := 0X
	END RealToString;

	PROCEDURE LRealToString* (x: LONGREAL; (*OUT*)VAR s: ARRAY OF CHAR);
	BEGIN 
		LongStr.RealToStr(x, s); s[31] := 0X
	END LRealToString;

	PROCEDURE NewString16* ( s-: ARRAY OF Char16): PStr16;
		VAR ps: PStr16; j: INTEGER;
	BEGIN
		NEW(ps, LEN(s)); j := 0;
		WHILE s[j] # 0 DO ps[j] := s[j]; INC(j) END;
		ps[j] := 0;
		RETURN ps
	END NewString16;
	
	PROCEDURE IntToStringZero* (x: LONGINT; len: INTEGER; (*OUT*)VAR s: ARRAY OF CHAR);
		VAR name: Name; j, l: INTEGER;
	BEGIN
		IntToString(x, name);
		l := Len(name);
		IF l < len THEN
			FOR j := 0 TO len-l-1 DO s[j] := '0' END; s[j] := 0X;
			Append(name, s)
		ELSE
			Copy(name, s)
		END
	END IntToStringZero;

	PROCEDURE NewString8* ( s-: ARRAY OF Char8; 
	(*OUT*)VAR ps: POINTER TO ARRAY OF Char8);
		VAR j: INTEGER;
	BEGIN
		NEW(ps, LEN(s)); j := 0;
		WHILE s[j] # 0X DO ps[j] := s[j]; INC(j) END;
		ps[j] := 0X;
	END NewString8;
	
	PROCEDURE StringToBool* ( s-: ARRAY OF CHAR; VAR x: BOOLEAN; VAR res: INTEGER);
		VAR i: INTEGER;
	BEGIN
		i := 0; res := 0;
		WHILE (s[i] # 0X) & ((s[i] = ' ') OR (s[i] = '$')) DO INC(i) END;
		CASE s[i] OF
		'T', 't': x := TRUE;
		|'F', 'f': x := FALSE; 
		ELSE
			res := 1
		END
	END StringToBool;
	
	PROCEDURE AppendAt* (src: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR; VAR index: INTEGER);
		VAR i: INTEGER;
	BEGIN
		i := 0;
		IF index < 0 THEN index := Len(dst) END;
		REPEAT dst[i+index] := src[i]; INC(i) UNTIL src[i-1] = CHR(0);
		INC(index, i-1);
	END AppendAt;

	PROCEDURE Scan* (src: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR; VAR pos: INTEGER);
		VAR i: INTEGER;
	BEGIN
		i := pos;
		WHILE (src[i] # 0X) & (src[i] # ' ') & (src[i] # 09X) DO
			dst[i-pos] := src[i];
			INC(i)
		END;
		dst[i-pos] := 0X;
		IF src[i] # 0X THEN 
			WHILE (src[i] = ' ') OR (src[i] = 09X) DO INC(i) END
		END;
		pos := i
	END Scan;

		(* ---------- Ta Timer routines ---------- *)
	PROCEDURE Time_Sec* (t: Time; VAR sec: UInt32);
		VAR ht, l32, h32: UInt32;
	BEGIN
		(* 27111902, 3577643008, 4294967296
		ht := (t - EPOCH_DIFF(11644473600)*10000000) DIV 10000000;
		sec := ht; *)
		IF t.dwHighDateTime = 0 THEN
			sec := t.dwLowDateTime DIV 10000000;
		ELSE
			l32 := t.dwLowDateTime; h32 := t.dwHighDateTime;
			ASM
				MOV EAX, l32
				MOV EDX, h32
				SUB EAX, 3577643008
				JNC ss
				DEC EDX
		ss:	SUB EDX, 27111902  (*t - 11644473600*1e7*)
				MOV EBX, 10000000
				DIV EBX
				MOV ht, EAX  (* ht := SHORT() DIV 1e7*)
			END;
			sec := ht
		END
	END Time_Sec;
	
	PROCEDURE Time_Mcs* (t: Time; VAR mcs: Int32);
		VAR l32, h32: UInt32; mc: Int32;
	BEGIN
		(* 27111902, 3577643008, 4294967296
			(t - EPOCH_DIFF(11644473600)*10000000) MOD 10000000 DIV 10 *)
		IF t.dwHighDateTime = 0 THEN
			mcs := t.dwLowDateTime MOD 10000000 DIV 10;
		ELSE
			l32 := t.dwLowDateTime; h32 := t.dwHighDateTime;
			ASM
				MOV EAX, l32
				MOV EDX, h32
				SUB EAX, 3577643008
				JNC sm
				DEC EDX
		sm:	SUB EDX, 27111902  (*t - 11644473600*1e7*)
				MOV EBX, 10000000
				DIV EBX		(* t MOD 10000000 *)
				MOV EAX, EDX
				MOV EDX, 0
				MOV EBX, 10
				DIV EBX
				MOV mc, EAX  (* mc := () DIV 10*)
			END;
			mcs := mc
		END
	END Time_Mcs;
	
	PROCEDURE SecMcs_Time* (sec: UInt32; mcs: Int32; from1970: BOOLEAN;
	VAR t: Time);
		VAR l32, h32, c1, c2: UInt32;
	BEGIN
		(* 2, 3054539008, 4294967296
		t64 := EPOCH_DIFF(11644473600);
		INC(t64, sec);
		t64 := t64*10000000;
		t64 := t64 + mcs*10; *)
		IF from1970 THEN
			c1 := 3054539008; c2 := 2
		ELSE
			c1 := 0; c2 := 0
		END;
		ASM
			MOV EAX, c1
			MOV EDX, c2      (* t64 := 11644473600 *)
			ADD EAX, sec
			ADC EDX, 0      (* INC(t64,sec) *)
			MOV c1, EBX
			MOV EBX, EDX
			MOV ECX, 10000000
			MUL ECX
			MOV l32, EAX
			MOV ECX, EDX
			MOV EAX, EBX
			MOV EBX, 10000000
			MUL EBX
			ADD EAX, ECX
			MOV EBX, EAX    (* t64 := t64*10000000 *)
			MOV EAX, mcs
			MOV ECX, 10
			MUL ECX
			ADD EAX, l32
			ADC EBX, 0      (* t64 := t64 + mcs*10 *)
			MOV l32, EAX
			MOV h32, EBX
			MOV EBX, c1
		END;
		t.dwLowDateTime := l32; t.dwHighDateTime := h32
	END SecMcs_Time;
	
	PROCEDURE Long_Time* (l: LONGINT; VAR t: Time);
		VAR li: UInt32;
	BEGIN
		t.dwHighDateTime := l DIV NUM_MCS_IN_1;
		li := l MOD NUM_MCS_IN_1;
		t.dwLowDateTime := li * 10;
	END Long_Time;
	
	PROCEDURE AddTime* (t1, t2: Time; VAR t: Time);
		VAR t1l, t1h, t2l, t2h, tl, th: UInt32;
	BEGIN
		t1l := t1.dwLowDateTime; t1h := t1.dwHighDateTime;
		t2l := t2.dwLowDateTime; t2h := t2.dwHighDateTime;
		ASM
			MOV EAX, t1l
			MOV EDX, t1h
			ADD EAX, t2l
			ADC EDX, t2h
			MOV tl, EAX
			MOV th, EDX
		END;
		t.dwLowDateTime := tl;
		t.dwHighDateTime := th;
	END AddTime;

	(*PROCEDURE AddTime* (t1, t2: Time; VAR t: Time);
	BEGIN
		t.dwLowDateTime := t1.dwLowDateTime + t2.dwLowDateTime;
		t.dwHighDateTime := t1.dwHighDateTime + t2.dwHighDateTime;
		IF (t.dwLowDateTime < t1.dwLowDateTime) OR (t.dwLowDateTime < t2.dwLowDateTime) THEN
			t.dwHighDateTime := t.dwHighDateTime + 1
		END
	END AddTime;*)

	PROCEDURE SubtractTime* (t1, t2: Time; VAR t: Time);
		VAR t1l, t1h, t2l, t2h, tl, th: UInt32;
	BEGIN
		t1l := t1.dwLowDateTime; t1h := t1.dwHighDateTime;
		t2l := t2.dwLowDateTime; t2h := t2.dwHighDateTime;
		ASM
			MOV EAX, t1l
			MOV EDX, t1h
			SUB EAX, t2l
			SBB EDX, t2h
			MOV tl, EAX
			MOV th, EDX
		END;
		t.dwLowDateTime := tl;
		t.dwHighDateTime := th;
	END SubtractTime;

	(*PROCEDURE SubtractTime* (t1, t2: Time; VAR t: Time);
	BEGIN
		t.dwLowDateTime := t1.dwLowDateTime - t2.dwLowDateTime;
		t.dwHighDateTime := t1.dwHighDateTime - t2.dwHighDateTime;
		IF (t1.dwLowDateTime < t.dwLowDateTime) OR (t1.dwLowDateTime < t2.dwLowDateTime) THEN
			t.dwHighDateTime := t.dwHighDateTime - 1
		END
	END SubtractTime;*)

	PROCEDURE Time_SystemTime* (t: Time; VAR lt: SystemTime; VAR res: INTEGER);
		VAR tm: Tm; sec: UInt32; mcs: Int32;
	BEGIN
		Time_Sec(t, sec);
		tm := Timelib.gmtime(SYSTEM.VAL(Int32, sec));
		IF tm = NIL THEN
			res := 1
		ELSE
			lt.wYear := tm.tm_year + 1900;
			lt.wMonth := tm.tm_mon + 1;
			lt.wDay := tm.tm_mday;
			lt.wDayOfWeek := tm.tm_wday + 1;
			lt.wHour := tm.tm_hour;
			lt.wMinute := tm.tm_min;
			lt.wSecond := tm.tm_sec;
			Time_Mcs(t, mcs);
			lt.wMilliseconds := mcs DIV 1000;
			res := 0
		END;
	END Time_SystemTime;

	PROCEDURE Time_LocalTime* (t: Time; VAR lt: SystemTime; VAR res: INTEGER);
		VAR tm: Tm; sec: UInt32; mcs: Int32;
	BEGIN
		Time_Sec(t, sec);
		tm := Timelib.localtime(SYSTEM.VAL(Int32, sec));
		IF tm = NIL THEN
			res := 1
		ELSE
			lt.wYear := tm.tm_year + 1900;
			lt.wMonth := tm.tm_mon + 1;
			lt.wDay := tm.tm_mday;
			lt.wDayOfWeek := tm.tm_wday + 1;
			lt.wHour := tm.tm_hour;
			lt.wMinute := tm.tm_min;
			lt.wSecond := tm.tm_sec;
			Time_Mcs(t, mcs);
			lt.wMilliseconds := mcs DIV 1000;
			res := 0
		END;
	END Time_LocalTime;

	PROCEDURE Sleep* (tp: Time);
		VAR sec: UInt32; mcs: Int32; rc: Int32;
	BEGIN
		Time_Sec(tp, sec);
		WHILE sec >= 1000 DO
			rc := Tl.usleep(1000000000);
			sec := sec - 1000
		END;
		Time_Mcs(tp, mcs);
		rc := Tl.usleep(SYSTEM.VAL(Int32, sec*1000000)+mcs)
	END Sleep;
	
	PROCEDURE GetTime* (VAR t: Time);
		VAR tv: Tl.timeval; tvz: Tl.timezone; rc: LONGINT;
	BEGIN
		rc := Tl.gettimeofday(tv, tvz);
		SecMcs_Time(tv.tv_sec, tv.tv_usec, TRUE, t)
	END GetTime;	

	PROCEDURE TimeEquals* (t1, t2: Time) : BOOLEAN;
	BEGIN
		RETURN (t1.dwHighDateTime = t2.dwHighDateTime)
			& (t1.dwLowDateTime = t2.dwLowDateTime)
	END TimeEquals;
	
	PROCEDURE TimePositive* (t: Time) : BOOLEAN;
	BEGIN
		RETURN ((t.dwHighDateTime # 0) OR (t.dwLowDateTime # 0))
			& ((t.dwHighDateTime # INFINITE_TIME.dwHighDateTime)
			OR (t.dwLowDateTime # INFINITE_TIME.dwLowDateTime))
	END TimePositive;
	
	PROCEDURE CompareTime* (t1, t2: Time) : REAL;
		VAR dt: REAL;
	BEGIN
		IF t1.dwHighDateTime > t2.dwHighDateTime THEN
			dt := 429496729.5*(t1.dwHighDateTime-t2.dwHighDateTime);
		ELSE
			dt := -429496729.5*(t2.dwHighDateTime-t1.dwHighDateTime);
		END;
		RETURN dt + 0.1*t1.dwLowDateTime - 0.1*t2.dwLowDateTime;
	END CompareTime;
	
	PROCEDURE Time_String* (t: Time; tfmt: SET; VAR s: ARRAY OF CHAR);
		VAR st: SystemTime; res, f0: INTEGER; si: ARRAY 256 OF CHAR; iv: LONGINT;
	BEGIN
		s[0] := 0X; si[0] := 0X;
		IF tloc IN tfmt THEN
			Time_LocalTime(t, st, res)
		ELSE
			Time_SystemTime(t, st, res)
		END;
		IF res = 0 THEN
			IF t0 IN tfmt THEN f0 := 2 ELSE f0 := 0 END;
			IF tDD IN tfmt THEN iv := st.wDay; IntToStringZero(iv, f0, si); Append(si, s); Append("/", s) END;
			IF tMM IN tfmt THEN iv := st.wMonth; IntToStringZero(iv, f0, si); Append(si, s); Append("/", s) END;
			IF tYY IN tfmt THEN iv := st.wYear; IntToStringZero(iv, f0, si); Append(si, s); Append(" ", s) 
			ELSIF tyy IN tfmt THEN 
				iv := st.wYear MOD 100; IntToStringZero(iv, f0, si); Append(si, s); Append(" ", s) 
			END;
			IF thh IN tfmt THEN iv := st.wHour; IntToStringZero(iv, f0, si); Append(si, s) END;
			IF tmm IN tfmt THEN iv := st.wMinute; IntToStringZero(iv, f0, si); Append(":", s); Append(si, s) END;
			IF tss IN tfmt THEN iv := st.wSecond; IntToStringZero(iv, f0, si); Append(":", s); Append(si, s) END;
			IF tms IN tfmt THEN iv := st.wMilliseconds; IntToString(iv, si); Append(".", s); Append(si, s) END;
		ELSE
			Copy("?", s); IntToString(res, si); Append(si, s)
		END;
	END Time_String;
	

	PROCEDURE T64_Time* (t64: Int64; VAR t: Time);
	BEGIN
		t.dwLowDateTime := t64.l;
		t.dwHighDateTime := t64.h;
	END T64_Time;

	PROCEDURE Time_T64* (t: Time; VAR t64: Int64);
	BEGIN
		t64.l := t.dwLowDateTime;
		t64.h := t.dwHighDateTime;
	END Time_T64;

	PROCEDURE ParseHostPort* ( name-: ARRAY OF CHAR; 
	 hostname-: ARRAY OF CHAR; VAR outname: ARRAY OF CHAR;
	VAR port: Int16);
		VAR j, l, m: INTEGER; name1: Name64;
	BEGIN
		IF hostname[0] = 0X THEN
			Copy(name, outname)
		ELSE
			Copy(hostname, outname)
		END;
		l := Len(outname); j := l-1;
		WHILE (j >= 0) & (outname[j] # ':') DO DEC(j) END;
		IF j >= 0 THEN
			FOR m := 0 TO l-j-1 DO name1[m] := outname[m+j+1] END;
			StringToInt(name1, m, l);
			IF l = 0 THEN
				port := Int_Int16(m);
			END;
			outname[j] := 0X
		END
	END ParseHostPort;
	
	PROCEDURE And (i1, i2: Int32; VAR out: Int32);
		VAR s1, s2, sout: SET;
	BEGIN
		s1 := SYSTEM.VAL(SET, i1);
		s2 := SYSTEM.VAL(SET, i2);
		sout := s1*s2;
		out := SYSTEM.VAL(Int32, sout);
	END And;

	PROCEDURE Xor (i1, i2: Int32; VAR out: Int32);
		VAR s1, s2, sout: SET;
	BEGIN
		s1 := SYSTEM.VAL(SET, i1);
		s2 := SYSTEM.VAL(SET, i2);
		sout := s1 + s2 - s1*s2;
		out := SYSTEM.VAL(Int32, sout);
	END Xor;

	PROCEDURE Crc32* ( buf-: ARRAY OF Char8; len: INTEGER; 
	crc, nbits: Int32): Int32;
		VAR j: INTEGER; k: Int32;
	BEGIN
		j := 0;
		WHILE (j # len) & (buf[j] # 0X) DO
			Xor(crc, ORD(buf[j]), k);
			And(k, 255, k);
			Xor(crc_table[k], SYSTEM.LSH(crc, -8), crc);
			INC(j)
		END;
		Xor(crc, -1, crc);
		IF nbits = 8 THEN
			And(crc, 255, crc)
		ELSIF nbits = 16 THEN
			And(crc, 65535, crc)
		END;
		RETURN crc
	END Crc32;
	
	(* taken from ObxRandom *)
	PROCEDURE Uniform* (): REAL;
		CONST a = 16807; m = 2147483647; q = m DIV a; r = m MOD a;
		VAR gamma: Int32;
	BEGIN
		gamma := a * (z MOD q) - r * (z DIV q);
		IF gamma > 0 THEN
			z := gamma
		ELSE
			z := gamma + m
		END;
		RETURN z * (1.0 / m)	(* value of the function *)
	END Uniform;	
		
	PROCEDURE Initialize;
		VAR i, j: INTEGER; cr: Int32; t_crt: Time;
	BEGIN
		INFINITE_TIME.dwHighDateTime := SYSTEM.VAL(SYSTEM.CARD32, -1);
		INFINITE_TIME.dwLowDateTime := SYSTEM.VAL(SYSTEM.CARD32, -1);
		ZERO_TIME.dwHighDateTime := 0;
		ZERO_TIME.dwLowDateTime := 0;
		FOR i := 0 TO 255 DO
			cr := i;
			FOR j := 0 TO 7 DO
				IF cr MOD 2 = 1 THEN
					Xor(SYSTEM.LSH(cr, -1), -306674912, cr)
				ELSE
					cr := SYSTEM.LSH(cr, -1)
				END
			END;
			crc_table[i] := cr
		END;
		compatible_types[boolTyp] := "TaBool";
		compatible_types[char8Typ] := "TaChar";
		compatible_types[char16Typ] := "TaWChar";
		compatible_types[int8Typ] := "TaOctet";
		compatible_types[int16Typ] := "TaShort";
		compatible_types[int32Typ] := "TaLong";
		compatible_types[real32Typ] := "TaFloat";
		compatible_types[real64Typ] := "TaDouble";
		compatible_types[int64Typ] := "TaLongLong";
		compatible_types[recTyp] := "RECORD";
		compatible_types[arrTyp] := "ARRAY";
		compatible_types[ptrTyp] := "POINTER";
		GetTime(t_crt);
		z := SYSTEM.VAL(Int32, t_crt.dwLowDateTime);
	END Initialize;

BEGIN
	Initialize
END Ta_Basics.
