MODULE Ta_CFSrchEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Ta_STAreas, Ta_TCPChannel, Log;

	CONST
		MAX_STATIONS* = 2; MAX_STRING_LENGTH = 128; N_STRINGS_AT_ONCE = 16;

	TYPE
		FoundString* = RECORD (Tb.AnyRec)
			result_str* : ARRAY 128 OF Tb.Char8;
			pos* : Tb.Int32;
		END;
		SearchResults* = RECORD (Tb.AnyRec)
			pattern* : ARRAY 128 OF Tb.Char8;
			file_name* : ARRAY 128 OF Tb.Char8;
			start_pos*, end_pos*, request_number* : Tb.Int32;
			found* : POINTER TO ARRAY OF FoundString;
			n_found*: Tb.Int32;
		END;
		
		Opts* = RECORD (Tb.AnyRec)
			protocol*, host*: Tb.Name16;
			pattern*, filename*: Tb.Name;
			repeat*: Tb.Int32;
		END;

		FlagPtr = Tb.EmptyPtr;
		
		
	VAR
		searchVar* : SearchResults;
		area: Ta.Area;
		searchTable: Ta.Table;
		
		(* configuration params *)
		START: FlagPtr;
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
		

	(* -----------------------  Query Cycle  ----------------------- *)
	PROCEDURE DoNextQuery (t: Ta.Table; prod: Ta.ProData; first_call: Tb.AnyPtr);
		VAR i, l: Tb.Int32; out: Tb.Int8; s: ARRAY 256 OF CHAR;
		text: ARRAY N_STRINGS_AT_ONCE*MAX_STRING_LENGTH OF Tb.Char8;
	BEGIN
		IF first_call # START THEN
			Log.String("~"); Log.Int(searchVar.request_number); 
			Log.String("nfound:"); Log.Int(searchVar.n_found); Log.String(" pos:"); 
			Log.Int(searchVar.start_pos); Log.Int(searchVar.end_pos); Log.Ln;
			FOR i := 0 TO searchVar.n_found-1 DO
				Log.String("#"); Log.Int(searchVar.found[i].pos); Log.Tab; 
				Tb.Char8_Char(searchVar.found[i].result_str, s);
				Log.String(s); Log.Ln
			END
		END;
		IF searchVar.end_pos < 0 THEN
			t.Clear(searchVar);
			IF opts.repeat>0 THEN
				searchVar.end_pos := 0;
				t.Wait(1000000);
				t.Call(DoNextQuery, START);
			ELSE
				area.Finish(3)
			END
		ELSE
			searchVar.start_pos := searchVar.end_pos;
			t.Request(searchVar, TRUE, DoNextQuery, NIL);
		END
	END DoNextQuery;
		
	PROCEDURE OnErrorMessage(t: Ta.Table; prod: Ta.ProData);
	BEGIN
		IF prod.res # 0 THEN
			Log.String("*** error:"); Log.Int(prod.res); Log.Int(prod.sid); Log.Ln;
		END
	END OnErrorMessage;

	PROCEDURE ReallocResults (dim: INTEGER; old_adr: Tb.Address; VAR adr: Tb.Address);
	BEGIN
		NEW(searchVar.found, dim);
		adr := 0
	END ReallocResults;
	
	PROCEDURE SearchClientInit;
		VAR tqos: Ts.TableQos; ti_found_string: Ti.TInfo;
	BEGIN
		tqos := Ts.defaultTQos;
		tqos.tp_timeout := 10000000;
		tqos.realloc_mode := Ts.REALLOC_EXACT;
		NEW(ti_found_string); 
		ti_found_string.ConnectTo(Tb.no_rec, "Ta_CFSrchEC", "FoundString", 
			ReallocResults);
		ti_found_string.Insert(0, "result_str", Tb.char8Typ, Tb.exported, 128, NIL);
		ti_found_string.Insert(1, "pos", Tb.int32Typ, Tb.exported, 1, NIL);
		NEW(searchVar.found, 1);
		searchTable := area.NewTable(searchVar, "Tests.FSRCH", "SearchResults",
			Ts.sendSID, 0, tqos, OnErrorMessage);
		searchTable.ti.Insert(0, "pattern", Tb.char8Typ, Tb.exported, 128, NIL);
		searchTable.ti.Insert(1, "file_name", Tb.char8Typ, Tb.exported, 128, NIL);
		searchTable.ti.Insert(2, "start_pos", Tb.int32Typ, Tb.exported, 1, NIL);
		searchTable.ti.Insert(3, "end_pos", Tb.int32Typ, Tb.exported, 1, NIL);
		searchTable.ti.Insert(4, "request_number", Tb.int32Typ, Tb.exported, 1, NIL);
		searchTable.ti.Insert(5, "found", Tb.recTyp, Tb.exported, -1, 
			ti_found_string);
		searchTable.ti.Insert(6, "n_found", Tb.int32Typ, Tb.exported, 1, NIL);
	END SearchClientInit;
		
	(* -----------------------  Client Search Initialize  ----------------------- *)
	PROCEDURE Srch*;
		VAR aqos: Ta.AreaQos; cqos: Ta_TCPChannel.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(16, "fsrch", MAX_STATIONS, aqos);
		cqos := Ta_TCPChannel.defaultCQos;
		cqos.protocol := opts.protocol;
		Tb.Copy(opts.host, cqos.hostname);
		cqos.port := 9000;
		cqos.send_strat := Tc.SEND_STRAT_CLIENT;

		area.AddStations(opts.host, 1, 1, cqos, res);
		ASSERT(res = 0, 100);
		cqos.hostname[0] := 0X;
		area.AddStations("127.0.0.1", 1, 1, cqos, res);
		ASSERT(res = 0, 100);

		SearchClientInit;
		
		Tb.Char_Char8(opts.pattern, searchVar.pattern);
		Tb.Char_Char8(opts.filename, searchVar.file_name);

		searchVar.start_pos := 0;
		searchVar.end_pos := 0;
		searchTable.Wait(1000000);
		searchTable.Call(DoNextQuery, START);
		area.Start
		
	END Srch;
	
	PROCEDURE SrchCmd* (s: ARRAY OF CHAR);
	BEGIN
		ti.InitFromCommandLine(s, 0);
		Srch
	END SrchCmd;

	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "srch.cfg";
		NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "protocol", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(1, "host", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(2, "pattern", Tb.charTyp, Tb.exported, 256, NIL);
		myti.Insert(3, "filename", Tb.charTyp, Tb.exported, 256, NIL);
		myti.Insert(4, "repeat", Tb.int32Typ, Tb.exported, 1, NIL);
		opts.protocol := "TCP";
		opts.host := "127.0.0.1";
		opts.pattern := "BlackBox";
		opts.filename := "StartupBlackBox.vbs";
		opts.repeat := 0;
		NEW(START); 

		(* module_name, program_name, entry_function *)
		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_CFSrchEC', 'CFSRCH', Srch, ti, dqos);
	END Initialize;
		
BEGIN
	Initialize		
END Ta_CFSrchEC.
