MODULE Ta_CGWeatherEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Ta_SOAPStations, Ta_STAreas, Log
,SYSTEM;

	CONST
		MAX_STATIONS* = 2; 
	TYPE
		CountryList* = RECORD
			Country* : ARRAY 128 OF Tb.Char16;
			City* : ARRAY 128 OF Tb.Char16;
		END;
		NDSet* = RECORD
			Table* : POINTER TO ARRAY OF CountryList;
		END;
		GetCitiesByCountry* = RECORD (Tb.AnyRec)
			CountryName* : ARRAY 128 OF Tb.Char16;
			NewDataSet* : NDSet;
		END;
		
		CWeather* = RECORD
			Location* : ARRAY 128 OF Tb.Char16;
			Time* : ARRAY 128 OF Tb.Char16;
			Wind* : ARRAY 128 OF Tb.Char16;
			Visibility* : ARRAY 128 OF Tb.Char16;
			SkyConditions* : ARRAY 128 OF Tb.Char16;
			Temperature* : Tb.Real32;
			DewPoint* : Tb.Real32;
			RelativeHumidity* : Tb.Int32;
			Pressure* : ARRAY 64 OF Tb.Char16;
			Status* : ARRAY 32 OF Tb.Char16;
		END;
		GetWeather* = RECORD (Tb.AnyRec)
			CityName* : ARRAY 128 OF Tb.Char16;
			CountryName* : ARRAY 128 OF Tb.Char16;
			CurrentWeather* : CWeather;
		END;

		Opts* = RECORD (Tb.AnyRec)
			port*: Tb.Int16;
			soap*: Tb.Name16;
			host*, country*, city*: Tb.Name64;
		END;

		FlagPtr = Tb.EmptyPtr;

		
	VAR
		getCitiesByCountryVar* : GetCitiesByCountry;
		getWeatherVar* : GetWeather;
		area: Ta.Area;
		getCitiesByCountryTable, getWeatherTable : Ta.Table;

		START: FlagPtr;
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
		zero: Tb.Char16;
		
			
	PROCEDURE ReallocCities (dim: INTEGER; old_adr: Tb.Address; VAR adr: Tb.Address);
	BEGIN
		NEW(getCitiesByCountryVar.NewDataSet.Table, dim);
		adr := 0
	END ReallocCities;
	
	(* -----------------------  Get List of Cities  ----------------------- *)
	PROCEDURE DoGetCitiesByCountry(t: Ta.Table; prod: Ta.ProData; first_call: Tb.AnyPtr);
		VAR i: INTEGER; l: Tb.Int32; s: Tb.Name;
	BEGIN
		IF first_call # START THEN
			l := 0; 
			IF getCitiesByCountryVar.NewDataSet.Table # NIL THEN
				l := LEN(getCitiesByCountryVar.NewDataSet.Table^)
			END;
			i := 0;
			WHILE (i < l) & (getCitiesByCountryVar.NewDataSet.Table[i].City[0] # zero) DO
				Log.String("   '"); 
				Tb.Char16_Char(getCitiesByCountryVar.NewDataSet.Table[i].City, s);
				Log.String(s);
				Log.String("' '");
				Tb.Char16_Char(getCitiesByCountryVar.NewDataSet.Table[i].Country, s);
				Log.String(s);
				Log.String("'"); Log.Ln; 
				INC(i)
			END;
			Tb.Char16_Char(getCitiesByCountryVar.CountryName, s);
			Log.String(s);
			Log.String(" ncities:"); Log.Int(i); Log.Ln;
		ELSE
			getCitiesByCountryTable.Request(getCitiesByCountryVar, FALSE, 
				DoGetCitiesByCountry, NIL);
		END;
	END DoGetCitiesByCountry;
	
	(* -----------------------  Get Weather for requested City  ----------------------- *)
	PROCEDURE DoGetWeather(t: Ta.Table; prod: Ta.ProData; first_call: Tb.AnyPtr);
		VAR s: Tb.Name;
	BEGIN
		IF first_call # START THEN
			Log.String("'"); Tb.Char16_Char(getWeatherVar.CountryName, s); 
			Log.String(s); Log.String("   '"); 
			Tb.Char16_Char(getWeatherVar.CityName, s); Log.String(s); Log.String("'"); Log.Ln; 

			Log.String("   Location: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.Location, s); Log.String(s); Log.Ln;
			Log.String("   Time: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.Time, s); Log.String(s); Log.Ln;
			Log.String("   Wind: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.Wind, s); Log.String(s); Log.Ln;
			Log.String("   Visibility: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.Visibility, s); Log.String(s); Log.Ln;
			Log.String("   SkyConditions: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.SkyConditions, s); Log.String(s); 
				Log.Ln;
			Log.String("   Temperature: "); Log.Real(getWeatherVar.CurrentWeather.Temperature); Log.Ln;
			Log.String("   DewPoint: "); Log.Real(getWeatherVar.CurrentWeather.DewPoint); Log.Ln;
			Log.String("   RelativeHumidity: "); Log.Int(getWeatherVar.CurrentWeather.RelativeHumidity); 
				Log.Ln;
			Log.String("   Pressure: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.Pressure, s); Log.String(s); Log.Ln;
			Log.String("   Status: "); 
			Tb.Char16_Char(getWeatherVar.CurrentWeather.Status, s); Log.String(s); Log.Ln;
		ELSE
			getWeatherTable.Request(getWeatherVar, FALSE, DoGetWeather, NIL);
		END;
	END DoGetWeather;
	
	(* -----------------------  ClientInit  ----------------------- *)
	PROCEDURE ClientInit ;
		VAR tqos: Ta_SOAPStations.TableQos; ti_CountryList, ti_NDSet, ti_CWeather: Ti.TInfo;
	BEGIN
		tqos := Ta_SOAPStations.defaultTQos;
		tqos.tp_timeout := 20000000;
		tqos.tp_send_timeout := 10000000;
		tqos.realloc_mode := Ts.DONT_REALLOC;
		IF opts.soap = "11" THEN
			tqos.match_mode := Ta_SOAPStations.SOAP_11_CLIENT
		ELSIF opts.soap = "12" THEN
			tqos.match_mode := Ta_SOAPStations.SOAP_12_CLIENT
		ELSIF opts.soap = "GET" THEN
			tqos.match_mode := Ta_SOAPStations.SOAP_GET_CLIENT
		ELSIF opts.soap = "POST" THEN
			tqos.match_mode := Ta_SOAPStations.SOAP_POST_CLIENT
		ELSE
			ASSERT(opts.soap = "GET", Tb.INVALID_VAL)
		END;
		Tb.Copy(opts.host, tqos.xmlns);
		
		NEW(getCitiesByCountryVar.NewDataSet.Table, 5);
		getCitiesByCountryTable := area.NewTable(getCitiesByCountryVar, "globalweather.asmx", 
			"GetCitiesByCountry", Ts.sendSID, 0, tqos, NIL);
		NEW(ti_CountryList); NEW(ti_NDSet); 
		ti_CountryList.ConnectTo(Tb.no_rec, "Ta_CGWeatherEC", "CountryList", ReallocCities);
		ti_CountryList.Insert(0, "Country", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CountryList.Insert(1, "City", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_NDSet.ConnectTo(Tb.no_rec, "Ta_CGWeatherEC", "NDSet", NIL);
		ti_NDSet.Insert(0, "Table", Tb.recTyp, Tb.exported, -1, ti_CountryList);
		getCitiesByCountryTable.ti.Insert(0, "CountryName", Tb.char16Typ, Tb.exported, 128, NIL);
		getCitiesByCountryTable.ti.Insert(1, "NewDataSet", Tb.recTyp, Tb.exported, 1, ti_NDSet);
		
		getWeatherTable := area.NewTable(getWeatherVar, "globalweather.asmx", 
			"GetWeather", Ts.sendSID, 0, tqos, NIL);
		NEW(ti_CWeather);
		ti_CWeather.ConnectTo(Tb.no_rec, "Ta_CGWeatherEC", "CWeather", NIL);
		ti_CWeather.Insert(0, "Location", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(1, "Time", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(2, "Wind", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(3, "Visibility", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(4, "SkyConditions", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(5, "Temperature", Tb.real32Typ, Tb.exported, 1, NIL);
		ti_CWeather.Insert(6, "DewPoint", Tb.real32Typ, Tb.exported, 1, NIL);
		ti_CWeather.Insert(7, "RelativeHumidity", Tb.int32Typ, Tb.exported, 1, NIL);
		ti_CWeather.Insert(8, "Pressure", Tb.char16Typ, Tb.exported, 64, NIL);
		ti_CWeather.Insert(9, "Status", Tb.char16Typ, Tb.exported, 32, NIL);
		getWeatherTable.ti.Insert(0, "CityName", Tb.char16Typ, Tb.exported, 128, NIL);
		getWeatherTable.ti.Insert(1, "CountryName", Tb.char16Typ, Tb.exported, 128, NIL);
		getWeatherTable.ti.Insert(2, "CurrentWeather", Tb.recTyp, Tb.exported, 1, ti_CWeather);
	END ClientInit;
		
	(* -----------------------  Server Search Initialize  ----------------------- *)
	PROCEDURE Start*;
		VAR aqos: Ta.AreaQos; cqos: Tc.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(22, "gweather", MAX_STATIONS, aqos);
		cqos := Tc.defaultCQos;
		cqos.prot_format := "Ta_SOAP";
		cqos.protocol := "TCP";
		Tb.Copy(opts.host, cqos.hostname);
		cqos.port := opts.port;
		cqos.send_strat := Tc.SEND_STRAT_CLIENT;
		cqos.enable_ping := FALSE;
		area.AddStations("", MAX_STATIONS, 0, cqos, res);
		ASSERT(res = 0, 100);

		ClientInit;
		IF opts.city[0] = 0X THEN
			Tb.Char_Char16(opts.country, getCitiesByCountryVar.CountryName);
			getWeatherVar.CityName[0] := zero;
			getCitiesByCountryTable.Wait(3000000);
			getCitiesByCountryTable.Call(DoGetCitiesByCountry, START);
		ELSE
			Tb.Char_Char16(opts.country, getWeatherVar.CountryName);
			Tb.Char_Char16(opts.city, getWeatherVar.CityName);
			getWeatherTable.Wait(3000000);
			getWeatherTable.Call(DoGetWeather, START);
		END;
			
		area.Start
	END Start;
	
	PROCEDURE Initialize;
	BEGIN
		NEW(ti); NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "port", Tb.int16Typ, Tb.exported, 1, NIL);
		myti.Insert(1, "soap", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(2, "host", Tb.charTyp, Tb.exported, 64, NIL);
		myti.Insert(3, "country", Tb.charTyp, Tb.exported, 64, NIL);
		myti.Insert(4, "city", Tb.charTyp, Tb.exported, 64, NIL);
		opts.port := 80; Tb.Copy("127.0.0.1", opts.host); Tb.Copy("POST", opts.soap);
		Tb.Copy("Russian Federation", opts.country); opts.city[0] := 0X;
		NEW(START);
		Ta.dir.InitModule('Ta_CGWeatherEC', 'CGWEATHER', Start, ti, Ta.defaultDQos);
	END Initialize;

BEGIN
	Initialize		
END Ta_CGWeatherEC.
