MODULE Ta_C_SrchEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Ta_STAreas, Ta_TCPChannel, Log, Files, Tf := Ta_Files;

	CONST
		MAX_STATIONS* = 2; MAX_STRING_LENGTH = 128; N_STRINGS_AT_ONCE = 16;

	TYPE
		SearchQuery* = RECORD (Tb.AnyRec)
			pattern* : ARRAY 128 OF Tb.Char8;
			text* : POINTER TO ARRAY OF Tb.Char8;
		END;
		FoundString* = RECORD (Tb.AnyRec)
			result_str* : ARRAY 128 OF Tb.Char8;
			pos* : Tb.Int32;
		END;
		SearchResults* = RECORD (Tb.AnyRec)
			found* : POINTER TO ARRAY OF FoundString;
			n_found*: Tb.Int32;
		END;

		Opts* = RECORD (Tb.AnyRec)
			protocol*, host*: Tb.Name16;
			pattern*, filename*: Tb.Name;
			repeat*: Tb.Int32;
		END;
		
		FlagPtr = Tb.EmptyPtr;

		
	VAR
		queryVar* : SearchQuery;
		resultsVar* : SearchResults;
		area: Ta.Area;
		queryTable, resultsTable: Ta.Table;
		file: Files.File;
		rd: Files.Reader;

		(* configuration vars - must be exported for Meta if ReadCfg *)
		START, RUN, FINISH: FlagPtr;
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;		

	(* -----------------------  Query Cycle  ----------------------- *)
	PROCEDURE DoNextQuery (t: Ta.Table; prod: Ta.ProData; first_call: Tb.AnyPtr);
		VAR i, l: Tb.Int32; out: Tb.Int8; s: ARRAY 256 OF CHAR;
		text: ARRAY N_STRINGS_AT_ONCE*MAX_STRING_LENGTH OF Tb.Char8;
	BEGIN
		IF first_call # START THEN
			Log.String("nfound:"); Log.Int(resultsVar.n_found); Log.Ln;
			FOR i := 0 TO resultsVar.n_found-1 DO
				Log.String("#"); Log.Int(resultsVar.found[i].pos); Log.Tab; 
				Tb.Char8_Char(resultsVar.found[i].result_str, s);
				Log.String(s); Log.Ln
			END
		END;
		IF first_call = FINISH THEN
			IF opts.repeat>0 THEN
				rd.SetPos(0);
				resultsTable.Wait(1000000);
				resultsTable.Call(DoNextQuery, START);
			ELSE
				file.Close;
				area.Finish(3)
			END
		ELSIF ~rd.eof THEN
			i := 0; l := 0;
			REPEAT
				rd.ReadByte(out); text[i] := Tb.C_C8(CHR(out));
				IF out = 0DH THEN INC(l) END; INC(i)
			UNTIL (l = N_STRINGS_AT_ONCE) OR (out = 0) OR rd.eof;
			text[i] := Tb.C_C8(CHR(0));
			NEW(queryVar.text, i+1);
			FOR l := 0 TO i DO queryVar.text[l] := text[l] END;
			queryTable.Write(queryVar, TRUE);
			resultsTable.Wait(Ta.waitAnyMessage);
			IF (out = 0) OR rd.eof THEN
				resultsTable.Call(DoNextQuery, FINISH);
			ELSE
				resultsTable.Call(DoNextQuery, RUN);
			END
		END
	END DoNextQuery;
		
	PROCEDURE OnErrorMessage(t: Ta.Table; prod: Ta.ProData);
	BEGIN
		IF prod.res # 0 THEN
			Log.String("*** error:"); Log.Int(prod.res); Log.Int(prod.sid); Log.Ln;
		END
	END OnErrorMessage;

	(* -----------------------  SearchClientInit  ----------------------- *)
	PROCEDURE ReallocResults (dim: INTEGER; old_adr: Tb.Address; VAR adr: Tb.Address);
	BEGIN
		NEW(resultsVar.found, dim);
		adr := 0
	END ReallocResults;
	
	PROCEDURE SearchClientInit;
		VAR tqos: Ts.TableQos; ti_found_string: Ti.TInfo;
	BEGIN
		tqos := Ts.defaultTQos;
		tqos.tp_timeout := 10000000;
		tqos.realloc_mode := Ts.REALLOC_LARGEST;
		NEW(queryVar.text, 1);
		queryTable := area.NewTable(queryVar, "Tests.SRCH", "SearchQuery",
			Ts.sendSID, 0, tqos, NIL);
		queryTable.ti.Insert(0, "pattern", Tb.char8Typ, Tb.exported, 128, NIL);
		queryTable.ti.Insert(1, "text", Tb.char8Typ, Tb.exported, -1, NIL);

		NEW(ti_found_string); 
		ti_found_string.ConnectTo(Tb.no_rec, "Ta_C_SrchEC", "FoundString", 
			ReallocResults);
		ti_found_string.Insert(0, "result_str", Tb.char8Typ, Tb.exported, 128, NIL);
		ti_found_string.Insert(1, "pos", Tb.int32Typ, Tb.exported, 1, NIL);
		NEW(resultsVar.found, 1);
		resultsTable := area.NewTable(resultsVar, "Tests.SRCH", "SearchResults",
			Ts.sendSID, 0, tqos, OnErrorMessage);
		resultsTable.ti.Insert(0, "found", Tb.recTyp, Tb.exported, -1, 
			ti_found_string);
		resultsTable.ti.Insert(1, "n_found", Tb.int32Typ, Tb.exported, 1, NIL);
	END SearchClientInit;
		
	(* -----------------------  Client Search Initialize  ----------------------- *)
	PROCEDURE Srch*;
		VAR aqos: Ta.AreaQos; cqos: Ta_TCPChannel.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(17, "srch", MAX_STATIONS, aqos);
		cqos := Ta_TCPChannel.defaultCQos;
		cqos.protocol := opts.protocol;
		Tb.Copy(opts.host, cqos.hostname);
		cqos.port := 9000;
		cqos.send_strat := Tc.SEND_STRAT_CLIENT;
		
		area.AddStations(opts.host, 1, 1, cqos, res);
		ASSERT(res = 0, 100);
		cqos.hostname[0] := 0X;
		area.AddStations("127.0.0.1", 1, 1, cqos, res);
		ASSERT(res = 0, 100);

		SearchClientInit;
		
		Tb.Char_Char8(opts.pattern, queryVar.pattern);

		Tf.OldFile(opts.filename, TRUE, file);
		IF file = NIL THEN
			Log.String(opts.filename); Log.String(" not found"); Log.Ln;
		ELSE
			rd := file.NewReader(NIL);
			resultsTable.Wait(1000000);
			resultsTable.Call(DoNextQuery, START);
			area.Start
		END
	END Srch;
	
	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "srch.cfg";
		NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "protocol", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(1, "host", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(2, "pattern", Tb.charTyp, Tb.exported, 256, NIL);
		myti.Insert(3, "filename", Tb.charTyp, Tb.exported, 256, NIL);
		myti.Insert(4, "repeat", Tb.int32Typ, Tb.exported, 1, NIL);
		opts.protocol := "TCP";
		opts.host := "127.0.0.1";
		opts.pattern := "BlackBox";
		opts.filename := "StartupBlackBox.vbs";
		opts.repeat := 0;
		NEW(START); NEW(RUN); NEW(FINISH);

		(* module_name, program_name, entry_function *)
		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_C_SrchEC', 'C_SRCH', Srch, ti, dqos);
	END Initialize;
	
BEGIN
	Initialize		
END Ta_C_SrchEC.
