MODULE Ta_CfgTInfo;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
	
	Configurable Type Info
**)

	IMPORT Tb := Ta_Basics, Ti := Ta_TInfo, Log, Files, Tf := Ta_Files, 
		Tm := Ta_TunedMappers, Tp := Ta_SimpleParsers;

	CONST
		MAX_NARGS* = 32;


	TYPE
		LinkedList = POINTER TO LinkedListDesc;
		TInfo* = POINTER TO TInfoDesc;
		LinkedListDesc = RECORD
			ti: TInfo;
			linkname: Tb.Name64;
			next: LinkedList
		END;
		TInfoDesc* = RECORD (Ti.TInfoDesc)
			assert_not_found* : BOOLEAN;
			args- : ARRAY MAX_NARGS OF Tb.Name;
			nargs-: INTEGER;
			configured-, cfg_loaded- : BOOLEAN;
			cfg_file*: Tb.Name;
			ll: LinkedList
		END;

		ParseData = RECORD (Tb.AnyRec)
			ti: TInfo;
			nlevel0: INTEGER
		END;
				
	(* ========== TInfo ========== *)
	PROCEDURE (ti: TInfo) LinkMeTo* (main: TInfo; linkname: ARRAY OF CHAR) ;
		VAR ll, li: LinkedList;
	BEGIN
		NEW(ll); ll.ti := ti; Tb.Copy(linkname, ll.linkname);
		IF main.ll = NIL THEN
			main.ll := ll
		ELSE
			li := main.ll;
			WHILE li.next # NIL DO li := li.next END;
			li.next := ll
		END
	END LinkMeTo;

	PROCEDURE SetNameValue (ti: TInfo;  name-: ARRAY OF CHAR; 
	 value-: ARRAY OF CHAR);
		VAR trace_lev: LONGINT; res: INTEGER; ok: BOOLEAN; 
		it: Ti.Item; ll: LinkedList;
	BEGIN
		IF Tb.Equal(name, "trace_level") THEN
			Tb.StringToLInt(value, trace_lev, res);
			ASSERT(res = 0, Tb.STRING_CONVERSION_ERROR);
			Tb.Trace(Tb.Long_Int32(trace_lev));
		ELSIF Tb.Equal(name, "assert_not_found") THEN
			Tb.StringToBool(value, ti.assert_not_found, res);
			ASSERT(res = 0, Tb.STRING_CONVERSION_ERROR);
		ELSE
			ti.Find(name, ti.item, it); ok := (it.typ # 0);
			IF ~ok THEN
				ll := ti.ll;
				WHILE (ll # NIL) & ~ok DO
					IF ll.linkname[0] = 0X THEN
						ll.ti.Find(name, ll.ti.item, it); ok := (it.typ # 0)
					END;
					ll := ll.next
				END
			END;
			IF ok THEN
				Ti.String_ItemVal(it, value, ok);
				IF Tb.TRAL_RUNTIME IN Tb.trace THEN
					Log.String("Name/Value: "); Log.String(name); Log.Char('='); 
					Log.String(value); Log.String(" ok="); Log.Bool(ok); 
					Log.String(" adr="); Log.Int(it.adr); Log.Ln
				END;
				ASSERT(ok, Tb.STRING_CONVERSION_ERROR)
			ELSE
				ASSERT(~ti.assert_not_found, Tb.NOT_FOUND)
			END
		END
	END SetNameValue;

	PROCEDURE GetItem (ti: TInfo; it: Ti.Item; n0, nlevel: INTEGER; 
	VAR nstack: Tp.NamesStackArray; VAR itout: Ti.Item; VAR ok: BOOLEAN);
		VAR it0, it1, it2: Ti.Item;
	BEGIN
		IF n0 > nlevel THEN
			itout := it; ok := TRUE
		ELSE
			ti.Find(nstack[n0].name, it, it0);
			CASE it0.typ OF
			| 0:
				ok := FALSE
			| Tb.arrTyp:
				it0.Index(nstack[n0].index, it1);
				IF (it1.typ = Tb.char8Typ) OR (it1.typ = Tb.char16Typ) THEN
					GetItem(ti, it0, n0+1, nlevel, nstack, itout, ok)
				ELSE
					GetItem(ti, it1, n0+1, nlevel, nstack, itout, ok)
				END
			| Tb.ptrTyp:
				it0.Deref(it1);
				IF it1.typ = Tb.arrTyp THEN
					it1.Index(nstack[n0].index, it2);
					GetItem(ti, it2, n0+1, nlevel, nstack, itout, ok)
				ELSE
					GetItem(ti, it1, n0+1, nlevel, nstack, itout, ok)
				END
			ELSE
				GetItem(ti, it0, n0+1, nlevel, nstack, itout, ok)
			END
		END
	END GetItem;
	
	PROCEDURE SetRecNameValue (ti: TInfo; VAR nstack: Tp.NamesStackArray; 
	nlevel: INTEGER;  value-: ARRAY OF CHAR);
		VAR it: Ti.Item; named, ok: BOOLEAN; ll: LinkedList; n0: INTEGER;
	BEGIN
		ll := ti.ll; named := FALSE;
		WHILE (ll # NIL) & ~named DO
			IF Tb.Equal(nstack[0].name, ll.linkname) THEN
				named := TRUE; n0 := 1
			ELSE
				ll := ll.next
			END
		END;
		ll := ti.ll; ok := FALSE;
		WHILE (ll # NIL) & ~named & ~ok DO
			IF ll.linkname[0] = 0X THEN
				ok := TRUE; n0 := 0
			ELSE
				ll := ll.next
			END
		END;
		IF ll # NIL THEN
			GetItem(ll.ti, ll.ti.item, n0, nlevel, nstack, it, ok);
			IF ok THEN
				IF Tb.TRAL_RUNTIME IN Tb.trace THEN
					Log.String("..Value: "); Log.Char('='); Log.String(value); Log.Ln
				END;
				Ti.String_ItemVal(it, value, ok);
				ASSERT(ok, Tb.STRING_CONVERSION_ERROR)
			ELSE
				ASSERT(~ti.assert_not_found, Tb.NOT_FOUND)
			END
		END
	END SetRecNameValue;

	PROCEDURE OnReadCfg (parse_flag: INTEGER; 
	VAR nstack: Tp.NamesStackArray; nlevel: INTEGER; value: ARRAY OF CHAR;
	VAR pcl: Tb.AnyRec; VAR res: INTEGER);
	BEGIN
		WITH pcl: ParseData DO
			CASE parse_flag OF
			| Tp.COUPLE_BEGIN:
				pcl.ti.nargs := 0;
				Tb.Copy(nstack[nlevel].name, pcl.ti.args[pcl.ti.nargs]);
				INC(pcl.ti.nargs);
			| Tp.COUPLE_ARG:
				Tb.Copy(nstack[nlevel].name, pcl.ti.args[pcl.ti.nargs]);
				INC(pcl.ti.nargs);
			| Tp.COUPLE_BOTH:
				IF (pcl.nlevel0 = 1) OR (nlevel <= pcl.nlevel0) THEN
					SetNameValue(pcl.ti, nstack[nlevel].name, value)
				ELSE
					SetRecNameValue(pcl.ti, nstack, nlevel, value)
				END
			ELSE
			END
		END
	END OnReadCfg;
	
	PROCEDURE (ti: TInfo) InitFromCommandLine* (command_line: ARRAY OF CHAR;
	ind_command_line: INTEGER) ;
		VAR lrd: Files.Reader; s: Tm.Scanner; res: INTEGER; pcl: ParseData;
	BEGIN
		ti.CalcInfo;
		pcl.ti := ti; pcl.nlevel0 := 1;
		lrd := Tf.NewLineReader(NIL, command_line, ind_command_line);
		s.ConnectTo(lrd); res := 0;
		Tp.ParseOpVal(s, OnReadCfg, pcl, res);
		ASSERT(res = 0, Tb.PARSE_ERROR);
	END InitFromCommandLine;

	PROCEDURE (ti: TInfo) InitFromCfg* (current_dir, cfg_file: ARRAY OF CHAR; 
	VAR cfg_loaded: BOOLEAN) ;
		VAR file: Files.File; rd: Files.Reader; s: Tm.Scanner; res: INTEGER; 
		pcd: ParseData; str: Tb.Name;
	BEGIN
		Tf.OldFile(ti.cfg_file, TRUE, file);
		IF file = NIL THEN
			Tb.Concat("Rsrc/", cfg_file, str);
			Tf.OldFile(str, TRUE, file)
		END;
		IF file # NIL THEN
			rd := file.NewReader(NIL);
			s.ConnectTo(rd); res := 0;
			pcd.ti := ti; pcd.nlevel0 := 0;
			Tp.ParseCfg(s, OnReadCfg, pcd, res);
			ASSERT(res = 0, Tb.PARSE_ERROR);
			file.Close;
			cfg_loaded := TRUE
		ELSE
			cfg_loaded := FALSE
		END
	END InitFromCfg;
	
	PROCEDURE (ti: TInfo) InitConfig* (command_line, mod_name, 
	current_dir: ARRAY OF CHAR; ind_command_line: INTEGER) ;
		VAR i, j, pos: INTEGER; ll: LinkedList;
	BEGIN
		ti.CalcInfo;
		ll := ti.ll;
		WHILE ll # NIL DO
			ll.ti.CalcInfo; ll := ll.next
		END;
		ti.configured := FALSE; ti.cfg_loaded := FALSE;
		Tb.Search("cfg_file", command_line, 0, 9, Tb.Len(command_line), pos);
		IF pos >= 0 THEN	(* config file name changed - load it *)
			i := pos+9;
			WHILE command_line[i] = ' ' DO INC(i) END;
			j := 0;
			WHILE (command_line[i] # 0X) & (command_line[i] # ' ') & (command_line[i] # 09X) DO
				IF (command_line[i] # '"') & (command_line[i] # "'") THEN
					ti.cfg_file[j] := command_line[i]; INC(j)
				END;
				INC(i)
			END;
			ti.cfg_file[j] := 0X;
		ELSIF ti.cfg_file[0] = '-' THEN
			Tb.Concat(mod_name, ".cfg", ti.cfg_file);
		END;
		IF ti.cfg_file[0] # 0X THEN
			ti.InitFromCfg(current_dir, ti.cfg_file, ti.cfg_loaded);
		END;
		IF ind_command_line >= 0 THEN
			ti.InitFromCommandLine(command_line, ind_command_line)
		END;
		ti.configured := TRUE
	END InitConfig;
	
END Ta_CfgTInfo.