MODULE Ta_Channels;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics;
	
	CONST
		(* ---------- Channel Flags bits ---------- *)
		FLAGS_MCAST* = 0; FLAGS_DGRAM* = 1; FLAGS_STREAM* = 2; FLAGS_LISTEN* = 3;
		FLAGS_MEMORY* = 4; FLAGS_MESSAGE* = 5; FLAGS_UNIX_DOMAIN* = 6;
		FLAGS_LIB* = 7; FLAGS_ICMP* = 8;
		
		TRANS_SEND* = 0; TRANS_RECV* = 1; 
			TRANS_SEND_DIRECT* = 2; TRANS_RECV_DIRECT* = 3; 
			TRANS_SEND_INDIRECT* = 4; TRANS_RECV_INDIRECT* = 5; 
			TRANS_SEND_VIA_HOST* = 6; TRANS_RECV_VIA_SEND* = 7;
			TRANS_CONNECTION* = 8; TRANS_LISTEN* = 9;
		LINK_NONE* = 0; LINK_CLOSED* = 1; LINK_ERR* = 2; LINK_INIT* = 3; LINK_WAIT* = 4; 
			LINK_LISTEN* = 5; LINK_OK* = 6; 
		POLLIN* = 0; POLLPRI* = 1; POLLOUT* = 2; POLLERR* = 3; POLLHUP* = 4; 
			POLLNVAL* = 5; POLL_NEXT* = 7; MAX_NUM_POLL_SD* = 4;
			POLLIN2* = POLLIN+POLL_NEXT; POLLPRI2* = POLLPRI+POLL_NEXT;
			POLLOUT2* = POLLOUT+POLL_NEXT; POLLERR2* = POLLERR+POLL_NEXT;
			POLLHUP2* = POLLHUP+POLL_NEXT; POLLNVAL2* = POLLNVAL+POLL_NEXT;

		MAX_MID* = 32000; PERMANENT_MID* = 32001;
		PING_MID* = 32100; PING_IGNORE_MID* = 32110; PING_WARNING_MID* = 32111; 
			PING_GOOD_MID* = 32112; PING_BAD_MID* = 32113; PING_MAX* = 32114;
	
		SEND_STRAT_SERVER* = 0; SEND_STRAT_CLIENT* = 1; SEND_STRAT_MIN_ZVAL* = 2; 		
			SEND_STRAT_CR_SW* = 3;

	
	TYPE
		ChannelQosBase* =  RECORD (Tb.Msg)
			hostname* : Tb.Name64;
			port* : Tb.Int16;
			group_addr* : Tb.Name16;
			protocol*: Tb.Name16;
			enable_ping* : BOOLEAN;
			tp_ping_timeout* : LONGINT;
			tp_discovery_timeout* : LONGINT;
			send_strat* : INTEGER;
			prot_format* : Tb.Name16;
			connect_on_request* : BOOLEAN;
			trust_src_sid*: BOOLEAN;
		END;
		ChannelQos* = RECORD (ChannelQosBase)
		END;
		SockChannelQos* =  RECORD (ChannelQosBase)
			max_send_polls* : INTEGER;
			so_keepalive* : BOOLEAN;	(* setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, 1) *)
			so_reuseaddr* : INTEGER;	(* -1 never, 0 if bind, 1 always *)
		END;
		StreamChannelQos* =  RECORD (SockChannelQos)
			tp_connect_timeout* : LONGINT;
			tp_disconnect_timeout* : LONGINT;
			tcp_nodelay* : BOOLEAN;	(* setsockopt(sd, SOL_TCP, TCP_NODELAY, 1) *)
		END;
		
		InitMsg* = RECORD (Tb.Msg)
			reinit*: BOOLEAN
		END;
		CloseMsg* = RECORD (Tb.Msg)
		END;

		Polling* = POINTER TO PollingDesc;
		PollingFactory* = POINTER TO PollingFactoryDesc;
		Channel* = POINTER TO ChannelDesc;
		ChannelFactory* = POINTER TO ChannelFactoryDesc;

		PollingEvents* = POINTER TO ARRAY OF SET;
		PollingDesc* =  RECORD
		END;
		PollingFactoryDesc* =  RECORD
		END;		
		ChannelDesc* =  RECORD
			flags*, trans_flags*, poll_events*: SET;
			link_stat*: INTEGER
		END;
		ChannelFactoryDesc* =  RECORD
		END;		

		Dest* = RECORD
			addr*: Tb.Int32;
			port*: Tb.Int16;
		END;
		

	VAR
		defaultCQos*: ChannelQos;
		
		
	(* ========== Polling interface ========== *)
	PROCEDURE (polling: Polling) FdSetChannel* (index: INTEGER; channel: Channel);
	BEGIN END FdSetChannel;
	PROCEDURE (polling: Polling) Poll* ((*OUT*)VAR events: PollingEvents; VAR res: INTEGER);
	BEGIN END Poll;
	(* ========== Polling Factory ========== *)
	PROCEDURE (pf: PollingFactory) NewPolling* (n_stations: INTEGER; (*OUT*)VAR poll: Polling;
	(*OUT*)VAR res: INTEGER);
	BEGIN END NewPolling;

	(* ========== Channel interface ========== *)
	PROCEDURE (channel: Channel) InitData* (is_datahost, is_send_station: BOOLEAN; 
	 name-: ARRAY OF CHAR; VAR cqos: ChannelQosBase);
	BEGIN END InitData;
	PROCEDURE (channel: Channel) GetFds* (ind_first: INTEGER; VAR ind_last: INTEGER;
	VAR fd_array: ARRAY OF Tb.Int32);
	BEGIN ind_last := ind_first END GetFds;
	PROCEDURE (channel: Channel) CloseData* (shutdown: BOOLEAN);
	BEGIN END CloseData;
	PROCEDURE (channel: Channel) Connect* (indirect_channel: Channel; 
	disconnect_only: BOOLEAN; VAR fd_changed: BOOLEAN);
	BEGIN END Connect;
	PROCEDURE (channel: Channel) Send* (indirect_channel: Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32);
	BEGIN END Send;
	PROCEDURE (channel: Channel) Recv* (indirect_channel: Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32;
	VAR dest: Dest);
	BEGIN END Recv;
	PROCEDURE (channel: Channel) MyDest* ( dest-: Dest): BOOLEAN;
	BEGIN RETURN FALSE END MyDest;
	PROCEDURE (channel: Channel) HandleMsg* (VAR msg: Tb.Msg); 
	BEGIN END HandleMsg;
	(* ========== Channel Factory ========== *)
	PROCEDURE (cf: ChannelFactory) NewChannel* (aid, sid, host_sid: INTEGER; 
	(*OUT*)VAR chan: Channel; (*OUT*)VAR res: INTEGER);
	BEGIN END NewChannel;

	(* ========== Qos Management ========== *)		
	PROCEDURE DefaultChannelQos* (VAR qos: ChannelQosBase);
	BEGIN
		qos.hostname := "";
		qos.port := 7700;
		qos.group_addr := "239.255.0.1";
		qos.protocol := "";
		qos.enable_ping := TRUE;
		qos.tp_ping_timeout := 2000000;
		qos.tp_discovery_timeout := -1;
		qos.send_strat := SEND_STRAT_MIN_ZVAL;
		qos.prot_format := "Ta_SRPS";
		qos.connect_on_request := FALSE;
		qos.trust_src_sid := FALSE;
	END DefaultChannelQos;
		
	PROCEDURE DefaultSockChannelQos* (VAR qos: SockChannelQos);
	BEGIN
		DefaultChannelQos(qos);
		qos.max_send_polls := 1;
		qos.so_keepalive := FALSE;
		qos.so_reuseaddr := 0;
	END DefaultSockChannelQos;
	PROCEDURE DefaultStreamChannelQos* (VAR qos: StreamChannelQos);
	BEGIN
		DefaultSockChannelQos(qos);
		qos.tp_connect_timeout := 2000000;
		qos.tp_disconnect_timeout := 60000000;
		qos.tcp_nodelay := FALSE;
	END DefaultStreamChannelQos;
	
	PROCEDURE CopyChannelQos* ( src-: ChannelQosBase; 
	VAR dst: ChannelQosBase);
	BEGIN
		dst.hostname := src.hostname;
		dst.port := src.port;
		dst.group_addr := src.group_addr;
		dst.protocol := src.protocol;
		dst.enable_ping := src.enable_ping;
		dst.tp_ping_timeout := src.tp_ping_timeout;
		dst.send_strat := src.send_strat;
		dst.prot_format := src.prot_format;
		dst.connect_on_request := src.connect_on_request;
		dst.trust_src_sid := src.trust_src_sid
	END CopyChannelQos;
	

BEGIN
	DefaultChannelQos(defaultCQos);
END Ta_Channels.