MODULE Ta_FactManager;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Ts := Ta_Stations,
		oberonRTS, Log, SYSTEM;
	IMPORT Ta_Poll;
	IMPORT Ta_MUDPChannel, Ta_UDPChannel, Ta_TCPChannel, Ta_ICMPChannel,
		Ta_MSHMChannel;
	IMPORT Ta_SRPSStations, Ta_PINGStations;

	PROCEDURE CreatePollingFactory* ( polling_name-: ARRAY OF CHAR;
	(*OUT*)VAR pfactory: Tc.PollingFactory; (*OUT*)VAR res: INTEGER);
		VAR str: Tb.Name64; mod: oberonRTS.Module; type: oberonRTS.Type; 
		addr: SYSTEM.ADDRESS; 
	BEGIN
		res := Tb.NOT_FOUND;
		IF polling_name # '' THEN
			str := "Ta_"; Tb.Append(polling_name, str);
			mod := oberonRTS.Search(str);
			IF mod # oberonRTS.nullModule THEN
				type := oberonRTS.ThisType(mod, "PollingFactoryDesc");
				IF type # oberonRTS.nullType THEN
					addr := oberonRTS.NewObj(type);
					pfactory := SYSTEM.VAL(Tc.PollingFactory, addr);
					IF pfactory # NIL THEN
						res := 0
					END
				END
			END
		END;
		IF (res # 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
			Log.String("* New '"); Log.String(polling_name); Log.String("' Polling err="); 
			Log.Int(res); Log.Ln
		END;
	END CreatePollingFactory;

	PROCEDURE CreateChannelFactory* ( protocol-: ARRAY OF CHAR;
	(*OUT*)VAR cfactory: Tc.ChannelFactory; (*OUT*)VAR res: INTEGER);
		VAR str: Tb.Name64; mod: oberonRTS.Module; type: oberonRTS.Type; 
		addr: SYSTEM.ADDRESS;
	BEGIN
		res := Tb.NOT_FOUND;
		IF protocol # '' THEN
			str := "Ta_"; Tb.Append(protocol, str); Tb.Append('Channel', str);
			mod := oberonRTS.Search(str);
			IF mod # oberonRTS.nullModule THEN
				type := oberonRTS.ThisType(mod, "ChannelFactoryDesc");
				IF type # oberonRTS.nullType THEN
					addr := oberonRTS.NewObj(type);
					cfactory := SYSTEM.VAL(Tc.ChannelFactory, addr);
					IF cfactory # NIL THEN
						res := 0
					END
				END
			END
		END;
		IF (res # 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
			Log.String("* New '"); Log.String(protocol); Log.String("'  Channel err="); 
			Log.Int(res); Log.Ln
		END;
	END CreateChannelFactory;

	PROCEDURE CreateCarrierFactory* ( prot_format-: ARRAY OF CHAR;
	(*OUT*)VAR crfactory: Ts.CarrierFactory; (*OUT*)VAR res: INTEGER);
		VAR str: Tb.Name64; mod: oberonRTS.Module; type: oberonRTS.Type; 
		addr: SYSTEM.ADDRESS; 
	BEGIN
		res := Tb.NOT_FOUND;
		IF prot_format # '' THEN
			Tb.Copy(prot_format, str); Tb.Append('Stations', str);
			mod := oberonRTS.Search(str);
			IF mod # oberonRTS.nullModule THEN
				type := oberonRTS.ThisType(mod, "CarrierFactoryDesc");
				IF type # oberonRTS.nullType THEN
					addr := oberonRTS.NewObj(type);
					crfactory := SYSTEM.VAL(Ts.CarrierFactory, addr);
					IF crfactory # NIL THEN
						res := 0
					END
				END
			END
		END;
		IF (res # 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
			Log.String("* New '"); Log.String(prot_format); Log.String("' Carrier err="); 
			Log.Int(res); Log.Ln
		END;
	END CreateCarrierFactory;

	PROCEDURE CreateMapperFactory* ( prot_format-: ARRAY OF CHAR;
	(*OUT*)VAR mfactory: Ts.MapperFactory; (*OUT*)VAR res: INTEGER);
		VAR str: Tb.Name64; mod: oberonRTS.Module; type: oberonRTS.Type; 
		addr: SYSTEM.ADDRESS; 
	BEGIN
		res := Tb.NOT_FOUND;
		IF prot_format # '' THEN
			Tb.Copy(prot_format, str);
			Tb.Append('Stations', str); mod := oberonRTS.Search(str);
			IF mod # oberonRTS.nullModule THEN
				type := oberonRTS.ThisType(mod, "MapperFactoryDesc");
				IF type # oberonRTS.nullType THEN
					addr := oberonRTS.NewObj(type);
					mfactory := SYSTEM.VAL(Ts.MapperFactory, addr);
					IF mfactory # NIL THEN
						res := 0
					END
				END
			END
		END;
		IF (res # 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
			Log.String("* New '"); Log.String(prot_format); Log.String("' Mapper err="); 
			Log.Int(res); Log.Ln
		END;
	END CreateMapperFactory;



END Ta_FactManager.