MODULE Ta_ICMPChannel;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, Tl := Ta_Lib, Tb := Ta_Basics, Tc := Ta_Channels, Log ;
	
	CONST 
		ICMP_PROTOCOL_NUM = 1; ICMP_MINLEN = 8; DATALEN = 48;
		(* ICMP recv codes *)
		ECHO_REPLY* = 0; ICMP_1* = 1; ICMP_2* = 2; DEST_UNREACHABLE* = 3; 
			SOURCE_QUENCH* = 4; REDIRECT* = 5; ICMP_6* = 6; ICMP_7* = 7;
			ECHO* = 8; ICMP_9* = 9; ICMP_10* = 10; TIME_EXCEEDED* = 11;
			PARAMETER_PROBLEM* = 12; TIMESTAMP* = 13; TIMESTAMP_REPLY* = 14;
			INFO_REQUEST* = 15; INFO_REPLY* = 16;
		ICMP_RECV_IGNORE = Tc.PING_IGNORE_MID - Tc.PING_MID; (*10*)
		ICMP_RECV_WARNING = Tc.PING_WARNING_MID - Tc.PING_MID; (*11*)
		ICMP_RECV_GOOD = Tc.PING_GOOD_MID - Tc.PING_MID; (*12*)
		ICMP_RECV_BAD = Tc.PING_BAD_MID - Tc.PING_MID; (*13*)

	TYPE
		ChannelQos* = RECORD (Tc.ChannelQosBase)
			ping_set_err*: BOOLEAN;
			sid_limit*: INTEGER;
		END;

		Results = RECORD
			t_sent: Tb.Time;
			icmp_seq, dst_sid, nbytes: Tb.Int16;
			icmp_type, icmp_recv: Tb.Int8;
		END;
		
		IcmpItem = RECORD
			sa_msg: Tl.sockaddr_in;
			sa_msg_len, last_rc: Tb.Int32;
			tv: Tb.Time;
			rs: Results;
		END;
		Channel = POINTER TO ChannelDesc;
		ChannelDesc = RECORD (Tc.ChannelDesc)
			cqos: ChannelQos;
			hostname: Tb.Name64;
			sd: Tl.SOCKET;
			sa_from: Tl.sockaddr_in;
			send_id, recv_id: Tb.Int16;
			items: POINTER TO ARRAY OF IcmpItem;
		END;
		
		IcmpSend = RECORD 
			icmp_type, icmp_code: Tb.Int8;
			icmp_cksum: Tb.Int16;
			icmp_id, icmp_seq: Tb.Int16;
			tv: Tb.Time;
			datap: ARRAY DATALEN OF Tb.Int8;
		END;
		PIcmpSend = POINTER  TO IcmpSend;
		
		ChannelFactoryDesc* = RECORD (Tc.ChannelFactoryDesc)
		END;		
		ChannelFactory* = POINTER TO ChannelFactoryDesc;


	VAR
		defaultCQos-: ChannelQos;

	(* ========== Channel interface ========== *)
	PROCEDURE (channel: Channel) InitData (is_datahost, is_send_station: BOOLEAN; 
	 name-: ARRAY OF CHAR; VAR cqos: Tc.ChannelQosBase);
		VAR my_errno: Tb.Int32; inaddr: Tl.in_addr; ok: BOOLEAN;
		h: Tl.phostent; rc, opt: Tb.Int32; v: Tb.UInt32; 
	BEGIN
		ok := TRUE; 
		channel.sd := 0;
		IF cqos.hostname[0] # 0X THEN
			Tb.Copy(cqos.hostname, channel.hostname)
		ELSE
			Tb.Copy(name, channel.hostname)
		END;
		IF cqos IS ChannelQos THEN
			channel.cqos := cqos(ChannelQos);
			ASSERT(channel.cqos.sid_limit > 0, Tb.CHECK_RANGE_ERROR)
		ELSE
			channel.cqos := defaultCQos;
			Tc.CopyChannelQos(cqos, channel.cqos);
		END;
		(* whether this channel is a sender, receiver or uses another one to send *)
		IF is_send_station THEN
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_RECV, 
				Tc.TRANS_SEND_DIRECT, Tc.TRANS_RECV_DIRECT}
		ELSIF ~is_datahost THEN
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_RECV, 
				Tc.TRANS_SEND_INDIRECT, Tc.TRANS_RECV_VIA_SEND}
		ELSE
			channel.trans_flags := {}
		END;
		
		(* parse and determine proper addrinfo *)
		rc := Tl.inet_aton(channel.hostname, inaddr);
		IF (rc = 0) & (ORD(channel.hostname[0]) >= ORD('A')) 
		& (ORD(channel.hostname[0]) <= ORD('z')) THEN
			h := Tl.gethostbyname(channel.hostname);
			IF (h # NIL) & (h.h_length > 0) THEN
				SYSTEM.MOVE(h.h_addr_list[0], SYSTEM.ADR(inaddr), SIZE(Tl.in_addr));
				rc := 1
			END;
		END;
		IF rc = 0 THEN
			ok := FALSE
		ELSE
			channel.sa_from.sin_addr.s_addr := inaddr.s_addr;
			channel.sa_from.sin_family := Tl.AF_INET;
			channel.sa_from.sin_port := 0
		END;
		
		(* create and set necessary socket and socket options *)
		IF ~ok THEN
			channel.link_stat := Tc.LINK_ERR;
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* undefined addr '"); Log.String(channel.hostname); 
				Log.String("'"); Log.Ln
			END;
		ELSIF is_send_station THEN
			channel.sd := Tl.socket(Tl.AF_INET, Tl.SOCK_RAW, ICMP_PROTOCOL_NUM);
			my_errno := Tb.Errno();
			IF channel.sd <= 0 THEN
				ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* socket rc="); Log.Int(channel.sd); 
					Log.String(" err="); Log.Int(my_errno); Log.Ln
				END;		
				channel.sd := 0
			END;
			v := 1;
			IF ok & (Tl.fcntl(channel.sd, Tl.F_SETFL, Tl.O_NONBLOCK) < 0) THEN
				my_errno := Tb.Errno(); ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* NONBLOCK socket err="); Log.Int(my_errno); Log.Ln
				END;		
			END;

			IF ok THEN
				NEW(channel.items, channel.cqos.sid_limit);
				channel.recv_id := SYSTEM.VAL(Tb.Int16, channel);
				IF channel.recv_id < 0 THEN 
					channel.recv_id := SYSTEM.VAL(Tb.Int16, -channel.recv_id-1);
				END;
				channel.link_stat := Tc.LINK_OK;
			ELSE
				channel.CloseData(FALSE);
				channel.link_stat := Tc.LINK_ERR;
			END;
			IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
				IF ~ok THEN Log.String("* ") END;
				Log.String("ICMP socket sd='"); Log.Int(channel.sd); Log.String("' host='"); 
				Log.String(channel.hostname); Log.String("'"); Log.Ln
			END;
		ELSE
			channel.link_stat := Tc.LINK_NONE;
		END;
	END InitData;
	
	PROCEDURE (channel: Channel) GetFds (ind_first: INTEGER; 
	VAR ind_last: INTEGER; VAR fd_array: ARRAY OF Tb.Int32);
	BEGIN
		IF channel.sd = 0 THEN
			ind_last := ind_first
		ELSE
			fd_array[ind_first] := SYSTEM.VAL(Tb.Int32, channel.sd);
			ind_last := ind_first+1
		END
	END GetFds;
	
	PROCEDURE (channel: Channel) CloseData (shutdown: BOOLEAN);
		VAR rc: Tb.Int32;
	BEGIN
		IF channel.sd # 0 THEN
			IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
				Log.String("CloseData sd="); Log.Int(channel.sd); Log.Ln
			END;
			rc := Tl.close(channel.sd);
			channel.sd := 0
		END;
		channel.link_stat := Tc.LINK_CLOSED;
	END CloseData;


	PROCEDURE InCksum(addr: Tb.Int32; len: INTEGER) : Tb.Int16;
		VAR nleft, wi: INTEGER; sum, hi: Tb.Int32; u, answer: Tb.Int16;
	BEGIN
		nleft := len; wi := 0; sum := 0;
		WHILE nleft > 1 DO
			SYSTEM.GET(addr+wi+wi, u);
			INC(sum, u); INC(wi);
			DEC(nleft, 2)
		END;
		IF nleft = 1 THEN
			SYSTEM.GET(addr+wi+wi-1, u);
			u := SYSTEM.LSH(u, 8);
			INC(sum, u);
		END;
		(* add back carry outs from top 16 bits to low 16 bits *)
		hi := SYSTEM.LSH(sum, -16);				(* add hi 16 to low 16 *)
		sum := hi + (sum-SYSTEM.LSH(hi,16));	(* sum = (sum >> 16) + (sum & 0xffff); *)
		sum := sum + SYSTEM.LSH(sum, -16);	(* sum += (sum >> 16); add carry *)
		answer := SYSTEM.VAL(Tb.Int16, -1-sum);	(* answer = ~sum; truncate to 16 bits *)
		RETURN answer
	END InCksum;

	PROCEDURE (channel: Channel) Send (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32);
		VAR my_errno, a: Tb.Int32; s_channel: Channel; l, sa_len: INTEGER; icp: IcmpSend;
		sa_from: Tl.sockaddr;
	BEGIN
		ASSERT(indirect_channel IS Channel, Tb.WRONG_TYPE);
		ASSERT(buffer # 0, Tb.NIL_POINTER);
		ASSERT(buffer_length >= 2, Tb.CHECK_RANGE_ERROR);
		nbytes := 0; s_channel := indirect_channel(Channel);
		IF (s_channel.sd > 0) & (Tc.POLLOUT IN s_channel.poll_events) THEN
			sa_len := SIZE(Tl.sockaddr_in);
			Tb.Clean(SYSTEM.ADR(icp), SIZE(IcmpSend));
			icp.icmp_type := ECHO;
			icp.icmp_code := 0;
			icp.icmp_cksum := 0;
			icp.icmp_id := s_channel.recv_id;
			SYSTEM.GET(buffer, icp.icmp_seq);
			(*Tb.GetTime(icp.tv);*)
			FOR l := 0 TO DATALEN-1 DO icp.datap[l] := SYSTEM.VAL(Tb.Int8, l+8) END;
			icp.icmp_cksum := InCksum(SYSTEM.ADR(icp), SIZE(IcmpSend));
			s_channel.items[channel.send_id].sa_msg := channel.sa_from;
			s_channel.items[channel.send_id].sa_msg_len := sa_len;
			s_channel.items[channel.send_id].tv := icp.tv;
			s_channel.items[channel.send_id].last_rc := 0;
			SYSTEM.MOVE(SYSTEM.ADR(channel.sa_from), SYSTEM.ADR(sa_from), 
				SIZE(Tl.sockaddr));
			nbytes := Tl.sendto(s_channel.sd, SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(icp)), 
				SIZE(IcmpSend), Tl.MSG_NOSIGNAL, sa_from, sa_len);
			my_errno := Tb.Errno();
			IF (nbytes < 0) & (my_errno = Tl.EWOULDBLOCK) THEN
				nbytes := 0
			END;			
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
			OR (nbytes < 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
				IF nbytes < 0 THEN Log.String("* sendto err="); Log.Int(my_errno) 
				ELSE Log.String("sendto") END; 
				Log.String(" sd="); Log.Int(s_channel.sd); Log.String(" sid="); Log.Int(channel.send_id);
				a := SYSTEM.VAL(Tb.Int32, channel.sa_from.sin_addr.s_addr);
				Log.String(" addr="); Log.Int(a);
				Log.String(" rc="); Log.Int(nbytes); Log.String(" len="); Log.Int(buffer_length); 
				Log.String(" sa_len="); Log.Int(sa_len); Log.Ln
			END;
		END;
	END Send;
	
	PROCEDURE CheckOutPacket(channel: Channel; buffer: Tb.Address; 
	buffer_length: Tb.Int32; VAR rs: Results);
		VAR hlen: Tb.Int8; icp: PIcmpSend;
	BEGIN
		(* unsigned int ip_hl:4; hlen = ip->ip_hl << 2; *)
		SYSTEM.GET(buffer, hlen);
		hlen := SYSTEM.VAL(Tb.Int8, (hlen MOD 16) * 4);	
		IF (hlen < ICMP_MINLEN) OR (hlen > buffer_length) THEN
			rs.icmp_type := 0;
			rs.icmp_recv := ICMP_RECV_IGNORE
		ELSE
			icp := SYSTEM.VAL(PIcmpSend, SYSTEM.VAL(SYSTEM.CARD32, buffer)+
				SYSTEM.VAL(SYSTEM.CARD32, hlen));
			rs.icmp_type := icp.icmp_type;
			IF rs.icmp_type IN {DEST_UNREACHABLE, SOURCE_QUENCH, 
			TIME_EXCEEDED, PARAMETER_PROBLEM} THEN
				rs.icmp_recv := ICMP_RECV_BAD;
			ELSIF rs.icmp_type = ECHO THEN	(* msg, not reply *)
				rs.icmp_recv := ICMP_RECV_IGNORE
			ELSIF icp.icmp_id # channel.recv_id THEN
				rs.icmp_recv := ICMP_RECV_WARNING; (* seems not my message *)
			ELSE
				rs.icmp_recv := ICMP_RECV_GOOD;
			END;
			rs.icmp_seq := icp.icmp_seq;
			rs.t_sent := icp.tv
		END;
	END CheckOutPacket;
	
	
	PROCEDURE (channel: Channel) Recv (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32;
	VAR dest: Tc.Dest);
		VAR my_errno: Tb.Int32; sa_msg: Tl.sockaddr; sa_msg_in: Tl.sockaddr_in; 
		sa_msg_len, a: Tb.Int32; r_channel: Channel; j, n: INTEGER; done: BOOLEAN; rs: Results;
	BEGIN
		ASSERT(indirect_channel IS Channel, Tb.WRONG_TYPE);
		nbytes := 0;
		IF Tc.TRANS_RECV_VIA_SEND IN channel.trans_flags THEN
			r_channel := indirect_channel(Channel)
		ELSE
			r_channel := channel
		END;
		IF (r_channel.sd > 0) & (Tc.POLLIN IN r_channel.poll_events) THEN
			sa_msg_len := SIZE(Tl.sockaddr);
			Tb.Clean(SYSTEM.ADR(sa_msg), sa_msg_len);
			nbytes := Tl.recvfrom(r_channel.sd, SYSTEM.VAL(Tl.pbyte, buffer), buffer_length,
				Tl.MSG_NOSIGNAL, sa_msg, sa_msg_len);
			my_errno := Tb.Errno();
			IF (nbytes < 0) & (my_errno = Tl.EWOULDBLOCK) THEN
				nbytes := 0
			END;
			IF nbytes > 0 THEN
				j := -1; 
				CheckOutPacket(r_channel, buffer, nbytes, rs);
				SYSTEM.MOVE(SYSTEM.ADR(sa_msg), SYSTEM.ADR(sa_msg_in), 
					SIZE(Tl.sockaddr));
				IF rs.icmp_recv = ICMP_RECV_IGNORE THEN
					done := FALSE;	(* ignore package, go on *)
				ELSE
					done := FALSE; n := 0;
					IF r_channel.cqos.ping_set_err OR (rs.icmp_recv # ICMP_RECV_BAD) THEN
						j := 0;
						WHILE (j < r_channel.cqos.sid_limit) & ~done DO
							IF r_channel.items[j].sa_msg_len > 0 THEN
								INC(n);
								IF r_channel.items[j].sa_msg.sin_addr.s_addr 
								= sa_msg_in.sin_addr.s_addr THEN
									done := TRUE
								END
							END;
							IF ~done THEN INC(j) END
						END;
						(*IF ~done & (n = 1) & (rs.icmp_recv # ICMP_RECV_WARNING) THEN
							j := channel.send_id;
							done := TRUE
						END;*)
					END
				END;
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) THEN
					Log.String("recvfrom"); 
					Log.String(" sd="); Log.Int(r_channel.sd); Log.String(" sid="); Log.Int(j);
					a := SYSTEM.VAL(Tb.Int32, sa_msg_in.sin_addr.s_addr);
					Log.String(" addr="); Log.Int(a);
					Log.String(" rc="); Log.Int(nbytes); Log.String(" done="); Log.Bool(done);
					Log.String(" icmp_type="); Log.Int(rs.icmp_type); 
					Log.String(" icmp_seq="); Log.Int(rs.icmp_seq); 
					Log.String(" icmp_recv="); Log.Int(rs.icmp_recv); Log.Ln
				END;
				IF done THEN
					rs.dst_sid := j; 
					rs.nbytes := SHORT(nbytes);
					SYSTEM.MOVE(SYSTEM.ADR(rs), buffer, SIZE(Results));
					IF j = channel.send_id THEN
						dest.addr := 0; dest.port := 0
					ELSE
						dest.addr := SYSTEM.VAL(Tb.Int32, sa_msg_in.sin_addr.s_addr); 
						dest.port := 0
					END
				ELSE
					nbytes := 1 
				END;
			ELSIF nbytes < 0 THEN
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* recvfrom err="); Log.Int(my_errno); 
					Log.String(" sd="); Log.Int(r_channel.sd); 
					Log.String(" rc="); Log.Int(nbytes); Log.Ln
				END;
			END
		END;
	END Recv;	

	PROCEDURE (channel: Channel) MyDest ( dest-: Tc.Dest): BOOLEAN;
	BEGIN 
		RETURN SYSTEM.VAL(Tb.Int32, channel.sa_from.sin_addr.s_addr) = dest.addr
	END MyDest;
	

	(* ========== Initialize ========== *)
	PROCEDURE (cf: ChannelFactory) NewChannel* (aid, sid, host_sid: INTEGER; 
	(*OUT*)VAR chan: Tc.Channel; (*OUT*)VAR res: INTEGER);
		VAR channel: Channel;
	BEGIN
		ASSERT((sid >= 0) & (sid < Tl.MAX_FD_SET), Tb.CHECK_RANGE_ERROR);
		NEW(channel);
		channel.flags := {Tc.FLAGS_ICMP};
		channel.trans_flags := {};
		channel.link_stat := Tc.LINK_NONE;
		channel.send_id := sid;
		chan := channel;
		res := 0
	END NewChannel;
	
BEGIN
	Tc.DefaultChannelQos(defaultCQos);
	defaultCQos.protocol := "ICMP";
	defaultCQos.ping_set_err := FALSE;
	defaultCQos.sid_limit := Tl.MAX_FD_SET;
END Ta_ICMPChannel.