MODULE Ta_Lib;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM;
	
	CONST
		WSA_VER22* = 514; WSADESCRIPTION_LEN = 256; WSASYS_STATUS_LEN = 128;
		AF_UNSPEC*       = 0;           (* unspecified *)
		AF_UNIX*         = 1;               (* local to host (pipes, portals) *)
		AF_INET*         = 2;               (* internetwork: UDP, TCP, etc. *)
		AF_IMPLINK*      = 3;            (* arpanet imp addresses *)
		AF_PUP*          = 4;               (* pup protocols: e.g. BSP *)
		AF_CHAOS*        = 5;            (* mit CHAOS protocols *)
		AF_IPX*          = 6;                (* IPX and SPX *)
		AF_NS*           = 6;                (* XEROX NS protocols *)
		AF_ISO*          = 7;                (* ISO protocols *)
		AF_OSI*          = AF_ISO;      (* OSI is ISO *)
		AF_ECMA*         = 8;             (* european computer manufacturers *)
		AF_DATAKIT*      = 9;            (* datakit protocols *)
		AF_CCITT*        = 10;            (* CCITT protocols, X.25 etc *)
		AF_SNA*          = 11;             (* IBM SNA *)
		AF_DECnet*       = 12;           (* DECnet *)
		AF_DLI*          = 13;              (* Direct data link interface *)
		AF_LAT*          = 14;              (* LAT *)
		AF_HYLINK*       = 15;           (* NSC Hyperchannel *)
		AF_APPLETALK*    = 16;       (* AppleTalk *)
		AF_NETBIOS*      = 17;          (* NetBios-style addresses *)
		AF_VOICEVIEW*    = 18;       (* VoiceView *)
		AF_MAX*          = 19;
		SOCK_STREAM* = 1; SOCK_DGRAM* = 2; SOCK_RAW* = 3; 
			SOCK_RDM* = 4; SOCK_SEQPACKET*  = 5;
		SO_REUSEADDR* = 2;
		SO_KEEPALIVE* = 9;
		SO_DONTROUTE* = 5;
		SOL_SOCKET* = 1; 			
		MAX_FD_SET* = 64;
		POLLIN* = 1; POLLPRI* = 2; POLLOUT* = 4; POLLERR* = 8; POLLHUP* = 16;
			POLLNVAL* = 32;
		O_NONBLOCK* = 2048;
		F_SETFL* = 4;
		MSG_NOSIGNAL* = 04000H;
		NI_MAXHOST* = 1025; NI_MAXSERV* = 32;
		NI_NUMERICHOST* = 1; NI_NUMERICSERV* = 2;
		IPPROTO_IP* = 0; IPPROTO_TCP* = 6;
		IP_ADD_MEMBERSHIP* = 35;
		IP_MULTICAST_TTL* = 33; IP_MULTICAST_LOOP* = 34;
		TCP_NODELAY* = 1;
		EWOULDBLOCK* = 11; EINPROGRESS* = 115; EISCONN* = 106;
		EPIPE* = 32;
		ECONNABORTED* = 103; ESHUTDOWN* = 108;

		STAT_LIBC_VER* = 3;

	
	TYPE
		timeval* = RECORD
			tv_sec*, tv_usec*: LONGINT
		END;
		timezone* = RECORD
			tz_minuteswest-, tz_dsttime-: SYSTEM.CARD32
		END;

		SOCKET* = LONGINT;
		in_addr* = RECORD
			s_addr*: LONGINT
		END;
		sockaddr_in* = RECORD
			sin_family* : INTEGER;
			sin_port*   : INTEGER;
			sin_addr*   : in_addr;
			sin_zero   : ARRAY 8 OF CHAR;
		END;
		sockaddr* = RECORD
			sa_family*: INTEGER;
			sa_data*: ARRAY 14 OF CHAR;
		END;
		phostent* = POINTER TO hostent;
		PtrSTR* = POINTER TO ARRAY OF CHAR;
		hostent* = RECORD
			h_name*: PtrSTR;
			h_aliases*: POINTER TO (*?*) ARRAY OF PtrSTR;
			h_addrtype*: INTEGER;
			h_length*: INTEGER;
			h_addr_list*: POINTER TO (*?*) ARRAY OF POINTER TO ARRAY OF in_addr;
		END;
		pbyte* = PtrSTR;	
		fd_set* = RECORD
			fd_count*: LONGINT;
			fd_array*: ARRAY MAX_FD_SET OF SOCKET;
		END;
		pollfd* = RECORD
			fd*: SOCKET;
			events*, revents*: INTEGER
		END;
		ip_mreq* = RECORD
			imr_multiaddr*, imr_interface*: in_addr
		END;
		psockaddr* = POINTER TO sockaddr;
		

		struct_stat* = RECORD
			dummy1: ARRAY 44 OF CHAR;
			st_size*: LONGINT;
			dummy2: ARRAY 40 OF CHAR;
		END;
		ipc_perm* = RECORD
			key-: LONGINT;
			uid-, gid-, cuid-, cgid-, mode-, seq-: INTEGER
		END;
		shmid_ds* = RECORD 
			shm_perm-: ipc_perm;
			shm_segsz-, shm_atime-, shm_dtime-, shm_ctime-: LONGINT;
			shm_cpid-, shm_lpid-, shm_nattch-, shm_unused-: INTEGER;
			dummy: ARRAY 2 OF LONGINT;
		END;

	PROCEDURE ["C"] / gettimeofday* (tv: timeval; VAR tz: timezone) : LONGINT;	
	PROCEDURE ["C"] / nanosleep* (req: timeval; VAR rem: timeval) : LONGINT;	
	PROCEDURE ["C"] / usleep* (mcs: LONGINT) : LONGINT;
	PROCEDURE ["C"] / getnameinfo*(sa-: sockaddr; salen: LONGINT;
	host: pbyte; hostlen: LONGINT; serv: pbyte; servlen, flags: LONGINT) : LONGINT;
	PROCEDURE ["C"] / select* (nfds: LONGINT; VAR readfds: fd_set; 
	VAR writefds: fd_set; VAR exceptfds: fd_set; VAR timeout: timeval): LONGINT;
	PROCEDURE ["C"] / inet_aton*(name-: PtrSTR; VAR inaddr: in_addr) : LONGINT;
	PROCEDURE ["C"] / gethostbyname* (name: PtrSTR): phostent;	
	PROCEDURE ["C"] / htons* (hostshort: INTEGER): INTEGER;
	PROCEDURE ["C"] / socket* (af: LONGINT; type: LONGINT; protocol: LONGINT): SOCKET;
	PROCEDURE ["C"] / fcntl* (s: SOCKET; cmd, arg: LONGINT): LONGINT;
	PROCEDURE ["C"] / setsockopt* (s: SOCKET; level: LONGINT; optname: LONGINT; 
	optval: pbyte; optlen: LONGINT): LONGINT;	
	PROCEDURE ["C"] / bind* (s: SOCKET; name-: sockaddr; namelen: LONGINT): LONGINT;
	PROCEDURE ["C"] / shutdown* (s: SOCKET; how: LONGINT): LONGINT;	
	PROCEDURE ["C"] / sendto* (s: SOCKET; buf: pbyte; len: LONGINT; flags: LONGINT; 
	to-: sockaddr; tolen: LONGINT): LONGINT;
	PROCEDURE ["C"] / send* (s: SOCKET; buf: pbyte; len: LONGINT; flags: LONGINT): LONGINT;
	PROCEDURE ["C"] / recvfrom* (s: SOCKET; buf: pbyte; len: LONGINT; flags: LONGINT; 
	from-: sockaddr; VAR fromlen: LONGINT): LONGINT;	
	PROCEDURE ["C"] / recv* (s: SOCKET; buf: pbyte; len: LONGINT; flags: LONGINT): LONGINT;	
	PROCEDURE ["C"] / close*(s: SOCKET) : LONGINT;
	PROCEDURE ["C"] / poll*(ufds-: pollfd; nfds, timeout: LONGINT) : LONGINT;
	PROCEDURE ["C"] / getpeername* (s: SOCKET; VAR name: sockaddr; 
	VAR namelen: LONGINT): LONGINT;	
	PROCEDURE ["C"] / listen*(s: SOCKET; backlog: INTEGER) : LONGINT;
	PROCEDURE ["C"] / accept* (s: SOCKET; msg-: sockaddr; 
	VAR msglen: LONGINT): LONGINT;	
	PROCEDURE ["C"] / connect* (s: SOCKET; msg-: sockaddr; 
	msglen: LONGINT): LONGINT;	

	PROCEDURE ["C"] / __xstat* (ver: INTEGER; file_name-: ARRAY OF CHAR; 
	st-: struct_stat): LONGINT;
	PROCEDURE ["C"] / ftok*(pathname-: ARRAY OF CHAR;
	proj_id: LONGINT) : LONGINT;
	PROCEDURE ["C"] / shmget*(key, size, shmflg: LONGINT) : LONGINT;
	PROCEDURE ["C"] / shmat*(shmid, shmaddr, shmflg: LONGINT) : LONGINT;
	PROCEDURE ["C"] / shmdt*(shmaddr: LONGINT) : LONGINT;
	PROCEDURE ["C"] / shmctl*(shmid, shmaddr: LONGINT; sds-: shmid_ds) : LONGINT;

END Ta_Lib.
