MODULE Ta_NodexEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Ta_STAreas, Log;
	
	CONST
		MAX_STATIONS = 3; START_DELAY = 10000000; DELTA_T = 4000000000.0;
		
	TYPE 
		NodeStatus* = RECORD (Tb.AnyRec)
			t_changed* : Tb.Int64;
			t_display* : Tb.Int64;
			counters_mask* : Tb.Int64;
			counter* : Tb.Int32;
			host_counter* : Tb.Int32;
		END;
		
		NodeStatistics* = RECORD
			number_of_lost* : INTEGER;
			number_of_lost_pct* : REAL;
			not_just_in_time* : INTEGER;
			not_just_in_time_pct* : REAL;
			n_t_overtime, n_2t_overtime, n_gaps*, n_msgs_at_once,
			n_lost*, n_done, must_recv* : INTEGER;
			nodex*: NodeStatus;
		END;
		
		Opts* = RECORD  (Tb.AnyRec)
			protocol*: Tb.Name16;
			hosts*: ARRAY 256 OF CHAR; 
		END;

		
	VAR
		tabs: ARRAY 2 OF Ta.Table;
		nos*: ARRAY MAX_STATIONS OF NodeStatistics;
		nodex, zeros: NodeStatus;
		t_start, t_crt, last_crtu: Tb.Time;
		sidhost, update_counter: INTEGER;
		started: BOOLEAN;
		deviation_sum, deviation_avg*, cycle_time_sum, cycle_time*: REAL;
		area: Ta.Area;
		get: Ta.GetDirectoryQos;
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
		

	(* -----------------------  Test Nodex  ----------------------- *)
	PROCEDURE Clear (VAR ns: NodeStatistics);
	BEGIN
		ns.nodex := zeros;
		ns.number_of_lost := 0; ns.number_of_lost_pct := 0; 
		ns.not_just_in_time := 0; ns.not_just_in_time_pct := 0; 
		ns.n_t_overtime := 0; ns.n_2t_overtime := 0; ns.n_gaps := 0; 
		ns.n_msgs_at_once := 0; ns.n_lost := 0; ns.n_done := 0; 
		ns.must_recv := 0; 
	END Clear;
	
	PROCEDURE HandleMessage(t: Ta.Table; prod: Ta.ProData);
		VAR prev_counter: Tb.Int32; is_1: BOOLEAN; t_crt: Tb.Time;
	BEGIN
		IF prod.res = 0 THEN
			nodex.host_counter := update_counter;
			IF ~started THEN
				Tb.GetTime(t_crt);
				started := (Tb.CompareTime(t_crt, t_start) >= START_DELAY)
			END;
			prev_counter := nos[t.ind].nodex.counter;
			nos[t.ind].nodex := nodex;
			INC(nos[t.ind].n_msgs_at_once);
			IF started THEN
				INC(nos[t.ind].n_done);
				IF ~((nodex.counter = prev_counter+1) 
				OR (nodex.counter = 1) & (prev_counter = 1000)) THEN
					INC(nos[t.ind].n_gaps)
				END
			END;
		ELSE
			Log.String("*** error:"); Log.Int(prod.res); Log.Ln;
		END
	END HandleMessage;
	
	PROCEDURE NodexUpdate(area: Ta.Area; par: Tb.AnyPtr);
		VAR t_ch, t_ch0: Tb.Time; dt: REAL; masks, his: SET; i: INTEGER;
		times: Ta.GetAreaRunTimes;
	BEGIN
		last_crtu := t_crt; Tb.GetTime(t_crt);
		masks := {};
		Tb.T64_Time(nos[2].nodex.t_changed, t_ch0);
		FOR i := 0 TO 1 DO
			(* check skips and overtime *)
			IF started THEN
				INC(nos[i].must_recv);
				IF (nos[i].must_recv > nos[i].n_done) THEN
					nos[i].number_of_lost := nos[i].must_recv - nos[i].n_done
				ELSE
					nos[i].number_of_lost := 0;
				END;
				IF nos[i].n_msgs_at_once # 1 THEN
					INC(nos[i].not_just_in_time)
				END;
				Tb.T64_Time(nos[i].nodex.t_changed, t_ch);
				dt := Tb.CompareTime(t_crt, t_ch);
				IF dt >= get.dqos.cycle THEN
					INC(nos[i].n_t_overtime)
				END;
				IF dt >= 2*get.dqos.cycle THEN
					INC(nos[i].n_2t_overtime);
				END;
				his := Tb.Int32_Set(Tb.Int64_Int32(nos[i].nodex.counters_mask));
				IF ~(sidhost IN his) THEN
					INC(nos[i].n_lost)
				END;
			END;
			(* prepare send *)
			Tb.T64_Time(nos[i].nodex.t_changed, t_ch);
			dt := Tb.CompareTime(t_ch0, t_ch);
			IF (dt >= -DELTA_T) & (dt <= DELTA_T) THEN
				IF sidhost > i THEN
					INCL(masks, i)
				ELSE
					INCL(masks, i+1)
				END
			END;
			nos[i].n_msgs_at_once := 0;
		END;
				
		(* send *)
		Tb.Time_T64(t_crt, nos[2].nodex.t_changed);
		Tb.Time_T64(t_crt, nos[2].nodex.t_display);
		nos[2].nodex.counters_mask := Tb.Long_Int64(Tb.Set_Int32(masks));
		INC(nos[2].nodex.counter);
		IF nos[2].nodex.counter > 1000 THEN
			nos[2].nodex.counter := 1
		END;
		(*tabs[0].Write(nos[2].nodex, FALSE);
		IF ~(Tc.FLAGS_MCAST IN area.Station(tabs[0].ind).chs.flags) THEN
			(* send twice if 0 is not multicast *)
			tabs[1].Write(nos[2].nodex, FALSE)
		END;*)
		tabs[0].WriteTo(nos[2].nodex, Ta.ALL);

		INC(update_counter);
		area.Message(Ts.toNONE, Ts.toNONE, times);
		dt := 100. * Tb.CompareTime(times.t_setpoint, times.t_crtu) / get.dqos.cycle;
		IF (dt > 1000) OR (dt < -1000) THEN
			dt := 0
		END;
		deviation_sum := deviation_sum + dt;
		deviation_avg := deviation_sum / update_counter;
		IF update_counter # 1 THEN
			cycle_time_sum := cycle_time_sum + Tb.CompareTime(last_crtu, t_crt);
			cycle_time := cycle_time_sum / (update_counter-1);
		END;
	END NodexUpdate;
		
	PROCEDURE NodexStationUpdate(area: Ta.Area; par: Tb.AnyPtr);
		VAR j, k: INTEGER;
	BEGIN
		Log.String("deviation:"); Log.Real(deviation_avg);
		Log.String("% cycle_time:"); Log.Real(cycle_time); Log.String("mcs");
		Log.Ln;

		k := 0;
		FOR j := 0 TO 1 DO
			INC(k);
			IF nos[j].must_recv # 0 THEN
				nos[j].number_of_lost_pct := 100.* nos[j].number_of_lost
					/ nos[j].must_recv;
				nos[j].not_just_in_time_pct := 100.* nos[j].not_just_in_time
					/ nos[j].must_recv;
			END;
			Log.Char("#"); Log.Int(k); Log.String(" number_of_lost:"); 
			Log.Int(nos[j].number_of_lost); Log.Real(nos[j].number_of_lost_pct); 
			Log.Char("%"); Log.Ln;
			Log.Char("#"); Log.Int(k); Log.String(" not_just_in_time:");
			Log.Int(nos[j].not_just_in_time); Log.Real(nos[j].not_just_in_time_pct); 
			Log.Char("%"); Log.Ln;
			Log.Char("#"); Log.Int(k); Log.String(" no_order:"); 
			Log.Int(nos[j].n_gaps); Log.String(" must_recv:"); 
			Log.Int(nos[j].must_recv); Log.Ln;
		END
	END NodexStationUpdate;
	
	PROCEDURE Stop*;
	BEGIN
		IF area # NIL THEN
			area.Finish(Ta.NORMAL_FINISH);
			area := NIL
		END
	END Stop;
	
	PROCEDURE Nodex*;
		VAR aqos: Ta.AreaQos; cqos: Tc.ChannelQos; tqos: Ts.TableQos; 
			i0, i1, res: INTEGER; 
	BEGIN
		myti.InitFromCommandLine(Ta.dir.command_line, Ta.dir.ind_command_line);
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(3, "nodex", MAX_STATIONS, aqos);
		area.Message(Ts.toNONE, Ts.toNONE, get);
		cqos := Tc.defaultCQos;
		cqos.protocol := opts.protocol;
		cqos.trust_src_sid := TRUE;
		area.AddUpdater(NodexUpdate, NIL, 0);
		area.AddUpdater(NodexStationUpdate, NIL, -10000000);
		area.AddStations(opts.hosts, Ta.APPEND_AS_LIST, -1, cqos, res);
		ASSERT(res = 0, 102);
		sidhost := area.HostSid();
		i0 := 0; i1 := 1;
		IF i0 >= sidhost THEN INC(i0) END;
		IF i1 >= sidhost THEN INC(i1) END;

		started := FALSE;
		Tb.GetTime(t_start); t_crt := t_start;

		tqos := Ts.defaultTQos;
		tqos.realloc_mode := Ts.DONT_REALLOC;
		tabs[0] := area.NewTable(nodex, "Tests.INodex", "NodeStatus", 
			i0, 0, tqos, HandleMessage);
		tabs[0].ind := 0;
		tabs[1] := area.NewTable(nodex, "Tests.INodex", "NodeStatus", 
			i1, 0, tqos, HandleMessage);
		tabs[1].ind := 1;
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		tabs[0].ti.Insert(0, "t_changed", Tb.int64Typ, Tb.exported, 1, NIL);
		tabs[0].ti.Insert(1, "t_display", Tb.int64Typ, Tb.exported, 1, NIL);
		tabs[0].ti.Insert(2, "counters_mask", Tb.int64Typ, Tb.exported, 1, NIL);
		tabs[0].ti.Insert(3, "counter", Tb.int32Typ, Tb.exported, 1, NIL);
		tabs[0].ti.Insert(4, "host_counter", Tb.int32Typ, Tb.exported, 1, NIL);
		tabs[1].ti.CopyFrom(tabs[0].ti);

		Clear(nos[0]); Clear(nos[1]); Clear(nos[2]);
		
		update_counter := 0;
		deviation_sum := 0;
		deviation_avg := 0;
		last_crtu := Tb.ZERO_TIME;
		cycle_time_sum := 0;
		cycle_time := 0;
		area.Start
	END Nodex;

	PROCEDURE NodexCmd* (s, hosts: ARRAY OF CHAR);
	BEGIN
		ti.InitFromCommandLine(s, 0);
		myti.InitFromCommandLine(s, 0);
		Tb.Copy(hosts, opts.hosts);
		Nodex
	END NodexCmd;

	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "-";
		NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "protocol", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(1, "hosts", Tb.charTyp, Tb.exported, 256, NIL);
		opts.protocol := "MUDP";

		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_NodexEC', 'NODEX', Nodex, ti, dqos);
	END Initialize;


BEGIN
	Initialize		
END Ta_NodexEC.
