MODULE Ta_PINGStations;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Ti := Ta_TInfo, Ts := Ta_Stations, SYSTEM, Log;
	
	CONST
		ICMP_RECV_IGNORE = Tc.PING_IGNORE_MID - Tc.PING_MID; (*10*)
		ICMP_RECV_WARNING = Tc.PING_WARNING_MID - Tc.PING_MID; (*11*)
		ICMP_RECV_GOOD = Tc.PING_GOOD_MID - Tc.PING_MID; (*12*)
		ICMP_RECV_BAD = Tc.PING_BAD_MID - Tc.PING_MID; (*13*)
		
		RESULTS_SIZE = 16;

	TYPE
		Carrier = POINTER TO CarrierDesc;
		Mapper = POINTER TO MapperDesc;

		Results = RECORD
			t_sent: Tb.Time;
			icmp_seq, dst_sid, nbytes: Tb.Int16;
			icmp_type, icmp_recv: Tb.Int8;
		END;

		CarrierDesc = RECORD (Ts.CarrierDesc)
			rs: Results;
			aid: INTEGER;
		END;
		
		CarrierFactoryDesc* = RECORD (Ts.CarrierFactoryDesc)
		END;		
		CarrierFactory* = POINTER TO CarrierFactoryDesc;

		MapperDesc = RECORD (Ts.MapperDesc)
			sid: INTEGER;
		END;
		
		MapperFactoryDesc* = RECORD (Ts.MapperFactoryDesc)
		END;		
		MapperFactory* = POINTER TO MapperFactoryDesc;

	
	(* ========== Carrier Factory ========== *)
	PROCEDURE (crf: CarrierFactory) NewCarrier* (pf: Tc.PollingFactory; 
	aid, n_stations: INTEGER; send_buffer_size, recv_buffer_size: Tb.Int32;  
	prot_format: ARRAY OF CHAR; (*OUT*)VAR cr: Ts.Carrier; (*OUT*)VAR res: INTEGER);
		VAR carrier: Carrier; polling: Tc.Polling;
	BEGIN
		ASSERT(n_stations > 0, Tb.CHECK_RANGE_ERROR);
		ASSERT(pf # NIL, Tb.NIL_POINTER);
		pf.NewPolling(n_stations, polling, res);
		IF res = 0 THEN
			NEW(carrier); carrier.aid := aid;
			carrier.SetNewCarrier(polling, send_buffer_size, recv_buffer_size, prot_format)
		END;
		cr := carrier
	END NewCarrier;

	(* ========== Carrier Interface ========== *)
	PROCEDURE (cr: Carrier) InternalizeHeader (VAR rd: Ts.Reader; (*OUT*)VAR hdr: Ts.Header; 
	(*OUT*)VAR res: Tb.Int32);
		VAR x: ARRAY RESULTS_SIZE OF Tb.Int8;
	BEGIN
		rd.SetPos(0);
		rd.ReadBytes(x, 0, RESULTS_SIZE);
		res := Ts.CANT_READ_HDR;
		IF ~rd.eof THEN
			SYSTEM.MOVE(SYSTEM.ADR(x), SYSTEM.ADR(cr.rs), RESULTS_SIZE);
			IF (cr.rs.icmp_recv >= ICMP_RECV_IGNORE)
			& (cr.rs.icmp_recv <= ICMP_RECV_BAD) THEN
				hdr.msg_type := Tc.PING_MID + cr.rs.icmp_type;
				hdr.mid := SYSTEM.VAL(Tb.Int16, (Tc.PING_MID + cr.rs.icmp_recv));
				hdr.mid_resp := cr.rs.icmp_seq;
				hdr.src_sid := cr.rs.dst_sid; hdr.aid := cr.aid;
				hdr.t_set := cr.rs.t_sent; hdr.t_got := rd.t_recv;
				hdr.basic_length := cr.rs.nbytes; hdr.buffer_length := cr.rs.nbytes;
				hdr.tlg_type := Ts.USER_TELEGRAM_TYPE;
				res := 0
			END
		END
	END InternalizeHeader;

	PROCEDURE (cr: Carrier) ExternalizeHeader (VAR wr: Ts.Writer;  hdr-: Ts.Header;
	(*OUT*)VAR res: Tb.Int32);
		VAR x: ARRAY 2 OF Tb.Int8;
	BEGIN
		SYSTEM.MOVE(SYSTEM.ADR(hdr.mid), SYSTEM.ADR(x), 2);
		wr.WriteBytes(x, 0, 2);
		IF wr.eof THEN
			res := Ts.CANT_WRITE_HDR
		ELSE
			res := 0
		END
	END ExternalizeHeader;
	
	(* ========== Mapper interface ========== *)
	PROCEDURE (mapper: Mapper) MatchHeader ( hdr-: Ts.Header): BOOLEAN;
	BEGIN
		RETURN (mapper.sid = hdr.src_sid)
	END MatchHeader;

	PROCEDURE (mapper: Mapper) WriteHeader (VAR hdr: Ts.Header;  wr-: Ts.Writer);
		VAR res: Tb.Int32;
	BEGIN
		wr.WriteHeader(hdr, res);
		mapper.eof := (res # 0)
	END WriteHeader;

	(* ========== Initialize ========== *)
	PROCEDURE (mf: MapperFactory) NewMapper* (ti: Ti.TInfo;
	 iface_name-: ARRAY OF CHAR; VAR tqos: Ts.TableQosBase;
	sid, oid: INTEGER; (*OUT*)VAR mapr: Ts.Mapper; (*OUT*)VAR res: INTEGER);
		VAR mapper: Mapper;
	BEGIN
		NEW(mapper);
		mapper.sid := sid; mapr := mapper; res := 0
	END NewMapper;
	

END Ta_PINGStations.