MODULE Ta_PingEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Tt := Ta_Tables, Ta_STAreas, Log;
	
	CONST MAX_STATIONS = 4;
	
	TYPE 
		PingOpts* = RECORD (Tb.AnyRec)
			cnt*: INTEGER;
			interval*: Tb.Int32;
			hosts*: ARRAY 256 OF CHAR;
		END;
		PingResults* = RECORD (Tb.AnyRec)
			icmp_type*, icmp_recv*: Tb.Int8;
			icmp_seq*: Tb.Int16;
			t_sent*: Tb.Time;
		END;
		Sts = RECORD
			t: Ta.Table;
			name: Tb.Name;
			n_transmitted, n_received, n_errs: INTEGER;
		END;
		SeqMsg = RECORD (Tb.AnyRec)
			icmp_seq: Tb.Int16;
		END;
		
	
	VAR
		counter: INTEGER;
		ti, myti: Ta_CfgTInfo.TInfo;
		piops*: PingOpts;
		pirs*: PingResults;
		sts: ARRAY MAX_STATIONS OF Sts;
		area: Ta.Area;
		
	PROCEDURE Stop*;
	BEGIN
		IF area # NIL THEN
			area.Finish(Ta.NORMAL_FINISH);
			area := NIL
		END
	END Stop;
	
	PROCEDURE PrintResults(sts: Sts);
		VAR f: REAL; t: Tb.Time; s: Tb.Name; mcs: Tb.Int32;
	BEGIN
		Log.String(sts.name); Log.Char(':'); Log.Int(sts.n_transmitted); 
		Log.String(" packets transmitted,"); Log.Int(sts.n_received); 
		Log.String(" packets received,"); Log.Int(sts.n_errs); Log.String(" errors,");
		f := 100.0*(piops.cnt-sts.n_received)/piops.cnt; IF f < 0 THEN f := 0 END;
		Log.Real(f); Log.String("% packet loss"); Log.Ln;
	END PrintResults;

	PROCEDURE SendPingMessages (area: Ta.Area; par: Tb.AnyPtr);
		VAR j, n: INTEGER; sm: SeqMsg;
	BEGIN
		n := area.NStations();
		IF counter >= piops.cnt THEN
			FOR j := 1 TO n-1 DO
				PrintResults(sts[j])
			END;
			Stop
		ELSE
			sm.icmp_seq := SHORT(counter+1);
			FOR j := 1 TO n-1 DO
				INC(sts[j].n_transmitted);
				sts[j].t.Write(sm, FALSE)
			END;
			INC(counter);
		END
	END SendPingMessages;
	
	PROCEDURE OnPingResponse (t: Ta.Table; prod: Ta.ProData);
	BEGIN
		WITH t: Tt.Table DO
			Log.Int(t.hdr.buffer_length); Log.String(" from ");
			Log.String(sts[t.ind].name); Log.String(": res="); 
			Log.Int(t.hdr.mid - Tc.PING_MID); 
			Log.String(" icmp_seq="); Log.Int(t.hdr.mid_resp); Log.Ln;
			CASE t.hdr.mid OF
			| Tc.PING_WARNING_MID, Tc.PING_GOOD_MID:
				INC(sts[t.ind].n_received);
			| Tc.PING_BAD_MID:
				INC(sts[t.ind].n_errs);
			ELSE
			END
		END
	END OnPingResponse;
	
	(* ---------- Main Entry Point ---------- *)
	PROCEDURE Ping*;
		VAR aqos: Ta.AreaQos; cqos: Tc.ChannelQos; tqos: Ts.TableQos;
		host: Tb.Name; res, j, n: INTEGER;
	BEGIN
		myti.InitFromCommandLine(Ta.dir.command_line, Ta.dir.ind_command_line);
		counter := 0;
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(1983, "ping", MAX_STATIONS, aqos);
		host := "127.0.0.1";
		cqos := Tc.defaultCQos;
		cqos.protocol := "ICMP";
		cqos.prot_format := "Ta_PING";
		(* disable implicit ping, we'll do it *)
		cqos.enable_ping := FALSE;
		tqos := Ts.defaultTQos;
		area.AddUpdater(SendPingMessages, NIL, -piops.interval);
		area.AddStations(host, 1, 0, cqos, res);
		ASSERT(res = 0, Tb.INIT_FAILED);
		(*cqos.ping_set_err := FALSE;*)
		j := 1; n := 0; Tb.Scan(piops.hosts, host, n);
		WHILE host[0] # 0X DO
			area.AddStations(host, 1, 0, cqos, res);
			ASSERT(res = 0, Tb.INIT_FAILED);
			sts[j].t := area.NewTable(pirs, "ping", "PingResults", j, 0, 
				tqos, OnPingResponse);
			sts[j].t.ind := j;
			sts[j].name := host;
			sts[j].n_transmitted := 0; sts[j].n_received := 0; sts[j].n_errs := 0; 
			(* TInsert RTTI: no action on BB, used for XDS compatibility *)
			sts[j].t.TInsert(0, "icmp_type", Tb.int8Typ, Tb.exported, 1, NIL);		
			sts[j].t.TInsert(1, "icmp_recv", Tb.int8Typ, Tb.exported, 1, NIL);		
			sts[j].t.TInsert(2, "icmp_seq", Tb.int16Typ, Tb.exported, 1, NIL);		
			sts[j].t.TInsert(3, "t_sent", Tb.timeTyp, Tb.exported, 1, NIL);	
			(**)
			INC(j); Tb.Scan(piops.hosts, host, n);
		END;
		area.Start
	END Ping;
	
	(* ---------- Call From BB commander ---------- *)
	PROCEDURE PingCmd* (s, hosts: ARRAY OF CHAR);
	BEGIN
		ti.InitFromCommandLine(s, 0);
		myti.InitFromCommandLine(s, 0);
		Tb.Copy(hosts, piops.hosts);
		Ping
	END PingCmd;
	
	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "-";
		NEW(myti);
		myti.ConnectTo(piops, "", "", NIL);
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "cnt", Tb.int32Typ, Tb.exported, 1, NIL);
		myti.Insert(1, "interval", Tb.int32Typ, Tb.exported, 1, NIL);
		myti.Insert(2, "hosts", Tb.charTyp, Tb.exported, 256, NIL);
		piops.interval := 1000000;
		piops.cnt := 3;

		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_PingEC', 'PINGS', Ping, ti, dqos);
	END Initialize;
	
BEGIN
	Initialize
END Ta_PingEC.