MODULE Ta_SFSrchEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Files, Tf := Ta_Files, Ta_STAreas, Ta_TCPChannel, Log;

	CONST
		MAX_STATIONS* = 5; MAX_STRING_LENGTH = 128; N_STRINGS_AT_ONCE = 64; 
			N_STRINGS_FROM_FILE = 1024;

	TYPE
		FoundString* = RECORD (Tb.AnyRec)
			result_str* : ARRAY 128 OF Tb.Char8;
			pos* : Tb.Int32;
		END;
		SearchResults* = RECORD (Tb.AnyRec)
			pattern* : ARRAY 128 OF Tb.Char8;
			file_name* : ARRAY 128 OF Tb.Char8;
			start_pos*, end_pos*, request_number* : Tb.Int32;
			found* : POINTER TO ARRAY OF FoundString;
			n_found*: Tb.Int32;
		END;
		
		Opts* = RECORD (Tb.AnyRec)
			protocol*: Tb.Name16;
		END;

	VAR
		searchVar* : SearchResults;
		area: Ta.Area;
		searchTable: Ta.Table;
		request_numbers: ARRAY MAX_STATIONS OF Tb.Int32;

		(* configuration params *)
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
			
	(* -----------------------  BM Search - taken from ADruS193_BM   ----------------------- *)
	(* ind means start pos in text *)
	PROCEDURE Search (p, s: ARRAY OF Tb.Char8;  ind, M, N: Tb.Int32; VAR r: Tb.Int32 );
		VAR  i, j, k: Tb.Int32; d: ARRAY 128 OF Tb.Int32;
	BEGIN 
		FOR  i := 0  TO  127  DO  d[i] := M  END; 
		FOR  j := 0  TO  M-2  DO  d[ ORD( p[j] ) ] := M-j-1  END;

		i := M; j := M;  k := i;
		LOOP IF  (j > 0) & (i <= N) & (s[ind+k-1] = p[j-1])  THEN
			DEC(k); DEC(j);
		ELSIF  (j > 0) & (i <= N)  THEN
			i := i + d[ ORD( s[ind+i-1] ) ];  j := M;  k := i;
		ELSE EXIT END END;
		
		IF  j = 0  THEN  r := ind+k  ELSE  r := -1  END;
	END Search;

	(* -----------------------  Get Query and call Search  ----------------------- *)
	PROCEDURE ProcessSearchQuery(t: Ta.Table; prod: Ta.ProData; ptr: Tb.AnyPtr);
		CONST POS_EOF = -1; NOT_FOUND = -2;
		VAR k, i, j, j0, ri, M, N, pos, nstr: Tb.Int32; out : Tb.Int8; done: BOOLEAN; 
		found : ARRAY N_STRINGS_AT_ONCE OF FoundString;
		text : ARRAY MAX_STRING_LENGTH OF Tb.Char8; 
		str: ARRAY 256 OF CHAR; file: Files.File; rd: Files.Reader;
	BEGIN
		Log.String("Process pattern: ");  Tb.Char8_Char(searchVar.pattern, str);
		M := Tb.Len(str); Log.String(str); Log.String(" file_name:"); 
		Tb.Char8_Char(searchVar.file_name, str); Log.String(str); 
		Log.String(" pos:"); Log.Int(searchVar.start_pos); 
		Log.Int(searchVar.end_pos); Log.Ln;
		Tf.OldFile(str, TRUE, file);
		IF file = NIL THEN
			searchVar.end_pos := NOT_FOUND
		ELSIF searchVar.start_pos < 0 THEN
			file.Close;
			searchVar.end_pos := POS_EOF
		ELSE 
			rd := file.NewReader(NIL);
			pos := searchVar.start_pos;
			searchVar.end_pos := pos;
			rd.SetPos(pos);
			(* read string and search patterns *)
			out := 1; done := FALSE; k := 0; j := 0; ri := 0; nstr := 0;  
			LOOP IF (out # 0AH) & (out # 0DH) & ~rd.eof & ~done THEN
				rd.ReadByte(out); 
				IF (out # 0AH) & (out # 0DH) & ~rd.eof THEN
					text[k] := Tb.C_C8(CHR(out)); INC(k); 
				END;
				IF k >= MAX_STRING_LENGTH THEN
					k := MAX_STRING_LENGTH-1;
					out := 0DH
				END
			ELSIF (ri = 0) & ~done THEN
				text[k] := Tb.C_C8(CHR(0));
				i := 0; j0 := j; N := k;
				(* search for each string *)
				REPEAT
					Search(searchVar.pattern, text, i, M, N-i, ri);
					IF ri >= 0 THEN
						found[j].pos := ri;
						Tb.Copy8(text, found[j].result_str);
						INC(j);
						IF j < N_STRINGS_AT_ONCE THEN
							i := ri+1;
						ELSE
							j := j0;
							done := TRUE;
							ri := -4
						END
					END
				UNTIL ri < 0;
				IF ~done THEN
					searchVar.end_pos := rd.Pos();
				END;
				IF rd.eof THEN
					searchVar.end_pos := POS_EOF;
					done := TRUE
				ELSIF nstr < N_STRINGS_FROM_FILE THEN
					rd.ReadByte(out); k := 0; INC(nstr); 
					IF rd.eof THEN
						searchVar.end_pos := POS_EOF;
						done := TRUE
					ELSIF (out # 0AH) & (out # 0DH) THEN
						text[k] := Tb.C_C8(CHR(out)); INC(k)
					END;
					searchVar.end_pos := rd.Pos();
				ELSE
					searchVar.end_pos := rd.Pos();
					done := TRUE
				END;
				ri := 0;
				INC(pos,N);
			ELSE EXIT END END;
			file.Close; 
			(* allocate and prepare result *)
			IF j = 0 THEN
				searchVar.found  := NIL;
			ELSIF (searchVar.found = NIL) OR (searchVar.n_found # j) THEN
				NEW(searchVar.found, j);
			END;
			FOR i := 0 TO j-1 DO
				searchVar.found[i] := found[i];
			END;
		END;
		INC(request_numbers[prod.sid]);
		searchVar.request_number := request_numbers[prod.sid];
		IF searchVar.end_pos = NOT_FOUND THEN j := 0 END;
		searchVar.n_found := j;
		Log.String("Search for text file:"); Tb.Char8_Char(searchVar.file_name, str);
		Log.String(str); Log.String(" pattern:"); Tb.Char8_Char(searchVar.pattern, str);
		Log.String(str); Log.String(" resulted in"); Log.Int(j); Log.String(" matches pos:"); 
		Log.Int(searchVar.start_pos); Log.Int(searchVar.end_pos); Log.Ln;
		t.Write(searchVar, FALSE)
	END ProcessSearchQuery;
	
	(* -----------------------  SearchServerInit  ----------------------- *)
	PROCEDURE ReallocResults (dim: INTEGER; old_adr: Tb.Address; VAR adr: Tb.Address);
	BEGIN
		NEW(searchVar.found, dim);
		adr := 0
	END ReallocResults;
	
	PROCEDURE SearchServerInit;
		VAR tqos: Ts.TableQos; ti_found_string: Ti.TInfo;
	BEGIN
		tqos := Ts.defaultTQos;
		tqos.tp_timeout := -1;
		tqos.realloc_mode := Ts.REALLOC_EXACT;
		NEW(ti_found_string); 
		ti_found_string.ConnectTo(Tb.no_rec, "Ta_SFSrchEC", "FoundString", 
			ReallocResults);
		ti_found_string.Insert(0, "result_str", Tb.char8Typ, Tb.exported, 128, NIL);
		ti_found_string.Insert(1, "pos", Tb.int32Typ, Tb.exported, 1, NIL);
		NEW(searchVar.found, 1);
		searchTable := area.NewTable(searchVar, "Tests.FSRCH", "SearchResults",
			Ts.callerSID, 0, tqos, NIL);
		searchTable.ti.Insert(0, "pattern", Tb.char8Typ, Tb.exported, 128, NIL);
		searchTable.ti.Insert(1, "file_name", Tb.char8Typ, Tb.exported, 128, NIL);
		searchTable.ti.Insert(2, "start_pos", Tb.int32Typ, Tb.exported, 1, NIL);
		searchTable.ti.Insert(3, "end_pos", Tb.int32Typ, Tb.exported, 1, NIL);
		searchTable.ti.Insert(4, "request_number", Tb.int32Typ, Tb.exported, 1, NIL);
		searchTable.ti.Insert(5, "found", Tb.recTyp, Tb.exported, -1, 
			ti_found_string);
		searchTable.ti.Insert(6, "n_found", Tb.int32Typ, Tb.exported, 1, NIL);
		
		searchTable.Repeat(ProcessSearchQuery, NIL);
		
	END SearchServerInit;
		
	(* -----------------------  Server Search Initialize  ----------------------- *)
	PROCEDURE Srch*;
		VAR aqos: Ta.AreaQos; cqos: Ta_TCPChannel.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(16, "fsrch", MAX_STATIONS, aqos);
		cqos := Ta_TCPChannel.defaultCQos;
		cqos.protocol := opts.protocol;
		cqos.hostname := "0.0.0.0";
		cqos.port := 9000;
		cqos.send_strat := Tc.SEND_STRAT_SERVER;
		cqos.enable_ping := FALSE;
		
		area.AddStations("", MAX_STATIONS, 0, cqos, res);
		ASSERT(res = 0, 100);

		SearchServerInit;
		area.Start
	END Srch;
	
	PROCEDURE SrchCmd* (s: ARRAY OF CHAR);
	BEGIN
		ti.InitFromCommandLine(s, 0);
		Srch
	END SrchCmd;

	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "srch.cfg";
		NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "protocol", Tb.charTyp, Tb.exported, 16, NIL);
		opts.protocol := "TCP";

		(* module_name, program_name, entry_function *)
		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_SFSrchEC', 'SFSRCH', Srch, ti, dqos);
	END Initialize;

BEGIN
	Initialize		
END Ta_SFSrchEC.
