MODULE Ta_SGWeatherEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Ta_SOAPStations, Ta_STAreas, Log
,SYSTEM;

	CONST
		MAX_STATIONS* = 3; 

	TYPE
		CountryList* = RECORD
			Country* : ARRAY 128 OF Tb.Char16;
			City* : ARRAY 128 OF Tb.Char16;
		END;
		NDSet* = RECORD
			Table* : POINTER TO ARRAY OF CountryList;
		END;
		GetCitiesByCountry* = RECORD (Tb.AnyRec)
			CountryName* : ARRAY 128 OF Tb.Char16;
			NewDataSet* : NDSet;
		END;
		
		CWeather* = RECORD
			Location* : ARRAY 128 OF Tb.Char16;
			Time* : ARRAY 128 OF Tb.Char16;
			Wind* : ARRAY 128 OF Tb.Char16;
			Visibility* : ARRAY 128 OF Tb.Char16;
			SkyConditions* : ARRAY 128 OF Tb.Char16;
			Temperature* : Tb.Real32;
			DewPoint* : Tb.Real32;
			RelativeHumidity* : Tb.Int32;
			Pressure* : ARRAY 64 OF Tb.Char16;
			Status* : ARRAY 32 OF Tb.Char16;
		END;
		GetWeather* = RECORD (Tb.AnyRec)
			CityName* : ARRAY 128 OF Tb.Char16;
			CountryName* : ARRAY 128 OF Tb.Char16;
			CurrentWeather* : CWeather;
		END;

		Opts* = RECORD (Tb.AnyRec)
			port*: Tb.Int16;
		END;
		
	VAR
		getCitiesByCountryVar* : GetCitiesByCountry;
		getWeatherVar* : GetWeather;
		area: Ta.Area;
		getCitiesByCountryTable, getWeatherTable : Ta.Table;

		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
			
	(* -----------------------  Get List of Cities  ----------------------- *)
	PROCEDURE ProcessGetCitiesByCountry(t: Ta.Table; prod: Ta.ProData; ptr: Tb.AnyPtr);
		VAR s: Tb.Name;
	BEGIN
		Log.String("ProcessGetCitiesByCountry CountryName:"); 
		Tb.Char16_Char(getCitiesByCountryVar.CountryName, s);
		Log.String(s); Log.Ln;
			
		Tb.Copy16(getCitiesByCountryVar.CountryName,
			getCitiesByCountryVar.NewDataSet.Table[0].Country);
		Tb.Char_Char16("Kitezh", getCitiesByCountryVar.NewDataSet.Table[0].City);
		Tb.Copy16(getCitiesByCountryVar.CountryName,
			getCitiesByCountryVar.NewDataSet.Table[1].Country);
		Tb.Char_Char16("Tmutarakanh", getCitiesByCountryVar.NewDataSet.Table[1].City);
		Tb.Copy16(getCitiesByCountryVar.CountryName,
			getCitiesByCountryVar.NewDataSet.Table[2].Country);
		Tb.Char_Char16("Vassuki", getCitiesByCountryVar.NewDataSet.Table[2].City);
				
		getCitiesByCountryTable.Write(getCitiesByCountryVar, FALSE);
	END ProcessGetCitiesByCountry;
	
	(* -----------------------  Get Weather for requested City  ----------------------- *)
	PROCEDURE ProcessGetWeather(t: Ta.Table; prod: Ta.ProData; ptr: Tb.AnyPtr);
		VAR s: Tb.Name;
	BEGIN
		Log.String("ProcessGetWeather CountryName:"); 
		Tb.Char16_Char(getWeatherVar.CountryName, s);
		Log.String(s); 
		Log.String(" CityName:"); 
		Tb.Char16_Char(getWeatherVar.CityName, s);
		Log.String(s);Log.Ln;

		Tb.Char_Char16("Moscow, Russia", getWeatherVar.CurrentWeather.Location);
		Tb.Char_Char16("23/12/2012 00:00:00", getWeatherVar.CurrentWeather.Time);
		Tb.Char_Char16("Very Strong", getWeatherVar.CurrentWeather.Wind);
		Tb.Char_Char16("Low", getWeatherVar.CurrentWeather.Visibility);
		Tb.Char_Char16("Burning Sky", getWeatherVar.CurrentWeather.SkyConditions);
		getWeatherVar.CurrentWeather.Temperature := 38.5;
		getWeatherVar.CurrentWeather.DewPoint := 23.4;
		getWeatherVar.CurrentWeather.RelativeHumidity := 78;
		Tb.Char_Char16("Very High", getWeatherVar.CurrentWeather.Pressure);
		Tb.Char_Char16("Testing", getWeatherVar.CurrentWeather.Status);

		getWeatherTable.Write(getWeatherVar, FALSE);
	END ProcessGetWeather;
	
	(* -----------------------  ServerInit  ----------------------- *)
	PROCEDURE ServerInit ;
		VAR tqos: Ts.TableQos; ti_CountryList, ti_NDSet, ti_CWeather: Ti.TInfo;
	BEGIN
		tqos := Ts.defaultTQos;
		tqos.tp_timeout := -1;
		tqos.realloc_mode := Ts.DONT_REALLOC;
		
		NEW(getCitiesByCountryVar.NewDataSet.Table, 3);
		getCitiesByCountryTable := area.NewTable(getCitiesByCountryVar, "globalweather.asmx", 
			"GetCitiesByCountry", Ts.callerSID, 0, tqos, NIL);
		NEW(ti_CountryList); NEW(ti_NDSet); 
		ti_CountryList.ConnectTo(Tb.no_rec, "Ta_SGWeatherEC", "CountryList", NIL);
		ti_CountryList.Insert(0, "Country", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CountryList.Insert(1, "City", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_NDSet.ConnectTo(Tb.no_rec, "Ta_SGWeatherEC", "NDSet", NIL);
		ti_NDSet.Insert(0, "Table", Tb.recTyp, Tb.exported, -1, ti_CountryList);
		getCitiesByCountryTable.ti.Insert(0, "CountryName", Tb.char16Typ, Tb.exported, 128, NIL);
		getCitiesByCountryTable.ti.Insert(1, "NewDataSet", Tb.recTyp, Tb.exported, 1, ti_NDSet);
		getCitiesByCountryTable.Repeat(ProcessGetCitiesByCountry, NIL);
		
		getWeatherTable := area.NewTable(getWeatherVar, "globalweather.asmx", 
			"GetWeather", Ts.callerSID, 0, tqos, NIL);
		NEW(ti_CWeather);
		ti_CWeather.ConnectTo(Tb.no_rec, "Ta_SGWeatherEC", "CWeather", NIL);
		ti_CWeather.Insert(0, "Location", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(1, "Time", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(2, "Wind", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(3, "Visibility", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(4, "SkyConditions", Tb.char16Typ, Tb.exported, 128, NIL);
		ti_CWeather.Insert(5, "Temperature", Tb.real32Typ, Tb.exported, 1, NIL);
		ti_CWeather.Insert(6, "DewPoint", Tb.real32Typ, Tb.exported, 1, NIL);
		ti_CWeather.Insert(7, "RelativeHumidity", Tb.int32Typ, Tb.exported, 1, NIL);
		ti_CWeather.Insert(8, "Pressure", Tb.char16Typ, Tb.exported, 64, NIL);
		ti_CWeather.Insert(9, "Status", Tb.char16Typ, Tb.exported, 32, NIL);
		getWeatherTable.ti.Insert(0, "CityName", Tb.char16Typ, Tb.exported, 128, NIL);
		getWeatherTable.ti.Insert(1, "CountryName", Tb.char16Typ, Tb.exported, 128, NIL);
		getWeatherTable.ti.Insert(2, "CurrentWeather", Tb.recTyp, Tb.exported, 1, ti_CWeather);
		getWeatherTable.Repeat(ProcessGetWeather, NIL);
	END ServerInit;
		
	(* -----------------------  Server Search Initialize  ----------------------- *)
	PROCEDURE Start*;
		VAR aqos: Ta.AreaQos; cqos: Tc.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(22, "gweather", MAX_STATIONS, aqos);
		cqos := Tc.defaultCQos;
		cqos.prot_format := "Ta_SOAP";
		cqos.protocol := "TCP";
		cqos.hostname := "0.0.0.0";
		cqos.port := opts.port;
		cqos.send_strat := Tc.SEND_STRAT_SERVER;
		cqos.enable_ping := FALSE;
		area.AddStations("", MAX_STATIONS, 0, cqos, res);
		ASSERT(res = 0, 100);

		ServerInit;
		area.Start
	END Start;
	
	PROCEDURE Initialize;
	BEGIN
		NEW(ti); NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "port", Tb.int16Typ, Tb.exported, 1, NIL);
		opts.port := 80;
		Ta.dir.InitModule('Ta_SGWeatherEC', 'SGWEATHER', Start, ti, Ta.defaultDQos);
	END Initialize;

BEGIN
	Initialize		
END Ta_SGWeatherEC.
