MODULE Ta_SHMLib;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, Tl := Ta_Lib, Tb := Ta_Basics, Log , Files, Tf := Ta_Files;

	CONST
		SHMMIN = 128; O777 = 511; O444 = 292; IPC_CREAT = 512; IPC_RMID = 0; IPC_STAT = 2;

	TYPE
		ShmData* = RECORD
			file_id-, shm_id-: Tb.Int32;
			shm_buffer-: Tb.Address
		END;

	VAR
		zerodata-: ShmData;
		
	(* ========== SharedMemoryRoutines ========== *)
	PROCEDURE GetFileSize* ( filename-: ARRAY OF CHAR; VAR flen: Tb.Int32);
		VAR rc: Tb.Int32; st: Tl.struct_stat;
	BEGIN
		rc := Tl.__xstat(Tl.STAT_LIBC_VER, filename, st);
		IF rc < 0 THEN flen := -1 ELSE flen := st.st_size END;
	END GetFileSize;
	
	PROCEDURE CreateFileOfSize* ( filename-: ARRAY OF CHAR; 
	flen: Tb.Int32; VAR ok: BOOLEAN);
		VAR file: Files.File; wr: Files.Writer; byte: Tb.Int8;
	BEGIN
		Tf.NewFile(filename, file);
		IF file = NIL THEN
			ok := FALSE;
		ELSE
			wr := file.NewWriter(NIL);
			byte := 0;
			WHILE flen > 0 DO
				wr.WriteByte(byte); DEC(flen)
			END;
			file.Close;
			ok := TRUE
		END
	END CreateFileOfSize;

	PROCEDURE CreateShm* ( filename-: ARRAY OF CHAR; memsize: Tb.Int32;
	VAR shd: ShmData);
	BEGIN
		(* SHMMIN - the smallest shared memory size *)
		ASSERT(memsize >= SHMMIN, Tb.INVALID_VAL);
		shd.file_id := Tl.ftok(filename, 1);
		IF shd.file_id >= 0 THEN
			shd.shm_id := Tl.shmget(shd.file_id, memsize, O777+IPC_CREAT);
			IF shd.shm_id >= 0 THEN
				shd.shm_buffer := Tl.shmat(shd.shm_id, 0, 0)
			ELSE
				shd.shm_buffer := 0
			END
		ELSE
			shd.shm_id := 0;
			shd.shm_buffer := 0
		END;
	END CreateShm;
	
	PROCEDURE ConnectShm* ( filename-: ARRAY OF CHAR; memsize: Tb.Int32;
	readonly: BOOLEAN; VAR shd: ShmData);
		VAR sds: Tl.shmid_ds; rc: Tb.Int32;
	BEGIN
		shd.file_id := Tl.ftok(filename, 1);
		IF shd.file_id >= 0 THEN
			IF readonly THEN
				shd.shm_id := Tl.shmget(shd.file_id, memsize, O444)
			ELSE
				shd.shm_id := Tl.shmget(shd.file_id, memsize, O777)
			END
		ELSE
			shd.shm_id := -1
		END;
		IF shd.shm_id >= 0 THEN
			shd.shm_buffer := Tl.shmat(shd.shm_id, 0, 0)
		ELSE
			shd.shm_buffer := 0
		END
	END ConnectShm;

	PROCEDURE DisconnectShmBuffer* (VAR shd: ShmData);
		VAR rc: Tb.Int32;
	BEGIN
		IF shd.shm_buffer # 0 THEN
			rc := Tl.shmdt(shd.shm_buffer);
			shd.shm_buffer := 0;
		END;
	END DisconnectShmBuffer;
	
	PROCEDURE DeleteShm* (VAR shd: ShmData; delete: BOOLEAN);
		VAR rc: Tb.Int32;
		sds: Tl.shmid_ds;
	BEGIN
		(*IF shd.file_id > 0 THEN
			rc := Tl.CloseHandle(shd.file_id);
			shd.file_id := 0
		END;*)
		shd.file_id := 0;
		IF shd.shm_id > 0 THEN
			IF delete THEN
				rc := Tl.shmctl(shd.shm_id, IPC_RMID, sds)
			END;
			shd.shm_id := 0
		END
	END DeleteShm;
	
END Ta_SHMLib.
