MODULE Ta_SPSrchEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Tt := Ta_Tables, Files, Tf := Ta_Files, Ta_STAreas, Ta_MUDPChannel, Log, SYSTEM;

	CONST
		MAX_STATIONS* = 5; MAX_STRING_LENGTH = 128; N_STRINGS_AT_ONCE = 64; 
			N_STRINGS_FROM_FILE = 1024;

	TYPE
		FoundString* = RECORD (Tb.AnyRec)
			result_str* : ARRAY 128 OF Tb.Char8;
			pos* : Tb.Int32;
		END;
		SearchResults* = RECORD (Tb.AnyRec)
			pattern* : ARRAY 128 OF Tb.Char8;
			file_name* : ARRAY 128 OF Tb.Char8;
			start_pos*, end_pos*, request_number* : Tb.Int32;
			found* : POINTER TO ARRAY OF FoundString;
			n_found*: Tb.Int32;
		END;
		
		Opts* = RECORD (Tb.AnyRec)
			protocol*: Tb.Name16;
			pattern*, filename*: Tb.Name;
		END;

		
	VAR
		searchVar* : ARRAY MAX_STATIONS OF SearchResults;
		area: Ta.Area;
		searchTable: ARRAY MAX_STATIONS OF Ta.Table;
		request_numbers: ARRAY MAX_STATIONS OF Tb.Int32;

		(* configuration params *)
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
			
	(* -----------------------  BM Search - taken from ADruS193_BM   ----------------------- *)
	(* ind means start posin text *)
	PROCEDURE Search (p, s: ARRAY OF Tb.Char8;  ind, M, N: Tb.Int32; VAR r: Tb.Int32 );
		VAR  i, j, k: Tb.Int32; d: ARRAY 128 OF Tb.Int32;
	BEGIN 
		FOR  i := 0  TO  127  DO  d[i] := M  END; 
		FOR  j := 0  TO  M-2  DO  d[ ORD( p[j] ) ] := M-j-1  END;

		i := M; j := M;  k := i;
		LOOP IF  (j > 0) & (i <= N) & (s[ind+k-1] = p[j-1])  THEN
			DEC(k); DEC(j);
		ELSIF  (j > 0) & (i <= N)  THEN
			i := i + d[ ORD( s[ind+i-1] ) ];  j := M;  k := i;
		ELSE EXIT END END;
		
		IF  j = 0  THEN  r := ind+k  ELSE  r := -1  END;
	END Search;

	(* -----------------------  Get Query and call Search  ----------------------- *)
	PROCEDURE ProcessSearchQuery(t: Ta.Table; prod: Ta.ProData; ptr: Tb.AnyPtr);
		CONST POS_EOF = -1; NOT_FOUND = -2;
		VAR l, k, i, j0, j, ri, nstr, pos, M, N: Tb.Int32; out : Tb.Int8; 
		permanent, done: BOOLEAN;
		found : ARRAY N_STRINGS_AT_ONCE OF FoundString;
		text : ARRAY MAX_STRING_LENGTH OF Tb.Char8;
		str: ARRAY 256 OF CHAR; file: Files.File; rd: Files.Reader; tt: Tt.Table;
	BEGIN
		tt := t(Tt.Table); 
		IF tt.hdr.dst_oid <= 100 THEN
			l := 0
		ELSE
			ASSERT(tt.hdr.dst_oid-100 < MAX_STATIONS, 105);
			l := tt.hdr.dst_oid-100
		END;
		permanent := (l = 0);
		IF permanent THEN
			IF searchVar[l].end_pos < 0 THEN
				searchVar[l].start_pos := 0;
				searchVar[l].end_pos := 0;
			END
		ELSIF prod.got_message THEN
			searchVar[l].start_pos := 0
		ELSIF searchVar[l].end_pos < 0 THEN
			t.Clear(searchVar[l]);	(* stop publishing until next message *)
			RETURN
		ELSE
			searchVar[l].start_pos := searchVar[l].end_pos
		END;
		Log.String("Process "); Log.Int(tt.hdr.dst_oid); Log.String(" pattern: "); 
		Tb.Char8_Char(searchVar[l].pattern, str);
		M := Tb.Len(str); Log.String(str); Log.String(" file_name:"); 
		Tb.Char8_Char(searchVar[l].file_name, str); Log.String(str); 
		Log.String(" pos:"); Log.Int(searchVar[l].start_pos); 
		Log.Int(searchVar[l].end_pos); Log.Ln;
		Tf.OldFile(str, TRUE, file);
		IF file = NIL THEN
			searchVar[l].end_pos := NOT_FOUND
		ELSE 
			rd := file.NewReader(NIL);
			pos := searchVar[l].start_pos; 
			searchVar[l].end_pos := pos;
			rd.SetPos(pos);
			
			(* read string and search patterns *)
			out := 1; done := FALSE; k := 0; j := 0; ri := 0; nstr := 0; 
			LOOP IF (out # 0AH) & (out # 0DH) & ~rd.eof & ~done THEN
				rd.ReadByte(out);  
				IF (out # 0AH) & (out # 0DH) & ~rd.eof THEN
					text[k] := Tb.C_C8(CHR(out)); INC(k); 
				END;
				IF k >= MAX_STRING_LENGTH THEN
					k := MAX_STRING_LENGTH-1;
					out := 0DH
				END
			ELSIF (ri = 0) & ~done THEN
				text[k] := Tb.C_C8(CHR(0));
				i := 0; j0 := j; N := k;
				(* search for each string *)
				REPEAT
					Search(searchVar[l].pattern, text, i, M, N-i, ri);
					IF ri >= 0 THEN
						found[j].pos := ri;
						Tb.Copy8(text, found[j].result_str);
						INC(j);
						IF j < N_STRINGS_AT_ONCE THEN
							i := ri+1;
						ELSE
							j := j0;
							done := TRUE;
							ri := -4
						END
					END
				UNTIL ri < 0;
				IF ~done THEN
					searchVar[l].end_pos := rd.Pos();
				END;					
				IF rd.eof THEN
					searchVar[l].end_pos := POS_EOF;
					done := TRUE
				ELSIF nstr < N_STRINGS_FROM_FILE THEN
					rd.ReadByte(out); k := 0; INC(nstr); 
					IF rd.eof THEN
						searchVar[l].end_pos := POS_EOF;
						done := TRUE
					ELSIF (out # 0AH) & (out # 0DH) THEN
						text[k] := Tb.C_C8(CHR(out)); INC(k)
					END;
					searchVar[l].end_pos := rd.Pos();
				ELSE
					searchVar[l].end_pos := rd.Pos();
					done := TRUE
				END;
				ri := 0;
				INC(pos,N);
			ELSE EXIT END END;
			file.Close; 
			(* allocate and prepare result *)
			IF j = 0 THEN
				searchVar[l].found  := NIL;
			ELSIF searchVar[l].n_found # j THEN
				NEW(searchVar[l].found, j);
			END;
			FOR i := 0 TO j-1 DO
				searchVar[l].found[i] := found[i];
			END;
		END;
		searchVar[l].n_found := j;
		INC(request_numbers[l]);
		searchVar[l].request_number := request_numbers[l];
		Log.String("Search for text file:"); Tb.Char8_Char(searchVar[l].file_name, str);
		Log.String(str); Log.String(" pattern:"); Tb.Char8_Char(searchVar[l].pattern, str);
		Log.String(str); Log.String(" resulted in"); Log.Int(j); Log.String(" matches pos:"); 
		Log.Int(searchVar[l].start_pos); Log.Int(searchVar[l].end_pos); Log.Ln;
		t.Write(searchVar[l], FALSE)
	END ProcessSearchQuery;
	
	(* -----------------------  SearchServerInit  ----------------------- *)
	PROCEDURE ReallocResults (dim: INTEGER; old_adr: Tb.Address; VAR adr: Tb.Address);
		VAR found: POINTER TO ARRAY OF FoundString;
	BEGIN
		NEW(found, dim);
		adr := SYSTEM.VAL(Tb.Address, found);
	END ReallocResults;
	
	PROCEDURE SearchServerInit;
		VAR tqos: Ts.TableQos; j: INTEGER; ti_found_string: Ti.TInfo;
	BEGIN
		tqos := Ts.defaultTQos;
		tqos.tp_timeout := 2000000;
		tqos.realloc_mode := Ts.REALLOC_EXACT;
		NEW(ti_found_string); 
		ti_found_string.ConnectTo(Tb.no_rec, "Ta_SPSrchEC", "FoundString", 
			ReallocResults);
		ti_found_string.Insert(0, "result_str", Tb.char8Typ, Tb.exported, 128, NIL);
		ti_found_string.Insert(1, "pos", Tb.int32Typ, Tb.exported, 1, NIL);
		FOR j := 0 TO area.NStations()-1 DO
			NEW(searchVar[j].found, 1);
			searchTable[j] := area.NewTable(searchVar[j], "Tests.PSRCH", "SearchResults",
				Ts.hostSID, 100+j, tqos, NIL);
			searchTable[j].ind := j;
			IF j = 0 THEN
				searchTable[j].ti.Insert(0, "pattern", Tb.char8Typ, Tb.exported, 128, NIL);
				searchTable[j].ti.Insert(1, "file_name", Tb.char8Typ, Tb.exported, 128, NIL);
				searchTable[j].ti.Insert(2, "start_pos", Tb.int32Typ, Tb.exported, 1, NIL);
				searchTable[j].ti.Insert(3, "end_pos", Tb.int32Typ, Tb.exported, 1, NIL);
				searchTable[j].ti.Insert(4, "request_number", Tb.int32Typ, Tb.exported, 1, NIL);
				searchTable[j].ti.Insert(5, "found", Tb.recTyp, Tb.exported, -1, 
					ti_found_string);
				searchTable[j].ti.Insert(6, "n_found", Tb.int32Typ, Tb.exported, 1, NIL);
				searchTable[0].Publish(TRUE, ProcessSearchQuery, NIL);
				Tb.Char_Char8(opts.pattern, searchVar[0].pattern);
				Tb.Char_Char8(opts.filename, searchVar[0].file_name);
			ELSE
				searchTable[j].ti.CopyFrom(searchTable[0].ti);
				searchTable[j].Publish(FALSE, ProcessSearchQuery, NIL);
			END
		END
	END SearchServerInit;
		
	(* -----------------------  Server Search Initialize  ----------------------- *)
	PROCEDURE Srch*;
		VAR aqos: Ta.AreaQos; cqos: Ta_MUDPChannel.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(19, "psrch", MAX_STATIONS, aqos);
		cqos := Ta_MUDPChannel.defaultCQos;
		cqos.protocol := opts.protocol;
		cqos.hostname := "0.0.0.0";
		cqos.port := 8000;
		cqos.send_strat := Tc.SEND_STRAT_SERVER;
		cqos.trust_src_sid := TRUE;
		
		area.AddStations("", MAX_STATIONS, 0, cqos, res);
		ASSERT(res = 0, 100);

		SearchServerInit;
		area.Start
	END Srch;

	PROCEDURE SrchCmd* (s: ARRAY OF CHAR);
	BEGIN
		ti.InitFromCommandLine(s, 0);
		Srch
	END SrchCmd;

	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "srch.cfg";
		NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "protocol", Tb.charTyp, Tb.exported, 16, NIL);
		myti.Insert(1, "pattern", Tb.charTyp, Tb.exported, 256, NIL);
		myti.Insert(2, "filename", Tb.charTyp, Tb.exported, 256, NIL);
		opts.protocol := "MUDP";
		opts.pattern := "BlackBox";
		opts.filename := "StartupBlackBox.vbs";

		(* module_name, program_name, entry_function *)
		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_SPSrchEC', 'SPSRCH', Srch, ti, dqos);
	END Initialize;

BEGIN
	Initialize		
END Ta_SPSrchEC.
