MODULE Ta_S_SrchEC;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Ta_CfgTInfo, Tc := Ta_Channels, Ta := Ta_Areas, Ts := Ta_Stations,
		Ti := Ta_TInfo, Tt := Ta_Tables, Ta_STAreas, Ta_TCPChannel, Log;

	CONST
		MAX_STATIONS* = 5; MAX_STRING_LENGTH = 128; N_STRINGS_AT_ONCE = 16;
		TEXT_SIZE = (MAX_STRING_LENGTH * N_STRINGS_AT_ONCE);

	TYPE
		SearchQuery* = RECORD (Tb.AnyRec)
			pattern* : ARRAY 128 OF Tb.Char8;
			text* : POINTER TO ARRAY OF Tb.Char8;
		END;
		FoundString* = RECORD (Tb.AnyRec)
			result_str* : ARRAY 128 OF Tb.Char8;
			pos* : Tb.Int32;
		END;
		SearchResults* = RECORD (Tb.AnyRec)
			found* : POINTER TO ARRAY OF FoundString;
			n_found*: Tb.Int32;
		END;

		Opts* = RECORD (Tb.AnyRec)
			protocol*: Tb.Name16;
		END;
		
	VAR
		queryVar* : SearchQuery;
		resultsVar* : SearchResults;
		area: Ta.Area;
		queryTable, resultsTable: Ta.Table;

		(* configuration params *)
		ti, myti: Ta_CfgTInfo.TInfo;
		opts*: Opts;
			
	(* -----------------------  BM Search - taken from ADruS193_BM   ----------------------- *)
	(* ind means start pos in text *)
	PROCEDURE Search (p, s: ARRAY OF Tb.Char8;  ind, M, N: Tb.Int32; VAR r: Tb.Int32 );
		VAR  i, j, k: Tb.Int32; d: ARRAY 128 OF Tb.Int32;
	BEGIN 
		FOR  i := 0  TO  127  DO  d[i] := M  END; 
		FOR  j := 0  TO  M-2  DO  d[ ORD( p[j] ) ] := M-j-1  END;

		i := M; j := M;  k := i;
		LOOP IF  (j > 0) & (i <= N) & (s[ind+k-1] = p[j-1])  THEN
			DEC(k); DEC(j);
		ELSIF  (j > 0) & (i <= N)  THEN
			i := i + d[ ORD( s[ind+i-1] ) ];  j := M;  k := i;
		ELSE EXIT END END;
		
		IF  j = 0  THEN  r := ind+k  ELSE  r := -1  END;
	END Search;

	(* -----------------------  Get Query and call Search  ----------------------- *)
	PROCEDURE ProcessSearchQuery(t: Ta.Table; prod: Ta.ProData; ptr: Tb.AnyPtr);
		VAR i, j, ri, M, N, dr: Tb.Int32; done: BOOLEAN; c: Tb.Char8;
		poss : ARRAY N_STRINGS_AT_ONCE*MAX_STRING_LENGTH OF Tb.Int32;
		str: ARRAY 256 OF CHAR;
	BEGIN
		i := 0; j := 0; M := Tb.Len8(queryVar.pattern); N := Tb.Len8(queryVar.text^);
		Log.String("Process pattern: "); Tb.Char8_Char(queryVar.pattern, str);
		Log.String(str); Log.Int(N); Log.Ln;
		REPEAT
			Search(queryVar.pattern, queryVar.text^, i, M, N-i, ri);
			IF ri >= 0 THEN
				ASSERT(j<N_STRINGS_AT_ONCE*MAX_STRING_LENGTH, 101);
				poss[j] := ri;
				INC(j);
				i := ri+1;
			END
		UNTIL ri < 0;		
		IF j = 0 THEN
			resultsVar.found  := NIL;
		ELSE
			NEW(resultsVar.found, j);
		END;
		resultsVar.n_found := j;
		FOR i := 0 TO j-1 DO
			resultsVar.found[i].pos := poss[i];
			ASSERT(i<N_STRINGS_AT_ONCE*MAX_STRING_LENGTH, 102);
			dr := 0; done := FALSE;
			WHILE ~done & (dr > -MAX_STRING_LENGTH) & (poss[i]+dr > 0) DO
				ASSERT((poss[i]+dr>=0)&(poss[i]+dr<N), 103);
				IF (queryVar.text[poss[i]+dr] = Tb.C_C8(CHR(0DH))) 
				OR (queryVar.text[poss[i]+dr] = Tb.C_C8(CHR(0AH))) THEN
					INC(dr);
					done := TRUE
				ELSE
					DEC(dr)
				END
			END;
			ri := 0; done := FALSE;
			WHILE ~done & (ri < MAX_STRING_LENGTH) DO
				ASSERT((poss[i]+dr+ri>=0)&(poss[i]+dr+ri<=N),104);
				c := queryVar.text[poss[i]+dr+ri];
				IF (c = Tb.C_C8(CHR(0DH))) OR (c = Tb.C_C8(CHR(0AH))) 
				OR (c = Tb.C_C8(CHR(0H))) THEN
					done := TRUE
				ELSE
					ASSERT((ri>=0)&(ri<MAX_STRING_LENGTH),105);
					resultsVar.found[i].result_str[ri] := c;
					INC(ri);
				END
			END;
			IF ri >= MAX_STRING_LENGTH THEN ri := MAX_STRING_LENGTH END;
			resultsVar.found[i].result_str[ri] := Tb.C_C8(CHR(0))
		END;
		Log.String("Search for text len:"); Log.Int(N); Log.String(" pattern:"); 		
		Tb.Char8_Char(queryVar.pattern, str); Log.String(str);
		Log.String(" resulted in"); Log.Int(j); Log.String(" matches"); Log.Ln;
		(* write to queryTable Caller *)
		resultsTable.WriteTo(resultsVar, {prod.sid})
	END ProcessSearchQuery;
	
	(* -----------------------  SearchServerInit  ----------------------- *)
	PROCEDURE SearchServerInit;
		VAR tqos: Ts.TableQos; ti_found_string: Ti.TInfo;
	BEGIN
		tqos := Ts.defaultTQos;
		tqos.tp_timeout := -1;
		tqos.realloc_mode := Ts.REALLOC_LARGEST;
		NEW(queryVar.text, 1);
		queryTable := area.NewTable(queryVar, "Tests.SRCH", "SearchQuery",
			Ts.callerSID, 0, tqos, NIL);
		queryTable.ti.Insert(0, "pattern", Tb.char8Typ, Tb.exported, 128, NIL);
		queryTable.ti.Insert(1, "text", Tb.char8Typ, Tb.exported, -1, NIL);

		NEW(ti_found_string); 
		ti_found_string.ConnectTo(Tb.no_rec, "Ta_S_SrchEC", "FoundString", NIL);
		ti_found_string.Insert(0, "result_str", Tb.char8Typ, Tb.exported, 128, NIL);
		ti_found_string.Insert(1, "pos", Tb.int32Typ, Tb.exported, 1, NIL);
		NEW(resultsVar.found, 1);
		resultsTable := area.NewTable(resultsVar, "Tests.SRCH", "SearchResults",
			Ts.sendSID, 0, tqos, NIL);
		resultsTable.ti.Insert(0, "found", Tb.recTyp, Tb.exported, -1, 
			ti_found_string);
		resultsTable.ti.Insert(1, "n_found", Tb.int32Typ, Tb.exported, 1, NIL);
		
		queryTable.Repeat(ProcessSearchQuery, NIL);
		
	END SearchServerInit;
		
	(* -----------------------  Server Search Initialize  ----------------------- *)
	PROCEDURE Srch*;
		VAR aqos: Ta.AreaQos; cqos: Ta_TCPChannel.ChannelQos; res: INTEGER;
	BEGIN
		aqos := Ta.defaultAQos;
		area := Ta.dir.NewArea(17, "srch", MAX_STATIONS, aqos);
		cqos := Ta_TCPChannel.defaultCQos;
		cqos.protocol := opts.protocol;
		cqos.hostname := "0.0.0.0";
		cqos.port := 9000;
		cqos.send_strat := Tc.SEND_STRAT_SERVER;
		
		area.AddStations("", MAX_STATIONS, 0, cqos, res);
		ASSERT(res = 0, 100);

		SearchServerInit;
		area.Start
	END Srch;
	
	PROCEDURE SrchCmd* (s: ARRAY OF CHAR);
	BEGIN
		ti.InitFromCommandLine(s, 0);
		Srch
	END SrchCmd;

	PROCEDURE Initialize;
		VAR dqos: Ta.DirectoryQos;
	BEGIN
		NEW(ti); ti.cfg_file := "srch.cfg";
		NEW(myti);
		myti.ConnectTo(opts, "", "", NIL);
		myti.LinkMeTo(ti, "");
		(* Insert RTTI: no action on BB, used for XDS compatibility *)
		myti.Insert(0, "protocol", Tb.charTyp, Tb.exported, 16, NIL);
		opts.protocol := "TCP";

		(* module_name, program_name, entry_function *)
		dqos := Ta.defaultDQos;
		dqos.cycle := 100000;
		Ta.dir.InitModule('Ta_S_SrchEC', 'S_SRCH', Srch, ti, dqos);
	END Initialize;

BEGIN
	Initialize		
END Ta_S_SrchEC.
