MODULE Ta_SimpleParsers;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tm := Ta_TunedMappers, Log;
	
	CONST
		MAX_STACK* = 32;
		COUPLE_BEGIN* = 1; COUPLE_END* = 2; COUPLE_BOTH* = 3; 
			COUPLE_ARG* = 4; COUPLE_DIM* = 5; 
		PARSE_DELIM* = -11; PARSE_MATCH* = -12; PARSE_STR* = -13;
		PARSE_QUOTES* = -14; PARSE_INT* = -15;
		
	TYPE
		NamesStack* = RECORD
			name-: Tb.Name;
			index-: INTEGER;
		END;
		NamesStackArray* = ARRAY OF NamesStack;
		
		OnParseCouple* = PROCEDURE (parse_flag: INTEGER; 
			VAR nstack: NamesStackArray; nlevel: INTEGER; value: ARRAY OF CHAR;
			VAR userdata: Tb.AnyRec; VAR res: INTEGER);
		
	(* ========== User Interface ========== *)
	(** command string params: cmd -option1 val -opt2 'val 2' -opt3 "3 val" -o4 {v 4} **)
	PROCEDURE^ ParseOpVal* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);

	(** config file data: xch[3].hostname=10.0.201.216 **)
	PROCEDURE^ ParseCfg* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);

	(** HTTP parameters passing: City=Moscow&Country=Russian+Federation **)
	PROCEDURE^ ParseLine* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);
	
	PROCEDURE^ ParseXML* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);

	(* ========== Implementations ========== *)
	PROCEDURE ParseOpVal* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);
		CONST MODE_COMMAND = COUPLE_BEGIN; MODE_ARG = COUPLE_ARG; 
			MODE_NAME = 8; MODE_VALUE = 9;
		VAR nm: ARRAY MAX_STACK OF NamesStack; zv: ARRAY 1 OF CHAR; value: Tb.Name; 
			mode, ind: INTEGER; delim: CHAR;
	BEGIN
		s.SetOpts({Tm.withCrLf, Tm.withSpaces, Tm.separateStrings, Tm.separateQuotes});
		s.Scan; ind := 0; nm[ind].index := 0; nm[ind].name[0] := 0X; zv[0] := 0X; res := 0; 
		mode := MODE_COMMAND; delim := 0X;
		WHILE (s.type # Tm.eot) & (res = 0) DO
			CASE mode OF
			| MODE_COMMAND, MODE_ARG:
				IF (s.string = '#') & (mode = MODE_COMMAND) THEN
					REPEAT s.Scan UNTIL (s.type = Tm.crlf) OR (s.type = Tm.eot)
				ELSIF s.string = '-' THEN
					mode := MODE_NAME; 
					ind := 1; nm[ind].index := 0; nm[ind].name[0] := 0X;
				ELSIF s.type = Tm.crlf THEN
					IF nm[0].name[0] # 0X THEN
						opc(COUPLE_END, nm, 0, zv, userdata, res)
					END;
					mode := MODE_COMMAND;
					ind := 0; nm[ind].index := 0; nm[ind].name[0] := 0X;
				ELSIF s.type # Tm.spaces THEN
					Tb.Append(s.string, nm[ind].name)
				ELSIF (s.type # Tm.special) & ((s.string = '"') OR (s.string = "'")) THEN
				ELSIF nm[ind].name[0] # 0X THEN
					opc(mode, nm, ind, zv, userdata, res);
					mode := MODE_ARG; 
					ind := 1; nm[ind].index := 0; nm[ind].name[0] := 0X;
				END;
			| MODE_NAME:
				IF s.type = Tm.string THEN
					Tb.Copy(s.string, nm[ind].name);
				ELSIF s.type = Tm.spaces THEN
					mode := MODE_VALUE; value[0] := 0X; delim := 0X
				ELSIF s.string = "." THEN
					opc(COUPLE_BOTH, nm, ind, zv, userdata, res);
					INC(ind); nm[ind].index := 0; nm[ind].name[0] := 0X;
				ELSIF (s.type # Tm.special) & ((s.string = '"') OR (s.string = "'")) THEN
				ELSE
					res := PARSE_STR
				END;
			| MODE_VALUE:
				IF s.type = Tm.spaces THEN
					IF delim # 0X THEN
						Tb.Append(s.string, value)
					ELSIF value[0] # 0X THEN
						opc(COUPLE_BOTH, nm, ind, value, userdata, res);
						mode := MODE_ARG;
						ind := 1; nm[ind].index := 0; nm[ind].name[0] := 0X;
					END
				ELSIF s.type = Tm.crlf THEN
					IF delim # 0X THEN
						res := PARSE_DELIM
					ELSIF value[0] # 0X THEN
						opc(COUPLE_BOTH, nm, ind, value, userdata, res);
						IF nm[0].name[0] # 0X THEN
							opc(COUPLE_END, nm, 0, zv, userdata, res)
						END;
						mode := MODE_COMMAND;
						ind := 0; nm[ind].index := 0; nm[ind].name[0] := 0X;
					END
				ELSIF (s.string[0] = '"') OR (s.string[0] = "'") THEN
					IF delim = 0X THEN
						delim := s.string[0]
					ELSIF delim = s.string[0] THEN
						delim := 0X;
						opc(COUPLE_BOTH, nm, ind, value, userdata, res);
						mode := MODE_ARG;
						ind := 1; nm[ind].index := 0; nm[ind].name[0] := 0X;
					ELSE
						Tb.Append(s.string, value)
					END
				ELSIF s.string[0] = '{' THEN
					IF delim = 0X THEN
						delim := s.string[0]
					ELSE
						Tb.Append(s.string, value)
					END
				ELSIF s.string[0] = '}' THEN
					IF delim = '{' THEN
						delim := 0X;
						opc(COUPLE_BOTH, nm, ind, value, userdata, res);
						mode := MODE_ARG;
						ind := 1; nm[ind].index := 0; nm[ind].name[0] := 0X;
					ELSE
						Tb.Append(s.string, value)
					END
				ELSE
					Tb.Append(s.string, value);
					(*IF delim = 0X THEN
						opc(COUPLE_BOTH, nm, ind, value, userdata, res);
						mode := MODE_ARG;
						ind := 1; nm[ind].index := 0; nm[ind].name[0] := 0X;
					END*)
				END
			END;
			s.Scan
		END;
		IF res = 0 THEN
			IF delim # 0X THEN
				res := PARSE_QUOTES
			ELSIF (nm[ind].name[0] # 0X) & (value[0] # 0X) THEN
				opc(COUPLE_BOTH, nm, ind, value, userdata, res);
				IF nm[0].name[0] # 0X THEN
					opc(COUPLE_END, nm, 0, zv, userdata, res)
				END
			END
		END
	END ParseOpVal;

	PROCEDURE ParseCfg* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);
		VAR nm: ARRAY MAX_STACK OF NamesStack; nst, ind: INTEGER;
		zv: ARRAY 1 OF CHAR; value: Tb.Name; intag: BOOLEAN;
	BEGIN
		s.SetOpts({Tm.withCrLf, Tm.withSpaces, Tm.separateStrings});
		nst := 0; nm[0].index := 0; nm[0].name[0] := 0X;
		s.Scan; res := 0; intag := FALSE;
		WHILE (s.type # Tm.eot) & (res = 0) DO
			IF ~intag THEN
				IF s.type = Tm.string THEN
					Tb.Copy(s.string, nm[nst].name)
				ELSIF (s.type = Tm.special) & (s.string = '=') THEN
					value[0] := 0X;
					intag := TRUE
				ELSIF (s.type = Tm.special) & (s.string = '.') THEN
					opc(COUPLE_BEGIN, nm, nst, zv, userdata, res);
					INC(nst); nm[nst].index := 0; nm[nst].name[0] := 0X
				ELSIF (s.type = Tm.braces) & (s.string = '[') THEN
					s.Scan;
					IF s.type # Tm.string THEN
						res := PARSE_DELIM
					ELSE
						Tb.StringToInt(s.string, ind, res);
						IF (res # 0) OR (ind < 0) THEN 
							res := PARSE_INT
						ELSE
							s.Scan;
							IF (s.type = Tm.braces) & (s.string = ']') THEN
								nm[nst].index := ind
							ELSE
								res := PARSE_DELIM
							END
						END
					END
				ELSIF (s.type = Tm.special) & (s.string = '#') THEN
					REPEAT s.Scan UNTIL s.type = Tm.crlf;
					nst := 0; nm[0].index := 0; nm[0].name[0] := 0X
				ELSIF s.type = Tm.crlf THEN
					nst := 0; nm[0].index := 0; nm[0].name[0] := 0X
				ELSIF s.type # Tm.spaces THEN
					res := PARSE_STR
				END 
			ELSE
				IF s.type = Tm.crlf THEN
					opc(COUPLE_BOTH, nm, nst, value, userdata, res);
					nst := 0; nm[0].index := 0; nm[0].name[0] := 0X;
					intag := FALSE
				ELSE
					Tb.Append(s.string, value)
				END
			END;
			s.Scan
		END;
		IF (s.type = Tm.eot) & (nm[nst].name[0] # 0X) THEN
			opc(COUPLE_BOTH, nm, nst, value, userdata, res)
		END
	END ParseCfg;
	
	PROCEDURE ParseLine* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);
		VAR nm: ARRAY 1 OF NamesStack; zv: ARRAY 1 OF CHAR;
	BEGIN
		s.SetOpts({Tm.withCrLf, Tm.withSpaces, Tm.separateStrings, Tm.bracesAsString,
			Tm.unquotePercent});
		nm[0].index := 0; zv[0] := 0X;
		s.Scan; res := 0;
		WHILE (s.type # Tm.eot) & (s.type # Tm.crlf) & (s.type # Tm.spaces) & (res = 0) DO
			Tb.Copy(s.string, nm[0].name);
			s.Scan;
			IF (s.type # Tm.special) OR (s.string # '=') THEN
				res := PARSE_DELIM
			ELSE
				s.Scan;
				IF s.type = Tm.string THEN
					opc(COUPLE_BOTH, nm, 0, s.string, userdata, res);
					s.Scan
				ELSE
					opc(COUPLE_BOTH, nm, 0, zv, userdata, res)
				END;
				IF s.type = Tm.special THEN
					s.Scan
				END
			END
		END
	END ParseLine;
	
	PROCEDURE Copy( src-: ARRAY OF CHAR; VAR dst: ARRAY OF CHAR;
	VAR changed: BOOLEAN);
		VAR j: INTEGER;
	BEGIN
		changed := FALSE; j := 0;
		WHILE src[j] # 0X DO
			IF dst[j] # src[j] THEN changed := TRUE END;
			dst[j] := src[j]; INC(j)
		END;
		IF dst[j] # 0X THEN changed := TRUE END;
		dst[j] := 0X
	END Copy;
	
	PROCEDURE ParseXML* (VAR s: Tm.Scanner; opc: OnParseCouple; 
	VAR userdata: Tb.AnyRec; VAR res: INTEGER);
		CONST 
			TAGS_SET = {Tm.xmlMap, Tm.withSpaces, Tm.unquoteAmp};
			STRINGS_SET = {Tm.xmlMap, Tm.xmlStrings, Tm.unquoteAmp};
		VAR nm: ARRAY MAX_STACK OF NamesStack; nst, lq: INTEGER;
		intag, changed: BOOLEAN; arg, value: Tb.Name;
	BEGIN
		s.SetOpts(TAGS_SET);
		FOR nst := 0 TO MAX_STACK-1 DO
			nm[nst].index := 0; nm[nst].name[0] := 0X
		END; 
		s.Scan; res := 0; nst := 0; intag := FALSE; 
		WHILE (s.type # Tm.eot) & (res = 0) & (nst >= 0) DO
			IF s.type = Tm.special THEN
				IF ~intag & (s.string = "<") THEN
					intag := TRUE;
					s.Scan;
					s.SetOpts(STRINGS_SET);
					Copy(s.string, nm[nst].name, changed); value[0] := 0X;
					IF changed THEN nm[nst].index := 0 ELSE INC(nm[nst].index) END;
					opc(COUPLE_BEGIN, nm, nst, value, userdata, res);
					INC(nst)
				ELSIF ~intag & (s.string = "</") THEN
					intag := TRUE;
					DEC(nst);
					IF nst >= 0 THEN
						s.Scan;
						s.SetOpts(STRINGS_SET);
						IF ~Tb.Equal(s.string, nm[nst].name) THEN
							res := PARSE_MATCH
						ELSE
							opc(COUPLE_END, nm, nst, value, userdata, res)
						END
					END
				ELSIF intag & (s.string = "/>") THEN
					intag := FALSE;
					s.SetOpts(TAGS_SET);
					DEC(nst);
					value[0] := 0X;
					opc(COUPLE_END, nm, nst, value, userdata, res)
				ELSIF intag & (s.string = ">") THEN
					intag := FALSE; value[0] := 0X;
					s.SetOpts(TAGS_SET)
				END
			ELSIF s.type = Tm.spaces THEN
				IF ~intag THEN
					Tb.Append(s.string, value)
				END
			ELSIF s.type = Tm.string THEN
				IF ~intag THEN
					Tb.Append(s.string, value)
				ELSE
					Tb.Copy(s.string, arg);
					s.Scan;
					IF s.string # "=" THEN
						res := PARSE_DELIM
					ELSE
						s.Scan;
						IF (s.type # Tm.quote) & (s.type # Tm.dquote) THEN
							res := PARSE_QUOTES
						ELSE
							lq := s.type;
							s.Scan;
							IF s.type # Tm.string THEN
								res := PARSE_STR
							ELSE
								Tb.Copy(arg, nm[nst].name); 
								opc(COUPLE_ARG, nm, nst, s.string, userdata, res)
							END;
							arg[0] := 0X;
							s.Scan;
							IF s.type # lq THEN
								res := PARSE_QUOTES
							END
						END
					END
				END
			END;
			s.Scan
		END
	END ParseXML;

END Ta_SimpleParsers.
