MODULE Ta_TInfo;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
	
	Basic Type Info functionality
**)

	IMPORT Tb := Ta_Basics, SYSTEM, oberonRTS, Log;

	CONST
		STOP_ITERATE* = 1; ITERATE_NEXT* = 2; ITERATE_SKIP* = 3; ITERATE_INTO* = 4;
			ITERATE_FOREACH* = 5;
		MAX_ALIGNMENT = 4;

	
	TYPE
		ReallocCommand* = PROCEDURE (dim: INTEGER; old_adr: Tb.Address; VAR adr: Tb.Address);
		
		TInfo* = POINTER TO TInfoDesc;
		PItem = POINTER TO Item;
		Name* = Tb.Name;
		Item* = RECORD (Tb.AnyRec)
			typ-: Tb.Type;
			adr-: Tb.Address;
			vis-: INTEGER;
			size, alignment, offset, length, typof: LONGINT;
			name: Name;
			tinfo: TInfo;
			next, into: PItem
		END;
		TInfoDesc* =  RECORD
			item-: Item;
			size, packed_size, alignment: LONGINT;
			module: oberonRTS.Module;
			type: oberonRTS.Type;
			typof: LONGINT;
			realloc_cmd: ReallocCommand;
		END;
		
		Iterator* = PROCEDURE ( name-: ARRAY OF CHAR;  item-: Item; 
		begin: BOOLEAN; VAR client_rec: Tb.AnyRec): INTEGER;
		
		
	VAR
		zeroitem-: Item;
		
	
	(* ========== Item ========== *)
	PROCEDURE PtrLen* (item-: Item): LONGINT;
		VAR j, len: LONGINT;
	BEGIN
		ASSERT(item.adr # 0, Tb.NIL_POINTER);
		ASSERT(item.into # NIL, Tb.NIL_POINTER);
		SYSTEM.GET(item.adr, j);
		IF j = 0 THEN
			len := 0
		ELSE
			SYSTEM.GET(j+4, len);
		END;
		RETURN len;
	END PtrLen;
	
	PROCEDURE ItemTypOf* (item: Item): LONGINT;
		VAR typof: LONGINT;
	BEGIN
		IF item.tinfo # NIL THEN
			typof := item.tinfo.typof
		ELSE
			IF item.typof = 0 THEN
				item.typof := Tb.Crc32(item.name, -1, 0, 32)
			END;
			typof := item.typof
		END;
		RETURN typof
	END ItemTypOf;
	
	PROCEDURE MakeArr* (item-: Item; dim: INTEGER);
		VAR it: PItem; addr: SYSTEM.ADDRESS; b: POINTER TO ARRAY OF BOOLEAN; 
		c8: POINTER TO ARRAY OF CHAR; c16: POINTER TO ARRAY OF Tb.Char16;
		i8: POINTER TO ARRAY OF Tb.Int8; i16: POINTER TO ARRAY OF INTEGER;
		i32: POINTER TO ARRAY OF LONGINT; r32: POINTER TO ARRAY OF Tb.Real32;
		r64: POINTER TO ARRAY OF Tb.Real64; i64: POINTER TO ARRAY OF Tb.Int64;
		adr2: Tb.Address;
	BEGIN
		ASSERT(item.adr # 0, Tb.NIL_POINTER);
		IF dim = 0 THEN
			addr := SYSTEM.VAL(SYSTEM.ADDRESS, 0);
			SYSTEM.PUT(item.adr, addr)
		ELSE
			ASSERT(item.into # NIL, Tb.NIL_POINTER);
			ASSERT(item.into.into # NIL, Tb.NIL_POINTER);
			it := item.into.into;
			CASE it.typ OF
			| Tb.boolTyp: NEW(b,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, b));
			| Tb.char8Typ: NEW(c8,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, c8));
			| Tb.char16Typ: NEW(c16,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, c16));
			| Tb.int8Typ: NEW(i8,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, i8));
			| Tb.int16Typ: NEW(i16,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, i16));
			| Tb.int32Typ: NEW(i32,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, i32));
			| Tb.real32Typ: NEW(r32,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, r32));
			| Tb.real64Typ: NEW(r64,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, r64));
			| Tb.int64Typ: NEW(i64,dim); SYSTEM.PUT(item.adr, SYSTEM.VAL(LONGINT, i64));
			| Tb.recTyp:
				ASSERT(it.tinfo # NIL, Tb.ZERO_VAL);
				ASSERT(it.tinfo.module # oberonRTS.nullModule, Tb.NOT_FOUND);
				ASSERT(it.tinfo.type # oberonRTS.nullType, Tb.NOT_FOUND);
				IF it.tinfo.realloc_cmd # NIL THEN
					it.tinfo.realloc_cmd(dim, item.adr, adr2);
					IF adr2 # 0 THEN
						addr := SYSTEM.VAL(SYSTEM.ADDRESS, adr2);
						SYSTEM.PUT(item.adr, addr)
					END
				ELSE
					ASSERT(dim = 1, Tb.INVALID_EXTENSION);
					addr := oberonRTS.NewObj(it.tinfo.type);
					SYSTEM.PUT(item.adr, addr)
				END;
			ELSE
				ASSERT(item.into.into.typ = Tb.boolTyp, Tb.WRONG_TYPE)
			END
		END
	END MakeArr;
	
	PROCEDURE (VAR it: Item) Len* () : LONGINT;
	BEGIN RETURN it.length END Len;
	
	PROCEDURE (VAR it: Item) Size* () : LONGINT;
	BEGIN RETURN it.size END Size;
	
	PROCEDURE (VAR it: Item) BoolVal* (): BOOLEAN;
		VAR bool_val: BOOLEAN;
	BEGIN
		ASSERT(it.typ = Tb.boolTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.GET(it.adr, bool_val);
		RETURN bool_val
	END BoolVal;
	
	PROCEDURE (VAR it: Item) CharVal* (): CHAR;
		VAR char_val: CHAR;
	BEGIN
		ASSERT(it.typ = Tb.charTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.GET(it.adr, char_val);
		RETURN char_val
	END CharVal;
	
	PROCEDURE (VAR it: Item) IntVal* (): LONGINT;
		VAR int32: LONGINT; int16: INTEGER; int8: SHORTINT;
	BEGIN
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		IF it.typ = Tb.int8Typ THEN
			SYSTEM.GET(it.adr, int8); int32 := int8;
		ELSIF it.typ = Tb.int16Typ THEN
			SYSTEM.GET(it.adr, int16); int32 := int16;
		ELSIF it.typ = Tb.int32Typ THEN
			SYSTEM.GET(it.adr, int32)
		ELSE
			ASSERT(it.typ = Tb.intTyp, Tb.WRONG_TYPE)
		END;
		RETURN int32
	END IntVal;
	
	PROCEDURE (VAR it: Item) RealVal* (): REAL;
		VAR real_val: REAL;
	BEGIN
		ASSERT(it.typ = Tb.realTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.GET(it.adr, real_val);
		RETURN real_val
	END RealVal;
	
	PROCEDURE (VAR it: Item) LRealVal* (): LONGREAL;
		VAR lreal_val: LONGREAL;
	BEGIN
		ASSERT(it.typ = Tb.real64Typ, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.GET(it.adr, lreal_val);
		RETURN lreal_val
	END LRealVal;

	PROCEDURE (VAR it: Item) SetVal* (): SET;
		VAR lv: LONGINT;
	BEGIN
		ASSERT(it.typ = Tb.setTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.GET(it.adr, lv);
		RETURN Tb.Int32_Set(lv);
	END SetVal;

	PROCEDURE (VAR it: Item) GetStringVal* (VAR str_val: ARRAY OF CHAR; VAR ok: BOOLEAN);
		VAR j: LONGINT; 
	BEGIN
		ASSERT(it.typ = Tb.arrTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		j := -1;
		REPEAT
			INC(j); 
			SYSTEM.GET(it.adr+j, str_val[j])
		UNTIL (j >= it.length) OR (str_val[j] = 0X);
		ok := (str_val[j] = 0X)
	END GetStringVal;
	
	PROCEDURE (VAR it: Item) GetString16Val* (VAR str_val: ARRAY OF Tb.Char16; 
	VAR ok: BOOLEAN);
		VAR j: LONGINT; 
	BEGIN
		ASSERT(it.typ = Tb.arrTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		j := -1;
		REPEAT
			INC(j); 
			SYSTEM.GET(it.adr+2*j, str_val[j])
		UNTIL (j >= it.length) OR (str_val[j] = 0);
		ok := (str_val[j] = 0)
	END GetString16Val;
	
	PROCEDURE (VAR it: Item) PutStringVal* ( str_val-: ARRAY OF CHAR; VAR ok: BOOLEAN);
		VAR j: LONGINT; 
	BEGIN
		ASSERT(it.typ = Tb.arrTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		j := 0; 
		WHILE (j < it.length) & (str_val[j] # 0X) DO
			SYSTEM.PUT(it.adr+j, str_val[j]);
			INC(j)
		END;
		IF j < it.length THEN SYSTEM.PUT(it.adr+j, str_val[j]); ok := TRUE ELSE ok := FALSE END;
	END PutStringVal;
	
	PROCEDURE (VAR it: Item) PutString16Val* ( str_val-: ARRAY OF Tb.Char16; 
	VAR ok: BOOLEAN);
		VAR j: LONGINT; 
	BEGIN
		ASSERT(it.typ = Tb.arrTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		j := 0; 
		WHILE (j < it.length) & (str_val[j] # 0) DO
			SYSTEM.PUT(it.adr+2*j, str_val[j]);
			INC(j)
		END;
		IF j < it.length THEN SYSTEM.PUT(it.adr+2*j, str_val[j]); ok := TRUE ELSE ok := FALSE END;
	END PutString16Val;
	
	PROCEDURE (VAR it: Item) PutBoolVal* (bool_val: BOOLEAN);
	BEGIN
		ASSERT(it.typ = Tb.boolTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.PUT(it.adr, bool_val);
	END PutBoolVal;
	
	PROCEDURE (VAR it: Item) PutCharVal* (char_val: CHAR);
	BEGIN
		ASSERT(it.typ = Tb.charTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.PUT(it.adr, char_val);
	END PutCharVal;

	PROCEDURE (VAR it: Item) PutIntVal* (int_val: LONGINT);
	BEGIN
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		IF it.typ = Tb.int8Typ THEN
			SYSTEM.PUT(it.adr, SYSTEM.VAL(Tb.Int8, int_val))
		ELSIF it.typ = Tb.int16Typ THEN
			SYSTEM.PUT(it.adr, SYSTEM.VAL(Tb.Int16, int_val))
		ELSIF it.typ = Tb.int32Typ THEN
			SYSTEM.PUT(it.adr, int_val)
		ELSE
			ASSERT(it.typ = Tb.longTyp, Tb.WRONG_TYPE)
		END
	END PutIntVal;
	
	PROCEDURE (VAR it: Item) PutLongVal* (long_val: LONGINT);
	BEGIN
		ASSERT(it.typ = Tb.longTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.PUT(it.adr, long_val);
	END PutLongVal;
	
	PROCEDURE (VAR it: Item) PutRealVal* (real_val: REAL);
	BEGIN
		ASSERT(it.typ = Tb.realTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.PUT(it.adr, real_val);
	END PutRealVal;
	
	PROCEDURE (VAR it: Item) PutLRealVal* (lreal_val: LONGREAL);
	BEGIN
		ASSERT(it.typ = Tb.realTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.PUT(it.adr, lreal_val);
	END PutLRealVal;
	
	PROCEDURE (VAR it: Item) PutSetVal* (set_val: SET);
	BEGIN
		ASSERT(it.typ = Tb.setTyp, Tb.WRONG_TYPE);
		ASSERT(it.vis = Tb.exported, Tb.NOT_EXPORTED);
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		SYSTEM.PUT(it.adr, Tb.Set_Int32(set_val));
	END PutSetVal;
	
	PROCEDURE (VAR it: Item) GetTypeName* (VAR mname: ARRAY OF CHAR; 
	VAR tname: ARRAY OF CHAR);
	BEGIN
		IF it.tinfo = NIL THEN
			mname[0] := 0X; tname[0] := 0X
		ELSE
			oberonRTS.NameOfModule(it.tinfo.module, mname);
			oberonRTS.NameOfType(it.tinfo.type, tname)
		END
	END GetTypeName;

	PROCEDURE (VAR it: Item) Deref* (VAR base: Item);
		VAR adr, baseadr: LONGINT;
	BEGIN
		ASSERT(it.adr # 0, Tb.NIL_POINTER);
		base := it.into^;
		SYSTEM.GET(it.adr, adr);
		IF adr = 0 THEN
			base.adr := 0
		ELSE
			SYSTEM.GET(adr, baseadr);
			base.adr := baseadr
		END;
		base.length := PtrLen(it);
	END Deref;

	PROCEDURE (VAR it: Item) Index* (j: LONGINT; VAR iit: Item);
	BEGIN
		iit := it.into^;
		iit.adr := it.adr + iit.offset + iit.size*j;
	END Index;
	

	PROCEDURE ItemVal_String* ( it-: Item; zerotrim: BOOLEAN;
	VAR str: ARRAY OF CHAR; VAR ok: BOOLEAN);
		VAR res: INTEGER; iit: Item; int_val: LONGINT; real_val: REAL;
			lreal_val: LONGREAL; lstr: ARRAY 256 OF Tb.Char16;
	BEGIN
		res := 0; ok := TRUE;
		CASE it.typ OF
		| Tb.arrTyp:
			it.Index(0, iit);
			IF iit.typ = Tb.char8Typ THEN
				it.GetStringVal(str, ok)
			ELSIF iit.typ = Tb.char16Typ THEN
				it.GetString16Val(lstr, ok);
				Tb.Char16_Char(lstr, str)
			ELSE
				res := Tb.INVALID_VAL
			END;
		| Tb.boolTyp:
			IF it.BoolVal() THEN Tb.Copy("TRUE", str) 
			ELSIF zerotrim THEN str[0] := 0X
			ELSE Tb.Copy("FALSE", str) END;
		| Tb.int8Typ, Tb.int16Typ, Tb.int32Typ:
			int_val := it.IntVal();
			IF zerotrim & (int_val = 0) THEN str[0] := 0X
			ELSE Tb.IntToString(int_val, str) END;
		| Tb.int64Typ:
			res := Tb.UNSUPPORTED;
		| Tb.real32Typ:
			real_val := it.RealVal();
			IF zerotrim & (real_val = 0) THEN str[0] := 0X
			ELSE Tb.RealToString(real_val, str) END;
		| Tb.real64Typ:
			lreal_val := it.LRealVal();
			IF zerotrim & (lreal_val = 0) THEN str[0] := 0X
			ELSE Tb.LRealToString(lreal_val, str) END;
		| Tb.setTyp:
			int_val := Tb.Set_Int32(it.SetVal());
			IF zerotrim & (int_val = 0) THEN str[0] := 0X
			ELSE Tb.IntToString(int_val, str) END;
		ELSE
			res := Tb.WRONG_TYPE
		END;
		IF res # 0 THEN ok := FALSE END
	END ItemVal_String;

	PROCEDURE String_ItemVal* (VAR it: Item;  str-: ARRAY OF CHAR; 
	VAR ok: BOOLEAN);
		VAR res, int_val: INTEGER; bool_val: BOOLEAN; long_val: LONGINT; 
			real_val: REAL; iit: Item;
			lreal_val: LONGREAL; lstr: ARRAY 256 OF Tb.Char16;
	BEGIN
		res := 0; ok := TRUE;
		CASE it.typ OF
		| Tb.arrTyp:
			it.Index(0, iit);
			IF iit.typ = Tb.char8Typ THEN
				it.PutStringVal(str, ok)
			ELSIF iit.typ = Tb.char16Typ THEN
				Tb.Char_Char16(str, lstr);
				it.PutString16Val(lstr, ok)
			ELSE
				res := Tb.INVALID_VAL
			END;
		| Tb.boolTyp:
			Tb.StringToBool(str, bool_val, res);
			it.PutBoolVal(bool_val);
		| Tb.int8Typ, Tb.int16Typ:
			Tb.StringToInt(str, int_val, res);
			it.PutIntVal(int_val);
		| Tb.int32Typ:
			Tb.StringToLInt(str, long_val, res);
			it.PutLongVal(long_val);
		| Tb.int64Typ:
			res := Tb.UNSUPPORTED;
		| Tb.real32Typ:
			Tb.StringToReal(str, real_val, res);
			it.PutRealVal(real_val);
		| Tb.real64Typ:
			Tb.StringToLReal(str, lreal_val, res);
			it.PutLRealVal(lreal_val);
		| Tb.setTyp:
			Tb.StringToLInt(str, long_val, res);
			it.PutSetVal(Tb.Int32_Set(Tb.Long_Int32(long_val)));
		ELSE
			res := Tb.WRONG_TYPE
		END;
		IF res # 0 THEN ok := FALSE END;
	END String_ItemVal;

	(* ========== TInfo ========== *)	
	PROCEDURE Align(align: LONGINT): LONGINT;
	BEGIN
		IF align > MAX_ALIGNMENT THEN align := MAX_ALIGNMENT END;
		RETURN align
	END Align;
	
	PROCEDURE CalculateItem (VAR item: Item);
		VAR size, packed_size, alignment, offset, mod: LONGINT; it: PItem;
	BEGIN
		CASE item.typ OF
		Tb.boolTyp, Tb.char8Typ, Tb.int8Typ:
			size := 1; alignment := 1
		| Tb.char16Typ, Tb.int16Typ:
			size := 2; alignment := Align(2);
		| Tb.int32Typ, Tb.real32Typ, Tb.setTyp:
			size := 4; alignment := Align(4);
		| Tb.real64Typ:
			size := 8; alignment := Align(8);
		| Tb.int64Typ:
			size := 8; alignment := Align(4);
		| Tb.recTyp:
			size := 0; packed_size := 0; alignment := 0; offset := 0; it := item.into;
			WHILE it # NIL DO
				CalculateItem(it^);
				mod := offset MOD it.alignment;
				IF mod # 0 THEN
					INC(offset, it.alignment - mod)
				END;
				it.offset := offset;
				IF it.typ = Tb.recTyp THEN
					INC(packed_size, it.tinfo.packed_size)
				ELSE
					INC(packed_size, it.size);
				END;
				INC(offset, it.size);
				IF it.alignment > alignment THEN
					alignment := it.alignment
				END;
				it := it.next
			END;
			IF alignment > 0 THEN
				mod := offset MOD alignment;
				IF mod # 0 THEN
					INC(offset, alignment - mod)
				END
			END;
			size := offset;
			IF item.tinfo # NIL THEN
				item.tinfo.size := size;
				item.tinfo.packed_size := packed_size;
				item.tinfo.alignment := alignment;
				IF item.tinfo.type # oberonRTS.nullType THEN
					size := oberonRTS.SizeOf(item.tinfo.type);
					IF size # item.tinfo.size THEN
						IF item.into = NIL THEN
							IF (Tb.TRAL_RUNTIME IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
								Log.String("* empty record '"); Log.String(item.name); 
								Log.String("', RTTI information must be Inserted"); Log.Ln
							END;
							ASSERT(size = item.tinfo.size, Tb.NIL_RESULT)
						ELSE
							IF (Tb.TRAL_RUNTIME IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
								Log.String("* "); Log.String(item.name); 
								Log.String(" size check, real:"); Log.Int(size); 
								Log.String(" calculated:"); Log.Int(item.tinfo.size); Log.Ln
							END;
							ASSERT(size = item.tinfo.size, Tb.INIT_FAILED)
						END
					END
				END
			END;
		| Tb.arrTyp:
			CalculateItem(item.into^);
			size := item.into.size;
			alignment := item.into.alignment
		| Tb.ptrTyp:
			CalculateItem(item.into^);
			size := 4;
			alignment := Align(4)
		ELSE
			ASSERT(item.typ = Tb.boolTyp, Tb.WRONG_TYPE)
		END;
		item.size := size * item.length;
		item.alignment := alignment
	END CalculateItem; 

	PROCEDURE Calculate (ti: TInfo);
		VAR size: LONGINT; name: Tb.Name;
	BEGIN
		IF ti.size = 0 THEN
			CalculateItem(ti.item);
		END
	END Calculate;
	
	PROCEDURE (ti: TInfo) LookupItem* (rec-: Tb.AnyRec; VAR item: Item);
	BEGIN
		item := ti.item;
		item.adr := SYSTEM.ADR(rec);
	END LookupItem;
	
	PROCEDURE (ti: TInfo) NewForItem* (VAR item: Item; VAR ptr: Tb.AnyPtr);
	BEGIN
	END NewForItem;

	PROCEDURE (ti: TInfo) Insert* (ind: INTEGER; name: ARRAY OF CHAR; 
	typ: Tb.Type; vis: INTEGER; length: LONGINT; tinfo: TInfo);
		VAR it, insitem, item1, item2: PItem; j: INTEGER;
	BEGIN
		ASSERT(ti.item.typ = Tb.recTyp, Tb.WRONG_TYPE);
		ASSERT((vis >= Tb.hidden) & (vis <= Tb.exported), 
			Tb.CHECK_RANGE_ERROR);
		ASSERT(length # 0, Tb.ZERO_VAL);
		NEW(insitem);
		insitem.vis := vis; Tb.Copy(name, insitem.name);
		IF length = 1 THEN
			insitem.typ := typ; insitem.length := 1;
			IF tinfo # NIL THEN
				insitem.tinfo := tinfo; insitem.into := tinfo.item.into
			END
		ELSIF length < 0 THEN
			insitem.typ := Tb.ptrTyp; insitem.length := 1;
			NEW(item1); item1.typ := Tb.arrTyp; item1.length := -length;
			item1.vis := vis; Tb.Copy(name, item1.name);
			insitem.into := item1;
			NEW(item2); item2.typ := typ; item2.length := 1;
			item2.vis := vis; Tb.Copy(name, item2.name);
			item1.into := item2;
			IF tinfo # NIL THEN
				insitem.tinfo := tinfo; item1.tinfo := tinfo;
				item2.tinfo := tinfo; item2.length := tinfo.item.length;
				item2.into := tinfo.item.into
			END
		ELSIF length > 1 THEN
			insitem.typ := Tb.arrTyp; insitem.length := length;
			NEW(item2); item2.typ := typ; item2.length := 1;
			item2.vis := vis; Tb.Copy(name, item2.name);
			insitem.into := item2;
			IF tinfo # NIL THEN
				insitem.tinfo := tinfo; item2.tinfo := tinfo; 
				item2.length := tinfo.item.length; item2.into := tinfo.item.into
			END
		END;
		IF (ti.item.into = NIL) OR (ind = 0) THEN
			insitem.next := ti.item.into; ti.item.into := insitem;
		ELSE
			j := 1; it := ti.item.into;
			WHILE (it.next # NIL) & (j # ind) DO
				INC(j); it := it.next
			END;
			insitem.next := it.next; it.next := insitem;
		END
	END Insert;

	PROCEDURE (ti: TInfo) Find* ( name-: ARRAY OF CHAR; srcitem-: Item;
	VAR item: Item);
		VAR it: PItem;
	BEGIN
		ASSERT(srcitem.typ = Tb.recTyp, Tb.WRONG_TYPE);
		it := srcitem.into;
		WHILE (it # NIL) & (name # it.name) DO
			it := it.next
		END;
		IF it # NIL THEN
			item := it^; 
			item.adr := srcitem.adr + it.offset;
		ELSE
			item.typ := 0
		END
	END Find;

	PROCEDURE (ti: TInfo) Iterate* (itera: Iterator; name-: ARRAY OF CHAR;
	item-: Item; VAR client_rec: Tb.AnyRec; VAR irc: INTEGER);
		VAR it: PItem; item0: Item; j, irc0: INTEGER;
		adr1: LONGINT;
	BEGIN
		ASSERT(ti.item.typ = Tb.recTyp, Tb.WRONG_TYPE);
		irc0 := itera(name, item, TRUE, client_rec); irc := irc0;
		CASE item.typ OF
		| Tb.recTyp:
			IF irc = ITERATE_INTO THEN
				it := item.into;
				WHILE (it # NIL) & (irc # STOP_ITERATE) DO
					item0 := it^;
					item0.adr := item.adr + it.offset;
					ti.Iterate(itera, it.name, item0, client_rec, irc);
					IF irc # STOP_ITERATE THEN it := it.next END
				END
			END
		| Tb.arrTyp:
			IF irc = ITERATE_INTO THEN
				item0 := item.into^;
				item0.adr := item.adr + item0.offset;
				ti.Iterate(itera, item0.name, item0, client_rec, irc);
			ELSIF irc = ITERATE_FOREACH THEN
				item0 := item.into^;
				FOR j := 0 TO item.length-1 DO
					item0.adr := item.adr + item0.offset + item0.size*j;
					ti.Iterate(itera, item0.name, item0, client_rec, irc)
				END
			END
		| Tb.ptrTyp:
			IF irc = ITERATE_INTO THEN
				ASSERT(item.into # NIL, Tb.INTERNAL_ERROR);
				item0 := item.into^;
				item0.length := PtrLen(item);
				item0.offset := 0;
				SYSTEM.GET(item.adr+item0.offset, adr1);
				IF adr1 = 0 THEN
					item0.adr := 0
				ELSE
					SYSTEM.GET(adr1, item0.adr);
				END;
				ti.Iterate(itera, item0.name, item0, client_rec, irc);
			END
		ELSE
		END;
		IF (irc0 >= ITERATE_SKIP) & (irc # STOP_ITERATE) THEN
			irc := itera(name, item, FALSE, client_rec)
		END
	END Iterate;
	
	PROCEDURE (ti: TInfo) ConnectTo* (VAR rec: Tb.AnyRec;
	 mname-: ARRAY OF CHAR;  tname-: ARRAY OF CHAR; 
	realloc_cmd: ReallocCommand);
	BEGIN
		IF tname[0] = 0X THEN
			ti.module := oberonRTS.nullModule;
			ti.type := oberonRTS.nullType;
			ti.typof := Tb.undef;
			ti.item.name[0] := 0X;
		ELSE
			ti.module := oberonRTS.Search(mname);
			ASSERT(ti.module # oberonRTS.nullModule, Tb.NOT_FOUND);
			ti.type := oberonRTS.ThisType(ti.module, tname);
			ASSERT(ti.type # oberonRTS.nullType, Tb.NOT_FOUND);
			Tb.Copy(mname, ti.item.name); Tb.Append(".", ti.item.name); 
			Tb.Append(tname, ti.item.name); ti.typof := Tb.Crc32(ti.item.name, -1, 0, 32)
		END;
		ti.item.typ := Tb.recTyp; ti.item.vis := Tb.exported;
		ti.item.length := 1; ti.item.tinfo := ti; 
		ti.item.adr := SYSTEM.ADR(rec);
		ti.realloc_cmd := realloc_cmd
	END ConnectTo;

	PROCEDURE (ti: TInfo) CalcInfo* ()  ;
	BEGIN 
		ASSERT(ti.item.typ = Tb.recTyp, Tb.WRONG_TYPE);
		Calculate(ti);
	END CalcInfo;

	PROCEDURE (ti: TInfo) CopyFrom* (src: TInfo);
		VAR adr: Tb.Address;
	BEGIN
		ASSERT(src.item.typ = Tb.recTyp, Tb.WRONG_TYPE);
		Calculate(src);
		adr := ti.item.adr; ti^ := src^; ti.item.adr := adr; ti.item.tinfo := ti
	END CopyFrom;
		
	PROCEDURE (ti: TInfo) InitConfig* (command_line, mod_name, 
	current_dir: ARRAY OF CHAR; ind_command_line: INTEGER)  ;
	BEGIN 
		Calculate(ti)
	END InitConfig;
	
END Ta_TInfo.