MODULE Ta_Tables;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Ts := Ta_Stations, Ti := Ta_TInfo, 
		Ta := Ta_Areas, Log, SYSTEM;
	
	CONST
		readOp = -10; clearOp = -11; tcallOp = -18; writeOp = -19; waitOp = -20;
		(* positive numbers mean LONGINT time expired *)
		maxOperations = 20; MAX_PREALLOCED = 5;
		RES_MAPPER_CANCELLED* = 1; SEND_TIMEOUT* = 2; RESPONSE_TIMEOUT* = 3;
			RESPONSE_MISMATCH* = 4;

	TYPE
		Tablist* = POINTER TO TablistDesc;
		StdTablist = POINTER TO StdTablistDesc;
		Table* = POINTER TO TableDesc;
		StdTable = POINTER TO StdTableDesc;
		Elem = POINTER TO ElemDesc;
		
		TablistDesc* =  RECORD
			is_updating*: BOOLEAN;
		END;
		StdTablistDesc = RECORD (TablistDesc)
			tablist, tail: StdTable;
			stations: Ts.Stations;
			n_stations: INTEGER;
			aid, host_sid, send_sid: INTEGER;
			mf: Ts.MapperFactory;
			current_mid: Tb.Int16;
			t_waiting_expired: Tb.Time;
		END;
		
		TableDesc* =  RECORD (Ta.TableDesc)
			mapper-: Ts.Mapper;
			sid-, isid-, oid-: INTEGER;
			tqos-: Ts.TableQos;
			hdr-: Ts.Header;
			started-: BOOLEAN
		END;
		StdTableDesc = RECORD (TableDesc)
			mid_resp, src_mid: Tb.Int16;
			next: StdTable;
			tbl: StdTablist;
			on_data: Ta.OnDataHandler;
			queue, tail: Elem;
			prealloced: ARRAY MAX_PREALLOCED OF Elem;
			item: Ti.Item;
		END;

		ElemDesc = RECORD
			next: Elem;
			table: StdTable;
			item: Ti.Item;
			op: INTEGER;
			tcmd: Ta.TableCommand;
			par: Tb.AnyPtr;
			sids: SET;
			mid, waiting: INTEGER;
			tp_waiting_expired, t_waiting_expired: Tb.Time;
			current_mid: Tb.Int16;
			used: BOOLEAN
		END;
		

	VAR
		dummy: Ti.Item; zero_prod: Ta.ProData;
		
	
	(* ========== Tablist interface ========== *)
	PROCEDURE (tbl: Tablist) NewTable* (VAR rec: Tb.AnyRec; iface_name-: ARRAY OF CHAR; 
	tname-: ARRAY OF CHAR; n_stations, sid, oid: INTEGER; 
	VAR tqos: Ts.TableQosBase; on_data: Ta.OnDataHandler): Ta.Table;
	BEGIN RETURN NIL END NewTable;
	PROCEDURE (tbl: Tablist) Start* (VAR ok: BOOLEAN);
	BEGIN END Start;
	PROCEDURE (tbl: Tablist) Process* (rd: Ts.Reader; sid: INTEGER; VAR done: BOOLEAN);
	BEGIN END Process;
	PROCEDURE (tbl: Tablist) ProcessExpired* (VAR done: BOOLEAN);
	BEGIN END ProcessExpired;
	PROCEDURE (tbl: Tablist) Message* (src_oid: INTEGER; VAR msg: Tb.Msg);
	BEGIN END Message; 

	PROCEDURE ^ ProcessTable (t: StdTable; VAR prod: Ta.ProData; 
	VAR t_waiting_expired: Tb.Time);


	(* ========== Tablist Implementation ========== *)
	PROCEDURE NewTablist* (mf: Ts.MapperFactory; 
	stations: Ts.Stations; aid, host_sid, send_sid: INTEGER; 
	(*OUT*)VAR tblst: Tablist; (*OUT*)VAR res: INTEGER);
		VAR tbl: StdTablist;
	BEGIN
		ASSERT((mf # NIL) & (stations # NIL), Tb.NIL_POINTER);
		NEW(tbl); tbl.mf := mf;
		tbl.stations := stations;
		tbl.aid := aid; tbl.host_sid := host_sid; tbl.send_sid := send_sid;
		tbl.current_mid := SHORT(ENTIER((Tc.MAX_MID-1) * Tb.Uniform()));
		tblst := tbl;
		res := 0
	END NewTablist;
	
	PROCEDURE (tbl: StdTablist) NewTable (VAR rec: Tb.AnyRec;  iface_name-: ARRAY OF CHAR; 
	 tname-: ARRAY OF CHAR; n_stations, sid, oid: INTEGER; 
	VAR tqos: Ts.TableQosBase; on_data: Ta.OnDataHandler): Ta.Table;
		VAR t: StdTable; res, j: INTEGER;
			modname, sname: Tb.Name64;
	BEGIN
		ASSERT((tqos.realloc_mode >= Ts.DONT_REALLOC)
			& (tqos.realloc_mode <= Ts.REALLOC_EXACT), Tb.CHECK_RANGE_ERROR);
		NEW(t); t.tbl := tbl; tbl.n_stations := n_stations;
		IF tbl.tablist = NIL THEN tbl.tablist := t; tbl.tail := t ELSE tbl.tail.next := t; tbl.tail := t END;
		t.sid := sid;
		CASE sid OF
		| Ts.sendSID:
			t.isid := tbl.send_sid;
		| Ts.hostSID, Ts.callerSID:
			t.isid := tbl.host_sid;
		ELSE
			t.isid := sid
		END; 
		ASSERT((t.isid >= 0) & (t.isid < tbl.n_stations), Tb.CHECK_RANGE_ERROR);
		t.oid := oid; t.on_data := on_data;
		Ts.CopyTableQos(tqos, t.tqos); 
		NEW(t.ti); 
		j := 0; res := 0;
		WHILE (res = 0) & (tname[j] # 0X) DO
			IF tname[j] = '.' THEN
				modname[j] := 0X; res := 1
			ELSE
				modname[j] := tname[j]
			END;
			INC(j)
		END;
		IF res # 0 THEN
			res := j;
			WHILE tname[j] # 0X DO
				sname[j-res] := tname[j];
				INC(j)
			END;
			sname[j-res] := 0X;
			t.ti.ConnectTo(rec, modname, sname, NIL)
		ELSE
			t.ti.ConnectTo(rec, Ta.dir.default_mod.name, tname, NIL)
		END;
		tbl.mf.NewMapper(t.ti, iface_name, tqos, sid, oid, t.mapper, res);
		IF res # 0 THEN
			t := NIL
		ELSE
			FOR j := 0 TO MAX_PREALLOCED-1 DO
				NEW(t.prealloced[j])
			END
		END;
		RETURN t
	END NewTable;

	PROCEDURE StartTable (t: StdTable; VAR ok: BOOLEAN);
	BEGIN
		t.ti.CalcInfo;
		t.mapper.ExploreObject();
		ok := ~t.mapper.cancelled;
		t.item := t.ti.item; t.started := ok
	END StartTable;
	
	PROCEDURE (tbl: StdTablist) Start (VAR ok: BOOLEAN);
		VAR t: StdTable;
	BEGIN
		t := tbl.tablist; ok := TRUE;
		WHILE (t # NIL) DO
			StartTable(t, ok);
			t := t.next
		END
	END Start;
	
	PROCEDURE DoRead (t: StdTable; rd: Ts.Reader; VAR changed: BOOLEAN; 
	VAR res: INTEGER);
	BEGIN
		changed := FALSE;
		t.mapper.ReadObject(t.item, rd, changed);
		IF t.mapper.cancelled THEN res := RES_MAPPER_CANCELLED ELSE res := 0 END
	END DoRead;
	
	PROCEDURE (tbl: StdTablist) Process (rd: Ts.Reader; sid: INTEGER; VAR done: BOOLEAN);
		VAR t: StdTable; t_waiting_expired: Tb.Time; prod: Ta.ProData;
	BEGIN
		t := tbl.tablist;
		done := FALSE;
		IF ~rd.eof THEN
			WHILE (t # NIL) & ~done DO
				IF t.mapper.MatchHeader(rd.hdr) THEN
					done := TRUE; prod := zero_prod;
					t.hdr := rd.hdr; t.hdr.buffer := 0;
					DoRead(t, rd, prod.changed, prod.res);
					prod.sid := sid; prod.got_message := TRUE;
					ProcessTable (t, prod, t_waiting_expired);
					IF t.on_data # NIL THEN
						t.on_data(t, prod)
					END;
					IF Tb.TRAL_MESSAGE IN Tb.trace THEN
						IF ~prod.got_message THEN Log.String("ignore ") END;
						Log.String("message match="); Log.Bool(done);
						Log.String(" msg_type="); Log.Int(rd.hdr.msg_type); 
						Log.String(" mid="); Log.Int(rd.hdr.mid); 
						Log.String(" sid="); Log.Int(sid);
						Log.String(" dst_sid="); Log.Int(rd.hdr.dst_sid); Log.Ln
					END
				END;
				t := t.next
			END;
			IF ~done THEN
				IF Tb.TRAL_MESSAGE IN Tb.trace THEN
					Log.String("? unmatched message sid_from="); Log.Int(sid);
					Log.String(" msg_type="); Log.Int(rd.hdr.msg_type); 
					Log.String(" iface_code="); Log.Int(rd.hdr.iface_code); 
					Log.String(" mid="); Log.Int(rd.hdr.mid); 
					Log.String(" dst_sid,src_sid,dst_oid="); Log.Int(rd.hdr.dst_sid); 
					Log.Int(rd.hdr.src_sid); Log.Int(rd.hdr.dst_oid); Log.Ln
				END
			END
		END
	END Process;

	PROCEDURE (tbl: StdTablist) ProcessExpired (VAR done: BOOLEAN);
		VAR t: StdTable; t_waiting_expired: Tb.Time; prod: Ta.ProData;
	BEGIN
		done := FALSE;
		IF Tb.TimePositive(tbl.t_waiting_expired) THEN
			t_waiting_expired := Tb.ZERO_TIME;
			t := tbl.tablist;
			WHILE (t # NIL) & ~done DO
				prod := zero_prod;
				ProcessTable (t, prod, t_waiting_expired);
				IF (prod.res # 0) & (t.on_data # NIL) THEN
					t.on_data(t, prod)
				END;
				t := t.next
			END;
			tbl.t_waiting_expired := t_waiting_expired
		END
	END ProcessExpired;
	
	PROCEDURE (tbl: StdTablist) Message (src_oid: INTEGER; VAR msg: Tb.Msg);
		VAR t: StdTable; done: BOOLEAN;
	BEGIN
		t := tbl.tablist; done := FALSE;
		WHILE (t # NIL) & ~done DO
			IF src_oid = t.oid THEN
				done := TRUE
			END;
			IF done OR (src_oid = Ts.toALL) THEN
				t.HandleMsg(msg)
			END;
			t := t.next
		END
	END Message; 

	
	(* ========== Table Implementation ========== *)
	PROCEDURE (t: StdTable) TInsert (ind: INTEGER; name: ARRAY OF CHAR; 
	typ: Tb.Type; vis, length: INTEGER; tinfo: Ti.TInfo);
	BEGIN
		t.ti.Insert(ind, name, typ, vis, length, tinfo)
	END TInsert;
	
	PROCEDURE Synch (t: StdTable; op: INTEGER): BOOLEAN;
		(* check whether it is necessary to enqueue this operation *)
		(* TRUE if Put would insert at beginning of queue *)
		VAR h: Elem; rc: BOOLEAN;
	BEGIN
		h := t.queue;
		IF h = t.tail THEN
			rc := TRUE
		ELSIF op >= clearOp THEN
			WHILE (h # t.tail) & (h.table # t) DO h := h.next END;
			rc := (h = t.tail)
		ELSE
			rc := FALSE
		END;
		RETURN rc
	END Synch;

	PROCEDURE (t: StdTable) Available (): BOOLEAN;
	BEGIN
		RETURN Synch(t, readOp)
	END Available;

	PROCEDURE ClearTable (t: StdTable);
	BEGIN
		t.mid_resp := 0
	END ClearTable;
	
	PROCEDURE SetWaitingExpiration (e: Elem; VAR t_waiting_expired: Tb.Time);
		VAR t_crt: Tb.Time;
	BEGIN
		IF Tb.TimeEquals(e.t_waiting_expired, Tb.ZERO_TIME) THEN
			IF Tb.TimeEquals(e.tp_waiting_expired, Tb.ZERO_TIME) THEN
				IF e.op = writeOp THEN
					IF e.table.tqos.tp_send_timeout < 0 THEN
						e.tp_waiting_expired := Tb.INFINITE_TIME;
					ELSE
						Tb.Long_Time(e.table.tqos.tp_send_timeout, e.tp_waiting_expired)
					END
				ELSE
					IF e.table.tqos.tp_timeout < 0 THEN
						e.tp_waiting_expired := Tb.INFINITE_TIME;
					ELSE
						Tb.Long_Time(e.table.tqos.tp_timeout, e.tp_waiting_expired)
					END
				END
			END;
			IF Tb.TimeEquals(e.tp_waiting_expired, Tb.INFINITE_TIME) THEN
				e.t_waiting_expired := Tb.INFINITE_TIME
			ELSE
				Tb.GetTime(t_crt);
				Tb.AddTime(t_crt, e.tp_waiting_expired, e.t_waiting_expired);
			END
		END;
		t_waiting_expired := e.t_waiting_expired
	END SetWaitingExpiration;
	
	PROCEDURE Put (t: StdTable; op: INTEGER; item: Ti.Item; 
	tcmd: Ta.TableCommand; par: Tb.AnyPtr; sids: SET; mid, waiting: INTEGER; 
	tp_waiting_expired: Tb.Time);
		VAR e, h, k: Elem; j: INTEGER; found: BOOLEAN;
	BEGIN	(* put command into a queue for later async processing *)
		found := FALSE; j := 0;
		WHILE (j < MAX_PREALLOCED) & ~found DO
			IF ~t.prealloced[j].used THEN
				e := t.prealloced[j];
				t.prealloced[j].used := TRUE;
				found := TRUE
			END;
			INC(j)
		END;
		IF ~found THEN 
			NEW(e);
		END;
		e.table := t; e.op := op; e.item := item; e.tcmd := tcmd; e.par := par; 
		e.sids := sids; e.current_mid := 0; e.mid := mid; e.waiting := waiting; 
		e.tp_waiting_expired := tp_waiting_expired; e.t_waiting_expired := Tb.ZERO_TIME;
		IF t.queue = NIL THEN
			SetWaitingExpiration(e, t.tbl.t_waiting_expired);
		END;
		h := t.queue;
		IF h = t.tail THEN
			e.next := h; t.queue := e
		ELSIF op >= clearOp THEN	(* clear, read: insert after last op on same table *)
			k := NIL;
			WHILE h # t.tail DO
				IF h.table = t THEN k := h END;
				h := h.next
			END;
			IF k = NIL THEN e.next := t.queue; t.queue := e
			ELSE e.next := k.next; k.next := e
			END
		ELSE
			(* append to the end of queue *)
			WHILE h.next # t.tail DO h := h.next END;
			e.next := h.next; h.next := e
		END
	END Put;
	
	PROCEDURE ClearItem ( name-: ARRAY OF CHAR;  item-: Ti.Item; 
	begin: BOOLEAN; VAR client_rec: Tb.AnyRec): INTEGER;
		VAR rc: INTEGER;
	BEGIN
		rc := Ti.ITERATE_NEXT;
		CASE item.typ OF
		Tb.recTyp, Tb.ptrTyp:
			IF begin THEN rc := Ti.ITERATE_INTO END
		| Tb.arrTyp:
			IF begin THEN rc := Ti.ITERATE_FOREACH END
		ELSE
			(*Tb.Clean(item.adr, item.Len()) SIZE() must be here *)
		END;
		RETURN rc
	END ClearItem;
	
	PROCEDURE DoClear (t: StdTable;  item-: Ti.Item);
		VAR res: INTEGER;
	BEGIN
		res := 0;
		ClearTable(t);
		t.ti.Iterate(ClearItem, "", item, Tb.no_rec, res)
	END DoClear;
	
	PROCEDURE (t: StdTable) Clear (VAR data: Tb.AnyRec);
		VAR item: Ti.Item;
	BEGIN
		t.ti.LookupItem(data, item);
		ASSERT(item.adr # 0, Tb.NIL_RESULT);
		IF Synch(t, clearOp) THEN 
			ASSERT(t.started, Tb.ALREADY_INITIALIZED);
			DoClear(t, item)
		ELSE Put(t, clearOp, item, NIL, NIL, {}, 0, Ta.noWait, Tb.ZERO_TIME)
		END
	END Clear;
	
	PROCEDURE (t: StdTable) Call (command: Ta.TableCommand; par: Tb.AnyPtr);
	BEGIN
		ASSERT(command # NIL, Tb.NIL_POINTER);
		IF Synch(t, tcallOp) THEN command(t, zero_prod, par)
		ELSE Put(t, tcallOp, dummy, command, par, {}, 0, Ta.noWait, Tb.ZERO_TIME)
		END
	END Call;
	
	PROCEDURE WriteHeader (t: StdTable; unsubscribe, toall: BOOLEAN);
		VAR hdr: Ts.Header;
	BEGIN
		IF t.tbl.current_mid >= Tc.MAX_MID THEN
			t.tbl.current_mid := 1
		ELSE
			INC(t.tbl.current_mid)
		END;
		hdr.mid := t.tbl.current_mid; 
		IF unsubscribe THEN
			hdr.mid_resp := hdr.mid;
		ELSE
			hdr.mid_resp := t.mid_resp
		END;
		IF toall THEN
			hdr.dst_sid := -1
		ELSE
			hdr.dst_sid := Tb.Int_Int16(t.sid)
		END;
		hdr.aid := t.tbl.stations[t.isid].aid;
		t.mapper.WriteHeader(hdr, t.tbl.stations[t.isid].wr);
		t.src_mid := hdr.mid;
	END WriteHeader;

	PROCEDURE WriteItem (t: StdTable;  item-: Ti.Item);
	BEGIN
		t.mapper.WriteObject(item, t.tbl.stations[t.isid].wr)
	END WriteItem;
	
	PROCEDURE DoSend (t: StdTable; sids: SET);
		VAR j: INTEGER; length: Tb.Int32;
	BEGIN
		IF sids = {} THEN
			(* send to one recipient *)
			t.tbl.stations[t.isid].Send(-1);
			IF ~t.tbl.stations[t.isid].wr.complete THEN
				t.src_mid := 0
			END
		ELSIF Tc.FLAGS_MCAST IN t.tbl.stations[t.isid].chs.flags THEN
			(* send to all multicast group *)
			t.tbl.stations[t.isid].Send(-1);
			IF ~t.tbl.stations[t.isid].wr.complete THEN
				t.src_mid := 0
			END
		ELSE
			(* send the same buffer of length to group *)
			length := t.tbl.stations[t.isid].wr.length;
			j := 0;
			WHILE (j < 32) & (j < t.tbl.n_stations) DO
				IF j IN sids THEN
					t.tbl.stations[j].Send(length);
					IF ~t.tbl.stations[j].wr.complete THEN
						t.src_mid := 0
					END
				END;
				INC(j)
			END
		END
	END DoSend;

	PROCEDURE BeginWrite (t: StdTable;  item-: Ti.Item; sids: SET);
	BEGIN
		WriteItem(t, item);
		WriteHeader(t, FALSE, (sids # {}));
		DoSend(t, sids);
	END BeginWrite;
	
	PROCEDURE (t: StdTable) Write ( data-: Tb.AnyRec; 
	wait_write_permission: BOOLEAN);
	BEGIN
		t.WriteSid(data, wait_write_permission, t.isid)
	END Write;

	PROCEDURE (t: StdTable) WriteSid ( data-: Tb.AnyRec; 
	wait_write_permission: BOOLEAN; sid: INTEGER);
		VAR is_ready: BOOLEAN; tp_waiting_expired: Tb.Time;
		item: Ti.Item;
	BEGIN
		t.isid := sid;
		t.ti.LookupItem(data, item);
		ASSERT(item.adr # 0, Tb.NIL_RESULT);
		IF t.tbl.is_updating & Synch(t, writeOp) THEN
			IF wait_write_permission THEN
				t.tbl.stations[t.isid].ReadyForSending(is_ready)
			ELSE
				is_ready := TRUE
			END;
			IF is_ready THEN
				ASSERT(t.started, Tb.ALREADY_INITIALIZED);
				BeginWrite(t, item, {})
			ELSE
				IF t.tqos.tp_send_timeout < 0 THEN
					tp_waiting_expired := Tb.INFINITE_TIME
				ELSE
					Tb.Long_Time(t.tqos.tp_send_timeout, tp_waiting_expired);
				END;
				Put(t, writeOp, item, NIL, NIL, {}, 0, Ta.waitWritePermission, tp_waiting_expired)
			END
		ELSE
			IF wait_write_permission THEN
				Put(t, writeOp, item, NIL, NIL, {}, 0, Ta.waitWritePermission, Tb.ZERO_TIME)
			ELSE
				Put(t, writeOp, item, NIL, NIL, {}, 0, Ta.noWait, Tb.ZERO_TIME)
			END
		END;
	END WriteSid;
	
	PROCEDURE (t: StdTable) WriteTo ( data-: Tb.AnyRec; sids: SET);
		VAR item: Ti.Item;
	BEGIN
		t.ti.LookupItem(data, item);
		ASSERT(item.adr # 0, Tb.NIL_RESULT);
		IF t.tbl.is_updating & Synch(t, writeOp) THEN
			ASSERT(t.started, Tb.ALREADY_INITIALIZED);
			BeginWrite(t, item, sids)
		ELSE
			Put(t, writeOp, item, NIL, NIL, sids, 0, Ta.noWait, Tb.ZERO_TIME)
		END
	END WriteTo;
	
	PROCEDURE (t: StdTable) Wait (waiting_expired: LONGINT);
		VAR waiting: INTEGER; tp_waiting_expired: Tb.Time;
	BEGIN
		IF waiting_expired > 0 THEN
			waiting := Ta.waitUntilTimeout
		ELSE
			ASSERT((waiting_expired > Ta.repeatForeachResponse) 
			& (waiting_expired < Ta.waitUntilTimeout), Tb.CHECK_RANGE_ERROR);
			waiting := SHORT(waiting_expired)
		END;
		IF waiting <= Ta.waitAnyMessage THEN
			waiting_expired := 0
		END;
		Tb.Long_Time(waiting_expired, tp_waiting_expired);
		Put(t, waitOp, dummy, NIL, NIL, {}, 0, waiting, tp_waiting_expired)
	END Wait;
	
	PROCEDURE (t: StdTable) Repeat (command: Ta.TableCommand; par: Tb.AnyPtr);
	BEGIN
		(* ClearTable(t) *)
		Put(t, tcallOp, dummy, command, par, {}, 0, 
			Ta.repeatForeachResponse, Tb.ZERO_TIME)
	END Repeat;
	
	PROCEDURE (t: StdTable) Request ( data-: Tb.AnyRec; 
	wait_write_permission: BOOLEAN; command: Ta.TableCommand; par: Tb.AnyPtr);
	BEGIN
		(* ClearTable(t) *)
		t.Write(data, wait_write_permission);
		Put(t, tcallOp, dummy, command, par, {}, 0, Ta.request, Tb.ZERO_TIME)
	END Request;
	
	PROCEDURE (t: StdTable) Publish (permanent: BOOLEAN; command: Ta.TableCommand; 
	par: Tb.AnyPtr);
		VAR mid_resp: INTEGER;
	BEGIN
		IF permanent THEN mid_resp := Tc.PERMANENT_MID ELSE mid_resp := 0 END;
		Put(t, tcallOp, dummy, command, par, {}, mid_resp, Ta.publish, Tb.ZERO_TIME)
	END Publish;
	
	PROCEDURE (t: StdTable) Subscribe (controlled: BOOLEAN; command: Ta.TableCommand; 
	par: Tb.AnyPtr);
		VAR src_mid: INTEGER;
	BEGIN
		IF controlled THEN src_mid := 0 ELSE src_mid := Tc.PERMANENT_MID END;
		Put(t, tcallOp, dummy, command, par, {}, src_mid, Ta.subscribe, Tb.ZERO_TIME)
	END Subscribe;
	
	PROCEDURE (t: StdTable) Unsubscribe (inform_publisher: BOOLEAN);
	BEGIN
		IF inform_publisher THEN
			ASSERT(t.started, Tb.ALREADY_INITIALIZED);
			WriteItem(t, t.ti.item);
			(* send last message with mid_resp = mid, src_mid is changed *)
			WriteHeader(t, TRUE, FALSE);
			DoSend(t, {});
		ELSE
			INC(t.src_mid)
		END
	END Unsubscribe;

	PROCEDURE ProcessTable (t: StdTable; VAR prod: Ta. ProData; 
	VAR t_waiting_expired: Tb.Time);
		VAR e: Elem; operations: INTEGER; freeze_process, waitOp_done, is_ready, 
		time_expired, broken_chain, time_to_call: BOOLEAN; t_crt: Tb.Time;
		oldisid: INTEGER;
	BEGIN
		operations := 0; oldisid := t.isid;
		IF prod.got_message THEN
			IF t.sid = Ts.callerSID THEN
				t.isid := prod.sid		(* for response to caller sid *)
			END
		END;
		freeze_process := FALSE; waitOp_done := FALSE; broken_chain := FALSE;

		WHILE (t.queue # NIL) & (operations < maxOperations) & ~freeze_process DO
			e := t.queue; t.queue := e.next; t.tail := t.queue; INC(operations);
			IF ~broken_chain THEN
				(* check expiration date *)
				time_expired := FALSE;
				time_to_call := FALSE;
				IF (e.waiting # Ta.noWait) & Tb.TimePositive(e.tp_waiting_expired) THEN
					Tb.GetTime(t_crt);
					IF Tb.CompareTime(t_crt, e.t_waiting_expired) < 0 THEN
						freeze_process := TRUE;	(* still waiting *)
					ELSE
						(* waiting expired - fire response timeout event *)
						e.t_waiting_expired := Tb.ZERO_TIME;
						e.tp_waiting_expired := Tb.ZERO_TIME;	(* to be reset in Expiration *)
						CASE e.waiting OF
						| Ta.waitUntilTimeout:
							time_expired := TRUE;
						| Ta.subscribe, Ta.repeatForeachResponse:
							prod.res := RESPONSE_TIMEOUT;
							IF Tb.TRAL_ERRORS IN Tb.trace THEN
								Log.String("*response subscribe timeout"); Log.Ln
							END
						| Ta.publish:
							(* publish periodially *)
							time_to_call := TRUE;
						ELSE
							(* timeout errors *)
							IF e.op = writeOp THEN
								prod.res := SEND_TIMEOUT;
								IF Tb.TRAL_ERRORS IN Tb.trace THEN
									Log.String("*send timeout"); Log.Ln
								END;
								broken_chain := TRUE;
							ELSE
								prod.res := RESPONSE_TIMEOUT;
								IF Tb.TRAL_ERRORS IN Tb.trace THEN
									Log.String("*response timeout"); Log.Ln
								END;
								broken_chain := TRUE;
							END;
						END
					END
				END
			END;
			IF ~time_expired & ~broken_chain THEN
				ASSERT(t.started, Tb.ALREADY_INITIALIZED);
				CASE e.op OF
				| clearOp: DoClear(t, e.item); t.src_mid := 0;
				| tcallOp: 
					CASE e.waiting OF
					| Ta.publish:
						(* init here to the previous mid *)
						IF e.mid = Tc.PERMANENT_MID THEN
							IF t.hdr.mid = t.hdr.mid_resp THEN
								IF prod.got_message THEN
									t.isid := oldisid;
									prod.got_message := FALSE	(* ignore stop *)
								END
							END;
							t.mid_resp := Tc.PERMANENT_MID;
							e.current_mid := Tc.PERMANENT_MID
						ELSIF prod.got_message THEN
							IF t.hdr.mid = t.hdr.mid_resp THEN
								IF Tb.TRAL_MESSAGE IN Tb.trace THEN
									Log.String("Stop publish mid="); Log.Int(t.hdr.mid);  Log.Ln
								END;
								DoClear(t, e.item);	(* stop publish *)
								IF prod.got_message THEN
									t.isid := oldisid;
									prod.got_message := FALSE
								END;
								time_to_call := FALSE
							ELSE
								IF Tb.TRAL_MESSAGE IN Tb.trace THEN
									Log.String("Start publish mid="); Log.Int(t.hdr.mid);  Log.Ln
								END;
								t.mid_resp := t.hdr.mid	(* start publish *)
							END;
							e.current_mid := t.hdr.mid
						END;
						IF t.mid_resp = 0 THEN
							time_to_call := FALSE
						END;
						IF prod.got_message OR time_to_call THEN
							e.tcmd(t, prod, e.par)
						END;
						freeze_process := TRUE;
					| Ta.subscribe:
						freeze_process := TRUE;
						IF prod.got_message THEN
							IF e.current_mid = 0 THEN e.current_mid := t.src_mid; END;
							IF e.mid = Tc.PERMANENT_MID THEN
								t.src_mid := Tc.PERMANENT_MID;
								e.current_mid := Tc.PERMANENT_MID;
								e.tcmd(t, prod, e.par);
							ELSIF t.src_mid # e.current_mid THEN
								IF Tb.TRAL_MESSAGE IN Tb.trace THEN
									Log.String("Subsciption changed mids="); Log.Int(t.src_mid); 
									Log.Int(e.current_mid); Log.Ln
								END;
								freeze_process := FALSE;	(* subscription changed *)
							ELSIF t.hdr.mid_resp # e.current_mid THEN
								prod.res := RESPONSE_MISMATCH;
								IF Tb.TRAL_ERRORS IN Tb.trace THEN
									Log.String("*response mismatch mid_resp="); 
									Log.Int(t.hdr.mid_resp);
									Log.String(" current_mid="); Log.Int(e.current_mid); 
									Log.Ln
								END
							ELSE
								e.tcmd(t, prod, e.par)
							END;
							e.t_waiting_expired := Tb.ZERO_TIME;
							e.tp_waiting_expired := Tb.ZERO_TIME
						END;
					| Ta.repeatForeachResponse:
						IF prod.got_message THEN
							t.mid_resp := t.hdr.mid;	(* fix request mid *)
							e.tcmd(t, prod, e.par);
							e.t_waiting_expired := Tb.ZERO_TIME;
							e.tp_waiting_expired := Tb.ZERO_TIME
						END;
						freeze_process := TRUE;
					| Ta.request:
						freeze_process := TRUE;
						IF prod.got_message THEN
							IF e.current_mid = 0 THEN e.current_mid := t.src_mid; END;
							IF t.src_mid # e.current_mid THEN
								IF Tb.TRAL_ERRORS IN Tb.trace THEN
									Log.String("?request mismatch src_mid="); 
									Log.Int(t.src_mid); Log.String(" current_mid="); Log.Int(e.current_mid); 
									Log.Ln
								END
							ELSIF t.hdr.mid_resp # e.current_mid THEN
								IF Tb.TRAL_ERRORS IN Tb.trace THEN
									Log.String("?request mismatch mid_resp="); 
									Log.Int(t.hdr.mid_resp); Log.String(" current_mid="); 
									Log.Int(e.current_mid); Log.Ln
								END
							ELSE
								freeze_process := FALSE;
								t.src_mid := 0;
								e.tcmd(t, prod, e.par)
							END
						END
					ELSE
						e.tcmd(e.table, prod, e.par)
					END;
				| writeOp: 
					IF (e.sids = {}) & (e.waiting = Ta.waitWritePermission) THEN
						t.tbl.stations[t.isid].ReadyForSending(is_ready)
					ELSE
						is_ready := TRUE
					END;
					IF is_ready THEN
						BeginWrite(t, e.item, e.sids);
						freeze_process := FALSE
					ELSE
						freeze_process := TRUE
					END
				| waitOp:
					IF e.waiting < Ta.waitUntilTimeout THEN
						IF waitOp_done OR ~prod.got_message THEN
							freeze_process := TRUE
						ELSE
							waitOp_done := TRUE;
							freeze_process := FALSE;
							IF e.waiting <= Ta.waitForResponse THEN
								IF t.hdr.mid_resp # t.src_mid THEN
									freeze_process := TRUE;
									prod.res := RESPONSE_MISMATCH;
									IF Tb.TRAL_ERRORS IN Tb.trace THEN
										Log.String("*response mismatch2 mid_resp="); 
										Log.Int(t.hdr.mid_resp);
										Log.String(" current_mid="); Log.Int(e.current_mid); 
										Log.Ln
									END
								END
							END
						END
					END
				ELSE
				END
			END;
			
			IF freeze_process THEN
				t.queue := e
			END;
			IF e # t.queue THEN
				e.used := FALSE
			END;
			t.tail := NIL
		END;
		IF (t.queue # NIL) & (t.queue.waiting # Ta.noWait) THEN
			SetWaitingExpiration(t.queue, t_waiting_expired)
		END;
	END ProcessTable;
	
	PROCEDURE (t: StdTable) HandleMsg (VAR msg: Tb.Msg);
	BEGIN
		t.mapper.HandleMsg(msg)
	END HandleMsg;

BEGIN
	zero_prod.sid := 0; zero_prod.res := 0; zero_prod.got_message := FALSE;
	zero_prod.changed := FALSE	
END Ta_Tables.
