MODULE Ta_TunedMappers;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Log, Files, SYSTEM;
		
	CONST
		BUFFER_LENGTH* = Tb.NAME_LENGTH;
		(** Scanner.opts **)
		withCrLf* = 1; withSpaces* = 2; separateStrings* = 3;
			specialAsString* = 4; bracesAsString* = 5; xmlMap* = 6; xmlStrings* = 7;
			unquotePercent* = 8; unquoteAmp* = 9; addLf* = 10; withoutLf* = 11;
			separateQuotes* = 12;
		(** Scanner.type **)
		undef* = 0;
		string* = 1; quote* = 2; dquote* = 3; crlf* = 4; spaces* = 5;
		braces* = 7; special* = 10;
		eot* = 30;
		

	TYPE
		Buffer* = ARRAY BUFFER_LENGTH OF Tb.Int8;
		
		Scanner* = RECORD
			type-: INTEGER;
			string-: Tb.Name;
			opts: SET;
			rd: Files.Reader;
			buffer: Buffer;
			pos, length: Tb.Int32;
			ch: CHAR
		END;
		
		Formatter* = RECORD
			eof-: BOOLEAN;
			wr: Files.Writer;
			buffer, save: Buffer;
			beg, length, save_length: Tb.Int32;
			opts: SET;
		END;
		
		
	(* ========== Scanner Interface ========== *)
	PROCEDURE^ (VAR s: Scanner) ConnectTo* (rd: Files.Reader) ;
	PROCEDURE^ (VAR s: Scanner) SetOpts* (opts: SET) ;
	PROCEDURE^ (VAR s: Scanner) Scan* ;

	(* ========== Formatter Interface ========== *)
	PROCEDURE^ (VAR f: Formatter) ConnectTo* (wr: Files.Writer) ;
	PROCEDURE^ (VAR f: Formatter) SetOpts* (opts: SET) ;
	PROCEDURE^ (VAR f: Formatter) WriteString* (x: ARRAY OF CHAR) ;
	PROCEDURE^ (VAR f: Formatter) WriteChar* (x: CHAR) ;
	PROCEDURE^ (VAR f: Formatter) WriteInt* (x: Tb.Int32) ;
	PROCEDURE^ (VAR f: Formatter) WriteReal* (x: Tb.Real32) ;
	PROCEDURE^ (VAR f: Formatter) WriteLn* ;
	PROCEDURE^ (VAR f: Formatter) FlushBuf* ;
	
	(* ========== Scanner Implementation ========== *)
	PROCEDURE (VAR s: Scanner) ConnectTo* (rd: Files.Reader) ;
	BEGIN
		ASSERT(rd # NIL, Tb.NIL_POINTER);
		s.type := undef; s.rd := rd; 
		s.pos := 0; s.length := 0; s.opts := {}; s.ch := 0X
	END ConnectTo;
	
	PROCEDURE (VAR s: Scanner) SetOpts* (opts: SET) ;
	BEGIN
		s.opts := opts
	END SetOpts;
	
	PROCEDURE ReadChar (VAR s: Scanner; VAR ch: CHAR);
		VAR pos: Tb.Int32; j: INTEGER;
	BEGIN
		IF s.pos >= s.length THEN
			pos := s.rd.Pos();
			s.rd.ReadBytes(s.buffer, 0, BUFFER_LENGTH);
			s.pos := 0;
			s.length := s.rd.Pos() - pos;
			IF s.length < BUFFER_LENGTH THEN
				s.buffer[s.length] := 0
			END
		END;
		IF s.pos < s.length THEN
			ch := SYSTEM.VAL(CHAR,s.buffer[s.pos]); INC(s.pos)
		ELSE
			ch := 0X
		END
	END ReadChar;
	
	PROCEDURE TypeOf (opts: SET; ch: CHAR): INTEGER;
		VAR type: INTEGER;
	BEGIN
		CASE ch OF
		| 0X:	type := eot;
		| "'":	type := quote;
		| '"':	type := dquote;
		| 0AX, 0DX:	type := crlf;
		| 09X, " ":	type := spaces;
		| "{", "}", "[", "]", "(", ")":	
			IF bracesAsString IN opts THEN
				type := string
			ELSE
				type := braces
			END;
		| "A".."Z", "a".."z", "0".."9", "_":	type := string;
		ELSE	
			IF specialAsString IN opts THEN
				type := string
			ELSIF xmlMap IN opts THEN
				CASE ch OF
				| ':', '.', '-', '%':
					type := string
				| '/':
					IF xmlStrings IN opts THEN
						type := string
					ELSE
						type := special
					END
				ELSE
					type := special
				END
			ELSE
				type := special
			END
		END;
		RETURN type
	END TypeOf;
	
	PROCEDURE HC2I(hc: CHAR; VAR k: INTEGER);
		CONST oA = 041H; oF = 046H; oa = 061H; of = 066H; o0 = 030H; o9 = 039H;
		VAR j: INTEGER;
	BEGIN
		j := ORD(hc);
		IF (j >= oA) & (j <= oF) THEN k := 10+j-oA
		ELSIF (j >= oa) & (j <= of) THEN k := 10+j-oa
		ELSIF (j >= o0) & (j <= o9) THEN k := j-o0
		ELSE k := -1
		END;
	END HC2I;
	
	PROCEDURE H2Char(cc: ARRAY OF CHAR; VAR ch: CHAR);
		VAR k0, k1: INTEGER;
	BEGIN
		HC2I(cc[0], k0); HC2I(cc[1], k1);
		IF (k0 < 0) OR (k1 < 0) THEN
			ch := '*'
		ELSE
			ch := CHR(16*k0+k1)
		END;
	END H2Char;
	
	PROCEDURE ReadTypedChar (VAR s: Scanner; VAR ch: CHAR; 
	VAR type: INTEGER);
		VAR c: CHAR; cc: ARRAY 5 OF CHAR;
	BEGIN
		IF unquotePercent IN s.opts THEN
			ReadChar(s, c);
			IF c = '+' THEN
				ch := ' ';
				type := string
			ELSIF c = '%' THEN
				ReadChar(s, cc[0]);
				ReadChar(s, cc[1]);
				H2Char(cc, ch);
				type := string
			ELSE
				ch := c;
				type := TypeOf(s.opts, ch)		
			END
		ELSIF unquoteAmp IN s.opts THEN
			ReadChar(s, c);
			IF c = '&' THEN
				ReadChar(s, cc[0]);
				ReadChar(s, cc[1]);
				ReadChar(s, cc[2]);
				IF cc[2] = ';' THEN 
					cc[3] := 0X
				ELSE 
					ReadChar(s, cc[3]); cc[4] := 0X 
				END;
				IF cc = "lt;" THEN ch := '<'
				ELSIF cc = "gt;" THEN ch := '>'
				ELSIF cc = "amp;" THEN ch := '&'
				ELSE cc := '*' END;
				type := special
			ELSE
				ch := c;
				type := TypeOf(s.opts, ch)		
			END
		ELSE
			ReadChar(s, ch);
			type := TypeOf(s.opts, ch)
		END
	END ReadTypedChar;
	
	PROCEDURE (VAR s: Scanner) Scan* ;
		VAR ch: CHAR; j, type: INTEGER; done: BOOLEAN;
	BEGIN
		IF s.ch = 0X THEN
			ReadTypedChar(s, ch, s.type)
		ELSE
			ch := s.ch;
			s.type := TypeOf(s.opts, ch)
		END;
		j := 0; type := s.type;
		IF ~(withCrLf IN s.opts) & (s.type = crlf)
		OR ~(withSpaces IN s.opts) & (s.type = spaces) THEN
			WHILE type = s.type DO
				ReadTypedChar(s, ch, type)
			END;
			s.type := type
		END;
		WHILE (type # eot) & (type = s.type) DO
			s.string[j] := ch;
			INC(j);
			ReadTypedChar(s, ch, type);
			IF (ch = 0AX) & (separateStrings IN s.opts) THEN
				type := undef
			END;
			IF (separateQuotes IN s.opts) 
			& ((ch = '"') OR (ch = "'") OR (ch = '{') OR (ch = '}')) THEN
				type := undef
			END;
			IF xmlMap IN s.opts THEN
				IF (s.string[j-1] = '>') & (type = s.type) THEN
					type := undef
				END
			END
		END;
		s.string[j] := 0X;
		s.ch := ch;
	END Scan;

	(* ========== Formatter Implementation ========== *)
	PROCEDURE (VAR f: Formatter) ConnectTo* (wr: Files.Writer) ;
	BEGIN
		f.eof := FALSE; f.wr := wr; f.beg := 0; f.length := 0; f.save_length := 0;
		f.opts := {}
	END ConnectTo;
 
	PROCEDURE (VAR f: Formatter) SetOpts* (opts: SET) ;
	BEGIN
		f.opts := opts
	END SetOpts;
	
	PROCEDURE (VAR f: Formatter) WriteString* (x: ARRAY OF CHAR) ;
		VAR j: INTEGER;
	BEGIN
		j := 0;
		WHILE x[j] # 0X DO f.WriteChar(x[j]); INC(j) END
	END WriteString;
	
	PROCEDURE Char1 (VAR f: Formatter; x: CHAR);
		VAR b: Tb.Int8;
	BEGIN
		IF f.eof OR (f.length >= BUFFER_LENGTH) THEN
			f.FlushBuf
		END;
		b := SYSTEM.VAL(Tb.Int8, ORD(x));
		IF f.eof THEN
			f.save[f.save_length] := b; INC(f.save_length)
		ELSE
			IF f.save_length > 0 THEN
				f.buffer := f.save; f.beg := 0; f.length := f.save_length;
				f.save_length := 0;
				f.FlushBuf
			END;
			f.buffer[f.length] := b; INC(f.length)
		END
	END Char1;
	
	PROCEDURE I2HC(k: INTEGER; VAR hc: CHAR);
		CONST oA = 041H; o0 = 030H;
		VAR j: INTEGER;
	BEGIN
		IF k < 10 THEN
			hc := CHR(o0+k)
		ELSE
			hc := CHR(oA+k-10)
		END
	END I2HC;
	
	PROCEDURE Char2H (x: CHAR; cc: ARRAY OF CHAR);
		VAR j: INTEGER;
	BEGIN
		j := ORD(x);
		I2HC(j DIV 16, cc[0]); I2HC(j MOD 16, cc[1]);
	END Char2H;

	PROCEDURE (VAR f: Formatter) WriteChar* (x: CHAR) ;
		VAR cc: ARRAY 3 OF CHAR;
	BEGIN
		IF unquotePercent IN f.opts THEN
			CASE x OF
			' ':
				Char1(f, '+');
			| '+', '|', '"', "'", '>', '<', '%':
				Char2H(x, cc);
				Char1(f, '%'); Char1(f, cc[0]); Char1(f, cc[1])
			ELSE
				Char1(f, x)
			END
		ELSIF unquoteAmp IN f.opts THEN
			CASE x OF
			| '<':
					Char1(f, '&'); Char1(f, 'l'); Char1(f, 't'); Char1(f, ';'); 
			| '>':
					Char1(f, '&'); Char1(f, 'g'); Char1(f, 't'); Char1(f, ';'); 
			| '&':
					Char1(f, '&'); Char1(f, 'a'); Char1(f, 'm'); Char1(f, 'p'); Char1(f, ';'); 
			ELSE
				Char1(f, x)
			END
		ELSE
			Char1(f, x)
		END
	END WriteChar;
	
	PROCEDURE (VAR f: Formatter) WriteInt* (x: Tb.Int32) ;
		VAR str: Tb.Name;
	BEGIN
		Tb.IntToString(x, str); f.WriteString(str)
	END WriteInt;
	
	PROCEDURE (VAR f: Formatter) WriteReal* (x: Tb.Real32) ;
		VAR str: Tb.Name;
	BEGIN
		Tb.RealToString(x, str); f.WriteString(str)
	END WriteReal;
	
	PROCEDURE (VAR f: Formatter) WriteLn* ;
	BEGIN
		f.WriteChar(0AX);	
		IF ~(withoutLf IN f.opts) THEN
			IF addLf IN f.opts THEN
				f.WriteChar(0DX)
			END
		END;
		f.FlushBuf
	END WriteLn;
	
	PROCEDURE (VAR f: Formatter) FlushBuf* ;
		VAR pos: Tb.Int32;
	BEGIN
		pos := f.wr.Pos();
		f.wr.WriteBytes(f.buffer, f.beg, f.length);
		INC(f.beg, f.wr.Pos()-pos);
		IF f.beg >= f.length THEN
			f.beg := 0; f.length := 0; f.eof := FALSE
		ELSE
			f.eof := TRUE
		END
	END FlushBuf;

END Ta_TunedMappers.