MODULE Ta_MSHMChannel;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, Tl := Ta_Lib, Tb := Ta_Basics, Tc := Ta_Channels, Ta_SHMLib, Log;

	CONST
		SHMMIN = 128; O777 = 511; IPC_CREAT = 512; IPC_RMID = 0; IPC_STAT = 2;
		BUF_HEADER = 1234567890; HDR_LEN_SIZE = 8;
	
	TYPE
		ChannelQos* = RECORD (Tc.ChannelQosBase)
			mem_bufsize*: Tb.Int32;
			shm_rm_onexit*: BOOLEAN;
		END;

		Channel = POINTER TO ChannelDesc;
		ChannelDesc = RECORD (Tc.ChannelDesc)
			cqos: ChannelQos;
			hostname: Tb.Name64;
			aid, sid, host_sid: INTEGER;
			recv_offset: Tb.Int32;
			snd, rcd: Ta_SHMLib.ShmData;
		END;
		
		ChannelFactoryDesc* = RECORD (Tc.ChannelFactoryDesc)
		END;		
		ChannelFactory* = POINTER TO ChannelFactoryDesc;


	VAR
		defaultCQos-: ChannelQos;


	(* ========== Channel interface ========== *)
	PROCEDURE (channel: Channel) InitData (is_datahost, is_send_station: BOOLEAN; 
	 name-: ARRAY OF CHAR; VAR cqos: Tc.ChannelQosBase);
		VAR str: Tb.Name64; ok: BOOLEAN;
	BEGIN
		IF cqos IS ChannelQos THEN
			channel.cqos := cqos(ChannelQos);
		ELSE
			channel.cqos := defaultCQos;
			Tc.CopyChannelQos(cqos, channel.cqos);
		END;
		IF ~is_datahost THEN
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_SEND_VIA_HOST,
				Tc.TRANS_RECV, Tc.TRANS_RECV_INDIRECT, Tc.TRANS_CONNECTION};
			channel.link_stat := Tc.LINK_INIT;
		ELSE
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_SEND_DIRECT,
				Tc.TRANS_RECV, Tc.TRANS_RECV_DIRECT};
			Tb.Copy("area", channel.hostname);
			Tb.IntToString(channel.aid, str); Tb.Append(str, channel.hostname); 
			Tb.Append("_st", channel.hostname);
			Tb.IntToString(channel.sid, str); Tb.Append(str, channel.hostname); 
			Tb.Append(".log", channel.hostname); 
			
			(* create file with proper size *)
			Ta_SHMLib.CreateFileOfSize(channel.hostname, 
				channel.cqos.mem_bufsize, ok);
	
			IF ok THEN
				Ta_SHMLib.CreateShm(channel.hostname, 
					channel.cqos.mem_bufsize, channel.snd)
			ELSE
				channel.snd := Ta_SHMLib.zerodata
			END;
			IF channel.snd.shm_buffer # 0 THEN
				Tb.Clean(channel.snd.shm_buffer, channel.cqos.mem_bufsize);
				channel.link_stat := Tc.LINK_OK
			ELSE
				channel.CloseData(FALSE);
				channel.link_stat := Tc.LINK_ERR
			END;
			IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
				IF channel.snd.shm_buffer = 0 THEN Log.String("* ") END;
				Log.String("MSHM send_id="); Log.Int(channel.snd.shm_id); 
				Log.String(" file='"); Log.String(channel.hostname); 
				Log.String("' buffer="); Log.Int(channel.snd.shm_buffer); 
				Log.String(" memsize="); Log.Int(channel.cqos.mem_bufsize); Log.Ln
			END
		END
	END InitData;
	
	PROCEDURE (channel: Channel) GetFds (ind_first: INTEGER; 
	VAR ind_last: INTEGER; VAR fd_array: ARRAY OF Tb.Int32);
	BEGIN
		ind_last := ind_first
	END GetFds;
	
	PROCEDURE (channel: Channel) CloseData (shutdown: BOOLEAN);
	BEGIN
		Ta_SHMLib.DisconnectShmBuffer(channel.snd);
		Ta_SHMLib.DisconnectShmBuffer(channel.rcd);
		Ta_SHMLib.DeleteShm(channel.rcd, FALSE);
		Ta_SHMLib.DeleteShm(channel.snd, channel.cqos.shm_rm_onexit);
		IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
			Log.String("MSHM CloseData"); Log.Ln
		END;
		channel.link_stat := Tc.LINK_CLOSED;
	END CloseData;


	PROCEDURE (channel: Channel) Connect (indirect_channel: Tc.Channel; 
	disconnect_only: BOOLEAN; VAR fd_changed: BOOLEAN);
		VAR str: Tb.Name64; rc: Tb.Int32;
	BEGIN
		IF channel.link_stat = Tc.LINK_INIT THEN
			Tb.Copy("area", channel.hostname);
			Tb.IntToString(channel.aid, str); Tb.Append(str, channel.hostname); 
			Tb.Append("_st", channel.hostname);
			Tb.IntToString(channel.sid, str); Tb.Append(str, channel.hostname); 
			Tb.Append(".log", channel.hostname); 
			Ta_SHMLib.GetFileSize(channel.hostname, rc);
			IF rc >= 0 THEN
				Ta_SHMLib.ConnectShm(channel.hostname, rc, TRUE, channel.rcd)
			ELSE
				channel.rcd := Ta_SHMLib.zerodata
			END;
			IF channel.rcd.shm_buffer # 0 THEN
				SYSTEM.GET(channel.rcd.shm_buffer+4, channel.recv_offset);
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("MSHM connect recv_id="); Log.Int(channel.rcd.shm_id); 
					Log.String(" file='"); Log.String(channel.hostname); 
					Log.String("' buffer="); Log.Int(channel.rcd.shm_buffer); Log.Ln
				END;
				channel.link_stat := Tc.LINK_OK
			END
		END
	END Connect;

	PROCEDURE ReadFromBuffer(dst_buf, src_buf: Tb.Address; recv_offset,
	buffer_length, len: Tb.Int32); 
		VAR i, j: Tb.Int32; byte: Tb.Int8;
	BEGIN
		FOR i := 0 TO buffer_length-1 DO
			j := (i + recv_offset) MOD len;
			SYSTEM.GET(src_buf+j, byte);
			SYSTEM.PUT(dst_buf+i, byte);
		END
	END ReadFromBuffer;
	
	PROCEDURE WriteToBuffer(dst_buf, src_buf: Tb.Address; VAR o_end: Tb.Int32;
	buffer_length, len: Tb.Int32);
		VAR offset, i, j: Tb.Int32; byte: Tb.Int8;
	BEGIN
		offset := o_end;
		FOR i := 0 TO buffer_length-1 DO
			j := (i + offset) MOD len;
			SYSTEM.GET(src_buf+i, byte);
			SYSTEM.PUT(dst_buf+j, byte);
		END;
		o_end := (o_end + buffer_length) MOD len
	END WriteToBuffer;
	
	PROCEDURE (channel: Channel) Send (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32);
		VAR j, space_size, len, o_begin, o_end, mhdr, mlen: Tb.Int32; s_channel: Channel;
	BEGIN
		ASSERT(indirect_channel IS Channel, Tb.WRONG_TYPE);
		s_channel := indirect_channel(Channel);
		IF (s_channel.snd.shm_buffer # 0) & (buffer_length # 0) THEN
			len := s_channel.cqos.mem_bufsize - 8;
			SYSTEM.GET(s_channel.snd.shm_buffer, o_begin);
			SYSTEM.GET(s_channel.snd.shm_buffer+4, o_end);
			(* 1. check for space *)
			ASSERT(buffer_length+HDR_LEN_SIZE <= len, Tb.CHECK_RANGE_ERROR);
			IF o_end >= o_begin THEN
				space_size := o_end - o_begin
			ELSE
				space_size := len - (o_begin - o_end)
			END;
			(* 2. clean old messages *)
			j := o_begin;
			WHILE (len-space_size < buffer_length+HDR_LEN_SIZE) DO
				ReadFromBuffer(SYSTEM.ADR(mhdr), s_channel.snd.shm_buffer+8, j, 4, len);
				ReadFromBuffer(SYSTEM.ADR(mlen), s_channel.snd.shm_buffer+8, j+4, 4, len);
				IF mhdr # BUF_HEADER THEN
					o_end := o_begin;
					space_size := 0
				ELSE
					mhdr := 0;
					WriteToBuffer(s_channel.snd.shm_buffer+8, SYSTEM.ADR(mhdr), j, 4, len);
					INC(j, mlen);
					IF j >= len THEN
						j := j MOD len
					END;
					DEC(space_size, mlen);
				END
			END;
			(* 3. change start position *)
			o_begin := j;
			SYSTEM.PUT(s_channel.snd.shm_buffer, o_begin);
			(* 4. copy new buffer *)
			mhdr := BUF_HEADER;
			WriteToBuffer(s_channel.snd.shm_buffer+8, SYSTEM.ADR(mhdr), o_end, 4, len);
			mlen := buffer_length;
			WriteToBuffer(s_channel.snd.shm_buffer+8, SYSTEM.ADR(mlen), o_end, 4, len);
			WriteToBuffer(s_channel.snd.shm_buffer+8, buffer, o_end, buffer_length, len);
			SYSTEM.PUT(s_channel.snd.shm_buffer+4, o_end);
			IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
				Log.String("mshm_send begin="); Log.Int(o_begin); Log.String(" end="); Log.Int(o_end); 
				Log.String(" len="); Log.Int(buffer_length); Log.String("+8"); Log.Ln
			END;
			nbytes := buffer_length+HDR_LEN_SIZE
		ELSE
			nbytes := 0
		END
	END Send;
	
	PROCEDURE (channel: Channel) Recv (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32;
	VAR dest: Tc.Dest);
		VAR len, o_begin, o_end, old_o_begin, mhdr, mlen, rlen: Tb.Int32; 
		done: BOOLEAN;
	BEGIN
		IF channel.rcd.shm_buffer # 0 THEN
			len := channel.cqos.mem_bufsize - 8;
			SYSTEM.GET(channel.rcd.shm_buffer, o_begin);
			SYSTEM.GET(channel.rcd.shm_buffer+4, o_end);
			done := FALSE; nbytes := 0;
			WHILE ~done & (o_begin # o_end) & (channel.recv_offset # o_end) DO
				(* 1. save old begin *)
				SYSTEM.GET(channel.rcd.shm_buffer, o_begin);
				old_o_begin := o_begin;
				(* 2. check if valid message *)
				ReadFromBuffer(SYSTEM.ADR(mhdr), channel.rcd.shm_buffer+8, 
					channel.recv_offset, 4, len);
				IF mhdr # BUF_HEADER THEN	
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("*mshm_recv header error begin="); Log.Int(o_begin); 
						Log.String(" end="); Log.Int(o_end); Log.String(" offset="); 
						Log.Int(channel.recv_offset); Log.String(" sid=");
						Log.Int(channel.sid); Log.Ln
					END;
					channel.recv_offset := o_end;
					nbytes := -4;
					done := TRUE
				ELSE	
					ReadFromBuffer(SYSTEM.ADR(mlen), channel.rcd.shm_buffer+8, 
						channel.recv_offset+4, 4, len);
					IF mlen < buffer_length THEN rlen := mlen ELSE rlen := buffer_length END;
					ReadFromBuffer(buffer, channel.rcd.shm_buffer+8, 
						channel.recv_offset+8, rlen, len);			
					SYSTEM.GET(channel.rcd.shm_buffer, o_begin);
					IF old_o_begin = o_begin THEN
						IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
							Log.String("mshm_recv begin="); Log.Int(o_begin); Log.String(" end="); 
							Log.Int(o_end); Log.String(" len="); Log.Int(rlen); 
							Log.String("+8 offset="); Log.Int(channel.recv_offset); Log.String(" sid=");
							Log.Int(channel.sid); Log.String(" buffer=");
							Log.Int(channel.rcd.shm_buffer);Log.Ln
						END;
						channel.recv_offset := (channel.recv_offset+mlen+HDR_LEN_SIZE) MOD len;
						nbytes := rlen;
						done := TRUE
					END
				END
			END
		END	
	END Recv;	

	PROCEDURE (channel: Channel) MyDest ( dest-: Tc.Dest): BOOLEAN;
	BEGIN 
		RETURN FALSE
	END MyDest;


	(* ========== Initialize ========== *)
	PROCEDURE (cf: ChannelFactory) NewChannel* (aid, sid, host_sid: INTEGER; 
	(*OUT*)VAR chan: Tc.Channel; (*OUT*)VAR res: INTEGER);
		VAR channel: Channel;
	BEGIN
		NEW(channel);
		(* SHM Win for XDS is not implementent due to internal compiler problems *)
		(*ASSERT(channel = NIL, Tb.UNSUPPORTED);*)
		channel.flags := {Tc.FLAGS_MEMORY, Tc.FLAGS_MCAST};
		channel.trans_flags := {};
		channel.link_stat := Tc.LINK_NONE;
		channel.aid := aid; channel.sid := sid; channel.host_sid := host_sid;
		chan := channel;
		res := 0
	END NewChannel;
	
BEGIN
	Tc.DefaultChannelQos(defaultCQos);
	defaultCQos.protocol := "MSHM";
	defaultCQos.mem_bufsize := 32768;
	defaultCQos.shm_rm_onexit := TRUE; 	
END Ta_MSHMChannel.