MODULE Ta_MTAreas;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	(* ----------  Multi Threaded implementation for XDS ---------- *)
	IMPORT Log, Tb := Ta_Basics, Tc := Ta_Channels, Ta_FactManager,
	Ta := Ta_Areas, Ts := Ta_Stations, Tt := Ta_Tables, oberonRTS, ProgramArgs, 
	TextIO, SeqFile, IOConsts, IOResult, SYSTEM, Processes, Semaphores;
	

	CONST MAX_N_PROCESSES* = 64;
	
	TYPE
		Area = POINTER TO AreaDesc;
		ThreadData = POINTER TO ThreadDataDesc;
		Directory = POINTER TO DirectoryDesc;

		AreaDesc = RECORD (Ta.AreaDesc)
			aid : INTEGER;		(* > 0 unique area identifier *)
			name : Tb.Name64;
			aqos : Ta.AreaQos;
			n_stations, max_stations, host_sid, send_sid: INTEGER;
			bu: Ts.Buffer;
			host_station, send_station: Ts.Station;
			tbl: Tt.Tablist;
			(* ---------- Runtime Data ---------- *)
			t_cycle, t_setpoint, t_crt, t_crtu, tp_cdt, t_expired: Tb.Time;
			(* other data *)
			on_update: Ta.AreaUpdater;
		END;
		
		ThreadDataDesc = RECORD
			pid: Processes.ProcessId;
			client_data: SYSTEM.ADDRESS;
			ind: INTEGER;
			t_cycle: Tb.Time;
			area: Area;
			started: BOOLEAN
		END;
			
		DirectoryDesc = RECORD (Ta.DirectoryDesc)
			loaded_as: INTEGER;
			finish_flag: INTEGER;
			dqos: Ta.DirectoryQos;
			n_processes: INTEGER;
			processes: POINTER TO ARRAY OF ThreadData;
			mainarea: Area;
			t_finish_at: Tb.Time;
		END;

	VAR
		dir: Directory;

	(* ========== Area Implementation ========== *)
	PROCEDURE DoBeforeRun (area: Area);
		VAR cst: Tb.Time; s: Tb.Name; i: INTEGER;
	BEGIN
		dir.finish_flag := 0; Tb.Long_Time(dir.dqos.cycle, area.t_cycle);
		Tb.GetTime(area.t_crtu); area.t_setpoint := area.t_crtu;
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Tb.Time_String(area.t_crtu, {Tb.tDD, Tb.tMM, Tb.tYY, 
				Tb.thh, Tb.tmm, Tb.tss, Tb.tms, Tb.tloc, Tb.tdwh, Tb.tdwl}, s);
			Log.String(s); Log.String(" Run cycle=");
			Tb.Time_String(area.t_cycle, {Tb.thh,Tb.tmm,Tb.tss,Tb.tms}, s);
			Log.String(s); Log.String(" loaded_as="); Log.Int(dir.loaded_as); Log.Ln
		END;
		IF dir.dqos.expired # 0 THEN
			Tb.Long_Time(dir.dqos.expired, cst);
			Tb.AddTime(area.t_setpoint, cst, area.t_expired);
			IF Tb.TRAL_RUNTIME IN Tb.trace THEN
				Tb.Time_String(area.t_expired, {Tb.tDD, Tb.tMM, Tb.tYY, 
					Tb.thh, Tb.tmm, Tb.tss, Tb.tms, Tb.tloc, Tb.tdwh, Tb.tdwl}, s);
				Log.String("expired at "); Log.String(s); 
				Log.Ln
			END
		END;
	END DoBeforeRun;
	
	PROCEDURE (area: Area) Start ();
		VAR j: INTEGER;
	BEGIN
		ASSERT(area = dir.mainarea, Tb.OBJECT_INTEGRITY);
		DoBeforeRun(area);
		FOR j := 1 TO dir.n_processes-1 DO
			Processes.Activate(dir.processes[j].pid); dir.processes[j].started := TRUE;
		END
	END Start;

	PROCEDURE (area: Area) Finish (finish_flag: INTEGER);
	BEGIN
		ASSERT(finish_flag > 0, Tb.CHECK_RANGE_ERROR);
		ASSERT(area = dir.mainarea, Tb.OBJECT_INTEGRITY);
		dir.finish_flag := finish_flag;
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Log.String("Finish"); Log.Int(dir.finish_flag); Log.Ln
		END
	END Finish;

	PROCEDURE (area: Area) UpdateData ();
	BEGIN
		Tb.GetTime(area.t_crtu);
		(*RecvMessages(area);
		UpdateData(area);
		SendMessages(area);*)
		Tb.GetTime(area.t_crt);
		Tb.SubtractTime(area.t_crt, area.t_crtu, area.tp_cdt);
	END UpdateData;
	
	PROCEDURE (area: Area) Update ();
		VAR s: Tb.Name;
	BEGIN
		ASSERT(area = dir.mainarea, Tb.OBJECT_INTEGRITY);
		area.UpdateData;
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Tb.Time_String(area.t_crtu, 
				{Tb.tDD, Tb.tMM, Tb.tYY, Tb.thh, Tb.tmm, Tb.tss, 
				Tb.tms, Tb.tloc}, s); Log.String(s); 
			Log.String(" updated proctime=");
			Tb.Time_String(area.tp_cdt, {Tb.thh, Tb.tmm, Tb.tss, Tb.tms}, s);
			Log.String(s); 
			Log.Ln
		END;
	END Update;

	PROCEDURE (area: Area) AddStations ( name-: ARRAY OF CHAR; 
	n_appended, host_sid: INTEGER;  sqos-: Tc.StationQos; VAR cqos: Tc.ChannelQos;
	(*OUT*)VAR res: INTEGER);
	BEGIN
		res := 0
	END AddStations;

	PROCEDURE (area: Area) NStations (): INTEGER; 
	BEGIN
		RETURN area.n_stations
	END NStations;

	PROCEDURE (area: Area) NewTable (VAR rec: Tb.AnyRec; 
	 iface_name-: ARRAY OF CHAR; dst_oid, src_oid: INTEGER;
	 tqos-: Ta.TableQos; on_data: Ta.OnDataHandler): Ta.Table ;
	BEGIN
		IF area.tbl = NIL THEN
			Tt.NewTablist(area.n_stations, area.tbl);
			ASSERT(area.tbl # NIL, Tb.NIL_POINTER)
		END;
		(* to be added to list *)
		RETURN area.tbl.NewTable(rec, iface_name, dst_oid, src_oid, tqos, on_data);
	END NewTable;

	(* ========== Threads Routines ========== *)
	PROCEDURE ThreadRun;
		VAR ttd: ThreadData;
	BEGIN
		ttd := SYSTEM.VAL(ThreadData, Processes.MyParam());
		WHILE dir.finish_flag = 0 DO
			ttd.area.UpdateData;
			Tb.Sleep(ttd.t_cycle);
		END;
		ttd.started := FALSE;
		IF Tb.TRAL_RUNTIME IN Tb.trace THEN
			Log.String("finished thread #"); Log.Int(ttd.ind); Log.Ln
		END;
	END ThreadRun;

	(* ========== Directory Implementation ========== *)
	PROCEDURE (d: Directory) NewArea (aid: INTEGER; name: ARRAY OF CHAR;
	max_stations: INTEGER;  aqos-: Ta.AreaQos; on_update: Ta.AreaUpdater): Ta.Area;
		VAR j: INTEGER; param: SYSTEM.ADDRESS;
	BEGIN
		ASSERT(max_stations <= MAX_N_PROCESSES, Tb.CHECK_RANGE_ERROR);
		NEW(d.mainarea); d.mainarea.aid := aid; Tb.Copy(name, d.mainarea.name);
		d.mainarea.max_stations := max_stations; d.mainarea.aqos := aqos; 
		d.mainarea.on_update := on_update; d.mainarea.host_sid := 0; 
		d.mainarea.send_sid := 1;

		d.n_processes := max_stations;
		NEW(d.processes, d.n_processes);
		FOR j := 1 TO d.n_processes-1 DO
			NEW(d.processes[j]);
			Tb.Long_Time(d.dqos.cycle, d.processes[j].t_cycle);
			NEW(d.processes[j].area); d.processes[j].area.aid := aid; 
			Tb.Copy(name, d.processes[j].area.name); 
			d.processes[j].area.max_stations := max_stations;
			d.processes[j].area.aqos := aqos; 
			d.processes[j].area.on_update := on_update;
			d.processes[j].area.host_sid := 0;
			d.processes[j].area.send_sid := j;
			d.processes[j].ind := j;
			param := SYSTEM.VAL(SYSTEM.ADDRESS, d.processes[j]);
			Processes.Create (ThreadRun, aqos.thread_mem_size, 0, 
				param, d.processes[j].pid);
		END;
		RETURN d.mainarea;
	END NewArea;
	
	PROCEDURE Run (d: Directory);
		VAR j: INTEGER; b: BOOLEAN; t_setpoint, cst: Tb.Time; 
	BEGIN
		WHILE d.finish_flag = 0 DO
			j := 0; b := FALSE; 
			IF d.dqos.with_compensation THEN
				t_setpoint := d.mainarea.t_setpoint;
				WHILE ~b & (j < 5) DO
					Tb.AddTime(t_setpoint, d.mainarea.t_cycle, t_setpoint);
					IF Tb.CompareTime(t_setpoint, d.mainarea.t_crtu) >= 0 THEN
						b := TRUE
					END;
					INC(j)
				END
			ELSE
				Tb.AddTime(d.mainarea.t_crtu, d.mainarea.t_cycle, t_setpoint); b := TRUE
			END;
			IF b THEN
				Tb.SubtractTime(t_setpoint, d.mainarea.t_crtu, cst);
				Tb.Sleep(cst);
				d.mainarea.t_setpoint := t_setpoint;
				d.mainarea.Update;
				IF (d.dqos.expired # 0) & (Tb.CompareTime(d.mainarea.t_crtu, 
				d.mainarea.t_expired) > 0) THEN
					d.mainarea.Finish(Ta.FINISH_EXPIRED)
				END
			END
		END;
		Tb.GetTime(d.mainarea.t_crtu);
		Tb.AddTime(d.mainarea.t_crtu, d.mainarea.t_cycle, d.t_finish_at);
		b := FALSE; j := 1;
		WHILE (j < d.n_processes) & ~b DO b := d.processes[j].started; INC(j) END;
		WHILE b & (Tb.CompareTime(d.mainarea.t_crtu, d.t_finish_at) < 0) DO
			Tb.Sleep(d.mainarea.t_cycle); Tb.GetTime(d.mainarea.t_crtu);
			b := FALSE; j := 1;
			WHILE (j < d.n_processes) & ~b DO b := d.processes[j].started; INC(j) END;
		END
	END Run; 
	
	PROCEDURE (d: Directory) InitHostModule ( dqos-: Ta.DirectoryQos);
	BEGIN
		ASSERT(oberonRTS.Search(d.default_mod.name) # oberonRTS.nullModule, 
			Tb.NOT_FOUND);
		d.dqos := dqos;
		d.default_mod.ti.ConnectTo(d.dqos);
		d.default_mod.ti.Insert(0, "cycle", Tb.longTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.Insert(1, "expired", Tb.longTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.Insert(2, "with_compensation", Tb.boolTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.Insert(3, "finish_within", Tb.longTyp, Tb.exported, 1, NIL);
		d.default_mod.ti.InitConfig(d.command_line, d.default_mod.name, 
			d.current_dir, d.ind_command_line);
		d.default_mod.initproc;
		Run(d)
	END InitHostModule;		
	
	PROCEDURE (d: Directory) LoadedAs () : INTEGER;
	BEGIN 
		RETURN d.loaded_as
	END LoadedAs;

	(* ========== Initialize and Install ========== *)
	PROCEDURE Initialize;
		VAR i: INTEGER; str, command_line: Tb.Name; cid: ProgramArgs.ChanId;
	BEGIN
		NEW(dir);

		i := 0;
		Tb.AppendAt(".", command_line, i);
		cid := ProgramArgs.ArgChan();
		WHILE ProgramArgs.IsArgPresent() DO
			TextIO.ReadToken(cid, str);
			Tb.AppendAt(" ", command_line, i);
			Tb.AppendAt(str, command_line, i)
		END;
		dir.loaded_as := Ta.AS_CONSOLE_APP;

		Ta.SetDir(dir, "", "", command_line, 1);
	END Initialize;

BEGIN
	Initialize		
END Ta_MTAreas.
