MODULE Ta_MUDPChannel;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, Tl := Ta_Lib, Tb := Ta_Basics, Tc := Ta_Channels, Log ,WinSock;
	
	CONST
		INADDR_ANY = 0;
	
	TYPE
		ChannelQos* = RECORD (Tc.SockChannelQos)
		END;

		Channel = POINTER TO ChannelDesc;
		ChannelDesc = RECORD (Tc.ChannelDesc)
			cqos: ChannelQos;
			hostname: Tb.Name64;
			port: Tb.Int16;
			sd: Tl.SOCKET;
			sa_from: Tl.sockaddr;
			n_send_polls: INTEGER;
		END;
		
		ChannelFactoryDesc* = RECORD (Tc.ChannelFactoryDesc)
		END;		
		ChannelFactory* = POINTER TO ChannelFactoryDesc;


	VAR
		defaultCQos-: ChannelQos;

	(* ========== Channel interface ========== *)
	PROCEDURE (channel: Channel) InitData (is_datahost, is_send_station: BOOLEAN; 
	 name-: ARRAY OF CHAR; VAR cqos: Tc.ChannelQosBase);
		VAR my_errno: Tb.Int32; inaddr: Tl.in_addr; ok, usethis: BOOLEAN; 
		h: Tl.phostent; rc, opt: Tb.Int32; v: Tb.UInt32; sa_from: Tl.sockaddr_in;
		sa_from_bind: Tl.sockaddr; mreq: Tl.ip_mreq; ttl: Tb.Int8;
	BEGIN
		ok := TRUE; 
		channel.sd := 0;
		channel.port := cqos.port;
		Tb.ParseHostPort(name, cqos.hostname, channel.hostname, channel.port);
		Tb.Copy(cqos.group_addr, channel.hostname);
		IF cqos IS ChannelQos THEN
			channel.cqos := cqos(ChannelQos)
		ELSE
			channel.cqos := defaultCQos;
			Tc.CopyChannelQos(cqos, channel.cqos);
		END;
		(* whether this channel is a sender, receiver or uses another one to send *)
		usethis := FALSE;
		IF is_datahost THEN
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_SEND_INDIRECT,
				Tc.TRANS_RECV, Tc.TRANS_RECV_DIRECT};
			usethis := TRUE;
		ELSIF is_send_station THEN
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_SEND_DIRECT,
				Tc.TRANS_RECV, Tc.TRANS_RECV_INDIRECT};
			usethis := TRUE;
		ELSE
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_SEND_INDIRECT,
				Tc.TRANS_RECV, Tc.TRANS_RECV_INDIRECT};
		END;
		
		(* parse and determine proper addrinfo *)
		inaddr.s_addr := WinSock.inet_addr(channel.hostname);
		IF inaddr.s_addr = WinSock.INADDR_NONE THEN
			rc := 0
		ELSE
			rc := 1
		END;
		IF rc = 0 THEN
			ok := FALSE
		ELSE
			sa_from.sin_addr.s_addr := inaddr.s_addr;
			sa_from.sin_family := Tl.AF_INET;
			sa_from.sin_port := WinSock.htons(channel.port);
			SYSTEM.MOVE(SYSTEM.ADR(sa_from), SYSTEM.ADR(channel.sa_from), 
				SIZE(Tl.sockaddr));
			sa_from.sin_addr.s_addr := WinSock.htons(INADDR_ANY);
			SYSTEM.MOVE(SYSTEM.ADR(sa_from), SYSTEM.ADR(sa_from_bind), 
				SIZE(Tl.sockaddr));
		END;
		
		(* create and set necessary socket and socket options *)
		IF ~ok THEN
			channel.link_stat := Tc.LINK_ERR;
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* undefined addr '"); Log.String(channel.hostname); 
				Log.String("'"); Log.Ln
			END;
		ELSIF usethis THEN
			channel.sd := WinSock.socket(Tl.AF_INET, Tl.SOCK_DGRAM, 0);
			my_errno := WinSock.WSAGetLastError();
			IF channel.sd <= 0 THEN
				ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* socket rc="); Log.Int(channel.sd); 
					Log.String(" err="); Log.Int(my_errno); Log.Ln
				END;		
				channel.sd := 0
			END;
			v := 1;
			opt := Tl.FIONBIO;
			IF ok & (WinSock.ioctlsocket(channel.sd, opt, 
			SYSTEM.VAL(WinSock.p_u_long, SYSTEM.ADR(v))) # 0) THEN
				my_errno := WinSock.WSAGetLastError(); ok := FALSE;
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* NONBLOCK socket err="); Log.Int(my_errno); Log.Ln
				END;		
			END;

			IF ok & ((channel.cqos.so_reuseaddr > 0) 
			OR (channel.cqos.so_reuseaddr = 0) & (channel.port # 0) & is_datahost) THEN
				IF WinSock.setsockopt(channel.sd, WinSock.SOL_SOCKET, WinSock.SO_REUSEADDR,
				SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(v)), 4) < 0 THEN
					my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* setsockopt REUSEADDR err="); Log.Int(my_errno); Log.Ln
					END;		
				END;
			END;
			IF ok & channel.cqos.so_keepalive THEN
				IF WinSock.setsockopt(channel.sd, WinSock.SOL_SOCKET, WinSock.SO_KEEPALIVE,
				SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(v)), 4) < 0 THEN
					my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* setsockopt KEEPALIVE err="); Log.Int(my_errno); Log.Ln
					END;		
				END;
			END;
			IF ok & (channel.port # 0) & is_datahost THEN
				IF WinSock.bind(channel.sd, sa_from_bind, SIZE(Tl.sockaddr_in)) < 0 THEN
					my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* bind err="); Log.Int(my_errno); Log.Ln
					END;		
				END;
			END; 
			IF ok THEN
				Tb.Clean(SYSTEM.ADR(mreq), SIZE(Tl.ip_mreq));
				mreq.imr_multiaddr := inaddr;
				IF WinSock.setsockopt(channel.sd, Tl.IPPROTO_IP, Tl.IP_ADD_MEMBERSHIP,
				SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(mreq)), SIZE(Tl.ip_mreq)) < 0 THEN
					my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* setsockopt ADD_MEMBERSHIP err="); Log.Int(my_errno); Log.Ln
					END;		
				END;
			END;
			IF ok THEN
				ttl := 32;
				IF WinSock.setsockopt(channel.sd, Tl.IPPROTO_IP, Tl.IP_MULTICAST_TTL,
				SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(ttl)), 1) < 0 THEN
					my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* setsockopt MULTICAST_TTL err="); Log.Int(my_errno); Log.Ln
					END;		
				END;
			END;
			IF ok THEN
				v := 1;
				IF WinSock.setsockopt(channel.sd, Tl.IPPROTO_IP, Tl.IP_MULTICAST_LOOP,
				SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(v)), v) < 0 THEN
					my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
					IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
						Log.String("* setsockopt MULTICAST_LOOP err="); Log.Int(my_errno); Log.Ln
					END;		
				END;
			END;
			IF ok THEN
				channel.link_stat := Tc.LINK_OK;
			ELSE
				channel.CloseData(FALSE);
				channel.link_stat := Tc.LINK_ERR;
			END;
			IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
				IF ~ok THEN Log.String("* ") END;
				Log.String("MUDP socket sd='"); Log.Int(channel.sd); Log.String("' host='"); 
				Log.String(channel.hostname); Log.String("' port="); Log.Int(channel.port); 
				IF (channel.port # 0) & is_datahost THEN Log.String(" bind") END; Log.Ln
			END;
		ELSE
			channel.link_stat := Tc.LINK_NONE;
		END;
	END InitData;
	
	PROCEDURE (channel: Channel) GetFds (ind_first: INTEGER; 
	VAR ind_last: INTEGER; VAR fd_array: ARRAY OF Tb.Int32);
	BEGIN
		IF channel.sd = 0 THEN
			ind_last := ind_first
		ELSE
			fd_array[ind_first] := SYSTEM.VAL(Tb.Int32, channel.sd);
			ind_last := ind_first+1
		END
	END GetFds;
	
	PROCEDURE (channel: Channel) CloseData (shutdown: BOOLEAN);
		VAR rc: Tb.Int32;
	BEGIN
		IF channel.sd # 0 THEN
			IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
				Log.String("CloseData sd="); Log.Int(channel.sd); Log.Ln
			END;
			rc := WinSock.closesocket(channel.sd);
			channel.sd := 0
		END;
		channel.link_stat := Tc.LINK_CLOSED;
	END CloseData;


	PROCEDURE (channel: Channel) Send (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32);
		VAR go: BOOLEAN; my_errno: Tb.Int32; s_channel: Channel;
	BEGIN
		ASSERT(indirect_channel IS Channel, Tb.WRONG_TYPE);
		nbytes := 0; s_channel := indirect_channel(Channel);
		IF s_channel.sd <= 0 THEN
			go := FALSE
		ELSIF Tc.POLLOUT IN s_channel.poll_events THEN
			go := TRUE
		ELSIF s_channel.cqos.max_send_polls > 0 THEN
			INC(s_channel.n_send_polls);
			IF s_channel.n_send_polls >= s_channel.cqos.max_send_polls THEN
				s_channel.n_send_polls := 0;
				go := TRUE
			ELSE
				go := FALSE
			END
		ELSE
			s_channel.n_send_polls := 0;
			go := FALSE;
		END;
		IF go THEN
			nbytes := WinSock.sendto(s_channel.sd, SYSTEM.VAL(Tl.pbyte, buffer), buffer_length,
				Tl.MSG_NOSIGNAL, s_channel.sa_from, SIZE(Tl.sockaddr_in));
			my_errno := WinSock.WSAGetLastError();
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
			OR (nbytes < 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
				IF nbytes < 0 THEN Log.String("* sendto err="); Log.Int(my_errno) 
				ELSE Log.String("sendto") END; 
				Log.String(" port="); Log.Int(s_channel.port); Log.String(" sd="); Log.Int(s_channel.sd); 
				Log.String(" rc="); Log.Int(nbytes); Log.String(" len="); Log.Int(buffer_length); 
				Log.String(" sa_len="); Log.Int(SIZE(Tl.sockaddr_in)); Log.Ln
			END;
		END;
	END Send;
	
	PROCEDURE N2H(n: Tb.Int16): Tb.Int16;
		VAR ne, he: Tb.Bits16;
	BEGIN
		ne := SYSTEM.VAL(Tb.Bits16, n);
		he[0] := ne[1]; he[1] := ne[0];
		RETURN SYSTEM.VAL(Tb.Int16, he);
	END N2H;
	
	PROCEDURE (channel: Channel) Recv (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32;
	VAR dest: Tc.Dest);
		VAR my_errno: Tb.Int32; sa_msg: Tl.sockaddr; sa_msg_len: Tb.Int32; r_channel: Channel;
	BEGIN
		ASSERT(indirect_channel IS Channel, Tb.WRONG_TYPE);
		nbytes := 0; r_channel := indirect_channel(Channel);
		IF (r_channel.sd > 0) & (Tc.POLLIN IN r_channel.poll_events) THEN
			sa_msg_len := SIZE(Tl.sockaddr);
			Tb.Clean(SYSTEM.ADR(sa_msg), sa_msg_len);
			nbytes := WinSock.recvfrom(r_channel.sd, SYSTEM.VAL(Tl.pbyte, buffer), buffer_length,
				Tl.MSG_NOSIGNAL, sa_msg, sa_msg_len);
			my_errno := WinSock.WSAGetLastError();
			IF (nbytes < 0) & (my_errno = Tl.EWOULDBLOCK) THEN
				nbytes := 0
			END;
			IF nbytes > 1 THEN
				(* unknown address in general *)
				dest.addr := 1; dest.port := 0
			END;
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
			OR (nbytes < 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
				IF nbytes < 0 THEN Log.String("* recvfrom err="); Log.Int(my_errno) 
				ELSE Log.String("recvfrom") END; 
				Log.String(" port="); Log.Int(channel.port); Log.String(" sd="); Log.Int(r_channel.sd); 
				Log.String(" rc="); Log.Int(nbytes); Log.Ln
			END;
		END;	
	END Recv;	

	PROCEDURE (channel: Channel) MyDest ( dest-: Tc.Dest): BOOLEAN;
	BEGIN 
		RETURN FALSE
	END MyDest;


	(* ========== Initialize ========== *)
	PROCEDURE (cf: ChannelFactory) NewChannel* (aid, sid, host_sid: INTEGER; 
	(*OUT*)VAR chan: Tc.Channel; (*OUT*)VAR res: INTEGER);
		VAR channel: Channel;
	BEGIN
		NEW(channel);
		channel.flags := {Tc.FLAGS_DGRAM, Tc.FLAGS_MCAST};
		channel.trans_flags := {};
		channel.link_stat := Tc.LINK_NONE;
		chan := channel;
		res := 0
	END NewChannel;
	
BEGIN
	Tc.DefaultSockChannelQos(defaultCQos);
	defaultCQos.protocol := "MUDP"
END Ta_MUDPChannel.