MODULE Ta_Poll;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tl := Ta_Lib, Tc := Ta_Channels, SYSTEM, Log ,WinSock;
	
	CONST
		MAX_FD_FOR_ONE = 4;
		
	TYPE
		PollStation = RECORD
			fd_array: ARRAY MAX_FD_FOR_ONE OF Tb.Int32;
			trans_flags: SET;
			r_ind, w_ind, e_ind, r_num, w_num, e_num: Tb.Int32;
		END;		
		Polling = POINTER TO PollingDesc;
		PollingDesc = RECORD (Tc.PollingDesc)
			n_stations: INTEGER;
			stations: POINTER TO ARRAY OF PollStation;
			nfds: Tb.Int32;
			rf, wf, ef: Tl.fd_set;
			events: Tc.PollingEvents;
			fd_changed: BOOLEAN
		END;

		PollingFactoryDesc* = RECORD (Tc.PollingFactoryDesc)
		END;		
		PollingFactory* = POINTER TO PollingFactoryDesc;
	

	(* ========== Polling interface ========== *)
	PROCEDURE (polling: Polling) FdSetChannel (index: INTEGER; channel: Tc.Channel);
		VAR i: INTEGER;
	BEGIN
		ASSERT((index >= 0) & (index < polling.n_stations), Tb.CHECK_RANGE_ERROR);
		ASSERT(channel # NIL, Tb.NIL_POINTER);
		channel.GetFds(0, i, polling.stations[index].fd_array);
		polling.stations[index].e_num := i;
		polling.stations[index].trans_flags := channel.trans_flags;
		polling.fd_changed := TRUE
	END FdSetChannel;

	PROCEDURE Prepare (polling: Polling);
		VAR j, k: INTEGER;
	BEGIN
		polling.nfds := 0;
		polling.rf.fd_count := 0; polling.wf.fd_count := 0; polling.ef.fd_count := 0; 
		WinSock.FD_ZERO(polling.rf); WinSock.FD_ZERO(polling.wf); WinSock.FD_ZERO(polling.ef);

		FOR j := 0 TO polling.n_stations-1 DO
			polling.stations[j].e_ind := polling.ef.fd_count;
			FOR k := 0 TO polling.stations[j].e_num-1 DO 
				IF polling.nfds <= polling.stations[j].fd_array[k] THEN 
					polling.nfds := polling.stations[j].fd_array[k]+1
				END;
				WinSock.FD_SET(polling.stations[j].fd_array[k], polling.ef)
			END;
			(* INC(polling.ef.fd_count, i); append ??? *)
			polling.stations[j].r_ind := polling.rf.fd_count; 
			IF (Tc.TRANS_RECV_DIRECT IN polling.stations[j].trans_flags)
			OR (Tc.TRANS_LISTEN IN polling.stations[j].trans_flags) THEN
				polling.stations[j].r_num := polling.stations[j].e_num;
				FOR k := 0 TO polling.stations[j].e_num-1 DO
					WinSock.FD_SET(polling.stations[j].fd_array[k], polling.rf)
				END
				(* INC(polling.rf.fd_count, i) append ??? *)
			ELSE
				polling.stations[j].r_num := 0
			END;
			polling.stations[j].w_ind := polling.wf.fd_count; 
			IF Tc.TRANS_SEND_DIRECT IN polling.stations[j].trans_flags THEN
				polling.stations[j].w_num := polling.stations[j].e_num;
				FOR k := 0 TO polling.stations[j].e_num-1 DO
					WinSock.FD_SET(polling.stations[j].fd_array[k], polling.wf)
				END
				(* INC(polling.wf.fd_count, i) *)
			ELSE
				polling.stations[j].w_num := 0
			END
		END
	END Prepare;
	
	PROCEDURE (polling: Polling) Poll ((*OUT*)VAR events: Tc.PollingEvents; VAR res: INTEGER);
		VAR j, k: INTEGER; tv: Tl.timeval; rc, my_errno: Tb.Int32;
		rff, wff, eff: Tl.fd_flags; rf, wf, ef: Tl.fd_set;
	BEGIN
		IF polling.fd_changed THEN
			Prepare(polling);
			polling.fd_changed := FALSE
		END;
		IF polling.nfds > 0 THEN
			tv.tv_sec := 0; tv.tv_usec := 0;
			rf := polling.rf; wf := polling.wf; ef := polling.ef;
			SYSTEM.MOVE(SYSTEM.ADR(rf), SYSTEM.ADR(rff), SIZE(Tl.fd_flags));
			SYSTEM.MOVE(SYSTEM.ADR(wf), SYSTEM.ADR(wff), SIZE(Tl.fd_flags));
			SYSTEM.MOVE(SYSTEM.ADR(ef), SYSTEM.ADR(eff), SIZE(Tl.fd_flags));
			rc := WinSock.select(polling.ef.fd_count, rff.ws_flags, wff.ws_flags, eff.ws_flags, tv);
			SYSTEM.MOVE(SYSTEM.ADR(rff), SYSTEM.ADR(rf), SIZE(Tl.fd_flags));
			SYSTEM.MOVE(SYSTEM.ADR(wff), SYSTEM.ADR(wf), SIZE(Tl.fd_flags));
			SYSTEM.MOVE(SYSTEM.ADR(eff), SYSTEM.ADR(ef), SIZE(Tl.fd_flags));
			my_errno := WinSock.WSAGetLastError();
			IF Tb.TRAL_POLLING IN Tb.trace THEN
				Log.String("poll rc="); Log.Int(rc); Log.String(" n_stations="); Log.Int(polling.n_stations); 
				Log.String(" nfds="); Log.Int(polling.nfds); Log.String(" err="); Log.Int(my_errno); Log.Ln
			END;
			IF rc > 0 THEN
				FOR j := 0 TO polling.n_stations-1 DO
					polling.events[j] := {};
					IF polling.stations[j].e_num > 0 THEN
						FOR k := 0 TO polling.stations[j].e_num-1 DO
							IF WinSock.__WSAFDIsSet(ef.fd_array[polling.stations[j].e_ind+k], 
							ef) > 0 THEN
								INCL(polling.events[j], Tc.POLLERR+k*Tc.POLL_NEXT)
							END
						END;
						IF polling.stations[j].r_ind >= 0 THEN
							FOR k := 0 TO polling.stations[j].r_num-1 DO
								IF WinSock.__WSAFDIsSet(rf.fd_array[polling.stations[j].r_ind+k], 
								rf) > 0 THEN
									INCL(polling.events[j], Tc.POLLIN+k*Tc.POLL_NEXT)
								END
							END
						END;
						IF polling.stations[j].w_ind >= 0 THEN
							FOR k := 0 TO polling.stations[j].w_num-1 DO
								IF WinSock.__WSAFDIsSet(wf.fd_array[polling.stations[j].w_ind+k], 
								wf) > 0 THEN
									INCL(polling.events[j], Tc.POLLOUT+k*Tc.POLL_NEXT)
								END
							END
						END;
						IF Tb.TRAL_POLLING IN Tb.trace THEN
							Log.String("poll sid="); Log.Int(j);
							Log.String(" poll_events=");
							Log.Int(Tb.Set_Int32(polling.events[j])); Log.Ln;
						END
					END
				END
			END;
			res := SHORT(rc)
		ELSE
			res := -3
		END;
		events := polling.events
	END Poll;


	PROCEDURE (pf: PollingFactory) NewPolling* (n_stations: INTEGER; (*OUT*)VAR poll: Tc.Polling; 
	(*OUT*)VAR res: INTEGER);
		VAR polling: Polling;
	BEGIN
		IF (n_stations <= 0) OR (n_stations >= Tl.MAX_FD_SET) THEN
			res := Tb.CHECK_RANGE_ERROR
		ELSE
			NEW(polling);
			polling.n_stations := n_stations;
			NEW(polling.stations, n_stations);
			NEW(polling.events, n_stations);
			polling.nfds := 0;
			poll := polling;
			res := 0
		END;
	END NewPolling;
	
END Ta_Poll.