MODULE Ta_SHMLib;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, Tl := Ta_Lib, Tb := Ta_Basics, Log ,WinSock, Files, Tf := Ta_Files;
	IMPORT WinBase, WinDef, WinNT;

	CONST
		SHMMIN = 128; O777 = 511; O444 = 292; IPC_CREAT = 512; IPC_RMID = 0; IPC_STAT = 2;

	TYPE
		ShmData* = RECORD
			file_id-, shm_id-: Tb.Int32;
			shm_buffer-: Tb.Address
		END;

	VAR
		zerodata-: ShmData;
		
	(* ========== SharedMemoryRoutines ========== *)
	PROCEDURE GetFileSize* ( filename-: ARRAY OF CHAR; VAR flen: Tb.Int32);
		VAR rc: Tb.Int32; st: Tl.file_attribute_data; 
	BEGIN
		(*rc := Tl.GetFileAttributesExA(filename, 
			Tl.GET_FILE_EX_INFO_STANDARD, st);*)
		rc := SYSTEM.VAL(Tb.Int32, WinBase.GetFileAttributesExA(filename, 
			WinBase.GetFileExInfoStandard, SYSTEM.VAL(WinDef.PVOID, 
				SYSTEM.ADR(st))));
		IF rc = 0 THEN flen := -1 ELSE flen := st.nFileSizeLow END;
	END GetFileSize;
	
	PROCEDURE CreateFileOfSize* ( filename-: ARRAY OF CHAR; 
	flen: Tb.Int32; VAR ok: BOOLEAN);
		VAR file: Files.File; wr: Files.Writer; byte: Tb.Int8;
	BEGIN
		Tf.NewFile(filename, file);
		IF file = NIL THEN
			ok := FALSE;
		ELSE
			wr := file.NewWriter(NIL);
			byte := 0;
			WHILE flen > 0 DO
				wr.WriteByte(byte); DEC(flen)
			END;
			file.Close;
			ok := TRUE
		END
	END CreateFileOfSize;

	PROCEDURE CreateShm* ( filename-: ARRAY OF CHAR; memsize: Tb.Int32;
	VAR shd: ShmData);
		VAR ofs: WinBase.OFSTRUCT; 
	BEGIN
		shd.file_id := WinBase.OpenFile(filename, ofs, WinBase.OF_READWRITE);
		IF shd.file_id > 0 THEN
			shd.shm_id := SYSTEM.VAL(Tb.Int32, WinBase.CreateFileMappingA 
				(SYSTEM.VAL(WinDef.HANDLE, shd.file_id), NIL, 
				WinNT.PAGE_READWRITE, 0, memsize, filename));
			IF shd.shm_id >= 0 THEN
				shd.shm_buffer := SYSTEM.VAL(Tb.Address, WinBase.MapViewOfFile
					(SYSTEM.VAL(WinDef.HANDLE, shd.shm_id), 
						WinNT.SECTION_MAP_WRITE, 0, 0, 0))
			ELSE
				shd.shm_buffer := 0
			END
		ELSE
			shd.shm_id := 0;
			shd.shm_buffer := 0
		END;
	END CreateShm;
	
	PROCEDURE ConnectShm* ( filename-: ARRAY OF CHAR; memsize: Tb.Int32;
	readonly: BOOLEAN; VAR shd: ShmData);
	BEGIN
		shd.shm_id := SYSTEM.VAL(Tb.Int32, WinBase.OpenFileMappingA
			(WinNT.SECTION_ALL_ACCESS, FALSE, filename));
		IF shd.shm_id > 0 THEN
			IF readonly THEN
				shd.shm_buffer := SYSTEM.VAL(Tb.Address, WinBase.MapViewOfFile
					(SYSTEM.VAL(WinDef.HANDLE, shd.shm_id), 
						WinNT.SECTION_MAP_READ, 0, 0, 0))
			ELSE
				shd.shm_buffer := SYSTEM.VAL(Tb.Address, WinBase.MapViewOfFile
					(SYSTEM.VAL(WinDef.HANDLE, shd.shm_id), 
						WinNT.SECTION_MAP_WRITE, 0, 0, 0))
			END
		ELSE
			shd.shm_buffer := 0
		END;
	END ConnectShm;

	PROCEDURE DisconnectShmBuffer* (VAR shd: ShmData);
		VAR rc: Tb.Int32;
	BEGIN
		IF shd.shm_buffer # 0 THEN
			rc := SYSTEM.VAL(Tb.Int32, WinBase.UnmapViewOfFile 
				(SYSTEM.VAL(WinDef.LPCVOID, shd.shm_buffer)));
			shd.shm_buffer := 0;
		END;
	END DisconnectShmBuffer;
	
	PROCEDURE DeleteShm* (VAR shd: ShmData; delete: BOOLEAN);
		VAR rc: Tb.Int32;
	BEGIN
		(*IF shd.file_id > 0 THEN
			rc := Tl.CloseHandle(shd.file_id);
			shd.file_id := 0
		END;*)
		shd.file_id := 0;
		IF shd.shm_id > 0 THEN
			rc := SYSTEM.VAL(Tb.Int32, WinBase.CloseHandle
				(SYSTEM.VAL(WinDef.HANDLE, shd.shm_id)));
			shd.shm_id := 0
		END
	END DeleteShm;
	
END Ta_SHMLib.
