MODULE Ta_Stations;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT Tb := Ta_Basics, Tc := Ta_Channels, Tl := Ta_Lib, Ti := Ta_TInfo, SYSTEM, Log ,WinSock;
	IMPORT oberonRTS;
	
	
	CONST 
		MIN_RECV_BUFFER_LENGTH* = 100; MAX_COUNT = 10000;
		BUFS_ARRAY_SIZE = 100; BUFS_ARRAY_INCREMENT = 50;
		UNABLE_TO_SEND* = -7; UNABLE_TO_RECV* = -8; CANT_READ_HDR* = -9;
			WRONG_HDR_DATA* = -10; CANT_WRITE_HDR* = -11;

		PING_TELEGRAM_TYPE* = 16; USER_TELEGRAM_TYPE* = 64;

		MSG_ACCEPT_ANY* = 0; MSG_MATCH_DST* = 1; MSG_MATCH_SID* = 2;
		DONT_REALLOC* = 0; REALLOC_LARGEST* = 1; REALLOC_EXACT* = 2;
		
		toALL* = -1; toNONE* = -2; sendSID* = -3; hostSID* = -4; callerSID* = -5; usedSID = -6;
		


	TYPE
		TableQosBase* =  RECORD (Tb.Msg)
			tp_timeout*, tp_send_timeout*: LONGINT;
			realloc_mode*: INTEGER;
		END;
		TableQos* = RECORD (TableQosBase)
		END;

		Carrier* = POINTER TO CarrierDesc;
		Station* = POINTER TO StationDesc;
		StdStation = POINTER TO StdStationDesc;
		Stations* = POINTER TO ARRAY OF Station;
		
		Header* = RECORD
			mid*, mid_resp*, src_sid*, dst_sid*, dst_oid*, iface_code*: Tb.Int16;
			msg_type*: Tb.Int32;
			buffer*: Tb.Address;
			buffer_length*: Tb.Int32;
			basic_length*: Tb.Int32;
			t_set*, t_got*: Tb.Time;
			aid*, tlg_type*: INTEGER;
			sendto_sids*: SET;
		END;

		Buffer = RECORD
			offset, length, pos: Tb.Int32;
			sid: INTEGER;
			rcomplete: BOOLEAN;
			t_recv: Tb.Time;
			dest: Tc.Dest;
			hdr: Header
		END;
		
		CarrierDesc* =  RECORD
			poll_res*: INTEGER;
			prot_format-: Tb.Name16;
			host_sid-: INTEGER;
			host_station, send_station: StdStation;
			polling: Tc.Polling;
			t_poll: Tb.Time;
			send_buffer_size, recv_buffer_size, send_offset, recv_offset: Tb.Int32;
			send_buffer, recv_buffer: POINTER TO ARRAY OF Tb.Int8;
			send_addr, recv_addr: Tb.Address;
			n_recv_bufs, n_send_bufs, bufs_array_size: INTEGER;
			events: Tc.PollingEvents;
			bufs: POINTER TO ARRAY OF Buffer;
			fd_ind: INTEGER;
		END;
		CarrierFactory* = POINTER TO CarrierFactoryDesc;
		CarrierFactoryDesc* =  RECORD
		END;		

		ChannelStatus* = RECORD
			flags-, trans_flags-, poll_events-: SET;
			link_stat-: INTEGER;
		END;
		
		Reader* = RECORD
			eof-: BOOLEAN;
			complete-: BOOLEAN;
			res-: Tb.Int32;
			t_recv-: Tb.Time;
			hdr-: Header;
			cr-: Carrier;
			pos-, length-, offset: Tb.Int32;
		END;
		
		Writer* = RECORD
			eof-: BOOLEAN;
			complete-: BOOLEAN;
			res-: Tb.Int32;
			t_send-: Tb.Time;
			cr-: Carrier;
			pos-, length-: Tb.Int32;
			hdr_written: BOOLEAN
		END;
		
		StationDesc* =  RECORD
			rd*: Reader;
			wr*: Writer;
			chs-: ChannelStatus;
			is_online-: BOOLEAN;
			aid-, sid-: INTEGER;
			name-: Tb.Name64;
			cqos-: Tc.ChannelQos;
		END;
		StdStationDesc = RECORD (StationDesc)
			channel: Tc.Channel;
			lib_cid, fd_ind: INTEGER;
			recv_rc: Tb.Int32;
			ping_seq: Tb.Int16;
			tp_discovery: Tb.Time
		END;

		Mapper* = POINTER TO MapperDesc;
		MapperDesc* =  RECORD
			cancelled*: BOOLEAN;
			eof*: BOOLEAN;
		END;
		MapperFactory* = POINTER TO MapperFactoryDesc;
		MapperFactoryDesc* =  RECORD
		END;		

		
	VAR
		defaultTQos-: TableQos;
		zero_hdr: Header;



	(* ========== Carrier Interface ========== *)
	PROCEDURE^ (cr: Carrier) SetNewCarrier* (polling: Tc.Polling; send_buffer_size, 
	recv_buffer_size: Tb.Int32; prot_format: ARRAY OF CHAR);
	PROCEDURE^ (cr: Carrier) Poll* ();
	PROCEDURE^ (cr: Carrier) NewStation* (old: Station; cf: Tc.ChannelFactory; 
	aid, sid: INTEGER;  name-: ARRAY OF CHAR;
	VAR cqos: Tc.ChannelQosBase; is_datahost, is_send_station: BOOLEAN) 
	: Station;
	PROCEDURE^ (cr: Carrier) GetRecv* (ind: INTEGER; VAR rd: Reader; 
	VAR sid: INTEGER);
	PROCEDURE (cr: Carrier) InternalizeHeader* (VAR rd: Reader; VAR hdr: Header;
	VAR res: Tb.Int32);
	BEGIN END InternalizeHeader;
	PROCEDURE (cr: Carrier) ExternalizeHeader* (VAR wr: Writer; hdr-: Header;
	VAR res: Tb.Int32);
	BEGIN END ExternalizeHeader;
	PROCEDURE (cr: Carrier) LittleEndianNetwork* (): BOOLEAN;
	BEGIN RETURN FALSE END LittleEndianNetwork;
	(* ========== Carrier Factory ========== *)
	PROCEDURE (crf: CarrierFactory) NewCarrier* (pf: Tc.PollingFactory; 
	aid, n_stations: INTEGER; send_buffer_size, recv_buffer_size: Tb.Int32;  
	prot_format: ARRAY OF CHAR; (*OUT*)VAR cr: Carrier; (*OUT*)VAR res: INTEGER);
	BEGIN END NewCarrier;

	(* ========== Station Interface ========== *)
	PROCEDURE (station: Station) Close* ();
	BEGIN END Close;
	PROCEDURE (station: Station) CheckPoll* ();
	BEGIN END CheckPoll;
	PROCEDURE (station: Station) Recv* ();
	BEGIN END Recv;
	PROCEDURE (station: Station) Send* (length: Tb.Int32);
	BEGIN END Send;
	PROCEDURE (station: Station) SendPing* (src_sid: INTEGER);
	BEGIN END SendPing;
	PROCEDURE (station: Station) ReadyForSending* (VAR is_ready: BOOLEAN);
	BEGIN END ReadyForSending;
	PROCEDURE (station: Station) HandleMsg* (VAR msg: Tb.Msg); 
	BEGIN END HandleMsg;
	(* Reader Interface *)
	PROCEDURE^ (VAR rd: Reader) SetPos* (pos: Tb.Int32) ;
	PROCEDURE^ (VAR rd: Reader) ReadBytes* (VAR x: ARRAY OF Tb.Int8;
	beg, length: Tb.Int32) ;
	PROCEDURE^ (VAR rd: Reader) Read8* (VAR x: Tb.Bits8) ;
	PROCEDURE^ (VAR rd: Reader) Read16* (VAR x: Tb.Bits16) ;
	PROCEDURE^ (VAR rd: Reader) Read32* (VAR x: Tb.Bits32) ;
	PROCEDURE^ (VAR rd: Reader) Read64* (VAR x: Tb.Bits64) ;
	(* Writer Interface *)
	PROCEDURE^ (VAR wr: Writer) SetPos* (pos: Tb.Int32) ;
	PROCEDURE^ (VAR wr: Writer) WriteHeader* ( hdr-: Header; (*OUT*)VAR res: Tb.Int32) ;
	PROCEDURE^ (VAR wr: Writer) WriteBytes* ( x-: ARRAY OF Tb.Int8;
	beg, length: Tb.Int32) ;
	PROCEDURE^ (VAR wr: Writer) Write8* (x: Tb.Bits8) ;
	PROCEDURE^ (VAR wr: Writer) Write16* (x: Tb.Bits16) ;
	PROCEDURE^ (VAR wr: Writer) Write32* (x: Tb.Bits32) ;
	PROCEDURE^ (VAR wr: Writer) Write64* (x: Tb.Bits64) ; 

	(* ========== Mapper interface ========== *)
	PROCEDURE (mapper: Mapper) HandleMsg* (VAR msg: Tb.Msg); 
	BEGIN END HandleMsg;
	PROCEDURE (mapper: Mapper) ExploreObject* ();
	BEGIN END ExploreObject;
	PROCEDURE (mapper: Mapper) MatchHeader* ( hdr-: Header)
	: BOOLEAN;
	BEGIN RETURN FALSE END MatchHeader;
	PROCEDURE (mapper: Mapper) ReadObject* (VAR item: Ti.Item; VAR rd: Reader;
	VAR changed: BOOLEAN);
	BEGIN END ReadObject;
	PROCEDURE (mapper: Mapper) WriteHeader* (VAR hdr: Header;  wr-: Writer);
	BEGIN END WriteHeader;
	PROCEDURE (mapper: Mapper) WriteObject* ( item-: Ti.Item;
	VAR wr: Writer);
	BEGIN END WriteObject;
	(* ========== Mapper Factory ========== *)
	PROCEDURE (mf: MapperFactory) NewMapper* (ti: Ti.TInfo;
	 iface_name-: ARRAY OF CHAR; VAR tqos: TableQosBase;
	sid, oid: INTEGER; (*OUT*)VAR mapr: Mapper; (*OUT*)VAR res: INTEGER);
	BEGIN END NewMapper;


	(* ========== Carrier Implementation ========== *)
	PROCEDURE (cr: Carrier) SetNewCarrier* (polling: Tc.Polling; send_buffer_size, 
	recv_buffer_size: Tb.Int32; prot_format: ARRAY OF CHAR) ;
	BEGIN
		ASSERT(send_buffer_size > 0, Tb.CHECK_RANGE_ERROR);
		ASSERT(recv_buffer_size > 0, Tb.CHECK_RANGE_ERROR);
		ASSERT(polling # NIL, Tb.NIL_POINTER);
		ASSERT(cr.polling = NIL, Tb.ALREADY_INITIALIZED);	
		NEW(cr.send_buffer, send_buffer_size);
		NEW(cr.recv_buffer, recv_buffer_size);
		cr.send_addr := SYSTEM.VAL(Tb.Address, SYSTEM.ADR(cr.send_buffer^));
		cr.recv_addr := SYSTEM.VAL(Tb.Address, SYSTEM.ADR(cr.recv_buffer^));
		cr.send_buffer_size := send_buffer_size;
		cr.recv_buffer_size := recv_buffer_size;
		cr.bufs_array_size := BUFS_ARRAY_SIZE;
		NEW(cr.bufs, cr.bufs_array_size);
		Tb.Copy(prot_format, cr.prot_format);
		cr.polling := polling; cr.host_sid := -1
	END SetNewCarrier;
	
	PROCEDURE BufsArrayResize (cr: Carrier);
		VAR bufs: POINTER TO ARRAY OF Buffer; 
			j, bufs_array_size: INTEGER;
	BEGIN
		bufs_array_size := cr.bufs_array_size+BUFS_ARRAY_INCREMENT;
		NEW(bufs, bufs_array_size);
		FOR j := 0 TO cr.bufs_array_size-1 DO
			bufs[j] := cr.bufs[j]
		END;
		NEW(cr.bufs, bufs_array_size);
		FOR j := 0 TO cr.bufs_array_size-1 DO
			cr.bufs[j] := bufs[j]
		END;
		cr.bufs_array_size := bufs_array_size;
	END BufsArrayResize;
	
	PROCEDURE (cr: Carrier) Poll* () ;
	BEGIN
		cr.n_recv_bufs := 0; cr.recv_offset := 0; cr.n_send_bufs := 0;
		cr.send_offset := 0;
		cr.polling.Poll(cr.events, cr.poll_res);
		Tb.GetTime(cr.t_poll);
	END Poll;

	PROCEDURE FinalizeChannel(addr: SYSTEM.ADDRESS);
		VAR channel: Tc.Channel;
	BEGIN
		channel := SYSTEM.VAL(Tc.Channel, addr);
		channel.CloseData(FALSE)
	END FinalizeChannel;

	PROCEDURE (cr: Carrier) NewStation* (old: Station; cf: Tc.ChannelFactory; 
	aid, sid: INTEGER;  name-: ARRAY OF CHAR;
	VAR cqos: Tc.ChannelQosBase; is_datahost, is_send_station: BOOLEAN)
	: Station ;
		VAR station: StdStation; res: INTEGER;
	BEGIN
		ASSERT(cqos.prot_format = cr.prot_format, Tb.WRONG_TYPE);
		IF old = NIL THEN
			NEW(station)
		ELSE
			station := old(StdStation);
			IF station.cqos.prot_format # cqos.prot_format THEN
				station.channel := NIL
			END
		END;
		IF station.channel = NIL THEN
			ASSERT(cf # NIL, Tb.NIL_POINTER);
			IF is_datahost THEN cr.host_sid := sid END;
			cf.NewChannel(aid, sid, cr.host_sid, station.channel, res);
			oberonRTS.InstallFinalizer(FinalizeChannel, station.channel)
		ELSE
			res := 0
		END;
		IF res = 0 THEN
			station.rd.cr := cr; station.wr.cr := cr; station.aid := aid; station.sid := sid; 
			station.lib_cid := -1; 
			Tb.Copy(name, station.name); Tc.CopyChannelQos(cqos, station.cqos);
			IF is_datahost THEN cr.host_station := station; cr.host_sid := station.sid END;
			IF is_send_station THEN cr.send_station := station END;
			station.channel.CloseData(FALSE);
(*Log.String("aaaa");Log.Ln;*)
			station.channel.InitData(is_datahost, is_send_station, name, cqos);
(*Log.String("bbbb");Log.Ln;*)
			station.chs.flags := station.channel.flags;
			station.chs.trans_flags := station.channel.trans_flags;
			station.fd_ind := cr.fd_ind; INC(cr.fd_ind);
			IF station.channel.link_stat >= Tc.LINK_INIT THEN
				cr.polling.FdSetChannel(station.fd_ind, station.channel)
			END;
			IF cqos.tp_discovery_timeout < 0 THEN
				station.tp_discovery := Tb.INFINITE_TIME
			ELSE
				Tb.Long_Time(cqos.tp_discovery_timeout, station.tp_discovery)
			END
		ELSE
			station := NIL
		END;
		IF (station = NIL) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
			Log.String("* station sid="); Log.Int(sid); Log.String(" protocol="); 
			Log.String(cqos.protocol); Log.String(" name="); 
			Log.String(name); Log.String(" hostname="); Log.String(cqos.hostname);
			Log.String(" port="); Log.Int(cqos.port); Log.Ln
		END;
		RETURN station
	END NewStation;

	PROCEDURE (cr: Carrier) GetRecv* (ind: INTEGER; VAR rd: Reader; 
	VAR sid: INTEGER) ;
	BEGIN
		rd.cr := cr;
		IF (ind >= cr.n_recv_bufs) OR (cr.bufs[ind].sid = usedSID) THEN
			rd.length := 0; rd.res := 0; rd.offset := 0; rd.pos := 0;
			rd.eof := TRUE; sid := toNONE
		ELSE
			rd.offset := cr.bufs[ind].offset; rd.length := cr.bufs[ind].length;
			rd.complete := cr.bufs[ind].rcomplete; rd.eof := FALSE; 
			rd.res := cr.bufs[ind].length; rd.pos := cr.bufs[ind].pos;
			rd.hdr := cr.bufs[ind].hdr; sid := cr.bufs[ind].sid
		END
	END GetRecv;
	
	
	(* ========== Station Implementation ========== *)
	PROCEDURE (station: StdStation) Close ();
	BEGIN
		station.channel.CloseData(FALSE);
	END Close;

	PROCEDURE (station: StdStation) CheckPoll ();
		VAR fd_changed: BOOLEAN; t, t1: Tb.Time;
	BEGIN
		IF (station.rd.cr.poll_res) <= 0 THEN
			station.channel.poll_events := {}
		ELSE
			station.channel.poll_events := station.rd.cr.events[station.fd_ind]
		END;
		station.chs.poll_events := station.channel.poll_events;
		IF Tc.TRANS_CONNECTION IN station.chs.trans_flags THEN
			ASSERT(station.rd.cr.host_station # NIL, Tb.NIL_POINTER);
			fd_changed := FALSE;
			station.channel.Connect(station.rd.cr.host_station.channel, 
				station.cqos.connect_on_request, fd_changed);
			IF fd_changed THEN
				station.rd.cr.polling.FdSetChannel(station.fd_ind, station.channel)
			END
		END;
		IF Tb.TimePositive(station.tp_discovery) THEN
			Tb.GetTime(t); Tb.SubtractTime(t, station.tp_discovery, t1);
			station.is_online := (Tb.CompareTime(t1, station.rd.t_recv) <= 0)
		ELSE
			station.is_online := TRUE
		END;
		station.chs.link_stat := station.channel.link_stat; station.recv_rc := 1;
		station.rd.pos := 0; station.wr.pos := 0; station.rd.length := 0; 
		station.wr.length := 0; station.rd.offset := 0;
		station.wr.eof := FALSE
	END CheckPoll;
	
	PROCEDURE RecvMessages (station: StdStation; channel: Tc.Channel; VAR bfr: Buffer;
	VAR count: INTEGER; VAR rc: Tb.Int32);
		VAR res, nlen: Tb.Int32; is_ok, inc_offset: BOOLEAN;
	BEGIN
		rc := 0; count := 0;
		IF Tc.TRANS_RECV_DIRECT IN channel.trans_flags THEN
			IF channel.link_stat = Tc.LINK_OK THEN
				REPEAT
					station.channel.Recv(channel, station.rd.cr.recv_addr+station.rd.cr.recv_offset,
						station.rd.cr.recv_buffer_size-station.rd.cr.recv_offset, rc, bfr.dest);
					IF rc > 1 THEN
						Tb.GetTime(station.rd.t_recv);
						nlen := 0; inc_offset := FALSE;
						REPEAT
							bfr.hdr.src_sid := -1; bfr.hdr.buffer_length := rc-nlen;
							bfr.offset := station.rd.cr.recv_offset + nlen;
							station.rd.offset := bfr.offset; 
							station.rd.pos := 0; station.rd.length := bfr.hdr.buffer_length;
							station.rd.cr.InternalizeHeader(station.rd, bfr.hdr, res);
							bfr.pos := station.rd.pos;
							bfr.length := bfr.hdr.buffer_length;
							bfr.rcomplete := (bfr.length < station.rd.cr.recv_buffer_size
								- bfr.offset);
							bfr.t_recv := station.rd.t_recv;
							bfr.sid := toALL;
							IF station.channel.MyDest(bfr.dest) THEN
								bfr.sid := station.sid; bfr.hdr.src_sid := Tb.Int_Int16(station.sid)
							ELSIF station.cqos.trust_src_sid THEN
								bfr.sid := bfr.hdr.src_sid
							END;
							is_ok := (res = 0) & (bfr.hdr.aid = station.aid);
							IF (bfr.length > 0) & (bfr.length <= rc) THEN
								INC(nlen, bfr.length)
							ELSE
								is_ok := FALSE;
								INC(nlen, rc)
							END;
							IF is_ok THEN
								IF bfr.hdr.tlg_type # USER_TELEGRAM_TYPE THEN
									(* pinger and other telegrams *)
								ELSIF bfr.sid = station.rd.cr.host_station.sid THEN
									(* avoid self-telegrams *)
								ELSE
									inc_offset := TRUE;
									IF bfr.sid = station.sid THEN
										INC(count)
									END;
									station.rd.cr.bufs[station.rd.cr.n_recv_bufs] := bfr;
									INC(station.rd.cr.n_recv_bufs);
									IF station.rd.cr.n_recv_bufs >= station.rd.cr.bufs_array_size THEN
										BufsArrayResize(station.rd.cr)
									END;
									IF Tb.TRAL_MESSAGE_DATA IN Tb.trace THEN
										Log.String("add msg sid="); Log.Int(bfr.sid); 
										Log.String(" station.sid="); Log.Int(station.sid); Log.Int(bfr.hdr.mid);
										Log.String("# n_recv_bufs="); Log.Int(station.rd.cr.n_recv_bufs); 
										Log.String(" length,offset="); Log.Int(bfr.length); 
										Log.Int(bfr.offset); Log.Ln
									END
								END
							ELSE
								IF Tb.TRAL_MESSAGE_DATA IN Tb.trace THEN
									Log.String("?skip msg sid="); Log.Int(bfr.sid); 
									Log.String(" station.sid="); Log.Int(station.sid); Log.Int(bfr.hdr.mid);
									Log.String("# res="); Log.Int(res); Log.Int(bfr.hdr.aid);
									Log.String(" length,rc="); Log.Int(bfr.length); 
									Log.Int(rc); Log.Ln
								END
							END
						UNTIL nlen >= rc;
						IF inc_offset THEN
							INC(station.rd.cr.recv_offset, rc)
						ELSE
							bfr.length := 0
						END
					ELSE
						bfr.length := 0
					END;
				UNTIL rc # 1
			END
		END
	END RecvMessages;
	
	PROCEDURE (station: StdStation) Recv ();
		VAR st: StdStation; bfr: Buffer; count: INTEGER;
	BEGIN
		st := NIL;
		IF Tc.TRANS_RECV_INDIRECT IN station.channel.trans_flags THEN
			IF station.rd.cr.host_station # NIL THEN
				IF Tb.CompareTime(station.rd.cr.t_poll, 
				station.rd.cr.host_station.rd.t_recv) > 0 THEN
					st := station.rd.cr.host_station
				END
			END
		ELSIF Tc.TRANS_RECV_VIA_SEND IN station.channel.trans_flags THEN
			IF station.rd.cr.send_station # NIL THEN
				IF Tb.CompareTime(station.rd.cr.t_poll, 
				station.rd.cr.send_station.rd.t_recv) > 0 THEN
					st := station.rd.cr.send_station
				END
			END
		ELSIF Tc.TRANS_RECV_DIRECT IN station.channel.trans_flags THEN
			st := station
		END;
		count := 0; bfr.length := 0;
		IF st # NIL THEN
			RecvMessages(station, st.channel, bfr, count, station.recv_rc);
			station.rd.res := station.recv_rc
		ELSE
			station.rd.res := 0
		END;
		station.rd.length := 0; station.rd.pos := 0;
	END Recv; 

	PROCEDURE (station: StdStation) Send (length: Tb.Int32);
		VAR channel: Tc.Channel; rc: Tb.Int32;
	BEGIN
		station.wr.res := UNABLE_TO_SEND;
		IF Tc.TRANS_SEND_INDIRECT IN station.channel.trans_flags THEN
			IF station.rd.cr.send_station = NIL THEN
				channel := station.channel
			ELSE
				channel := station.rd.cr.send_station.channel
			END
		ELSIF Tc.TRANS_SEND_VIA_HOST IN station.channel.trans_flags THEN
			IF station.rd.cr.host_station = NIL THEN
				channel := station.channel
			ELSE
				channel := station.rd.cr.host_station.channel
			END
		ELSE
			channel := station.channel
		END;
		IF Tc.TRANS_SEND_DIRECT IN channel.trans_flags THEN
			IF channel.link_stat = Tc.LINK_OK THEN
				IF length < 0 THEN length := station.wr.length END;
				station.channel.Send(channel, station.rd.cr.send_addr+station.rd.cr.send_offset, 
					length, rc);
				Tb.GetTime(station.wr.t_send); station.wr.res := rc;
				station.wr.complete := (rc = length);
				station.wr.pos := 0; station.wr.length := 0;
			END
		END
	END Send;

	PROCEDURE (station: StdStation) SendPing (src_sid: INTEGER);
		VAR hdr: Header; res: Tb.Int32;
	BEGIN
		INC(station.ping_seq);
		IF station.ping_seq >= Tc.MAX_MID THEN
			station.ping_seq := 1
		END;
		hdr.mid := station.ping_seq; hdr.src_sid := Tb.Int_Int16(src_sid);
		hdr.dst_sid := Tb.Int_Int16(station.sid); 
		hdr.tlg_type := PING_TELEGRAM_TYPE;
		station.wr.WriteHeader(hdr, res); 
		station.Send(station.wr.length);
	END SendPing;
	
	PROCEDURE (station: StdStation) ReadyForSending (VAR is_ready: BOOLEAN);
		VAR channel: Tc.Channel; fd_changed: BOOLEAN;
	BEGIN
		IF Tc.TRANS_SEND_INDIRECT IN station.channel.trans_flags THEN
			IF station.rd.cr.send_station = NIL THEN
				channel := station.channel
			ELSE
				channel := station.rd.cr.send_station.channel
			END
		ELSE
			channel := station.channel
		END;
		IF channel.link_stat # Tc.LINK_OK THEN
			is_ready := FALSE
		ELSIF Tc.FLAGS_MEMORY IN channel.flags THEN
			is_ready := TRUE
		ELSIF ~(Tc.POLLOUT IN channel.poll_events) THEN
			is_ready := FALSE
		ELSE
			is_ready := TRUE
		END;
		IF station.cqos.connect_on_request
		& (Tc.TRANS_CONNECTION IN station.chs.trans_flags) THEN
			ASSERT(station.rd.cr.host_station # NIL, Tb.NIL_POINTER);
			fd_changed := FALSE;
			station.channel.Connect(station.rd.cr.host_station.channel, FALSE, fd_changed);
			IF fd_changed THEN
				station.rd.cr.polling.FdSetChannel(station.fd_ind, station.channel)
			END
		END
	END ReadyForSending;
	
	PROCEDURE (station: StdStation) HandleMsg (VAR msg: Tb.Msg); 
		VAR fd_changed: BOOLEAN;
	BEGIN
		WITH msg: Tc.ChannelQosBase DO
			station.channel.CloseData(TRUE);
			station.channel.InitData(station.rd.cr.host_station = station,
				station.rd.cr.send_station = station, station.name, msg);
			station.rd.cr.polling.FdSetChannel(station.fd_ind, station.channel);
			IF msg.tp_discovery_timeout < 0 THEN
				station.tp_discovery := Tb.INFINITE_TIME
			ELSE
				Tb.Long_Time(msg.tp_discovery_timeout, station.tp_discovery)
			END
		| msg: Tc.InitMsg DO
			IF Tc.TRANS_CONNECTION IN station.chs.trans_flags THEN
				IF msg.reinit THEN
					station.channel.Connect(station.rd.cr.host_station.channel, TRUE,
						fd_changed)
				END;
				station.channel.Connect(station.rd.cr.host_station.channel, FALSE,
					fd_changed);
				station.rd.cr.polling.FdSetChannel(station.fd_ind, station.channel)				
			END
		| msg: Tc.CloseMsg DO
			IF Tc.TRANS_CONNECTION IN station.chs.trans_flags THEN
				station.channel.Connect(station.rd.cr.host_station.channel, TRUE,
					fd_changed);
				IF fd_changed THEN
					station.rd.cr.polling.FdSetChannel(station.fd_ind, station.channel)
				END
			END
		ELSE
		END;
	END HandleMsg;

	(* ========== Reader Implementation ========== *)
	PROCEDURE (VAR rd: Reader) SetPos* (pos: Tb.Int32) ;
	BEGIN
		ASSERT((pos >= 0) & (pos <= rd.length), Tb.CHECK_RANGE_ERROR);
		rd.pos := pos;
		rd.eof := (pos = rd.length)
	END SetPos;
	
	PROCEDURE (VAR rd: Reader) ReadBytes* (VAR x: ARRAY OF Tb.Int8;
	beg, length: Tb.Int32) ;
		VAR count: Tb.Int32;
	BEGIN
		ASSERT(beg >= 0, Tb.INVALID_VAL);
		IF length > 0 THEN
			count := length;
			IF count + rd.pos >= rd.length THEN
				count := rd.length - rd.pos
			END;
			IF count > 0 THEN
				ASSERT((rd.offset+rd.pos >= 0) & 
				(rd.offset+rd.pos+count <= rd.cr.recv_buffer_size), 
					Tb.INTERNAL_ERROR);
				ASSERT(rd.cr.recv_addr # 0, Tb.NIL_POINTER);
				SYSTEM.MOVE(rd.cr.recv_addr+rd.offset+rd.pos, 
					SYSTEM.ADR(x)+beg, count);
				INC(rd.pos, count);
				rd.eof := FALSE
			ELSE
				rd.eof := TRUE
			END
		ELSE ASSERT(length = 0, Tb.INVALID_VAL)
		END
	END ReadBytes;

	PROCEDURE (VAR rd: Reader) Read8* (VAR x: Tb.Bits8) ;
	BEGIN
		rd.ReadBytes(x, 0, 1)
	END Read8;

	PROCEDURE (VAR rd: Reader) Read16* (VAR x: Tb.Bits16) ;
		VAR ne: Tb.Bits16;	(* netwok byte ordering *)
	BEGIN
		IF Tb.littleEndian = rd.cr.LittleEndianNetwork() THEN
			rd.ReadBytes(x, 0, 2)
		ELSE
			rd.ReadBytes(ne, 0, 2);
			x[0] := ne[1]; x[1] := ne[0];
		END;
	END Read16;

	PROCEDURE (VAR rd: Reader) Read32* (VAR x: Tb.Bits32) ;
		VAR ne: Tb.Bits32;	(* netwok byte ordering *)
	BEGIN
		IF Tb.littleEndian = rd.cr.LittleEndianNetwork() THEN
			rd.ReadBytes(x, 0, 4)
		ELSE
			rd.ReadBytes(ne, 0, 4);
			x[0] := ne[3]; x[1] := ne[2]; x[2] := ne[1]; x[3] := ne[0];
		END;
	END Read32;

	PROCEDURE (VAR rd: Reader) Read64* (VAR x: Tb.Bits64) ;
		VAR ne: Tb.Bits64;	(* netwok byte ordering *)
	BEGIN
		IF Tb.littleEndian = rd.cr.LittleEndianNetwork() THEN
			rd.ReadBytes(x, 0, 8)
		ELSE
			rd.ReadBytes(ne, 0, 8);
			x[0] := ne[7]; x[1] := ne[6]; x[2] := ne[5]; x[3] := ne[4];
			x[4] := ne[3]; x[5] := ne[2]; x[6] := ne[1]; x[7] := ne[0];
		END;
	END Read64;

	(* ========== Writer Implementation ========== *)
	PROCEDURE (VAR wr: Writer) SetPos* (pos: Tb.Int32) ;
	BEGIN
		ASSERT(pos >= 0, Tb.CHECK_RANGE_ERROR);
		IF wr.cr.send_offset + pos > wr.cr.send_buffer_size THEN
			wr.eof := TRUE;
			wr.pos := wr.cr.send_buffer_size - wr.cr.send_offset;
		ELSE
			wr.eof := FALSE;
			wr.pos := pos;
		END;
		IF wr.pos > wr.length THEN
			wr.length := wr.pos
		END
	END SetPos;
	
	PROCEDURE (VAR wr: Writer) WriteHeader* ( hdr-: Header; (*OUT*)VAR res: Tb.Int32) ;
	BEGIN
		wr.pos := 0; 
		wr.cr.ExternalizeHeader(wr, hdr, res)
	END WriteHeader;

	PROCEDURE (VAR wr: Writer) WriteBytes* ( x-: ARRAY OF Tb.Int8;
	beg, length: Tb.Int32) ;
		VAR count: Tb.Int32;
	BEGIN
		ASSERT(beg >= 0, Tb.INVALID_VAL);
		IF length > 0 THEN
			count := length;
			IF count + wr.cr.send_offset + wr.pos 
			>= wr.cr.send_buffer_size THEN
				count := wr.cr.send_buffer_size - wr.cr.send_offset - wr.pos
			END;
			IF count > 0 THEN
				SYSTEM.MOVE(SYSTEM.ADR(x)+beg, 
					wr.cr.send_addr+wr.cr.send_offset+wr.pos, count);
				INC(wr.pos, count);
				IF wr.pos > wr.length THEN
					wr.length := wr.pos
				END;
				wr.eof := FALSE
			ELSE
				wr.eof := TRUE
			END
		ELSE ASSERT(length = 0, Tb.INVALID_VAL)
		END
	END WriteBytes;
	
	PROCEDURE (VAR wr: Writer) Write8* (x: Tb.Bits8) ;
	BEGIN
		wr.WriteBytes(x, 0, 1)
	END Write8;

	PROCEDURE (VAR wr: Writer) Write16* (x: Tb.Bits16) ;
		VAR ne: Tb.Bits16;	(* netwok byte ordering *)
	BEGIN
		IF Tb.littleEndian = wr.cr.LittleEndianNetwork() THEN
			wr.WriteBytes(x, 0, 2)
		ELSE
			ne[0] := x[1]; ne[1] := x[0];
			wr.WriteBytes(ne, 0, 2)
		END;
	END Write16;

	PROCEDURE (VAR wr: Writer) Write32* (x: Tb.Bits32) ;
		VAR ne: Tb.Bits32;	(* netwok byte ordering *)
	BEGIN
		IF Tb.littleEndian = wr.cr.LittleEndianNetwork() THEN
			wr.WriteBytes(x, 0, 4)
		ELSE
			ne[0] := x[3]; ne[1] := x[2]; ne[2] := x[1]; ne[3] := x[0];
			wr.WriteBytes(ne, 0, 4)
		END;
	END Write32;

	PROCEDURE (VAR wr: Writer) Write64* (x: Tb.Bits64) ;
		VAR ne: Tb.Bits64;	(* netwok byte ordering *)
	BEGIN
		IF Tb.littleEndian = wr.cr.LittleEndianNetwork() THEN
			wr.WriteBytes(x, 0, 8)
		ELSE
			ne[0] := x[7]; ne[1] := x[6]; ne[2] := x[5]; ne[3] := x[4];
			ne[4] := x[3]; ne[5] := x[2]; ne[6] := x[1]; ne[7] := x[0];
			wr.WriteBytes(ne, 0, 8)
		END;
	END Write64; 


	PROCEDURE DefaultTableQos* (VAR qos: TableQosBase);
	BEGIN
		qos.tp_timeout := 3000000;
		qos.tp_send_timeout := 2000000;
		qos.realloc_mode := REALLOC_LARGEST
	END DefaultTableQos;

	PROCEDURE CopyTableQos* ( src-: TableQosBase;
	VAR dst: TableQosBase);
	BEGIN
		dst.tp_timeout := src.tp_timeout;
		dst.tp_send_timeout := src.tp_send_timeout;
		dst.realloc_mode := src.realloc_mode
	END CopyTableQos;


	(* ========== WSA Windows Initialization ========== *)		
	PROCEDURE Initialize;
		VAR rc: Tb.Int32; wsadata: Tl.WSAData;
	BEGIN
		rc := WinSock.WSAStartup(Tl.WSA_VER22, wsadata);
		ASSERT(rc = 0, Tb.INIT_FAILED)
	END Initialize;

BEGIN
	Initialize;
	DefaultTableQos(defaultTQos)
END Ta_Stations.
