MODULE Ta_TCPChannel;
(** 
	project	= "Transparent Architecture"
	organization	= "http://sourceforge.net/"
	contributors	= "Dmitry V.Dagaev"
	license = "LGPL"
**)

	IMPORT SYSTEM, Tl := Ta_Lib, Tb := Ta_Basics, Tc := Ta_Channels, Log ,WinSock;
	
	CONST
		LISTEN_BACKLOG = 5;
	
	TYPE
		ChannelQos* = RECORD (Tc.StreamChannelQos)
		END;

		Channel = POINTER TO ChannelDesc;
		ChannelDesc = RECORD (Tc.ChannelDesc)
			cqos: ChannelQos;
			hostname: Tb.Name64;
			port: Tb.Int16;
			sd, cd: Tl.SOCKET;
			sa_from: Tl.sockaddr;
			send_strat, n_send_polls: INTEGER;
			sid_above: BOOLEAN;
			t_recv: Tb.Time;
			disconnect_next: BOOLEAN
		END;
		
		ChannelFactoryDesc* = RECORD (Tc.ChannelFactoryDesc)
		END;		
		ChannelFactory* = POINTER TO ChannelFactoryDesc;


	VAR
		defaultCQos-: ChannelQos;

	(* ========== Channel interface ========== *)
	PROCEDURE InitChannel (channel: Channel; is_datahost: BOOLEAN;
	VAR sd: Tl.SOCKET; VAR ok: BOOLEAN);
		VAR my_errno: Tb.Int32; v: Tb.UInt32; opt: Tb.Int32;
	BEGIN
		sd := WinSock.socket(Tl.AF_INET, Tl.SOCK_STREAM, 0);
		my_errno := WinSock.WSAGetLastError();
		IF sd <= 0 THEN
			ok := FALSE; 
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* socket rc="); Log.Int(sd); 
				Log.String(" err="); Log.Int(my_errno); Log.Ln
			END;		
			sd := 0
		END;
		v := 1;
		opt := Tl.FIONBIO;
		IF ok & (WinSock.ioctlsocket(sd, opt, 
		SYSTEM.VAL(WinSock.p_u_long, SYSTEM.ADR(v))) # 0) THEN
			my_errno := WinSock.WSAGetLastError(); ok := FALSE;
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* NONBLOCK socket err="); Log.Int(my_errno); Log.Ln
			END;		
		END;

		IF ok & channel.cqos.tcp_nodelay THEN
			IF WinSock.setsockopt(sd, Tl.IPPROTO_TCP, Tl.TCP_NODELAY,
			SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(v)), 4) < 0 THEN
				my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* setsockopt TCP_NODELAY err="); Log.Int(my_errno); Log.Ln
				END;		
			END;
		END;
		IF ok & ((channel.cqos.so_reuseaddr > 0) 
		OR (channel.cqos.so_reuseaddr = 0) & (channel.port # 0) & is_datahost) THEN
			IF WinSock.setsockopt(sd, WinSock.SOL_SOCKET, WinSock.SO_REUSEADDR,
			SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(v)), 4) < 0 THEN
				my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* setsockopt REUSEADDR err="); Log.Int(my_errno); Log.Ln
				END;		
			END;
		END;
		IF ok & channel.cqos.so_keepalive THEN
			IF WinSock.setsockopt(sd, WinSock.SOL_SOCKET, WinSock.SO_KEEPALIVE,
			SYSTEM.VAL(Tl.pbyte, SYSTEM.ADR(v)), 4) < 0 THEN
				my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* setsockopt KEEPALIVE err="); Log.Int(my_errno); Log.Ln
				END;		
			END;
		END;
		IF ok & (channel.port # 0) & is_datahost THEN
			IF WinSock.bind(sd, channel.sa_from, SIZE(Tl.sockaddr_in)) < 0 THEN
				my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* bind err="); Log.Int(my_errno); Log.Ln
				END;		
			END;
		END; 
		IF ok & is_datahost THEN
			IF WinSock.listen(sd, LISTEN_BACKLOG) < 0 THEN
				my_errno := WinSock.WSAGetLastError(); ok := FALSE; 
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* listen err="); Log.Int(my_errno); Log.Ln
				END;		
			END;
		END; 
	END InitChannel;
	
	PROCEDURE (channel: Channel) InitData (is_datahost, is_send_station: BOOLEAN; 
	 name-: ARRAY OF CHAR; VAR cqos: Tc.ChannelQosBase);
		VAR inaddr: Tl.in_addr; ok, usecd, usesd: BOOLEAN;
		h: Tl.phostent; rc: Tb.Int32; sa_from: Tl.sockaddr_in;
	BEGIN
		ok := TRUE; 
		channel.sd := 0; channel.cd := 0;
		channel.port := cqos.port;
		Tb.ParseHostPort(name, cqos.hostname, channel.hostname, channel.port);
		IF cqos IS ChannelQos THEN
			channel.cqos := cqos(ChannelQos)
		ELSE
			channel.cqos := defaultCQos;
			Tc.CopyChannelQos(cqos, channel.cqos);
		END;
		IF channel.cqos.hostname[0] = 0X THEN
			Tb.Copy(name, channel.cqos.hostname)
		END;
		(* whether this channel is a sender, receiver or uses another one to send *)
		channel.send_strat := cqos.send_strat;
		usecd := FALSE; usesd := FALSE; channel.trans_flags := {};
		IF is_datahost THEN
			usesd := (channel.send_strat # Tc.SEND_STRAT_CLIENT);
			channel.trans_flags := {Tc.TRANS_LISTEN}
		ELSE
			usecd := (channel.send_strat = Tc.SEND_STRAT_CLIENT)
				OR (channel.send_strat = Tc.SEND_STRAT_MIN_ZVAL) 
				& ~channel.sid_above;
				(*OR (channel.send_strat = Tc.SEND_STRAT_CR_SW);*)
			channel.trans_flags := {Tc.TRANS_SEND, Tc.TRANS_SEND_DIRECT,
				Tc.TRANS_RECV, Tc.TRANS_RECV_DIRECT, Tc.TRANS_CONNECTION}
		END;
		
		IF ~usecd & ~usesd THEN
			channel.link_stat := Tc.LINK_NONE
		ELSE
			(* parse and determine proper addrinfo *)
			inaddr.s_addr := WinSock.inet_addr(channel.hostname);
			IF inaddr.s_addr = WinSock.INADDR_NONE THEN
				rc := 0
			ELSE
				rc := 1
			END;
			IF (rc = 0) & (ORD(channel.hostname[0]) >= ORD('A')) 
			& (ORD(channel.hostname[0]) <= ORD('z')) THEN
				h := WinSock.gethostbyname(channel.hostname);
				IF (h # NIL) & (h.h_length > 0) THEN
					SYSTEM.MOVE(h.h_addr_list^, SYSTEM.ADR(inaddr), SIZE(Tl.in_addr));
					rc := 1
				END;
			END;
			IF rc = 0 THEN
				ok := FALSE;
				channel.link_stat := Tc.LINK_ERR;
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* undefined addr '"); Log.String(channel.hostname); 
					Log.String("'"); Log.Ln
				END
			ELSE
				sa_from.sin_addr.s_addr := inaddr.s_addr;
				sa_from.sin_family := Tl.AF_INET;
				sa_from.sin_port := WinSock.htons(channel.port);
				SYSTEM.MOVE(SYSTEM.ADR(sa_from), SYSTEM.ADR(channel.sa_from), 
					SIZE(Tl.sockaddr));
		
				(* create and set necessary socket and socket options *)
				IF usecd THEN
					InitChannel(channel, is_datahost, channel.cd, ok)
				END;
				IF usesd THEN
					InitChannel(channel, is_datahost, channel.sd, ok)
				END;
		
				IF ~ok THEN
					channel.CloseData(FALSE);
					channel.link_stat := Tc.LINK_ERR
				ELSIF is_datahost THEN
					channel.link_stat := Tc.LINK_LISTEN
				ELSE
					channel.link_stat := Tc.LINK_INIT
				END;
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					IF ~ok THEN Log.String("* ") END; Log.String("TCP socket");
					IF usecd THEN Log.String(" cd="); Log.Int(channel.cd) END; 
					IF usesd THEN Log.String(" sd="); Log.Int(channel.sd) END; 
					Log.String(" host='"); Log.String(channel.hostname); 
					Log.String("' port="); Log.Int(channel.port); 
					IF (channel.port # 0) & is_datahost THEN Log.String(" listen") END; Log.Ln
				END
			END
		END
	END InitData;
	
	PROCEDURE (channel: Channel) GetFds (ind_first: INTEGER; 
	VAR ind_last: INTEGER; VAR fd_array: ARRAY OF Tb.Int32);
	BEGIN
		IF channel.sd = 0 THEN
			ind_last := ind_first
		ELSE
			fd_array[ind_first] := SYSTEM.VAL(Tb.Int32, channel.sd);
			ind_last := ind_first+1
		END;
		IF channel.cd # 0 THEN
			fd_array[ind_last] := SYSTEM.VAL(Tb.Int32, channel.cd);
			INC(ind_last)
		END
	END GetFds;
	
	PROCEDURE CloseSD (channel: Channel; close, shutdown: BOOLEAN);
		VAR rc: Tb.Int32;
	BEGIN
		IF channel.sd # 0 THEN
			IF shutdown THEN
				rc := WinSock.shutdown(channel.sd, 2);
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("? ShutdownData sd="); Log.Int(channel.sd); Log.Ln
				END;
			END;
			IF close THEN
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("CloseData sd="); Log.Int(channel.sd); Log.Ln
				END;
				rc := WinSock.closesocket(channel.sd);
				channel.sd := 0
			END
		END
	END CloseSD;
	
	PROCEDURE CloseCD (channel: Channel; close, shutdown: BOOLEAN);
		VAR rc: Tb.Int32;
	BEGIN
		IF channel.cd # 0 THEN
			IF shutdown THEN
				rc := WinSock.shutdown(channel.cd, 2);
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("? ShutdownData cd="); Log.Int(channel.cd); Log.Ln
				END;
			END;
			IF close THEN
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("CloseData cd="); Log.Int(channel.cd); Log.Ln
				END;
				rc := WinSock.closesocket(channel.cd);
				channel.cd := 0
			END
		END;
	END CloseCD;

	PROCEDURE Disconnect(channel: Channel; VAR fd_changed: BOOLEAN);
	BEGIN
		IF channel.cd # 0 THEN
			CloseCD(channel, TRUE, TRUE);
			channel.link_stat := Tc.LINK_CLOSED;
			channel.InitData(FALSE, FALSE, channel.cqos.hostname, channel.cqos);
			fd_changed := TRUE
		ELSIF channel.sd # 0 THEN
			CloseSD(channel, TRUE, FALSE);
			channel.link_stat := Tc.LINK_CLOSED;
			fd_changed := TRUE
		END
	END Disconnect;
	
	PROCEDURE (channel: Channel) CloseData (shutdown: BOOLEAN);
		VAR rc: Tb.Int32;
	BEGIN
		CloseSD(channel, TRUE, FALSE);
		CloseCD(channel, TRUE, shutdown);
		channel.link_stat := Tc.LINK_CLOSED
	END CloseData;

	PROCEDURE TryDisconnectChannel (channel: Channel;
	VAR fd_changed: BOOLEAN);
		VAR t_crt: Tb.Time; rc: Tb.Int32; s: Tb.Name64;
	BEGIN
		IF channel.link_stat = Tc.LINK_OK THEN
			IF channel.disconnect_next THEN
				Disconnect(channel, fd_changed);
				channel.disconnect_next := FALSE
			ELSIF Tc.POLLHUP IN channel.poll_events THEN
				IF Tc.POLLIN IN channel.poll_events THEN
					IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
						Log.String("Wait for Close Socket on HUP poll="); 
						Log.Int(Tb.Set_Int32(channel.poll_events)); Log.Ln
					END
				ELSE
					Disconnect(channel, fd_changed)
				END
			ELSIF channel.cqos.tp_disconnect_timeout >= 0 THEN
				Tb.GetTime(t_crt);
				IF Tb.CompareTime(t_crt, channel.t_recv) >= 
				channel.cqos.tp_disconnect_timeout THEN
					Disconnect(channel, fd_changed)
				END
			END
		END
	END TryDisconnectChannel;
	
	PROCEDURE TryAcceptChannel (host_channel, channel: Channel;
	VAR fd_changed: BOOLEAN);
		VAR my_errno: Tb.Int32; sa_msg: Tl.sockaddr; sa_msg_len: Tb.Int32;
			sd: Tl.SOCKET; v: Tb.UInt32; opt: Tb.Int32;
	BEGIN
		sa_msg_len := SIZE(Tl.sockaddr);
		Tb.Clean(SYSTEM.ADR(sa_msg), sa_msg_len);
		sd := WinSock.accept(host_channel.sd, sa_msg, sa_msg_len);
		my_errno := WinSock.WSAGetLastError();
		IF (sd # 0) & (sd # SYSTEM.VAL(Tl.SOCKET, -1)) THEN
			channel.sd := sd;
			channel.link_stat := Tc.LINK_OK;
			v := 1;
			opt := Tl.FIONBIO;
			IF WinSock.ioctlsocket(sd, opt, 
			SYSTEM.VAL(WinSock.p_u_long, SYSTEM.ADR(v))) # 0 THEN
				my_errno := WinSock.WSAGetLastError();
				IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
				OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
					Log.String("* NONBLOCK socket err="); 
					Log.Int(my_errno); Log.Ln
				END;		
			ELSE
				fd_changed := TRUE;
				Tb.GetTime(channel.t_recv);
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("accept new sd="); Log.Int(sd);
					Log.String(" port="); Log.Int(host_channel.port); 
					Log.String(" sd="); Log.Int(host_channel.sd); Log.Ln
				END
			END;
		ELSIF my_errno = Tl.EWOULDBLOCK THEN
			sd := 0
		ELSE
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
			OR (Tb.TRAL_ERRORS IN Tb.trace) THEN
				Log.String("* accept err="); Log.Int(my_errno);
				Log.String(" port="); Log.Int(host_channel.port); 
				Log.String(" sd="); Log.Int(host_channel.sd); Log.Ln
			END
		END
	END TryAcceptChannel;

	PROCEDURE TryConnectChannel (channel: Channel);
		VAR rc: Tb.Int32; my_errno: Tb.Int32; t_crt: Tb.Time;
	BEGIN
		IF channel.link_stat = Tc.LINK_INIT THEN
			Tb.GetTime(channel.t_recv);
			rc := WinSock.connect(channel.cd, channel.sa_from, SIZE(Tl.sockaddr_in));
			my_errno := WinSock.WSAGetLastError();
			IF (rc >= 0) OR (my_errno = Tl.EISCONN) THEN
				channel.link_stat := Tc.LINK_OK;
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("connected "); Log.String(channel.hostname); 
					Log.String(" port="); Log.Int(channel.port); 
					Log.String(" cd="); Log.Int(channel.cd);  Log.Ln
				END
			ELSIF (my_errno = Tl.EWOULDBLOCK) 
			OR (my_errno = Tl.EINPROGRESS) THEN
				channel.link_stat := Tc.LINK_WAIT;
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("connecting "); Log.String(channel.hostname); 
					Log.String(" port="); Log.Int(channel.port); 
					Log.String(" cd="); Log.Int(channel.cd);  Log.Ln
				END
			END;
		ELSE
			(* Tc.LINK_WAIT *)
			IF Tc.POLLHUP IN channel.poll_events THEN
				channel.link_stat := Tc.LINK_INIT
			ELSIF Tc.POLLOUT IN channel.poll_events THEN
				Tb.GetTime(channel.t_recv);
				(* rc := WinSock.getpeername(channel.cd, sa_msg, sa_msg_len);*)
				(* The second connect here *)
				rc := WinSock.connect(channel.cd, channel.sa_from, SIZE(Tl.sockaddr_in));
				my_errno := WinSock.WSAGetLastError();
				IF (rc >= 0) OR (my_errno = Tl.EISCONN) THEN
					channel.link_stat := Tc.LINK_OK;
					IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
						Log.String("connected "); Log.String(channel.hostname); 
						Log.String(" port="); Log.Int(channel.port); 
						Log.String(" cd="); Log.Int(channel.cd);  Log.Ln
					END					
				END
			END;
			IF channel.link_stat = Tc.LINK_WAIT THEN
				Tb.GetTime(t_crt);
				IF Tb.CompareTime(t_crt, channel.t_recv) >= 
				channel.cqos.tp_connect_timeout THEN
					(* Do connect again *)
					channel.link_stat := Tc.LINK_INIT
				END
			END
		END
	END TryConnectChannel;
	
	PROCEDURE (channel: Channel) Connect (indirect_channel: Tc.Channel; 
	disconnect_only: BOOLEAN; VAR fd_changed: BOOLEAN);
		VAR host_channel: Channel;
	BEGIN
		ASSERT(indirect_channel IS Channel, Tb.WRONG_TYPE);
		host_channel := indirect_channel(Channel);
		CASE channel.link_stat OF
		|Tc.LINK_OK:
			TryDisconnectChannel(channel, fd_changed);
		|Tc.LINK_CLOSED, Tc.LINK_NONE:
			IF ~disconnect_only THEN
				channel.link_stat := Tc.LINK_NONE;
				IF (host_channel.link_stat = Tc.LINK_LISTEN)
				& (Tc.POLLIN IN host_channel.poll_events) THEN
					TryAcceptChannel(host_channel, channel, fd_changed)
				END
			END;
		|Tc.LINK_INIT, Tc.LINK_WAIT:
			IF ~disconnect_only THEN
				TryConnectChannel(channel)
			END;
		ELSE
		END
	END Connect;
	

	PROCEDURE (channel: Channel) Send (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32);
		VAR go: BOOLEAN; my_errno: Tb.Int32; sd: Tl.SOCKET;
	BEGIN
		nbytes := 0; sd := channel.sd;
		IF sd <= 0 THEN sd := channel.cd END;
		IF sd <= 0 THEN
			go := FALSE
		ELSIF Tc.POLLOUT IN channel.poll_events THEN
			go := TRUE
		ELSIF channel.cqos.max_send_polls > 0 THEN
			INC(channel.n_send_polls);
			IF channel.n_send_polls >= channel.cqos.max_send_polls THEN
				channel.n_send_polls := 0;
				go := TRUE
			ELSE
				go := FALSE
			END
		ELSE
			channel.n_send_polls := 0;
			go := FALSE
		END;
		IF go THEN
			nbytes := WinSock.send(sd, SYSTEM.VAL(Tl.pbyte, buffer), buffer_length,
				Tl.MSG_NOSIGNAL);
			my_errno := WinSock.WSAGetLastError();
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
			OR (nbytes < 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
				IF nbytes < 0 THEN Log.String("* send err="); Log.Int(my_errno) 
				ELSE Log.String("send") END; 
				Log.String(" port="); Log.Int(channel.port); Log.String(" sd="); 
				Log.Int(sd); Log.String(" rc="); Log.Int(nbytes); Log.String(" len="); 
				Log.Int(buffer_length); Log.Ln
			END;
			IF (nbytes < 0) & (my_errno = Tl.ECONNRESET) OR
			(my_errno = Tl.ECONNABORTED) OR (my_errno = Tl.ESHUTDOWN) THEN
				channel.disconnect_next := TRUE
			END
		END
	END Send;
	
	PROCEDURE (channel: Channel) Recv (indirect_channel: Tc.Channel;
	buffer: Tb.Address; buffer_length: Tb.Int32; VAR nbytes: Tb.Int32;
	VAR dest: Tc.Dest);
		VAR my_errno: Tb.Int32; sa_msg: Tl.sockaddr; sa_msg_len: Tb.Int32;
			sd: Tl.SOCKET;
	BEGIN
		nbytes := 0; sd := channel.sd;
		IF sd <= 0 THEN sd := channel.cd END;
		IF (sd > 0) & (Tc.POLLIN IN channel.poll_events) THEN
			nbytes := WinSock.recv(sd, SYSTEM.VAL(Tl.pbyte, buffer), buffer_length,
				Tl.MSG_NOSIGNAL);
			my_errno := WinSock.WSAGetLastError();
			Tb.GetTime(channel.t_recv);
			IF nbytes = 0 THEN
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("safe disconnect"); Log.Ln
				END;
				channel.disconnect_next := TRUE
			ELSIF nbytes = 1 THEN
				IF Tb.TRAL_TRANSPORT IN Tb.trace THEN
					Log.String("unsafe disconnect"); Log.Ln
				END;
				channel.disconnect_next := TRUE
			ELSIF nbytes < 0 THEN
				IF my_errno = Tl.EWOULDBLOCK THEN
					nbytes := 0
				ELSIF (my_errno = Tl.ECONNRESET) OR
				(my_errno = Tl.ECONNABORTED) OR (my_errno = Tl.ESHUTDOWN) THEN
					channel.disconnect_next := TRUE
				END
			END;
			IF (Tb.TRAL_TRANSPORT IN Tb.trace) 
			OR (nbytes < 0) & (Tb.TRAL_ERRORS IN Tb.trace) THEN
				IF nbytes < 0 THEN Log.String("* recv err="); Log.Int(my_errno) 
				ELSE Log.String("recv") END; 
				Log.String(" port="); Log.Int(channel.port); Log.String(" sd="); Log.Int(sd); 
				Log.String(" rc="); Log.Int(nbytes); Log.Ln
			END;
		END;	
	END Recv;	

	PROCEDURE (channel: Channel) MyDest ( dest-: Tc.Dest): BOOLEAN;
	BEGIN 
		RETURN TRUE (*(channel.cd # 0)*)
	END MyDest;


	(* ========== Initialize ========== *)
	PROCEDURE (cf: ChannelFactory) NewChannel* (aid, sid, host_sid: INTEGER; 
	(*OUT*)VAR chan: Tc.Channel; (*OUT*)VAR res: INTEGER);
		VAR channel: Channel;
	BEGIN
		NEW(channel);
		channel.flags := {Tc.FLAGS_STREAM};
		channel.trans_flags := {};
		channel.link_stat := Tc.LINK_NONE; 
		channel.sid_above := (sid > host_sid) & (host_sid >= 0);
		chan := channel;
		res := 0
	END NewChannel;
	
BEGIN
	Tc.DefaultStreamChannelQos(defaultCQos);
	defaultCQos.protocol := "TCP"
END Ta_TCPChannel.